
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       program cleanbg
!__________________________________________________________________________________________________________________________________
!
! Clean background by interpolating sources off the background according to their footprints.
!
! GETSF • Multi-Scale Multi-Wavelength Source & Filament Extraction • Alexander Men'shchikov, DAp IRFU CEA Saclay
!__________________________________________________________________________________________________________________________________
!
       implicit      none

       logical       lfnam1, lfnam2, lunix, ldone
                                          
       character*3   dot
       character*4   cpmx
       character*6   cfitsversion, cverbose
       character*7   clibname, cx, cy
       character*8   ctime, cfpptime                
       character*10  cdate
       character*21  compda
       character*80  object, bunit, ctype1, ctype2, creator, history
       character*500 arg1, arg2, arg3, arg4, arg5, arg6, originback, sourcemask, cleanback, outname, worker

       integer       firstb, lastc, lasts, i, j, iorigin, iobjs, iclean, ndate, iotty, nx, ny, nx1, ny1, nx2, ny2, ia1, ia2, iwork
     &             , ia3, ia4, ia5, ia6, iz1, iz2, jz1, jz2, k, l, ij, ji, im1, jm1, ii, iadd, jadd, nave, lb, nav, ia, ja, iim, iip
     &             , jjm, jjp, iim1, iip1, iim2, iip2, jjm1, jjp1, jjm2, jjp2, isp1, ion, irc, dnx, dny, dnxa, dnxb, dnya, dnyb
     &             , nmax, blank, nsort, ncx, ncy, nmed, nradmf, npmx, nn, mm, icb, iml, ipl, jml, jpl !!, n1, n2, m1, m2, n, m

       parameter   ( iotty = 6, iadd = 0, jadd = 0 )  !<-- number of additional pixels to offset (unused now)

       real          fitsvers

       real*8        selectk, dx, dy, dx1, dy1, dx2, dy2, xr, yr, fint, crpix1, crpix2, crval1, crval2, rx, ry, ra, dec, beam, beam1
     &             , beam2, almostzero, df, f1, f2, funmin, funmax, crota1, crota2, equinox, bzero, bscale, wave, datamin
     &             , datamax, cd11, cd12, cd21, cd22, length1, length2, length3, length4, rw2, rx2, ry2, rxy2, funw(4), func(4)
     &             , fnorm, powers, pwr, radmf

       integer     , allocatable :: cimask(:,:)
       real*8      , allocatable :: original(:,:), fclean(:,:), fwork(:,:), mask(:,:), fwork1(:,:), fclean1(:,:), fclean2(:,:)
     &             , fclean3(:,:), fclean4(:,:), fwork2(:,:), fwork3(:,:), fwork4(:,:), fun(:), funmed(:,:)

       parameter   ( almostzero = 1.0d-12, dot = '•', nave = 1 )

       external     osystem, ftvers, firstb, lastc, lasts, selectk, rfits, wfits, when, printerr, getfitshead, sortarr
!__________________________________________________________________________________________________________________________________
!
! Determine operating system

       call osystem ( lunix )
       call ftvers ( fitsvers )
       lb = 1
       write (cfitsversion,'(f6.3)') fitsvers
       if (cfitsversion .eq. ' 5.030') then
         lb = 2
       endif
       clibname = 'CFITSIO'
       cfpptime = __TIME__
       cfpptime = cfpptime(1:5)
       compda = __DATE__//' '//cfpptime

! Get command line parameters (file names).

       call getarg ( 1, arg1 )
       call getarg ( 2, arg2 )
       call getarg ( 3, arg3 )
       call getarg ( 4, arg4 )
       call getarg ( 5, arg5 )
       call getarg ( 6, arg6 )

       ia1 = lastc ( arg1 )
       ia2 = lastc ( arg2 )
       ia3 = lastc ( arg3 )
       ia4 = lastc ( arg4 )
       ia5 = lastc ( arg5 )
       ia6 = lastc ( arg6 )

       cverbose = '-verb2'
       if (ia6 .gt. 0 .and. arg6(1:ia6-1) .eq. '-verb') cverbose = arg6(1:ia6)
       if (ia5 .gt. 0 .and. arg5(1:ia5-1) .eq. '-verb') cverbose = arg5(1:ia5)
       if (ia4 .gt. 0 .and. arg4(1:ia4-1) .eq. '-verb') cverbose = arg4(1:ia4)
       if (ia3 .gt. 0 .and. arg3(1:ia3-1) .eq. '-verb') cverbose = arg3(1:ia3)

       if (cverbose .eq. '-verb2') then
         write (iotty,'( )')
         if (arg1(1:ia1) .eq. ':') write (iotty,'(a$)') '  '
         write (iotty,'(a$)') ' CLEANBG '//dot//' Remove Masked Structures '//dot//' '//compda
       endif
       if (arg1(1:ia1) .eq. ':') stop
       if (cverbose .eq. '-verb2') write (iotty,'()')
       if (cverbose .eq. '-verb2') write (iotty,'(a)') ' Alexander Men’shchikov, DAp IRFU CEA Saclay, France.'
       if (cverbose .eq. '-verb2') write (iotty,'(a)') ' Using '//clibname(lb:7)//' library version'//cfitsversion
     &                                               //' by William D Pence.'

       if (ia3 .eq. 0) then
         write (iotty,'(/a)') ' USAGE: cleanbg <nradmf> <image> <strucmask> [-o <outname>] [-verb{0|1|2}]'
         stop 99
       endif
       read (arg1(1:ia1),*,err=9) radmf
       nradmf = nint ( radmf )
       goto 10
  9    continue       
         write (iotty,'(/a)') '   CLEANBG: ERROR: Trouble reading command-line parameter ''NRADMF''.'
         stop 99
 10    continue
       originback = arg2(1:ia2)
       sourcemask = arg3(1:ia3)
       iorigin = lastc ( originback  )
       iobjs  = lastc ( sourcemask )

! Check if the input file exists and open it.

       inquire ( file=originback(1:iorigin), exist=lfnam1 )
       if (.not.lfnam1 .and. originback(iorigin-4:iorigin) .ne. '.fits') then
         originback = originback(1:iorigin)//'.fits'
         iorigin = lastc ( originback )
         inquire ( file=originback(1:iorigin), exist=lfnam1 )
       endif

! Check if the input file exists and open it.

       inquire ( file=sourcemask(1:iobjs), exist=lfnam2 )
       if (.not.lfnam2 .and. sourcemask(iobjs-4:iobjs) .ne. '.fits') then
         sourcemask = sourcemask(1:iobjs)//'.fits'
         iobjs = lastc ( sourcemask )
         inquire ( file=sourcemask(1:iobjs), exist=lfnam2 )
       endif

       isp1 = lasts ( originback ) + 1

       outname = originback(1:max(iorigin-5,1))//'.cb.fits'
       if (arg3(1:2) .eq. '-o') outname = arg4
       if (arg4(1:2) .eq. '-o') outname = arg5
       if (arg5(1:2) .eq. '-o') outname = arg6
       ion = lastc ( outname )
       if (outname(max(ion-4,1):ion) .ne. '.fits') then
         outname = outname(1:ion)//'.fits'
         ion = lastc ( outname )
       endif              
       cleanback = outname(1:ion)
!!       icb = index ( outname(1:ion), '.cb.' )
!!       if (icb .eq. 0) then
!!         if (outname(max(ion-8,1):ion) .ne. '.cb.fits') then
!!           cleanback = outname(1:max(ion-5,1))//'.cb.fits'
!!         endif
!!       endif
       iclean = lastc ( cleanback )
       icb = index ( cleanback(1:iclean), '.cb.' )
       worker = cleanback
       worker(icb:icb+3) = '.wk.'
       iwork = lastc ( worker )

       if (cverbose .eq. '-verb1') write (iotty,'(/a)') ' CLEANBG: '//arg1(1:ia1)//' '//originback(1:iorigin-5)
     &                                                //' '//sourcemask(1:iobjs-5)
       if (.not.lfnam1) then
         write (iotty,'(/a)') '   CLEANBG: ERROR: File '''//originback(1:iorigin-5)//''' not found.'
         stop 1
       endif
       if (.not.lfnam2) then
         write (iotty,'(/a)') '   CLEANBG: ERROR: File '''//sourcemask(1:iobjs-5)//''' not found.'
         stop 2
       endif

! Determine numbers of pixels in the FITS image.

       call getfitshead ( originback(1:iorigin), nx1, ny1, dx1, dy1, bunit )

       allocate ( original(nx1,ny1), stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a)') '   CLEANBG: ERROR: Trouble allocating memory (5).'
         stop 5
       endif

       write (cx,'(i7)') nx1
       write (cy,'(i7)') ny1
       ncx = int ( log10 ( dble ( nx1 ) ) ) + 1
       ncy = int ( log10 ( dble ( ny1 ) ) ) + 1

       if (cverbose .eq. '-verb2') write (iotty,'(/a)') '   Reading ('//cx(7-ncx+1:7)//' x '//cy(7-ncy+1:7)//') '''
     &                                                //originback(1:iorigin)//''''

       call rfits ( nx1, ny1, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, original, dx1, dy1, object, ra, dec
     &            , originback(1:iorigin), iotty, 0, creator, beam1, funmin, funmax, blank, crota1, crota2
     &            , cd11, cd12, cd21, cd22, equinox, bzero, bscale, wave, datamin, datamax, history, cverbose )

       if (cverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7))') '   Minmax values in the image:', funmin, funmax

! Determine numbers of pixels in the FITS image.

       call getfitshead ( sourcemask(1:iobjs), nx2, ny2, dx2, dy2, bunit )

       allocate ( mask(nx2,ny2), stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a)') '   CLEANBG: ERROR: Trouble allocating memory (6).'
         stop 6
       endif

       write (cx,'(i7)') nx2
       write (cy,'(i7)') ny2
       ncx = int ( log10 ( dble ( nx2 ) ) ) + 1
       ncy = int ( log10 ( dble ( ny2 ) ) ) + 1

       if (cverbose .eq. '-verb2') write (iotty,'(a)') '   Reading ('//cx(7-ncx+1:7)//' x '//cy(7-ncy+1:7)//') '''
     &                                               //sourcemask(1:iobjs)//''''

       call rfits ( nx2, ny2, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, mask, dx2, dy2, object, ra, dec
     &            , sourcemask(1:iobjs), iotty, 0, creator, beam2, funmin, funmax, blank, crota1, crota2
     &            , cd11, cd12, cd21, cd22, equinox, bzero, bscale, wave, datamin, datamax, history, cverbose )

       if (cverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7))') '   Minmax values in the image:', funmin, funmax

       if (beam1 .eq. beam2) then
         beam = beam1
       else
         beam = 0.0
       endif

       if (nx1 .ne. nx2 .or. ny1 .ne. ny2 .or. abs ( dx1 - dx2 ) .gt. 0.001d0 * dx2 .or. abs ( dy1 - dy2 ) .gt. 0.001d0 * dy2) then
         write (iotty, '(/a)') '   CLEANBG: ERROR: The images have different parameters.'
         stop 99
       else
         nx = nx1
         ny = ny1
         dx = dx1
         dy = dy1
         rx = crpix1
         ry = crpix2
         xr = crval1
         yr = crval2
       endif
              
       nmax = max ( nx, ny )
       dnx = 0
       dny = 0
       if (nx .gt. ny) dny = nx - ny
       if (ny .gt. nx) dnx = ny - nx
       dnxa = dnx / 2
       dnya = dny / 2
       dnxb = dnxa
       dnyb = dnya
       if (mod ( dnx, 2 ) .gt. 0) dnxa = dnxa + 1
       if (mod ( dny, 2 ) .gt. 0) dnya = dnya + 1

       allocate ( fwork1(nx,ny), fwork2(nx,ny), fwork3(nx,ny), fwork4(nx,ny), fclean(nx,ny), fwork(nx,ny)
     &          , fclean1(nx,ny), fclean2(nx,ny), fclean3(nx,ny), fclean4(nx,ny), fun(nx*ny), funmed(nx,ny), stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a)') '   CLEANBG: ERROR: Trouble allocating memory (10).'
         stop 10
       endif

       if (cverbose .eq. '-verb2') write (iotty,'(a)') '   Cleaning image by interpolating masked structures'

       do j=1,ny
         do i=1,nx
           fwork1 (i,j) = 0.0d0
           fwork2 (i,j) = 0.0d0
           fwork3 (i,j) = 0.0d0
           fwork4 (i,j) = 0.0d0
           fclean (i,j) = original(i,j)
           fclean1(i,j) = original(i,j)
           fclean2(i,j) = original(i,j)
           fclean3(i,j) = original(i,j)
           fclean4(i,j) = original(i,j)

           if (isnan ( mask(i,j) )) then
             mask(i,j) = 0.0d0
           endif
         enddo
       enddo

! Here the 1-pixel border around the entire mask image is zeroed (if not zero yet), aiding in interpolations below.
       
       do i=1,nx
         mask(i,1 ) = 0.0d0
         mask(i,ny) = 0.0d0
         original(i,1 ) = 0.0d0
         original(i,ny) = 0.0d0
       enddo
       do j=1,ny
         mask( 1,j) = 0.0d0
         mask(nx,j) = 0.0d0
         original( 1,j) = 0.0d0
         original(nx,j) = 0.0d0
       enddo
!-------------------------------------------------
! Find background and interpolate in the X-direction.
!-------------------------------------------------
       iz1 = 0
       iz2 = 0
       im1 = 1
       do j=1,ny
         do i=1,nx
           if (mask(im1,j) .le. almostzero .and. mask(i,j) .gt. almostzero) iz1 = max ( im1 - iadd, 1  )
           if (mask(im1,j) .gt. almostzero .and. mask(i,j) .le. almostzero) iz2 = min ( i   + iadd, nx )
           if (iz2 .gt. 0) then
             jjm  = max (  j  - nave, 1  )               
             jjp  = min (  j  + nave, ny )
             iim1 = max ( iz1 - nave, 1  )
             iip1 = min ( iz1 + nave, nx )
             iim2 = max ( iz2 - nave, 1  )
             iip2 = min ( iz2 + nave, nx )

! Average start interpolation pixel value 'f1'.

             nav = 0
             f1 = 0.0d0
             do ja=jjm,jjp
               do ia=iim1,iip1
                 if (mask(ia,ja) .le. almostzero) then
                   nav = nav + 1
                   f1 = f1 + original(ia,ja)
                 endif
               enddo
             enddo
             f1 = f1 / dble ( nav )

! Average end interpolation pixel value 'f2'.

             nav = 0
             f2 = 0.0d0
             do ja=jjm,jjp
               do ia=iim2,iip2
                 if (mask(ia,ja) .le. almostzero) then
                   nav = nav + 1
                   f2 = f2 + original(ia,ja)
                 endif
               enddo
             enddo        
             f2 = f2 / dble ( nav )
             
! Interpolate.

             length1 = dble ( iz2 - iz1 + 1 )
             df = (f2 - f1) / dble ( iz2 - iz1 )
             do k=iz1,iz2                 
               fint = f1 + df * dble ( k - iz1 )
               fwork1 (k,j) = length1
               fclean1(k,j) = fint
             enddo 

!!! If there is averaging of interpolation pixels 'f1' and 'f2', i.e. if the averaging window is greater than 1 pixel,
!!! spread the interpolation value over the entire window (outside the mask!). This smoothes the image and prevents
!!! appearance of jumps in values at the edge of the mask.
!!
!!             do ja=jjm,jjp
!!               do ia=iim1,iip1
!!                 if (mask(ia,ja) .le. almostzero) then
!!                   fwork1 (ia,ja) = length1
!!                   fclean1(ia,ja) = f1
!!                 endif
!!               enddo
!!             enddo
!!             
!!             do ja=jjm,jjp
!!               do ia=iim2,iip2
!!                 if (mask(ia,ja) .le. almostzero) then
!!                   fwork1 (ia,ja) = length1
!!                   fclean1(ia,ja) = f2
!!                 endif
!!               enddo
!!             enddo        

             fclean2(iz1,j) = f1
             fclean3(iz1,j) = f1
             fclean4(iz1,j) = f1
             fclean2(iz2,j) = f2
             fclean3(iz2,j) = f2
             fclean4(iz2,j) = f2
             iz1 = 0
             iz2 = 0 
           endif
           im1 = i
         enddo
         iz1 = 0
         iz2 = 0
         im1 = 1
       enddo
!-------------------------------------------------
! Find background and interpolate in the Y-direction.
!-------------------------------------------------
       jz1 = 0
       jz2 = 0
       jm1 = 1
       do i=1,nx
         do j=1,ny
           if (mask(i,jm1) .le. almostzero .and. mask(i,j) .gt. almostzero) jz1 = max ( jm1 - jadd, 1  )
           if (mask(i,jm1) .gt. almostzero .and. mask(i,j) .le. almostzero) jz2 = min ( j   + jadd, ny )
           if (jz2 .gt. 0) then
             iim  = max (  i  - nave, 1  )               
             iip  = min (  i  + nave, nx )
             jjm1 = max ( jz1 - nave, 1  )
             jjp1 = min ( jz1 + nave, ny )
             jjm2 = max ( jz2 - nave, 1  )
             jjp2 = min ( jz2 + nave, ny )

! Average start interpolation pixel value 'f1'.

             nav = 0
             f1 = 0.0d0
             do ia=iim,iip
               do ja=jjm1,jjp1
                 if (mask(ia,ja) .le. almostzero) then
                   nav = nav + 1
                   f1 = f1 + original(ia,ja)
                 endif
               enddo
             enddo
             f1 = f1 / dble ( nav )

! Average end interpolation pixel value 'f2'.

             nav = 0
             f2 = 0.0d0
             do ia=iim,iip
               do ja=jjm2,jjp2
                 if (mask(ia,ja) .le. almostzero) then
                   nav = nav + 1
                   f2 = f2 + original(ia,ja)
                 endif
               enddo
             enddo        
             f2 = f2 / dble ( nav )

! Interpolate.

             length2 = dble ( jz2 - jz1 + 1 )
             df = (f2 - f1) / dble ( jz2 - jz1 )
             do k=jz1,jz2
               fint = f1 + df * dble ( k - jz1 )
               fwork2 (i,k) = length2
               fclean2(i,k) = fint
             enddo 

!!! If there is averaging of interpolation pixels 'f1' and 'f2', i.e. if the averaging window is greater than 1 pixel,
!!! spread the interpolation value over the entire window (outside the mask!). This smoothes the image and prevents
!!! appearance of jumps in values at the edge of the mask.
!!
!!             do ia=iim,iip
!!               do ja=jjm1,jjp1
!!                 if (mask(ia,ja) .le. almostzero) then
!!                   fwork2 (ia,ja) = length2
!!                   fclean2(ia,ja) = f1
!!                 endif
!!               enddo
!!             enddo
!!
!!             do ia=iim,iip
!!               do ja=jjm2,jjp2
!!                 if (mask(ia,ja) .le. almostzero) then
!!                   fwork2 (ia,ja) = length2
!!                   fclean2(ia,ja) = f2
!!                 endif
!!               enddo
!!             enddo        

             fclean1(i,jz1) = f1
             fclean3(i,jz1) = f1
             fclean4(i,jz1) = f1
             fclean1(i,jz2) = f2
             fclean3(i,jz2) = f2
             fclean4(i,jz2) = f2
             jz1 = 0
             jz2 = 0 
           endif
           jm1 = j
         enddo
         jz1 = 0
         jz2 = 0
         jm1 = 1
       enddo
!-----------------------------------------------------------------------------
! Find background and interpolate in the forward diagonal direction (upper part).
!-----------------------------------------------------------------------------
       iz1 = 0
       iz2 = 0
       jz1 = 0
       jz2 = 0
       im1 = 1
       do j=nmax,(1-dnya),-1
         jm1 = max ( j - 1, 1 )
         do i=(1-dnxa),nmax-j+1
           ji = j + i - 1
           if (i .ge. 1 .and. i .le. nx .and. ji .ge. 1 .and. ji .le. ny) then
             if (mask(im1,jm1) .le. almostzero .and. mask(i,ji) .gt. almostzero) then 
               iz1 = max ( im1 - iadd, 1 )
               jz1 = max ( jm1 - jadd, 1 )
             endif
             if (mask(im1,jm1) .gt. almostzero .and. mask(i,ji) .le. almostzero) then 
               iz2 = min ( i  + iadd, nx )
               jz2 = min ( ji + jadd, ny )
             endif
             if (jz2 .gt. 0) then
               iim1 = max ( iz1 - nave, 1  )
               iip1 = min ( iz1 + nave, nx )
               iim2 = max ( iz2 - nave, 1  )
               iip2 = min ( iz2 + nave, nx )
               jjm1 = max ( jz1 - nave, 1  )
               jjp1 = min ( jz1 + nave, ny )
               jjm2 = max ( jz2 - nave, 1  )
               jjp2 = min ( jz2 + nave, ny )

! Average start interpolation pixel value 'f1'.

               nav = 0
               f1 = 0.0d0
               do ja=jjm1,jjp1
                 do ia=iim1,iip1
                   if (mask(ia,ja) .le. almostzero) then
                     nav = nav + 1
                     f1 = f1 + original(ia,ja)
                   endif
                 enddo
               enddo
               f1 = f1 / dble ( nav )

! Average end interpolation pixel value 'f2'.

               nav = 0
               f2 = 0.0d0
               do ja=jjm2,jjp2
                 do ia=iim2,iip2
                   if (mask(ia,ja) .le. almostzero) then
                     nav = nav + 1
                     f2 = f2 + original(ia,ja)
                   endif
                 enddo
               enddo        
               f2 = f2 / dble ( nav )

! Interpolate.

               length3 = sqrt ( dble ( iz2 - iz1 + 1 )**2 + dble ( jz2 - jz1 + 1 )**2 )
               df = (f2 - f1) / dble ( jz2 - jz1 )
               l = 0
               do k=jz1,jz2             
                 fint = f1 + df * dble ( k - jz1 )
                 fwork3 (iz1+l,k) = length3
                 fclean3(iz1+l,k) = fint
                 l = l + 1
               enddo 

!!! If there is averaging of interpolation pixels 'f1' and 'f2', i.e. if the averaging window is greater than 1 pixel,
!!! spread the interpolation value over the entire window (outside the mask!). This smoothes the image and prevents
!!! appearance of jumps in values at the edge of the mask.
!!
!!               do ja=jjm1,jjp1
!!                 do ia=iim1,iip1
!!                   if (mask(ia,ja) .le. almostzero) then
!!                     fwork3 (ia,ja) = length3
!!                     fclean3(ia,ja) = f1
!!                   endif
!!                 enddo
!!               enddo
!!               
!!               do ja=jjm2,jjp2
!!                 do ia=iim2,iip2
!!                   if (mask(ia,ja) .le. almostzero) then
!!                     fwork3 (ia,ja) = length3
!!                     fclean3(ia,ja) = f2
!!                   endif
!!                 enddo
!!               enddo        
  
               fclean1(iz1,jz1) = f1
               fclean2(iz1,jz1) = f1
               fclean4(iz1,jz1) = f1
               fclean1(iz2,jz2) = f2
               fclean2(iz2,jz2) = f2
               fclean4(iz2,jz2) = f2
               iz1 = 0
               iz2 = 0 
               jz1 = 0
               jz2 = 0 
             endif
             im1 = i
             jm1 = ji
           endif
         enddo
         iz1 = 0
         iz2 = 0
         jz1 = 0
         jz2 = 0
         im1 = 1
       enddo
!-----------------------------------------------------------------------------
! Find background and interpolate in the forward diagonal direction (lower part).
!-----------------------------------------------------------------------------
       iz1 = 0
       iz2 = 0
       jz1 = 0
       jz2 = 0
       jm1 = 1
       do i=nmax,(2-dnxa),-1
         im1 = max ( i - 1, 1 )
         do j=(1-dnya),nmax-i+1
           ij = i + j - 1
           if (ij .ge. 1 .and. ij .le. nx .and. j .ge. 1 .and. j .le. ny) then
             if (mask(im1,jm1) .le. almostzero .and. mask(ij,j) .gt. almostzero) then 
               iz1 = max ( im1 - iadd, 1 )
               jz1 = max ( jm1 - jadd, 1 )
             endif
             if (mask(im1,jm1) .gt. almostzero .and. mask(ij,j) .le. almostzero) then 
               iz2 = min ( ij + iadd, nx )
               jz2 = min ( j  + jadd, ny )
             endif
             if (iz2 .gt. 0) then
               iim1 = max ( iz1 - nave, 1  )
               iip1 = min ( iz1 + nave, nx )
               iim2 = max ( iz2 - nave, 1  )
               iip2 = min ( iz2 + nave, nx )
               jjm1 = max ( jz1 - nave, 1  )
               jjp1 = min ( jz1 + nave, ny )
               jjm2 = max ( jz2 - nave, 1  )
               jjp2 = min ( jz2 + nave, ny )

! Average start interpolation pixel value 'f1'.

               nav = 0
               f1 = 0.0d0
               do ja=jjm1,jjp1
                 do ia=iim1,iip1
                   if (mask(ia,ja) .le. almostzero) then
                     nav = nav + 1
                     f1 = f1 + original(ia,ja)
                   endif
                 enddo
               enddo
               f1 = f1 / dble ( nav )

! Average end interpolation pixel value 'f2'.

               nav = 0
               f2 = 0.0d0
               do ja=jjm2,jjp2
                 do ia=iim2,iip2
                   if (mask(ia,ja) .le. almostzero) then
                     nav = nav + 1
                     f2 = f2 + original(ia,ja)
                   endif
                 enddo
               enddo        
               f2 = f2 / dble ( nav )

! Interpolate.

               length3 = sqrt ( dble ( iz2 - iz1 + 1 )**2 + dble ( jz2 - jz1 + 1 )**2 )
               df = (f2 - f1) / dble ( iz2 - iz1 )
               l = 0
               do k=iz1,iz2             
                 fint = f1 + df * dble ( k - iz1 )
                 fwork3 (k,jz1+l) = length3
                 fclean3(k,jz1+l) = fint
                 l = l + 1
               enddo 

!!! If there is averaging of interpolation pixels 'f1' and 'f2', i.e. if the averaging window is greater than 1 pixel,
!!! spread the interpolation value over the entire window (outside the mask!). This smoothes the image and prevents
!!! appearance of jumps in values at the edge of the mask.
!!
!!               do ja=jjm1,jjp1
!!                 do ia=iim1,iip1
!!                   if (mask(ia,ja) .le. almostzero) then
!!                     fwork3 (ia,ja) = length3
!!                     fclean3(ia,ja) = f1
!!                   endif
!!                 enddo
!!               enddo
!!               
!!               do ja=jjm2,jjp2
!!                 do ia=iim2,iip2
!!                   if (mask(ia,ja) .le. almostzero) then
!!                     fwork3 (ia,ja) = length3
!!                     fclean3(ia,ja) = f2
!!                   endif
!!                 enddo
!!               enddo        
  
               fclean1(iz1,jz1) = f1
               fclean2(iz1,jz1) = f1
               fclean4(iz1,jz1) = f1
               fclean1(iz2,jz2) = f2
               fclean2(iz2,jz2) = f2
               fclean4(iz2,jz2) = f2
               iz1 = 0
               iz2 = 0 
               jz1 = 0
               jz2 = 0 
             endif
             im1 = ij
             jm1 = j
           endif
         enddo
         iz1 = 0
         iz2 = 0
         jz1 = 0
         jz2 = 0
         jm1 = 1
       enddo
!-----------------------------------------------------------------------------
! Find background and interpolate in the backward diagonal direction (upper part).
!-----------------------------------------------------------------------------
       iz1 = 0
       iz2 = 0
       jz1 = 0
       jz2 = 0
       im1 = nx
       do j=nmax,(1-dnya),-1
         jm1 = max ( j - 1, 1 )
         do i=(1-dnxa),nmax-j+1
           ii = nx - i + 1
           ji = j + i - 1
           if (ii .ge. 1 .and. ii .le. nx .and. ji .ge. 1 .and. ji .le. ny) then
             if (mask(im1,jm1) .le. almostzero .and. mask(ii,ji) .gt. almostzero) then 
               iz1 = min ( im1 + iadd, nx )
               jz1 = max ( jm1 - jadd, 1  )
             endif
             if (mask(im1,jm1) .gt. almostzero .and. mask(ii,ji) .le. almostzero) then 
               iz2 = max ( ii - iadd, 1  )
               jz2 = min ( ji + jadd, ny )
             endif
             if (jz2 .gt. 0) then
               iim1 = max ( iz1 - nave, 1  )
               iip1 = min ( iz1 + nave, nx )
               iim2 = max ( iz2 - nave, 1  )
               iip2 = min ( iz2 + nave, nx )
               jjm1 = max ( jz1 - nave, 1  )
               jjp1 = min ( jz1 + nave, ny )
               jjm2 = max ( jz2 - nave, 1  )
               jjp2 = min ( jz2 + nave, ny )

! Average start interpolation pixel value 'f1'.

               nav = 0
               f1 = 0.0d0
               do ja=jjm1,jjp1
                 do ia=iim1,iip1
                   if (mask(ia,ja) .le. almostzero) then
                     nav = nav + 1
                     f1 = f1 + original(ia,ja)
                   endif
                 enddo
               enddo
               f1 = f1 / dble ( nav )

! Average end interpolation pixel value 'f2'.

               nav = 0
               f2 = 0.0d0
               do ja=jjm2,jjp2
                 do ia=iim2,iip2
                   if (mask(ia,ja) .le. almostzero) then
                     nav = nav + 1
                     f2 = f2 + original(ia,ja)
                   endif
                 enddo
               enddo        
               f2 = f2 / dble ( nav )

! Interpolate.

               length4 = sqrt ( dble ( iz2 - iz1 + 1 )**2 + dble ( jz2 - jz1 + 1 )**2 )
               df = (f2 - f1) / dble ( jz2 - jz1 )
               l = 0
               do k=jz1,jz2             
                 fint = f1 + df * dble ( k - jz1 )
                 fwork4 (iz1-l,k) = length4
                 fclean4(iz1-l,k) = fint
                 l = l + 1
               enddo 

!!! If there is averaging of interpolation pixels 'f1' and 'f2', i.e. if the averaging window is greater than 1 pixel,
!!! spread the interpolation value over the entire window (outside the mask!). This smoothes the image and prevents
!!! appearance of jumps in values at the edge of the mask.
!!
!!               do ja=jjm1,jjp1
!!                 do ia=iim1,iip1
!!                   if (mask(ia,ja) .le. almostzero) then
!!                     fwork4 (ia,ja) = length4
!!                     fclean4(ia,ja) = f1
!!                   endif
!!                 enddo
!!               enddo
!!               
!!               do ja=jjm2,jjp2
!!                 do ia=iim2,iip2
!!                   if (mask(ia,ja) .le. almostzero) then
!!                     fwork4 (ia,ja) = length4
!!                     fclean4(ia,ja) = f2
!!                   endif
!!                 enddo
!!               enddo        
               
               fclean1(iz1,jz1) = f1
               fclean2(iz1,jz1) = f1
               fclean3(iz1,jz1) = f1
               fclean1(iz2,jz2) = f2
               fclean2(iz2,jz2) = f2
               fclean3(iz2,jz2) = f2
               iz1 = 0
               iz2 = 0 
               jz1 = 0
               jz2 = 0 
             endif
             im1 = ii
             jm1 = ji
           endif
         enddo
         iz1 = 0
         iz2 = 0
         jz1 = 0
         jz2 = 0
         im1 = nx
       enddo
!-----------------------------------------------------------------------------
! Find background and interpolate in the backward diagonal direction (lower part).
!-----------------------------------------------------------------------------
       iz1 = 0
       iz2 = 0
       jz1 = 0
       jz2 = 0
       jm1 = 1
       do i=(1-dnxa),nmax-1
         im1 = min ( i + 1, nx )
         do j=(1-dnya),i
           ij = i - j + 1
           if (ij .ge. 1 .and. ij .le. nx .and. j .ge. 1 .and. j .le. ny) then
             if (mask(im1,jm1) .le. almostzero .and. mask(ij,j) .gt. almostzero) then 
               iz1 = min ( im1 + iadd, nx )
               jz1 = max ( jm1 - jadd, 1  )
             endif
             if (mask(im1,jm1) .gt. almostzero .and. mask(ij,j) .le. almostzero) then 
               iz2 = max ( ij - iadd, 1  )
               jz2 = min ( j  + jadd, ny )
             endif
             if (iz2 .gt. 0) then
               iim1 = max ( iz1 - nave, 1  )
               iip1 = min ( iz1 + nave, nx )
               iim2 = max ( iz2 - nave, 1  )
               iip2 = min ( iz2 + nave, nx )
               jjm1 = max ( jz1 - nave, 1  )
               jjp1 = min ( jz1 + nave, ny )
               jjm2 = max ( jz2 - nave, 1  )
               jjp2 = min ( jz2 + nave, ny )

! Average start interpolation pixel value 'f1'.

               nav = 0
               f1 = 0.0d0
               do ja=jjm1,jjp1
                 do ia=iim1,iip1
                   if (mask(ia,ja) .le. almostzero) then
                     nav = nav + 1
                     f1 = f1 + original(ia,ja)
                   endif
                 enddo
               enddo
               f1 = f1 / dble ( nav )

! Average end interpolation pixel value 'f2'.

               nav = 0
               f2 = 0.0d0
               do ja=jjm2,jjp2
                 do ia=iim2,iip2
                   if (mask(ia,ja) .le. almostzero) then
                     nav = nav + 1
                     f2 = f2 + original(ia,ja)
                   endif
                 enddo
               enddo        
               f2 = f2 / dble ( nav )

! Interpolate.

               length4 = sqrt ( dble ( iz2 - iz1 + 1 )**2 + dble ( jz2 - jz1 + 1 )**2 )
               df = (f2 - f1) / dble ( iz2 - iz1 )
               l = 0
               do k=iz1,iz2,-1          
                 fint = f1 + df * dble ( k - iz1 )
                 fwork4 (k,jz1+l) = length4
                 fclean4(k,jz1+l) = fint
                 l = l + 1
               enddo 

!!! If there is averaging of interpolation pixels 'f1' and 'f2', i.e. if the averaging window is greater than 1 pixel,
!!! spread the interpolation value over the entire window (outside the mask!). This smoothes the image and prevents
!!! appearance of jumps in values at the edge of the mask.
!!
!!               do ja=jjm1,jjp1
!!                 do ia=iim1,iip1
!!                   if (mask(ia,ja) .le. almostzero) then
!!                     fwork4 (ia,ja) = length4
!!                     fclean4(ia,ja) = f1
!!                   endif
!!                 enddo
!!               enddo
!!  
!!               do ja=jjm2,jjp2
!!                 do ia=iim2,iip2
!!                   if (mask(ia,ja) .le. almostzero) then
!!                     fwork4 (ia,ja) = length4
!!                     fclean4(ia,ja) = f2
!!                   endif
!!                 enddo
!!               enddo        
               
               fclean1(iz1,jz1) = f1
               fclean2(iz1,jz1) = f1
               fclean3(iz1,jz1) = f1
               fclean1(iz2,jz2) = f2
               fclean2(iz2,jz2) = f2
               fclean3(iz2,jz2) = f2
               iz1 = 0
               iz2 = 0 
               jz1 = 0
               jz2 = 0 
             endif
             im1 = ij
             jm1 = j
           endif
         enddo
         iz1 = 0
         iz2 = 0
         jz1 = 0
         jz2 = 0
         jm1 = 1
       enddo
!-----------------------------------------------------------------------------
! Averaging all the interpolations.
!-----------------------------------------------------------------------------
       nsort = 4
       do j=1,ny
         do i=1,nx
           funw(1) = fwork1(i,j)
           funw(2) = fwork2(i,j)
           funw(3) = fwork3(i,j)
           funw(4) = fwork4(i,j)
           fwork(i,j) = funw(1) + funw(2) + funw(3) + funw(4)

           func(1) = fclean1(i,j)
           func(2) = fclean2(i,j)
           func(3) = fclean3(i,j)
           func(4) = fclean4(i,j)

           fnorm = abs ( func(1) + func(2) + func(3) + func(4) ) / 4.0d0
           if (fnorm .lt. almostzero) fnorm = 1.0d0
           
! Sort the array of the four interpolated values (FUNC) according to the array of four interpolation lengths (FUNW).

           call sortarr ( nsort, funw, func )
           
! Average only those directions that have not too distant outside pixels providing the interpolated values.
! This is especially important for crowded sources, whose footprints overlap.

!!           if (funw(2) .lt. 2.0d0 * funw(1) .and. funw(3) .lt. 2.0d0 * funw(1) .and. funw(4) .lt. 2.0d0 * funw(1)) then 
!!!!             fclean(i,j) = (func(1) + func(2) + func(3) + func(4)) / 4.0d0
!!             fclean(i,j) = max ( func(1), func(2), func(3), func(4) )
!!           else if (funw(2) .lt. 2.0d0 * funw(1) .and. funw(3) .lt. 2.0d0 * funw(1)) then 
!!!!             fclean(i,j) = (func(1) + func(2) + func(3)) / 3.0d0
!!             fclean(i,j) = max ( func(1), func(2), func(3) )
!!           else if (funw(2) .lt. 2.0d0 * funw(1)) then 
!!!!             fclean(i,j) = (func(1) + func(2)) / 2.0d0
!!             fclean(i,j) = max ( func(1), func(2) )
!!           else
!!             fclean(i,j) = func(1)
!!           endif
!!           fclean(i,j) = selectk ( (nsort + 1) / 2, nsort, func )
!!           fclean(i,j) = max ( func(1), func(2), func(3), func(4) )

           pwr = 9.0d0

           if (funw(2) .lt. 2.0d0 * funw(1) .and. funw(3) .lt. 2.0d0 * funw(1) .and. funw(4) .lt. 2.0d0 * funw(1)) then 

             powers = ((func(1) / fnorm)**pwr + (func(2) / fnorm)**pwr + (func(3) / fnorm)**pwr + (func(4) / fnorm)**pwr) / 4.0d0
             fclean(i,j) = sign ( fnorm, powers ) * abs ( powers )**(1.0d0 / pwr)

           else if (funw(2) .lt. 2.0d0 * funw(1) .and. funw(3) .lt. 2.0d0 * funw(1)) then 

             powers = ((func(1) / fnorm)**pwr + (func(2) / fnorm)**pwr + (func(3) / fnorm)**pwr) / 3.0d0
             fclean(i,j) = sign ( fnorm, powers ) * abs ( powers )**(1.0d0 / pwr)

           else if (funw(2) .lt. 2.0d0 * funw(1)) then 

             powers = ((func(1) / fnorm)**pwr + (func(2) / fnorm)**pwr) / 2.0d0
             fclean(i,j) = sign ( fnorm, powers ) * abs ( powers )**(1.0d0 / pwr)

           else
             fclean(i,j) = func(1)
           endif
         enddo
       enddo
       
       if (nradmf .gt. 0) then
         write (cpmx,'(i4.4)') 2 * nradmf + 1
         npmx = int ( log10 ( dble ( 2 * nradmf + 1 ) ) ) + 1
         if (cverbose .eq. '-verb2') then
           write (iotty,'(a,i1,a)') '   Median filtering in a circular window of D = '//cpmx(4-npmx+1:4)//' pixels'
         endif

! Circular mask.

         if (dble (nradmf) .le. 2.0d0) then
           rw2 = (dble (nradmf) + 0.5d0)**2
         else
           rw2 = (dble (nradmf))**2
         endif

         allocate ( cimask(-nradmf:nradmf,-nradmf:nradmf), stat=irc )
         if (irc .ne. 0) then
           write (iotty,'(/a)') '   CLEANBG: ERROR: Trouble allocating memory (160).'
           stop 160
         endif

         do l=-nradmf,nradmf
           ry2 = (dble ( l ))**2
           do k=-nradmf,nradmf
             rx2 = (dble ( k ))**2
             rxy2 = rx2 + ry2
             if (rxy2 .le. rw2) then
               cimask(k,l) = 1 
             else
               cimask(k,l) = 0
             endif
           enddo
         enddo

         do j=1,ny
!!           n1 = max ( j - nradmf, 1 )
!!           n2 = min ( j + nradmf, ny )
           do i=1,nx
             funmed(i,j) = fclean(i,j)
             if (mask(i,j) .gt. almostzero) then

!!               m1 = max ( i - nradmf, 1 )
!!               m2 = min ( i + nradmf, nx )
!!               nmed = 0
!!               do m=m1,m2
!!                 do n=n1,n2
!!                   if (mask(m,n) .gt. almostzero) then
!!                     nmed = nmed + 1
!!                     fun(nmed) = fclean(m,n)
!!                   endif
!!                 enddo
!!               enddo

! Next spiral scanning of the sliding window is to automatically shrink the window radius when it comes to the mask edges.
! This avoids jumps of the filtered areas adjucent to the edges of the masked area, it makes the result smooth at egdes.

               nmed = 1
               fun(nmed) = fclean(i,j)

               do l=1,nradmf

                 ldone = .false.

                 do nn=min(j+l,ny),max(j-l+1,1),-1
                   ipl = min ( i + l, nx )
                   if (mask(ipl,nn) .gt. almostzero .and. cimask(ipl-i,nn-j) .ge. 1) then
                     nmed = nmed + 1
                     fun(nmed) = fclean(ipl,nn)
                   else
                     ldone = .true.
                   endif
                 enddo
                 do mm=min(i+l,nx),max(i-l+1,1),-1
                   jml = max ( j - l, 1 )
                   if (mask(mm,jml) .gt. almostzero .and. cimask(mm-i,jml-j) .ge. 1) then
                     nmed = nmed + 1
                     fun(nmed) = fclean(mm,jml)
                   else
                     ldone = .true.
                   endif
                 enddo
                 do nn=max(j-l,1),min(j+l-1,ny),1
                   iml = max ( i - l, 1 )
                   if (mask(iml,nn) .gt. almostzero .and. cimask(iml-i,nn-j) .ge. 1) then
                     nmed = nmed + 1
                     fun(nmed) = fclean(iml,nn)
                   else
                     ldone = .true.
                   endif
                 enddo
                 do mm=max(i-l,1),min(i+l-1,nx),1
                   jpl = min ( j + l, ny )
                   if (mask(mm,jpl) .gt. almostzero .and. cimask(mm-i,jpl-j) .ge. 1) then
                     nmed = nmed + 1
                     fun(nmed) = fclean(mm,jpl)
                   else
                     ldone = .true.
                   endif
                 enddo
                 if (ldone) exit
               enddo

               if (nmed .gt. 1) then
                 funmed(i,j) = selectk ( (nmed + 1) / 2, nmed, fun )
               endif
             endif
           enddo
         enddo

         do j=1,ny
           do i=1,nx
             fclean(i,j) = funmed(i,j)
           enddo
         enddo

         deallocate ( cimask )
       endif
       
       deallocate ( fwork1, fwork2, fwork3, fwork4, fclean1, fclean2, fclean3, fclean4, fun, funmed )

! Get the current date and time.

       call when ( lunix, ctime, cdate, ndate, 4 )

!!       if (cverbose .eq. '-verb2') write (iotty,'(a)') '   Writing masked interpolation '''//worker(1:iwork)//''''
!!
!!       call wfits ( cfitsversion, nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, fwork, dx, dy, object, crval1
!!     &            , crval2, worker(1:iwork), cdate, ctime, creator, beam, blank, crota1, crota2, cd11, cd12, cd21
!!     &            , cd22, equinox, bzero, bscale, wave, datamin, datamax, history )

       if (cverbose .eq. '-verb2') write (iotty,'(a)') '   Writing cleaned background '''//cleanback(1:iclean)//''''

       call wfits ( cfitsversion, nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, fclean, dx, dy, object, crval1
     &            , crval2, cleanback(1:iclean), cdate, ctime, creator, beam, blank, crota1, crota2, cd11, cd12
     &            , cd21, cd22, equinox, bzero, bscale, wave, datamin, datamax, history )

       deallocate ( original, fclean, fwork, mask )

       if (cverbose .eq. '-verb2') write (iotty,'(/a)') ' Done.'

!!       stop !<-- commented out because it would lead to run-time messages about denormalized values, when using gfortran.
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine rfits 
       
     &            ( nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname, iotty, iolog, creator
     &            , beam, funmin, funmax, blank, rot1, rot2, cd11, cd12, cd21, cd22, equinox, bzero, bscale, wave, datamn
     &            , datamx, history, cverbose )
!__________________________________________________________________________________________________________________________________
!
! GETSF • Multi-Scale Multi-Wavelength Source & Filament Extraction • Alexander Men'shchikov, DAp IRFU CEA Saclay
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend, anynull
       character*(*) ctype1, ctype2, creator, fname, bunit, object, history, cverbose
       character*3   cnan
       character*25  cdatetime
       character*80  comment
       integer       i, j, nx, ny, stat, blksize, group, naxis, pcount, gcount, bitpix, unit, iotty, iolog, maxdim, rw
     &             , nfound, blank, stat05, stat06, stat07, stat08, stat09, stat10, stat11, stat12, stat13, stat14, stat15, stat16
     &             , stat17, stat18, stat19, stat20, stat21, stat22, stat23, stat24, stat25, stat26, stat27, stat28, stat29, stat30
     &             , stat31, stat32, stat0
       parameter   ( maxdim = 2 )
       integer       naxes(maxdim)
       real*8        dx, dy, ra, dec, beam, funmin, funmax, xr, yr, rp1, rp2, rot1, rot2, epoch, equinox, bzero, bscale, wave
     &             , dxdeg, dydeg, cd11, cd12, cd21, cd22, almostzero, as2deg, datamx, datamn, sign1, arg1, arg2, argm
     &             , sinrota, cosrota, pi, angle
       real*8        fun(nx,ny)
       parameter   ( pi = 3.14159265358979d0, as2deg = 1.0d0 / 3600.0d0, almostzero = 1.0d-30 )
       external      ftopen, ftgkne, ftghpr, ftgkys, ftgkyd, ftgkye, ftg2de, ftclos, printerr, ftgkyj, ftg2dd
!__________________________________________________________________________________________________________________________________
!
! Initialize variables.

       stat = 0
       unit   = 20
       group  = 1
       rw     = 0
       stat05 = 0
       stat06 = 0
       stat07 = 0
       stat08 = 0
       stat09 = 0
       stat10 = 0
       stat11 = 0
       stat12 = 0
       stat13 = 0
       stat14 = 0
       stat15 = 0
       stat16 = 0
       stat17 = 0
       stat18 = 0
       stat19 = 0
       stat20 = 0
       stat21 = 0
       stat22 = 0
       stat23 = 0
       stat24 = 0
       stat25 = 0
       stat26 = 0
       stat27 = 0
       stat28 = 0
       stat29 = 0
       stat30 = 0
       stat31 = 0
       stat32 = 0

! Open the FITS file.

       call ftopen ( unit, fname, rw, blksize, stat )

       if (stat .ne. 0) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         stop 99
       endif

! Determine the size of the image.

       call ftgkne ( unit, 'NAXIS', 1, 2, naxes, nfound, stat )

! Check that it found both NAXIS1 and NAXIS2 keywords.

       if (nfound .ne. 2) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         stop 99
       endif

! Get the main primary header keywords which define the array structure.

       call ftghpr ( unit, maxdim, simple, bitpix, naxis, naxes, pcount, gcount, extend, stat )

! Read all parameters.

       call ftgkys ( unit, 'CREATOR', creator, comment, stat )
       
! I don't understant why blank is assigned to my array values (if blank is not zero) - I have no undefined values in arrays. Maybe
! there is a bug in the version of FITSIO I use. For now I will ignore the blank value when reading FITS files.

       blank = 0

       datamx = 0.0d0
       datamn = 0.0d0
       bunit = ' '
       ctype1 = ' '
       ctype2 = ' '
       rp1 = 0.0d0
       rp2 = 0.0d0
       rot1 = 0.0d0
       rot2 = 0.0d0
       cd11 = 0.0d0
       cd12 = 0.0d0
       cd21 = 0.0d0
       cd22 = 0.0d0
       xr = 0.0d0
       yr = 0.0d0
       dx = 0.0d0
       dy = 0.0d0
       object = ' '
       ra = 0.0d0
       dec = 0.0d0
       epoch = 0.0d0
       equinox = 0.0d0
       beam = 0.0d0
       bzero = 0.0d0
       bscale = 1.0d0
       wave = 1.0d2
       history = ' '

       call ftgkys ( unit, 'DATE', cdatetime , comment, stat05 )
       call ftgkyd ( unit, 'DATAMAX', datamx , comment, stat06 )
       call ftgkyd ( unit, 'DATAMIN', datamn , comment, stat07 )
       call ftgkyj ( unit, 'BLANK'  , blank  , comment, stat08 )
       call ftgkys ( unit, 'BUNIT'  , bunit  , comment, stat09 )
       call ftgkyd ( unit, 'BZERO'  , bzero  , comment, stat10 )
       call ftgkyd ( unit, 'BSCALE' , bscale , comment, stat11 )
       call ftgkys ( unit, 'CTYPE1' , ctype1 , comment, stat12 )
       call ftgkys ( unit, 'CTYPE2' , ctype2 , comment, stat13 )
       call ftgkyd ( unit, 'CRPIX1' , rp1    , comment, stat14 )
       call ftgkyd ( unit, 'CRPIX2' , rp2    , comment, stat15 )
       call ftgkyd ( unit, 'CROTA1' , rot1   , comment, stat16 )
       call ftgkyd ( unit, 'CROTA2' , rot2   , comment, stat17 )
       call ftgkyd ( unit, 'CD1_1'  , cd11   , comment, stat18 )
       call ftgkyd ( unit, 'CD1_2'  , cd12   , comment, stat19 )
       call ftgkyd ( unit, 'CD2_1'  , cd21   , comment, stat20 )
       call ftgkyd ( unit, 'CD2_2'  , cd22   , comment, stat21 )
       call ftgkyd ( unit, 'CRVAL1' , xr     , comment, stat22 )
       call ftgkyd ( unit, 'CRVAL2' , yr     , comment, stat23 )
       call ftgkyd ( unit, 'CDELT1' , dxdeg  , comment, stat24 )
       call ftgkyd ( unit, 'CDELT2' , dydeg  , comment, stat25 )
       call ftgkys ( unit, 'OBJECT' , object , comment, stat26 )
       call ftgkyd ( unit, 'RA'     , ra     , comment, stat27 )
       call ftgkyd ( unit, 'DEC'    , dec    , comment, stat28 )
       call ftgkyd ( unit, 'EPOCH'  , epoch  , comment, stat29 )
       call ftgkyd ( unit, 'EQUINOX', equinox, comment, stat30 )
       call ftgkyd ( unit, 'WAVE'   , wave   , comment, stat31 )
       call ftgkyd ( unit, 'BEAM'   , beam   , comment, stat )
       call ftgkys ( unit, 'HISTORY', history, history, stat32 )

       if (stat .ne. 0) then
         beam = 0.0d0
         stat = 0
       endif

       stat0 = stat05 + stat06 + stat07 + stat09 + stat10 + stat11 + stat12 + stat13 + stat14 + stat15 + stat16 + stat17 + stat18 
     &       + stat19 + stat20 + stat21 + stat22 + stat23 + stat24 + stat25 + stat26 + stat27 + stat28 + stat30 + stat31

       if (cverbose .ne. '-verb0') then
         if (stat0  .gt. 0) write (iotty,'()')
         if (stat05 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATE absent or error'
         if (stat06 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMAX absent or error'
         if (stat07 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMIN absent or error'
         if (stat09 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BUNIT absent or error'
         if (stat10 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BZERO absent or error'
         if (stat11 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BSCALE absent or error'
         if (stat12 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE1 absent or error'
         if (stat13 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE2 absent or error'
         if (stat14 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX1 absent or error'
         if (stat15 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX2 absent or error'
         if (stat16 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA1 absent or error'
         if (stat17 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA2 absent or error'
         if (stat18 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_1 absent or error'
         if (stat19 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_2 absent or error'
         if (stat20 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_1 absent or error'
         if (stat21 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_2 absent or error'
         if (stat22 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL1 absent or error'
         if (stat23 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL2 absent or error'
         if (stat24 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT1 absent or error'
         if (stat25 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT2 absent or error'
         if (stat26 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword OBJECT absent or error'
         if (stat27 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword RA absent or error'
         if (stat28 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DEC absent or error'
         if (stat30 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword EQUINOX absent or error'
         if (stat31 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword WAVE absent or error'
       endif
       if (stat26 .ne. 0 .or. object(1:10) .eq. '          ') object = '<Name>'

! Read the FITS file into the 2D array.

       call ftg2dd ( unit, group, dble ( blank ), naxes(1), naxes(1), naxes(2), fun, anynull, stat )

       if (stat .ne. 0) write (iotty,'(a)') ' Error reading FITS file'

! Close the file and free the unit number.

       call ftclos ( unit, stat )

! Check for any error, and if so print out error messages.

       if (stat .gt. 0) call printerr ( stat )

       if (stat17 .ne. 0) then
         rot2 = 0.0d0
       endif
       rot1 = rot2

       if (stat29 .eq. 0 .and. stat30 .ne. 0) then
         equinox = epoch
       endif

! CDELT* is absent or error.

       if (stat24 .ne. 0 .or. stat25 .ne. 0) then

! CD1* and CD2* are defined correctly: determine CDELT* and ROTA*.

         if (stat18 .eq. 0 .and. stat19 .eq. 0 .and. stat20 .eq. 0 .and. stat21 .eq. 0) then
           sign1 = sign ( 1.0d0, cd11 * cd22 - cd12 * cd21 )
           dxdeg = sign1 * sqrt ( cd11**2 + cd21**2 )
           dydeg = sqrt ( cd12**2 + cd22**2 )
           arg1 = sign1 * cd12 / cd22
           arg2 = -sign1 * cd21 / cd11
           argm = (arg1 + arg2) / 2.0d0
           if (abs ( arg1 - arg2 ) / argm .gt. 1.0d-10) then
             write (iotty,'(/a/)') '   RFITS: WARNING: Coordinate axes are not orthogonal'
           endif
           rot2 = atan ( argm )
           sinrota = sin ( rot2 )
           cosrota = cos ( rot2 )
           rot2 = rot2 * 180.0d0 / pi
           if (cosrota .ge. 0.0d0 .and. sinrota .ge. 0.0d0) then
             rot1 = rot2
           elseif (cosrota .lt. 0.0d0 .and. sinrota .ge. 0.0d0) then
             rot1 = 180.0d0 - rot2
           elseif (cosrota .lt. 0.0d0 .and. sinrota .lt. 0.0d0) then
             rot1 = 180.0d0 + rot2
           elseif (cosrota .ge. 0.0d0 .and. sinrota .lt. 0.0d0) then
             rot1 = - rot2
           endif
           rot2 = rot1
         endif
       endif

! CD1* and CD2* are absent or error.

       if (stat18 .ne. 0 .or. stat19 .ne. 0 .or. stat20 .ne. 0 .or. stat21 .ne. 0) then

! CDELT* are defined correctly: determine CD1* and CD2*.

         if (stat24 .eq. 0 .and. stat25 .eq. 0) then
           angle = rot2 * pi / 180.0d0
           cd11 = dxdeg * cos ( angle )
           cd12 = abs ( dydeg ) * sign ( 1.0d0, dxdeg ) * sin ( angle )
           cd21 = - abs ( dxdeg ) * sign ( 1.0d0, dydeg ) * sin ( angle )
           cd22 = dydeg * cos ( angle )
           if (abs ( cd12 ) .lt. almostzero) cd12 = 0.0d0
           if (abs ( cd21 ) .lt. almostzero) cd21 = 0.0d0
         endif
       endif

       dx = abs ( dxdeg ) / as2deg
       dy = abs ( dydeg ) / as2deg

! Print out the min and max values.

       nx = naxes(1)
       ny = naxes(2)
       funmin = 1.0d30
       funmax =-1.0d30
       cnan = 'NAN'
       do j=1,ny
         do i=1,nx
           if (.not.isnan ( fun(i,j) )) then  
             funmin = min ( funmin, fun(i,j) )
             funmax = max ( funmax, fun(i,j) )
           endif  
           if (int ( fun(i,j) ) .eq. blank .and. blank .ne. 0) then  
             read (cnan,*) fun(i,j)
           endif
         enddo
       enddo

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine wfits 
       
     &            ( cfitsversion, nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname, cdate, ctime
     &            , creator, beam, blank, rot1, rot2, cd11, cd12, cd21, cd22, equinox, bzero, bscale, wave, datamn, datamx
     &            , history )
!__________________________________________________________________________________________________________________________________
!
! GETSF • Multi-Scale Multi-Wavelength Source & Filament Extraction • Alexander Men'shchikov, DAp IRFU CEA Saclay
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend
       character*(*) fname, ctype1, ctype2, object, creator, ctime, cdate, bunit, cfitsversion, history
       character*19  cdatetime
       integer       i, j, nx, ny, status, blocksze, group, naxis, pcount, gcount, bitpix, unit, naxes(2), blank
       real*8        xr, yr, dx, dy, rp1, rp2, ra, dec, beam, bzero, bscale, as2deg, rot1, rot2, equinox, datamn
     &             , datamx, dxdeg, dydeg, wave, cd11, cd12, cd21, cd22, almostzero, pi
       real*8        fun(nx,ny)
       parameter   ( pi = 3.14159265358979d0, as2deg = 1.0d0 / 3600.0d0, almostzero = 1.0d-30 )
       external      fdelete, ftgiou, ftinit, ftphpr, ftpkys, ftpcom, ftpkyd, ftpkyj, ftphis, ftp2dd, ftclos, ftfiou, printerr
!__________________________________________________________________________________________________________________________________
!
! Create a FITS primary array containing a 2-D image. Initialize parameters about the FITS image.

       status   = 0
       blocksze = 1
       simple   = .true.
       extend   = .false.
       group    = 1
       naxis    = 2
       pcount   = 0
       gcount   = 1
       naxes(1) = nx
       naxes(2) = ny

! Compute offset and scaling factor so as to represent real numbers most efficiently and accurately with given 'bitpix'.

       datamn = 1.0d30
       datamx =-1.0d30
       do j=1,ny
         do i=1,nx
           if (.not.isnan ( fun(i,j) )) then
             if (fun(i,j) .lt. datamn ) datamn = fun(i,j)
             if (fun(i,j) .gt. datamx ) datamx = fun(i,j)
           endif
         enddo
       enddo

! When packing numbers using the formulas below, I got values of the function different in the 5-th digit from DATAMIN and
!  DATAMAX placed in the header when writing the data. From now on I will use bzero = 0 bscale = 1 and bitpix = -32, as these
! seem to cure the problem.

cc!      bitpix = 16
cc!      bzero  = 0.5d0 * (funmin + funmax)
cc!      bscale = dabs ( funmax - bzero ) / 32767.0d0

       bitpix = -32     !<-- single precision floating-point values
       bzero  = 0.0d0
       bscale = 1.0d0
       blank = blank

       cdatetime = cdate(7:10)//'-'//cdate(4:5)//'-'//cdate(1:2)//'T'//ctime

! Delete the file if it already exists.

       if (cfitsversion .eq. '5.030') then
         call fdelete ( fname, status )
       endif

! Get an unused Logical Unit Number to use to open the FITS file.

       call ftgiou ( unit, status )

! Create the new empty FITS file.

       if (cfitsversion .eq. '5.030') then
         call ftinit ( unit, fname, blocksze, status )
       else
         call ftinit ( unit, '!'//fname, blocksze, status )
       endif

! Write the required header keywords.

       call ftphpr ( unit, simple, bitpix, naxis, naxes, pcount, gcount, extend, status )

! Write an optional comment.

       call ftpkys ( unit, 'CREATOR', creator,'Alexander Men''shchikov, DAp IRFU CEA Saclay', status )
       call ftpcom ( unit,' ',status)
       call ftpcom ( unit,' ',status)
       call ftpkys ( unit, 'DATE'  , cdatetime, 'creation date and time', status )

! Write all parameters.

       dxdeg = - dx * as2deg
       dydeg = dy * as2deg

       call ftpkyd ( unit, 'BZERO'  , bzero , 13 , 'zero point in scaling equation', status )
       call ftpkyd ( unit, 'BSCALE' , bscale, 13 , 'linear factor in scaling equation' , status )
       call ftpkyd ( unit, 'DATAMAX', datamx, 13 , 'maximum data value', status )
       call ftpkyd ( unit, 'DATAMIN', datamn, 13 , 'minimum data value', status )
!!       if  (blank .gt. 0)
!!     & call ftpkyj ( unit, 'BLANK'  , blank      , 'value used for undefined array elements', status )
       call ftpkys ( unit, 'BUNIT'  , bunit      , 'physical units of the array values', status )
       call ftpkys ( unit, 'CTYPE1' , ctype1     , 'name of the coordinate axis', status )
       call ftpkys ( unit, 'CTYPE2' , ctype2     , 'name of the coordinate axis', status )
       call ftpkyd ( unit, 'CRPIX1' , rp1   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CRPIX2' , rp2   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CROTA1' , rot1  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CROTA2' , rot2  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CD1_1'  , cd11  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD1_2'  , cd12  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_1'  , cd21  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_2'  , cd22  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CRVAL1' , xr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CRVAL2' , yr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CDELT1' , dxdeg , 13 , 'coordinate increment along axis', status )
       call ftpkyd ( unit, 'CDELT2' , dydeg , 13 , 'coordinate increment along axis', status )
       call ftpkys ( unit, 'OBJECT' , object     , 'object identifier', status )
       call ftpkyd ( unit, 'RA'     , ra     , 13, 'right ascension', status )
       call ftpkyd ( unit, 'DEC'    , dec    , 13, 'declination', status )
       call ftpkyd ( unit, 'EQUINOX', equinox, 13, 'equinox', status )
       call ftpkyd ( unit, 'WAVE'   , wave   , 13, 'wavelength (microns)', status )
       call ftphis ( unit, history, status )

       if (beam .ne. 0.0d0) call ftpkyd ( unit, 'BEAM', beam, 13, 'convolved with the beam (arcsec)', status )

! Write the array to the FITS file.

       call ftp2dd ( unit, group, nx, nx, ny, fun, status )

! Close the file.

       call ftclos ( unit, status )

! Free the unit number.

       call ftfiou ( unit, status )

! Check for any error, and if so print out error messages

       if (status .gt. 0) call printerr ( status )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||