
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       program ellipses
!__________________________________________________________________________________________________________________________________
!
! Draw extraction ellipses on top of original FITS images.
!
! GETSF • Multi-Scale Multi-Wavelength Source & Filament Extraction • Alexander Men'shchikov, DAp IRFU CEA Saclay  
!__________________________________________________________________________________________________________________________________
!
       implicit      none

       logical       lfname1, lfname2, lunix, lfootprints, lonlywhite, inellipse
                                            
       character*2   cnwave
       character*3   dot
       character*5   cattype
       character*6   cfitsversion, cverbose
       character*7   clibname, cx, cy, regcol
       character*8   ctime, cfpptime, regcolor
       character*10  cdate
       character*21  compda
       character*80  object, creator, ctype1, ctype2, bunit, exmeth, history
       character*500 arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, argA, argB, argC, argD, argE, argF, argG, filename
     &             , outname, excat
       character*15000 cline
       character*15000, allocatable :: headline(:)

       integer       i, j, l, k, isp1, fnlen1, ion, ndate, iotty, nx, ny, firstb, lastc, lasts, lhl, irc, lb, iexm, iexc, ia1, ia2
     &             , ia3, ia4, ia5, ia6, ia7, ia8, ia9, iaA, iaB, iaC, iaD, iaE, iaF, iaG, blank, nhead, nextr, nxe, nye, icl
     &             , ncolumnx, ncolumny, ncolumna, ncolumnb, ncolumnt, ncolumnfp, ncolumnft, ncolumnr, ncolumnfpe, ncolumnfte, icol
     &             , no, m, nxco, nyco, nr, l1, l2, m1, m2, nhcols, ncolumng, ncolumnfm, flagmon, ncx, ncy, naverage
     &             , ncolumnfa, ncolumnfb, ncolumngm, nplot, ncolumnxas, ncolumnyas

       parameter   ( iotty = 6 )

       real          fitsvers
       real*8        dx, dy, wave, crpix1, crpix2, crval1, crval2, ra, dec, beam, funmin, funmax, goodmin, elongmax
     &             , ellvalfact, crota1, crota2, equinox, bzero, bscale, datamin, datamax, x0o, y0o, majdiam, mindiam, angle, radx
     &             , rady, obsbeam, thickness, sigmono, sigmin, fxpbest, fxperro, fxtbest, fxterro, major2, minor1, minor2
     &             , semimajor2, snr1min, snr2min, semiminor2, dbll, dblm, deltax, deltay, sizefact, major1, goodness, goodmono
     &             , cd11, cd12, cd21, cd22, deltaxdot, deltaydot, meanradius, meansqradius, fmajdiam, fmindiam

       real*8, allocatable :: funa(:,:), funb(:,:)

       parameter   ( dot = '•' )

       external      firstb, lastc, rfits, wfits, when, lastbs, getnumpix, osystem, ftvers, lasts, getfitshead, skipcomm
     &             , linescolumns, inellipse, ds9region
!__________________________________________________________________________________________________________________________________
!
! Determine operating system

       call osystem ( lunix )
       call ftvers ( fitsvers )
       lb = 1
       write (cfitsversion,'(f6.3)') fitsvers
       if (cfitsversion .eq. ' 5.030') then
         lb = 2
       endif
       clibname = 'CFITSIO'
       cfpptime = __TIME__
       cfpptime = cfpptime(1:5)
       compda = __DATE__//' '//cfpptime

! Get the current date and time.

       call when ( lunix, ctime, cdate, ndate, 4 )

! Get command line parameters (file names).

       call getarg (  1, arg1 )
       call getarg (  2, arg2 )
       call getarg (  3, arg3 )
       call getarg (  4, arg4 )
       call getarg (  5, arg5 )
       call getarg (  6, arg6 )
       call getarg (  7, arg7 )
       call getarg (  8, arg8 )
       call getarg (  9, arg9 )
       call getarg ( 10, argA )
       call getarg ( 11, argB )
       call getarg ( 12, argC )
       call getarg ( 13, argD )
       call getarg ( 14, argE )
       call getarg ( 15, argF )
       call getarg ( 16, argG )

       ia1 = lastc ( arg1 )
       ia2 = lastc ( arg2 )
       ia3 = lastc ( arg3 )
       ia4 = lastc ( arg4 )
       ia5 = lastc ( arg5 )
       ia6 = lastc ( arg6 )
       ia7 = lastc ( arg7 )
       ia8 = lastc ( arg8 )
       ia9 = lastc ( arg9 )
       iaA = lastc ( argA )
       iaB = lastc ( argB )
       iaC = lastc ( argC )
       iaD = lastc ( argD )
       iaE = lastc ( argE )
       iaF = lastc ( argF )
       iaG = lastc ( argG )
       
       cverbose = '-verb2'
       if (iaG .gt. 0 .and. argG(1:iaG-1) .eq. '-verb') cverbose = argG(1:iaG)
       if (iaF .gt. 0 .and. argF(1:iaF-1) .eq. '-verb') cverbose = argF(1:iaF)
       if (iaE .gt. 0 .and. argE(1:iaE-1) .eq. '-verb') cverbose = argE(1:iaE)
       if (iaD .gt. 0 .and. argD(1:iaD-1) .eq. '-verb') cverbose = argD(1:iaD)
       if (iaC .gt. 0 .and. argC(1:iaC-1) .eq. '-verb') cverbose = argC(1:iaC)
       if (iaB .gt. 0 .and. argB(1:iaB-1) .eq. '-verb') cverbose = argB(1:iaB)
       if (iaA .gt. 0 .and. argA(1:iaA-1) .eq. '-verb') cverbose = argA(1:iaA)
       
       if (cverbose .eq. '-verb2') then
         write (iotty,'( )')
         if (arg1(1:ia1) .eq. ':') write (iotty,'(a$)') ' '
         write (iotty,'(a$)') ' ELLIPSES '//dot//' Produce Ellipses of Extracted Sources '//dot//' '//compda
       endif
       if (arg1(1:ia1) .eq. ':') stop
       if (cverbose .eq. '-verb2') write (iotty,'()')
       if (cverbose .eq. '-verb2') write (iotty,'( a)') ' Alexander Men’shchikov, DAp IRFU CEA Saclay, France.'
       if (cverbose .eq. '-verb2') write (iotty,'( a)') ' Using '//clibname(lb:7)//' library version'//cfitsversion
     &                                                //' by William D Pence.'
       goodmin = 1.0d0
       sigmin = 1.0d0
       snr1min = 2.0d0
       snr2min = 2.0d0
       elongmax = 2.0d0
       sizefact = 1.0d0
       ellvalfact = 1.1d0
       thickness = 3.0d0

       if (ia6 .gt. 0) then
         exmeth = arg1(1:ia1)
         iexm = ia1
         cnwave = '  '
         read (arg2(1:ia2),'(a)',err=1) cnwave
         if (cnwave(2:2) .eq. ' ') cnwave = '0'//cnwave(1:1)
         if (arg2(1:ia2) .eq. '0') cnwave = ' '
         goto 2
  1      continue
           write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble getting CNWAVE.'
           stop 99
  2      continue
         read (arg3(1:ia3),*,err=3) obsbeam
         goto 4
  3      continue
           write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble getting OBSBEAM.'
           stop 99
  4      continue
         read (arg4(1:ia4),*,err=5) goodmin
         goto 6
  5      continue
           write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble getting GOODMIN.'
           stop 99
  6      continue
         read (arg5(1:ia5),*,err=7) sigmin
         goto 8
  7      continue
           write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble getting SIGMIN.'
           stop 99
  8      continue
         read (arg6(1:ia6),*,err=9) snr1min
         goto 10
  9      continue
           write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble getting SNR1MIN.'
           stop 99
 10      continue
         read (arg7(1:ia7),*,err=11) snr2min
         goto 12
 11      continue
           write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble getting SNR2MIN.'
           stop 99
 12      continue
         excat = arg8(1:ia8)
         filename = arg9(1:ia9)
         if (iaA .gt. 0 .and. argA(1:2) .ne. '-o') then
           read (argA,*,err=13) sizefact
           goto 14
 13        continue
             write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble getting SIZEFACT.'
             stop 99
 14        continue
         endif
         lonlywhite = .false.
         if (iaB .gt. 0 .and. argB(1:2) .ne. '-o') then
           read (argB,*,err=15) ellvalfact
           if (argB(1:1) .eq. '+') then
             lonlywhite = .true.
           endif
           goto 16
 15        continue
             write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble getting ELLVALFACT.'
             stop 99
 16        continue
         endif
         lfootprints = .false.
         if (iaC .gt. 0 .and. argC(1:2) .ne. '-o') then
           read (argC,*,err=17) thickness
           if (argC(1:1) .eq. '+') then
             lfootprints = .true.
           endif
           if (argC(1:1) .eq. '-') then
             thickness = -2.0d0
           endif
           goto 18
 17        continue
             write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble getting THICKNESS.'
             stop 99
 18        continue
         endif
         regcol = 'none'
         regcolor = 'none'
         if (iaD .gt. 0 .and. argD(1:2) .ne. '-o') then
           read (argD,*,err=19) regcolor
           goto 20
 19        continue
             write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble getting REGCOLOR.'
             stop 99
 20        continue
         endif
         if (regcolor(1:3) .eq. 'gre' .or. regcolor(1:5) .eq. 'green'   ) regcol = '#3ce04b'  !<- 'green'
         if (regcolor(1:3) .eq. 'blu' .or. regcolor(1:4) .eq. 'blue'    ) regcol = '#7383ff'  !<- 'blue'
         if (regcolor(1:3) .eq. 'ora' .or. regcolor(1:6) .eq. 'orange'  ) regcol = '#ff9241'  !<- 'orange'
         if (regcolor(1:3) .eq. 'pur' .or. regcolor(1:6) .eq. 'purple'  ) regcol = '#a14eff'  !<- 'purple'
         if (regcolor(1:3) .eq. 'cya' .or. regcolor(1:4) .eq. 'cyan'    ) regcol = '#66ffff'  !<- 'cyan'
         if (regcolor(1:3) .eq. 'mag' .or. regcolor(1:7) .eq. 'magenta' ) regcol = '#bcf60c'  !<- 'magenta'
         if (regcolor(1:3) .eq. 'lim' .or. regcolor(1:4) .eq. 'lime'    ) regcol = '#7383ff'  !<- 'lime'
         if (regcolor(1:3) .eq. 'pin' .or. regcolor(1:4) .eq. 'pink'    ) regcol = '#ffb5b5'  !<- 'pink'
         if (regcolor(1:3) .eq. 'tea' .or. regcolor(1:4) .eq. 'teal'    ) regcol = '#00a0a0'  !<- 'teal'
         if (regcolor(1:3) .eq. 'lav' .or. regcolor(1:8) .eq. 'lavender') regcol = '#e6beff'  !<- 'lavender'
         if (regcolor(1:3) .eq. 'bro' .or. regcolor(1:5) .eq. 'brown'   ) regcol = '#ba7334'  !<- 'brown'
         if (regcolor(1:3) .eq. 'bei' .or. regcolor(1:5) .eq. 'beige'   ) regcol = '#f2f2ca'  !<- 'beige'
         if (regcolor(1:3) .eq. 'mar' .or. regcolor(1:6) .eq. 'maroon'  ) regcol = '#aa0000'  !<- 'maroon'
         if (regcolor(1:3) .eq. 'min' .or. regcolor(1:4) .eq. 'mint'    ) regcol = '#aaffc3'  !<- 'mint'
         if (regcolor(1:3) .eq. 'oli' .or. regcolor(1:5) .eq. 'olive'   ) regcol = '#a5b500'  !<- 'olive'
         if (regcolor(1:3) .eq. 'apr' .or. regcolor(1:7) .eq. 'apricot' ) regcol = '#ffd8b1'  !<- 'apricot'
         if (regcolor(1:3) .eq. 'gra' .or. regcolor(1:4) .eq. 'gray'    ) regcol = '#909090'  !<- 'gray'
         if (regcolor(1:3) .eq. 'red' .or. regcolor(1:3) .eq. 'red'     ) regcol = '#f0194b'  !<- 'red'
         if (regcolor(1:3) .eq. 'yel' .or. regcolor(1:6) .eq. 'yellow'  ) regcol = '#fff019'  !<- 'yellow'
         if (regcolor(1:3) .eq. 'whi' .or. regcolor(1:5) .eq. 'white'   ) regcol = '#ffffff'  !<- 'white'
         if (regcolor(1:3) .eq. 'bla' .or. regcolor(1:5) .eq. 'black'   ) regcol = '#000000'  !<- 'black'
         icol = lastc ( regcol )
       else
         write (iotty,'( )')
         write (iotty,'(a)') ' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ USAGE ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~'
         write (iotty,'( )')
         write (iotty,'(a)') ' ellipses <exmeth> <cnwave> <obsbeam> <goodmin> <sigmin> <snr1min> <snr2min>'
     &                     //' <excat> <imagein> [<sizefact> [+]<ellvalfact> [+|-]<thickness> regcolor]'
     &                     //' [-o <imageout>] [-verb{0|1|2}]'
         write (iotty,'( )')   
         write (iotty,'(a)') ' This utility overplots the half-maximum or footprint ellipses in a FITS image'
         write (iotty,'(a)') ' <imagein>, creating <imageout> as an output FITS image and also producing DS9'
         write (iotty,'(a)') ' region files containing the ellipses that can be ooverlaid on the image.'
         write (iotty,'( )')
         write (iotty,'(a)') ' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PARAMETERS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~'
         write (iotty,'( )')   
         write (iotty,'(a)') ' <exmeth> ........... extraction method (getsf or getsources)'
         write (iotty,'(a)') ' <cnwave> ........... number of wavelength (in a multi-band catalog)'
         write (iotty,'(a)') ' <obsbeam> .......... observational beam (arcsec)'
         write (iotty,'(a)') ' <goodmin> .......... minimum monochromatic goodness (default: 1)'
         write (iotty,'(a)') ' <sigmin> ........... minimum monochromatic significance (default: 1)'
         write (iotty,'(a)') ' <snr1min> .......... minimum S/N (peak) (default: 2)'
         write (iotty,'(a)') ' <snr2min> .......... minimum S/N (flux) (default: 2)'
         write (iotty,'(a)') ' <excat> ............ source extraction catalog (GETSF|GETSOURCES-formatted)'
         write (iotty,'(a)') ' <imagein> .......... image for overplotting ellipses'
         write (iotty,'(a)') ' <sizefact> ......... scaling factor for ellipse sizes'
         write (iotty,'(a)') ' [+]<ellvalfact> .... scaling factor for ellipse intensity'
         write (iotty,'(a)') ' [+|-]<thickness> ... width of the ellipses ("+" plots footprint ellipses)'
         write (iotty,'(a)') ' <regcolor> ......... color of ellipses in DS9 region files'
         write (iotty,'(a)') ' <imageout> ......... output FITS image with ellipses'
         write (iotty,'(a)') ' -verb[1|2|3] ....... verbosity level'
         write (iotty,'( )')
         write (iotty,'(a)') ' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ NOTES ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~'
         write (iotty,'( )')
         write (iotty,'(a)') ' If <regcolor> is omitted on command line, DS9 region files are not created.'
         write (iotty,'(a)') ' If <regcolor> is specified, it must take one of the values: gre[en] blu[e]'
         write (iotty,'(a)') ' ora[nge] pur[ple] cya[n] mag[enta] lim[e] pin[k] tea[l] lav[ender] bro[wn]'
         write (iotty,'(a)') ' bei[ge] mar[oon] min[t] oli[ve] apr[icot] gra[y] red yel[low] whi[te] bla[ck]'
         write (iotty,'( )')
         write (iotty,'(a)') ' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ EXAMPLE ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~'
         write (iotty,'( )')
         write (iotty,'(a)') ' ellipses getsf 3 18.2 1 1 2 2 X.sw.sources.ok.cat X.sw.250.obs.bs 1 1.1 3 ora -o X.sw.250.obs.bs.ell'
         write (iotty,'( )')
         write (iotty,'(a)') ' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~'
         stop 99
       endif  

       iexc = lastc ( excat )
       fnlen1 = lastc ( filename )

! Check if the input file exists and open it.

       inquire ( file=excat(1:iexc), exist=lfname1 )
       
! Check if the input file exists and open it.

       inquire ( file=filename(1:fnlen1), exist=lfname2 )
       if (.not.lfname2 .and. filename(fnlen1-4:fnlen1) .ne. '.fits') then
         filename = filename(1:fnlen1)//'.fits'
         fnlen1 = lastc ( filename )
         inquire ( file=filename(1:fnlen1), exist=lfname2 )
       endif

       isp1 = lasts ( filename ) + 1

       outname = filename(isp1:fnlen1-5)//'.m.fits'
       if (argA(1:2) .eq. '-o') outname = argB
       if (argB(1:2) .eq. '-o') outname = argC
       if (argC(1:2) .eq. '-o') outname = argD
       if (argD(1:2) .eq. '-o') outname = argE
       if (argE(1:2) .eq. '-o') outname = argF
       ion = lastc ( outname )
       if (outname(max(1,ion-4):ion) .ne. '.fits') then
         outname = outname(1:ion)//'.fits'
         ion = lastc ( outname )
       endif

       if (cverbose .eq. '-verb1') write (iotty,'(/a)') ' ELLIPSES: '//arg1(1:ia1)//' '//arg2(1:ia2)//' '//arg3(1:ia3)
     &                                                //' '//arg4(1:ia4)//' '//excat(1:iexc)
               
       if (exmeth(1:iexm) .eq. 'sextractor' ) cattype = 'sextr'
       if (exmeth(1:iexm) .eq. 'getsf'      ) cattype = 'getsf'
       if (exmeth(1:iexm) .eq. 'clumpfind'  ) cattype = 'clump'
       if (exmeth(1:iexm) .eq. 'gaussclumps') cattype = 'gauss'
       if (exmeth(1:iexm) .eq. 'csar'       ) cattype = 'csar '
       if (exmeth(1:iexm) .eq. 'fellwalker' ) cattype = 'fellw'
       if (exmeth(1:iexm) .eq. 'reinhold'   ) cattype = 'reinh'
       if (exmeth(1:iexm) .eq. 'derivatives') cattype = 'deriv'
       if (exmeth(1:iexm) .eq. 'mre-gcl'    ) cattype = 'mregc'
       if (exmeth(1:iexm) .eq. 'truth'      ) cattype = 'truth'
              
       if (index ( excat(1:iexc), '.smf'   ) .gt. 0) cattype = 'smflx'
       if (index ( excat(1:iexc), '.final' ) .gt. 0) cattype = 'final'
       if (index ( excat(1:iexc), '.sem'   ) .gt. 0) cattype = 'sem'
       if (index ( excat(1:iexc), '.semif' ) .gt. 0) cattype = 'semif'
     
!!       if (cverbose .eq. '-verb2') then
!!         write (iotty,*)
!!         write (iotty,*) arg1(1:ia1)
!!         write (iotty,*) arg2(1:ia2)
!!         write (iotty,*) arg3(1:ia3)
!!         write (iotty,*) arg4(1:ia4)
!!         write (iotty,*) arg5(1:ia5)
!!         write (iotty,*) arg6(1:ia6)
!!         write (iotty,*) arg7(1:ia7)
!!         write (iotty,*) arg8(1:ia8)
!!         write (iotty,*) arg9(1:ia9)
!!         write (iotty,*) argA(1:iaA)
!!         write (iotty,*) argB(1:iaB)
!!         write (iotty,*) argC(1:iaC)
!!         write (iotty,*) argD(1:iaD)
!!         write (iotty,*) argE(1:iaE)
!!         write (iotty,*) argF(1:iaF)
!!         write (iotty,*) sizefact
!!         write (iotty,*) ellvalfact
!!         write (iotty,*) thickness
!!         write (iotty,*) regcolor
!!         write (iotty,*) cattype
!!         write (iotty,*) lfootprints
!!         write (iotty,*) cverbose
!!       endif
     
       if (.not.lfname1) then
         write (iotty,'(/a)') '   ELLIPSES: ERROR: File '''//excat(1:iexc)//''' not found.'
         stop 1
       endif
       if (.not.lfname2) then
         write (iotty,'(/a)') '   ELLIPSES: ERROR: File '''//filename(1:fnlen1-5)//''' not found.'
         stop 2
       endif

! Determine numbers of pixels in the FITS image.

       call getfitshead ( filename(1:fnlen1), nx, ny, dx, dy, bunit )

! Allocate the image arrays.
       
       allocate ( funa(nx,ny), funb(nx,ny), headline(1000), stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a)') '   ELLIPSES: ERROR: Trouble allocating memory (12).'
         stop 12
       endif

       write (cx,'(i7)') nx
       write (cy,'(i7)') ny
       ncx = int ( log10 ( dble ( nx ) ) ) + 1
       ncy = int ( log10 ( dble ( ny ) ) ) + 1

       if (cverbose .eq. '-verb2') write (iotty,'(/a)') '   Reading ('//cx(7-ncx+1:7)//' x '//cy(7-ncy+1:7)//') '''
     &                                                //filename(1:fnlen1)//''''
       
       call rfits ( nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, funa, dx, dy, object, ra, dec
     &            , filename(1:fnlen1), iotty, 0, creator, beam, funmin, funmax, blank, crota1, crota2, cd11, cd12
     &            , cd21, cd22, equinox, bzero, bscale, wave, datamin, datamax, history, cverbose )

       if (cverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7))') '   Minmax values in the image:', funmin, funmax

       do j=1,ny
         do i=1,nx
           funb(i,j) = funa(i,j)
         enddo
       enddo
       
       if (cverbose .eq. '-verb2') write (iotty,'(a)') '   Reading catalog '''//excat(1:iexc)//''''

       open ( 14, file=excat(1:iexc), status='old' )

       call skipcomm     ( 14, excat(1:iexc), nhead )
       call linescolumns ( 14, excat(1:iexc), nhead, nextr )

       rewind ( 14 )
       do k=1,nhead
         read (14,'(a)') headline(k)
         lhl = lastc ( headline(k) )
         if (lhl .gt. 5) nhcols = k
       enddo
       lhl = lastc ( headline(nhcols) )

       ncolumnx = 0
       ncolumny = 0
       ncolumnxas = 0
       ncolumnyas = 0
       ncolumng = 0
       ncolumngm = 0
       ncolumna = 0
       ncolumnb = 0
       ncolumnfa = 0
       ncolumnfb = 0
       ncolumnt = 0
       ncolumnr = 0
       ncolumnfm = 0
       ncolumnfp = 0
       ncolumnft = 0
       ncolumnfpe = 0
       ncolumnfte = 0

       if (cattype .eq. 'truth') then
         
         if (ncolumnx  .eq. 0) ncolumnx  = index ( headline(nhcols)(1:lhl), ' XCO_PIXEL'  )
         if (ncolumnx  .gt. 0 .and. ncolumnx .lt. 5) ncolumnx = 0        
         if (ncolumny  .eq. 0) ncolumny  = index ( headline(nhcols)(1:lhl), ' YCO_PIXEL'  )
         if (ncolumny  .gt. 0 .and. ncolumny .lt. 5) ncolumny = 0        
         if (ncolumna  .eq. 0) ncolumna  = index ( headline(nhcols)(1:lhl), ' AFWHM_ARCS' )
         if (ncolumnb  .eq. 0) ncolumnb  = index ( headline(nhcols)(1:lhl), ' BFWHM_ARCS' )
         if (ncolumnt  .eq. 0) ncolumnt  = index ( headline(nhcols)(1:lhl), ' FWHM_ARCS'  )
         if (ncolumnfp .eq. 0) ncolumnfp = index ( headline(nhcols)(1:lhl), ' PEAKFLUX'   )
         if (ncolumnft .eq. 0) ncolumnft = index ( headline(nhcols)(1:lhl), ' TOTALFLUX'  )
         if (ncolumnr  .eq. 0) ncolumnr  = index ( headline(nhcols)(1:lhl), ' SN_RATIO'   )
         
         if (ncolumnx .eq. 0 .or. ncolumny .eq. 0 .or. ncolumna .eq. 0 .or. ncolumnb .eq. 0 .or. 
     &       ncolumnt .eq. 0 .or. ncolumnfp .eq. 0 .or. ncolumnft .eq. 0 .or. ncolumnr .eq. 0) then
           
           write (iotty,'(/a)') '   ELLIPSES: ERROR: Some columns for the method TRUTH not found in:'
           write (iotty,'(i5,a)') nhcols, '"'//headline(nhcols)(1:lhl)//'"'
           stop 99
         endif
         ncolumnxas = ncolumnx
         ncolumnyas = ncolumny
       else
         if (ncolumnx .eq. 0) ncolumnx = index ( headline(nhcols)(1:lhl), ' XCO_P' )
         if (ncolumnx .gt. 0 .and. ncolumnx .lt. 5) ncolumnx = 0        
         if (ncolumny .eq. 0) ncolumny = index ( headline(nhcols)(1:lhl), ' YCO_P' )
         if (ncolumny .gt. 0 .and. ncolumny .lt. 5) ncolumny = 0        

         if (ncolumnxas .eq. 0) ncolumnxas = index ( headline(nhcols)(1:lhl), ' WCS_ACOOR' )
         if (ncolumnxas .eq. 0) ncolumnxas = ncolumnx
         if (ncolumnyas .eq. 0) ncolumnyas = index ( headline(nhcols)(1:lhl), ' WCS_DCOOR' )
         if (ncolumnyas .eq. 0) ncolumnyas = ncolumny

         if (ncolumng .eq. 0) ncolumng = index ( headline(nhcols)(1:lhl), '  GOOD' )
         if (ncolumnfm .eq. 0) ncolumnfm = index ( headline(nhcols)(1:lhl), ' FM'//cnwave )
         
         if (ncolumna .eq. 0) ncolumna = index ( headline(nhcols)(1:lhl), ' AFWHM'//cnwave )
         if (ncolumna .eq. 0) ncolumna = index ( headline(nhcols)(1:lhl), ' AFWH'//cnwave )
         if (ncolumnb .eq. 0) ncolumnb = index ( headline(nhcols)(1:lhl), ' BFWHM'//cnwave )
         if (ncolumnb .eq. 0) ncolumnb = index ( headline(nhcols)(1:lhl), ' BFWH'//cnwave )
         
         if (ncolumnfa .eq. 0) ncolumnfa = index ( headline(nhcols)(1:lhl), ' FOOA'//cnwave )
         if (ncolumnfb .eq. 0) ncolumnfb = index ( headline(nhcols)(1:lhl), ' FOOB'//cnwave )

         if (ncolumnt .eq. 0) ncolumnt = index ( headline(nhcols)(1:lhl), ' THETA'//cnwave )
         if (ncolumnt .eq. 0) ncolumnt = index ( headline(nhcols)(1:lhl), ' THEP'//cnwave )

         if (ncolumnfp  .eq. 0) ncolumnfp  = index ( headline(nhcols)(1:lhl), ' FXP_BST'//cnwave )
         if (ncolumnfp  .eq. 0) ncolumnfp  = index ( headline(nhcols)(1:lhl), ' FXP_BEST'//cnwave )
         if (ncolumnfpe .eq. 0) ncolumnfpe = index ( headline(nhcols)(1:lhl), ' FXP_ERR'//cnwave )
         if (ncolumnfpe .eq. 0) ncolumnfpe = index ( headline(nhcols)(1:lhl), ' FXP_ERRO'//cnwave )

         if (ncolumnft  .eq. 0) ncolumnft  = index ( headline(nhcols)(1:lhl), ' FXT_BST'//cnwave )
         if (ncolumnft  .eq. 0) ncolumnft  = index ( headline(nhcols)(1:lhl), ' FXT_BEST'//cnwave )
         if (ncolumnfte .eq. 0) ncolumnfte = index ( headline(nhcols)(1:lhl), ' FXT_ERR'//cnwave )
         if (ncolumnfte .eq. 0) ncolumnfte = index ( headline(nhcols)(1:lhl), ' FXT_ERRO'//cnwave )

         if (ncolumnr  .eq. 0) ncolumnr  = index ( headline(nhcols)(1:lhl), ' SIGNM'//cnwave )
         if (ncolumnr  .eq. 0) ncolumnr  = index ( headline(nhcols)(1:lhl), ' SIG_MONO'//cnwave )
         if (ncolumnr  .eq. 0) ncolumnr  = index ( headline(nhcols)(1:lhl), ' SN_RATIO'//cnwave )
         if (ncolumngm .eq. 0) ncolumngm = index ( headline(nhcols)(1:lhl), ' GOODM'//cnwave )
         if (ncolumngm .eq. 0) ncolumngm = index ( headline(nhcols)(1:lhl), ' GOOD' )
         
         if (ncolumnx  .eq. 0 .or. ncolumny  .eq. 0 .or. ncolumnxas .eq. 0 .or. ncolumnyas .eq. 0 .or. ncolumna .eq. 0 .or. 
     &       ncolumnb .eq. 0 .or. ncolumnt .eq. 0 .or. ncolumnfp .eq. 0 .or. ncolumnft .eq. 0 .or. ncolumnr .eq. 0 .or. 
     &       (cattype .eq. 'final' .and. ncolumng .eq. 0) .or. (cattype .eq. 'final' .and. ncolumnfm .eq. 0) .or. 
     &       (cattype .eq. 'final' .and. ncolumngm .eq. 0)) then
           
           write (iotty,'(/a)') '   ELLIPSES: ERROR: Some columns not found in:'
           write (iotty,'(i5,a)') nhcols, '"'//headline(nhcols)(1:lhl)//'"'
           write (iotty,'(13i5)') ncolumnx, ncolumny, ncolumnxas, ncolumnyas, ncolumna, ncolumnb, ncolumnt, ncolumnfp, ncolumnft
     &                          , ncolumnr, ncolumng, ncolumnfm, ncolumngm
           stop 99
         endif
       endif
       
       open ( 10, file=excat(1:iexc)//'.good', status='unknown' )
   
       do k=1,nhead
         lhl = lastc ( headline(k) )
         write (10,'(a)') headline(k)(1:lhl)
       enddo

       nplot = 0
       
       do k=1,nextr

         read (14,'(a)') cline
         icl = lastc ( cline )
         
         if (cline(1:1) .eq. '!' .or. cline(1:1) .eq. '#') goto 77
         
         read (cline,*) no
         read (cline(ncolumnx:),*) x0o
         read (cline(ncolumny:),*) y0o
         
         goodness = 1.0d99
         if (ncolumng .gt. 0) then
           read (cline(ncolumng:),*) goodness
         endif
         goodmono = 1.0d99
         if (ncolumngm .gt. 0) then
           read (cline(ncolumngm:),*) goodmono
         endif
         flagmon = 0
         if (ncolumnfm .gt. 0) then
           read (cline(ncolumnfm:),*) flagmon
         endif
         
         read (cline(ncolumna:), *) majdiam
         read (cline(ncolumnb:), *) mindiam
         read (cline(ncolumnt:), *) angle
         read (cline(ncolumnfp:),*) fxpbest
         read (cline(ncolumnft:),*) fxtbest
         read (cline(ncolumnr:), *) sigmono

         if (ncolumnfa .gt. 0) then
           read (cline(ncolumnfa:), *) fmajdiam
         else
           fmajdiam = 2.3d0 * majdiam
         endif
         if (ncolumnfb .gt. 0) then
           read (cline(ncolumnfb:), *) fmindiam
         else
           fmindiam = 2.3d0 * mindiam
         endif

         if (lfootprints) then
           read (cline(ncolumnfa:),*) majdiam
           read (cline(ncolumnfb:),*) mindiam
           fmajdiam = 2.3d0 * majdiam
           fmindiam = 2.3d0 * mindiam
         endif

         majdiam = majdiam * sizefact
         mindiam = mindiam * sizefact
         
         if (cattype .eq. 'gauss') angle = - angle - 90.0d0
                
         if (cattype .ne. 'truth') then
           read (cline(ncolumnfpe:),*) fxperro
           read (cline(ncolumnfte:),*) fxterro
         else
           fxperro = 0.0d0
           fxterro = 0.0d0
         endif
         
         if (thickness .gt. 0.0d0) then
           nxe = int ( max ( majdiam, mindiam ) / dx * 2.0d0 ) + 10 !<-- number of pixels in the ellipse window in the X-dir
           nye = int ( max ( majdiam, mindiam ) / dy * 2.0d0 ) + 10 !<-- number of pixels in the ellipse window in the Y-dir
         else
           nxe = int ( obsbeam / dx * 2.0d0 ) + 2
           nye = int ( obsbeam / dy * 2.0d0 ) + 2
         endif
         nxco = nint ( x0o )  !<-- ellipse center's X-coordinate expressed in pixels
         nyco = nint ( y0o )  !<-- ellipse center's Y-coordinate expressed in pixels

         radx = majdiam / 2.0d0  !<-- major radius of the current ellipse
         rady = mindiam / 2.0d0  !<-- minor radius of the current ellipse

         if (abs ( sigmono ) .ge. sigmin .and. 
     &       abs ( goodmono ) .ge. goodmin .and. 
     &       fxpbest .gt. snr1min * fxperro .and. 
     &       fxtbest .gt. snr2min * fxterro .and.
     &       fmajdiam .lt. elongmax * fmindiam .and.
     &       fmajdiam .gt. 1.15d0 * majdiam) then
     
           nplot = nplot + 1

           if (flagmon .eq. 0) write (10,'(a)') cline(1:icl)
           
! Plot the inner (black) side of the ellipse.
           
           deltax = dx * abs ( thickness )
           deltay = dy * abs ( thickness )
           deltaxdot = deltax / 2.0d0
           deltaydot = deltay / 2.0d0
           major1 = majdiam - deltax
           minor1 = mindiam - deltay
           major2 = majdiam
           minor2 = mindiam
           semimajor2 = major2 / 2.0d0
           semiminor2 = minor2 / 2.0d0
           
           nr = nint ( semimajor2 / min ( dx, dy ) ) + 2
           l1 = max ( nxco - nr, 1 )
           l2 = min ( nxco + nr, nx )
           m1 = max ( nyco - nr, 1 )
           m2 = min ( nyco + nr, ny )

           do m=m1,m2
             dblm = dble ( m )
             do l=l1,l2
               dbll = dble ( l )

               if (thickness .gt. 0.0d0) then
                 
                 if (mindiam / 2.0d0 .ge. deltax / 3.0d0) then
                   
                   if (inellipse ( dbll, dblm, dx, dy, x0o, y0o, major2, minor2, angle )) then
                     if (.not.inellipse ( dbll, dblm, dx, dy, x0o, y0o, major1, minor1, angle )) then
                       if (lonlywhite) then
                         funb(l,m) = ellvalfact
                       else
                         funb(l,m) = ellvalfact * funmin
                       endif
                     endif
                   endif
                 else
                   if (inellipse ( dbll, dblm, dx, dy, x0o, y0o, major2, minor2, angle )) then
                     if (lonlywhite) then
                       funb(l,m) = ellvalfact
                     else
                       funb(l,m) = ellvalfact * funmin
                     endif
                   endif
                 endif
               else
                 if (abs ( l - nxco ) * dx .le. deltaxdot .and. abs ( m - nyco ) * dy .le. deltaydot) then
                   if (lonlywhite) then
                     funb(l,m) = ellvalfact
                   else
                     funb(l,m) = ellvalfact * funmin
                   endif
                 endif
               endif
             enddo
           enddo

! Plot the outer (white) side of the ellipse.

           major1 = majdiam
           minor1 = mindiam
           major2 = majdiam + deltax
           minor2 = mindiam + deltay
           semimajor2 = major2 / 2.0d0
           semiminor2 = minor2 / 2.0d0
           
           nr = nint ( semimajor2 / min ( dx, dy ) ) + 2
           l1 = max ( nxco - nr, 1 )
           l2 = min ( nxco + nr, nx )
           m1 = max ( nyco - nr, 1 )
           m2 = min ( nyco + nr, ny )
           meanradius = 0.0d0
           meansqradius = 0.0d0
           naverage = 0
           
           do m=m1,m2
             dblm = dble ( m )
             do l=l1,l2
               dbll = dble ( l )
               
               if (thickness .gt. 0.0d0) then
                 
                 if (inellipse ( dbll, dblm, dx, dy, x0o, y0o, major2, minor2, angle )) then
                   if (.not.inellipse ( dbll, dblm, dx, dy, x0o, y0o, major1, minor1, angle )) then
                     if (lonlywhite) then
                       funb(l,m) = ellvalfact
                     else
                       funb(l,m) = ellvalfact * max ( funmax, 1.0d-10 )
                     endif
                     meanradius = meanradius + sqrt ( (dbll - x0o)**2 + (dblm - y0o)**2 ) * dx
                     meansqradius = meansqradius + ((dbll - x0o)**2 + (dblm - y0o)**2) * dx * dy
                     naverage = naverage + 1
                   endif
                 endif
               else
                 if ((abs ( l - nxco ) * dx .le. 3.0d0 * deltaxdot .and. abs ( m - nyco ) * dy .le. 3.0d0 * deltaydot) .and. 
     &               (abs ( l - nxco ) * dx .gt. deltaxdot .or. abs ( m - nyco ) * dy .gt. deltaydot)) then
                   if (lonlywhite) then
                     funb(l,m) = ellvalfact
                   else
                     funb(l,m) = ellvalfact * max ( funmax, 1.0d-10 )
                   endif
                 endif
               endif
             enddo
           enddo
           meanradius = meanradius / dble ( naverage )
           meansqradius = sqrt ( meansqradius / dble ( naverage ) )
         endif
  77     continue     
       enddo
       
       write (cx,'(i7)') nplot
       ncx = int ( log10 ( dble ( nplot ) ) ) + 1

       if (cverbose .eq. '-verb2') write (iotty,'(a)') '   Plotted ellipses of '//cx(7-ncx+1:7)//' selected sources'

       if (cverbose .eq. '-verb2') write (iotty,'(a)') '   Writing output image '''//outname(1:ion)//''''
      
       call wfits ( cfitsversion, nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, funb, dx, dy, object, crval1
     &            , crval2, outname(1:ion), cdate, ctime, creator, beam, blank, crota1, crota2, cd11, cd12, cd21, cd22, equinox
     &            , bzero, bscale, wave, datamin, datamax, history )
         
! Save DS9 region files with source ellipses.

       if (regcol(1:icol) .ne. 'none') then

         call ds9region ( 14, dx, dy, nhead, nhcols, lhl, excat(1:iexc), headline, nextr, cnwave, obsbeam, ncolumnx, ncolumny
     &                  , ncolumnxas, ncolumnyas, ncolumng, ncolumngm, ncolumnfm, ncolumna, ncolumnb, ncolumnt, ncolumnfp
     &                  , ncolumnft, ncolumnr, ncolumnfa, ncolumnfb, sizefact, ncolumnfpe, ncolumnfte, thickness, sigmin, goodmin
     &                  , snr1min, snr2min, elongmax, lfootprints, regcol(1:icol), 'phy', 'fk5' )
       endif

       close ( 10 )
       close ( 14 )

       deallocate ( funa, funb, headline )
      
       if (cverbose .eq. '-verb2') write (iotty,'(/a)') ' Done.'

!!       stop !<-- commented out because it would lead to run-time messages about denormalized values, when using gfortran.
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine rfits 
       
     &            ( nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname, iotty, iolog, creator
     &            , beam, funmin, funmax, blank, rot1, rot2, cd11, cd12, cd21, cd22, equinox, bzero, bscale, wave, datamn, datamx
     &            , history, cverbose )
!__________________________________________________________________________________________________________________________________
!
! GETSF • Multi-Scale Multi-Wavelength Source & Filament Extraction • Alexander Men'shchikov, DAp IRFU CEA Saclay
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend, anynull, myfile
       character*(*) ctype1, ctype2, creator, fname, bunit, object, history, cverbose
       character*3   cnan
       character*25  cdatetime
       character*80  comment
       integer       i, j, nx, ny, stat, blksize, group, naxis, pcount, gcount, bitpix, unit, iotty, iolog, maxdim
     &             , rw, nfound, blank, stat05, stat06, stat07, stat08, stat09, stat10, stat11, stat12, stat13, stat14, stat15
     &             , stat16, stat17, stat18, stat19, stat20, stat21, stat22, stat23, stat24, stat25, stat26, stat27, stat28, stat29
     &             , stat30, stat31, stat32, stat0
       parameter   ( maxdim = 2 )
       integer       naxes(maxdim)
       real*8        dx, dy, ra, dec, beam, funmin, funmax, xr, yr, rp1, rp2, rot1, rot2, epoch, equinox, bzero, bscale, wave
     &             , dxdeg, dydeg, cd11, cd12, cd21, cd22, almostzero, as2deg, datamx, datamn, sign1, arg1, arg2, argm
     &             , sinrota, cosrota, pi, angle
       real*8        fun(nx,ny)
       parameter   ( pi = 3.14159265358979d0, as2deg = 1.0d0 / 3600.0d0, almostzero = 1.0d-30 )
       external      ftopen, ftgkne, ftghpr, ftgkys, ftgkyd, ftgkye, ftg2de, ftclos, printerr, ftgkyj, ftg2dd
!__________________________________________________________________________________________________________________________________
!
! Initialize variables.

       stat = 0
       unit   = 20
       group  = 1
       rw     = 0
       stat05 = 0
       stat06 = 0
       stat07 = 0
       stat08 = 0
       stat09 = 0
       stat10 = 0
       stat11 = 0
       stat12 = 0
       stat13 = 0
       stat14 = 0
       stat15 = 0
       stat16 = 0
       stat17 = 0
       stat18 = 0
       stat19 = 0
       stat20 = 0
       stat21 = 0
       stat22 = 0
       stat23 = 0
       stat24 = 0
       stat25 = 0
       stat26 = 0
       stat27 = 0
       stat28 = 0
       stat29 = 0
       stat30 = 0
       stat31 = 0
       stat32 = 0

! Open the FITS file.

       call ftopen ( unit, fname, rw, blksize, stat )

       if (stat .ne. 0) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         stop 99
       endif

! Determine the size of the image.

       call ftgkne ( unit, 'NAXIS', 1, 2, naxes, nfound, stat )

! Check that it found both NAXIS1 and NAXIS2 keywords.

       if (nfound .ne. 2) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         stop 99
       endif

! Get the main primary header keywords which define the array structure.

       call ftghpr ( unit, maxdim, simple, bitpix, naxis, naxes, pcount, gcount, extend, stat )

! Read all parameters.

       call ftgkys ( unit, 'CREATOR', creator, comment, stat )
       
       if (stat .eq. 0 .and. (comment(1:21) .eq. 'Alexander Menshchikov' .or. comment(1:22) .eq. 'Alexander Men''shchikov')) then
         myfile = .true.
       else
         myfile = .false.
         stat = 0
         creator = 'ELLIPSES'
       endif

! I don't understant why blank is assigned to my array values (if blank is not zero) - I have no undefined values in arrays. Maybe
! there is a bug in the version of FITSIO I use. For now I will ignore the blank value when reading FITS files.

       blank = 0

       datamx = 0.0d0
       datamn = 0.0d0
       bunit = ' '
       ctype1 = ' '
       ctype2 = ' '
       rp1 = 0.0d0
       rp2 = 0.0d0
       rot1 = 0.0d0
       rot2 = 0.0d0
       cd11 = 0.0d0
       cd12 = 0.0d0
       cd21 = 0.0d0
       cd22 = 0.0d0
       xr = 0.0d0
       yr = 0.0d0
       dx = 0.0d0
       dy = 0.0d0
       object = ' '
       ra = 0.0d0
       dec = 0.0d0
       epoch = 0.0d0
       equinox = 0.0d0
       beam = 0.0d0
       bzero = 0.0d0
       bscale = 1.0d0
       wave = 1.0d2
       history = ' '

       call ftgkys ( unit, 'DATE', cdatetime , comment, stat05 )
       call ftgkyd ( unit, 'DATAMAX', datamx , comment, stat06 )
       call ftgkyd ( unit, 'DATAMIN', datamn , comment, stat07 )
       call ftgkyj ( unit, 'BLANK'  , blank  , comment, stat08 )
       call ftgkys ( unit, 'BUNIT'  , bunit  , comment, stat09 )
       call ftgkyd ( unit, 'BZERO'  , bzero  , comment, stat10 )
       call ftgkyd ( unit, 'BSCALE' , bscale , comment, stat11 )
       call ftgkys ( unit, 'CTYPE1' , ctype1 , comment, stat12 )
       call ftgkys ( unit, 'CTYPE2' , ctype2 , comment, stat13 )
       call ftgkyd ( unit, 'CRPIX1' , rp1    , comment, stat14 )
       call ftgkyd ( unit, 'CRPIX2' , rp2    , comment, stat15 )
       call ftgkyd ( unit, 'CROTA1' , rot1   , comment, stat16 )
       call ftgkyd ( unit, 'CROTA2' , rot2   , comment, stat17 )
       call ftgkyd ( unit, 'CD1_1'  , cd11   , comment, stat18 )
       call ftgkyd ( unit, 'CD1_2'  , cd12   , comment, stat19 )
       call ftgkyd ( unit, 'CD2_1'  , cd21   , comment, stat20 )
       call ftgkyd ( unit, 'CD2_2'  , cd22   , comment, stat21 )
       call ftgkyd ( unit, 'CRVAL1' , xr     , comment, stat22 )
       call ftgkyd ( unit, 'CRVAL2' , yr     , comment, stat23 )
       call ftgkyd ( unit, 'CDELT1' , dxdeg  , comment, stat24 )
       call ftgkyd ( unit, 'CDELT2' , dydeg  , comment, stat25 )
       call ftgkys ( unit, 'OBJECT' , object , comment, stat26 )
       call ftgkyd ( unit, 'RA'     , ra     , comment, stat27 )
       call ftgkyd ( unit, 'DEC'    , dec    , comment, stat28 )
       call ftgkyd ( unit, 'EPOCH'  , epoch  , comment, stat29 )
       call ftgkyd ( unit, 'EQUINOX', equinox, comment, stat30 )
       call ftgkyd ( unit, 'WAVE'   , wave   , comment, stat31 )
       call ftgkyd ( unit, 'BEAM'   , beam   , comment, stat )
       call ftgkys ( unit, 'HISTORY', history, history, stat32 )

       if (stat .ne. 0) then
         beam = 0.0d0
         stat = 0
       endif

       stat0 = stat05 + stat06 + stat07 + stat09 + stat10 + stat11 + stat12 + stat13 + stat14 + stat15 + stat16 + stat17 + stat18
     &       + stat19 + stat20 + stat21 + stat22 + stat23 + stat24 + stat25 + stat26 + stat27 + stat28 + stat30 + stat31

       if (cverbose .ne. '-verb0') then
         if (stat0  .gt. 0) write (iotty,'()')
         if (stat05 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATE absent or error'
         if (stat06 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMAX absent or error'
         if (stat07 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMIN absent or error'
         if (stat09 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BUNIT absent or error'
         if (stat10 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BZERO absent or error'
         if (stat11 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BSCALE absent or error'
         if (stat12 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE1 absent or error'
         if (stat13 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE2 absent or error'
         if (stat14 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX1 absent or error'
         if (stat15 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX2 absent or error'
         if (stat16 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA1 absent or error'
         if (stat17 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA2 absent or error'
         if (stat18 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_1 absent or error'
         if (stat19 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_2 absent or error'
         if (stat20 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_1 absent or error'
         if (stat21 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_2 absent or error'
         if (stat22 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL1 absent or error'
         if (stat23 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL2 absent or error'
         if (stat24 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT1 absent or error'
         if (stat25 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT2 absent or error'
         if (stat26 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword OBJECT absent or error'
         if (stat27 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword RA absent or error'
         if (stat28 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DEC absent or error'
         if (stat30 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword EQUINOX absent or error'
         if (stat31 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword WAVE absent or error'
       endif
       if (stat26 .ne. 0 .or. object(1:10) .eq. '          ') object = '<Name>'

! Read the FITS file into the 2D array.

       call ftg2dd ( unit, group, dble ( blank ), naxes(1), naxes(1), naxes(2), fun, anynull, stat )

       if (stat .ne. 0) write (iotty,'(a)') ' Error reading FITS file'

! Close the file and free the unit number.

       call ftclos ( unit, stat )

! Check for any error, and if so print out error messages.

       if (stat .gt. 0) call printerr ( stat )

       if (stat17 .ne. 0) then
         rot2 = 0.0d0
       endif
       rot1 = rot2

       if (stat29 .eq. 0 .and. stat30 .ne. 0) then
         equinox = epoch
       endif

! CDELT* is absent or error.

       if (stat24 .ne. 0 .or. stat25 .ne. 0) then

! CD1* and CD2* are defined correctly: determine CDELT* and ROTA*.

         if (stat18 .eq. 0 .and. stat19 .eq. 0 .and. stat20 .eq. 0 .and. stat21 .eq. 0) then
           sign1 = sign ( 1.0d0, cd11 * cd22 - cd12 * cd21 )
           dxdeg = sign1 * sqrt ( cd11**2 + cd21**2 )
           dydeg = sqrt ( cd12**2 + cd22**2 )
           arg1 = sign1 * cd12 / cd22
           arg2 = -sign1 * cd21 / cd11
           argm = (arg1 + arg2) / 2.0d0
           if (abs ( arg1 - arg2 ) / argm .gt. 1.0d-10) then
             write (iotty,'(/a/)') '   RFITS: WARNING: Coordinate axes are not orthogonal'
           endif
           rot2 = atan ( argm )
           sinrota = sin ( rot2 )
           cosrota = cos ( rot2 )
           rot2 = rot2 * 180.0d0 / pi
           if (cosrota .ge. 0.0d0 .and. sinrota .ge. 0.0d0) then
             rot1 = rot2
           elseif (cosrota .lt. 0.0d0 .and. sinrota .ge. 0.0d0) then
             rot1 = 180.0d0 - rot2
           elseif (cosrota .lt. 0.0d0 .and. sinrota .lt. 0.0d0) then
             rot1 = 180.0d0 + rot2
           elseif (cosrota .ge. 0.0d0 .and. sinrota .lt. 0.0d0) then
             rot1 = - rot2
           endif
           rot2 = rot1
         endif
       endif

! CD1* and CD2* are absent or error.

       if (stat18 .ne. 0 .or. stat19 .ne. 0 .or. stat20 .ne. 0 .or. stat21 .ne. 0) then

! CDELT* are defined correctly: determine CD1* and CD2*.

         if (stat24 .eq. 0 .and. stat25 .eq. 0) then
           angle = rot2 * pi / 180.0d0
           cd11 = dxdeg * cos ( angle )
           cd12 = abs ( dydeg ) * sign ( 1.0d0, dxdeg ) * sin ( angle )
           cd21 = - abs ( dxdeg ) * sign ( 1.0d0, dydeg ) * sin ( angle )
           cd22 = dydeg * cos ( angle )
           if (abs ( cd12 ) .lt. almostzero) cd12 = 0.0d0
           if (abs ( cd21 ) .lt. almostzero) cd21 = 0.0d0
         endif
       endif

       dx = abs ( dxdeg ) / as2deg
       dy = abs ( dydeg ) / as2deg

! Print out the min and max values.

       nx = naxes(1)
       ny = naxes(2)
       funmin = 1.0d30
       funmax =-1.0d30
       cnan = 'NAN'
       do j=1,ny
         do i=1,nx
           if (.not.isnan ( fun(i,j) )) then  
             funmin = min ( funmin, fun(i,j) )
             funmax = max ( funmax, fun(i,j) )
           endif  
           if (int ( fun(i,j) ) .eq. blank .and. blank .ne. 0) then  
             read (cnan,*) fun(i,j)
           endif
         enddo
       enddo

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine wfits 
       
     &            ( cfitsversion, nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname
     &            , cdate, ctime, creator, beam, blank, rot1, rot2, cd11, cd12, cd21, cd22, equinox, bzero, bscale
     &            , wave, datamn, datamx, history )
!__________________________________________________________________________________________________________________________________
!
! GETSF • Multi-Scale Multi-Wavelength Source & Filament Extraction • Alexander Men'shchikov, DAp IRFU CEA Saclay
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend
       character*(*) fname, ctype1, ctype2, object, creator, ctime, cdate, bunit, cfitsversion, history
       character*19  cdatetime
       integer       i, j, nx, ny, status, blocksze, group, naxis, pcount, gcount, bitpix, unit, naxes(2), blank
       real*8        xr, yr, dx, dy, rp1, rp2, ra, dec, beam, bzero, bscale, as2deg, rot1, rot2, equinox, datamn
     &             , datamx, dxdeg, dydeg, wave, cd11, cd12, cd21, cd22, almostzero, pi
       real*8        fun(nx,ny)
       parameter   ( pi = 3.14159265358979d0, as2deg = 1.0d0 / 3600.0d0, almostzero = 1.0d-30 )
       external      fdelete, ftgiou, ftinit, ftphpr, ftpkys, ftpcom, ftpkyd, ftpkyj, ftphis, ftp2dd, ftclos, ftfiou, printerr
!__________________________________________________________________________________________________________________________________
!
! Create a FITS primary array containing a 2-D image. Initialize parameters about the FITS image.

       status   = 0
       blocksze = 1
       simple   = .true.
       extend   = .false.
       group    = 1
       naxis    = 2
       pcount   = 0
       gcount   = 1
       naxes(1) = nx
       naxes(2) = ny

! Compute offset and scaling factor so as to represent real numbers most efficiently and accurately with given 'bitpix'.

       datamn = 1.0d30
       datamx =-1.0d30
       do j=1,ny
         do i=1,nx
           if (.not.isnan ( fun(i,j) )) then
             if (fun(i,j) .lt. datamn ) datamn = fun(i,j)
             if (fun(i,j) .gt. datamx ) datamx = fun(i,j)
           endif
         enddo
       enddo

! When packing numbers using the formulas below, I got values of the function different in the 5-th digit from DATAMIN and
!  DATAMAX placed in the header when writing the data. From now on I will use bzero = 0 bscale = 1 and bitpix = -32, as these
! seem to cure the problem.

cc!      bitpix = 16
cc!      bzero  = 0.5d0 * (funmin + funmax)
cc!      bscale = dabs ( funmax - bzero ) / 32767.0d0

       bitpix = -32     !<-- single precision floating-point values
       bzero  = 0.0d0
       bscale = 1.0d0
       blank = blank

       cdatetime = cdate(7:10)//'-'//cdate(4:5)//'-'//cdate(1:2)//'T'//ctime

! Delete the file if it already exists.

       if (cfitsversion .eq. '5.030') then
         call fdelete ( fname, status )
       endif

! Get an unused Logical Unit Number to use to open the FITS file.

       call ftgiou ( unit, status )

! Create the new empty FITS file.

       if (cfitsversion .eq. '5.030') then
         call ftinit ( unit, fname, blocksze, status )
       else
         call ftinit ( unit, '!'//fname, blocksze, status )
       endif

! Write the required header keywords.

       call ftphpr ( unit, simple, bitpix, naxis, naxes, pcount, gcount, extend, status )

! Write an optional comment.

       call ftpkys ( unit, 'CREATOR', creator,'Alexander Menshchikov, DAp IRFU CEA Saclay', status )
       call ftpcom ( unit,' ',status)
       call ftpkys ( unit, 'DATE'  , cdatetime, 'creation date and time', status )

! Write all parameters.

       dxdeg = - dx * as2deg
       dydeg = dy * as2deg

       call ftpkyd ( unit, 'BZERO'  , bzero , 13 , 'zero point in scaling equation', status )
       call ftpkyd ( unit, 'BSCALE' , bscale, 13 , 'linear factor in scaling equation' , status )
       call ftpkyd ( unit, 'DATAMAX', datamx, 13 , 'maximum data value', status )
       call ftpkyd ( unit, 'DATAMIN', datamn, 13 , 'minimum data value', status )
!!       if  (blank .gt. 0)
!!     & call ftpkyj ( unit, 'BLANK'  , blank      , 'value used for undefined array elements', status )
       call ftpkys ( unit, 'BUNIT'  , bunit      , 'physical units of the array values', status )
       call ftpkys ( unit, 'CTYPE1' , ctype1     , 'name of the coordinate axis', status )
       call ftpkys ( unit, 'CTYPE2' , ctype2     , 'name of the coordinate axis', status )
       call ftpkyd ( unit, 'CRPIX1' , rp1   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CRPIX2' , rp2   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CROTA1' , rot1  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CROTA2' , rot2  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CD1_1'  , cd11  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD1_2'  , cd12  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_1'  , cd21  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_2'  , cd22  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CRVAL1' , xr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CRVAL2' , yr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CDELT1' , dxdeg , 13 , 'coordinate increment along axis', status )
       call ftpkyd ( unit, 'CDELT2' , dydeg , 13 , 'coordinate increment along axis', status )
       call ftpkys ( unit, 'OBJECT' , object     , 'object identifier', status )
       call ftpkyd ( unit, 'RA'     , ra     , 13, 'right ascension', status )
       call ftpkyd ( unit, 'DEC'    , dec    , 13, 'declination', status )
       call ftpkyd ( unit, 'EQUINOX', equinox, 13, 'equinox', status )
       call ftpkyd ( unit, 'WAVE'   , wave   , 13, 'wavelength (microns)', status )
       call ftphis ( unit, history, status )

       if (beam .ne. 0.0d0) call ftpkyd ( unit, 'BEAM', beam, 13, 'convolved with the beam (arcsec)', status )

! Write the array to the FITS file.

       call ftp2dd ( unit, group, nx, nx, ny, fun, status )

! Close the file.

       call ftclos ( unit, status )

! Free the unit number.

       call ftfiou ( unit, status )

! Check for any error, and if so print out error messages

       if (status .gt. 0) call printerr ( status )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||