
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       program expanda
!__________________________________________________________________________________________________________________________________
!
! Expand masked image area towards the edges of the image.
!
! GETSF • Multi-Scale Multi-Wavelength Source & Filament Extraction • Alexander Men'shchikov, DAp IRFU CEA Saclay  
!__________________________________________________________________________________________________________________________________
!
       implicit      none

       logical       lfnam1, lfnam2, lunix, ldir1p, ldir1m, ldir2p, ldir2m, ldir3pp, ldir3mm, ldir4pm, ldir4mp
                                          
       character*3   dot
       character*6   cfitsversion, cverbose         
       character*7   clibname, cx, cy
       character*8   ctime, cfpptime
       character*10  cdate
       character*21  compda
       character*80  object, bunit, ctype1, ctype2, creator, history
       character*500 sourcemask, outname, origname, arg1, arg2, arg3, arg4, arg5, arg6

       integer       firstb, lastc, lasts, i, j, iobjs, ndate, iotty, nx, ny, nx1, ny1, nx2, ny2, ia1, ia2, ia3, ia4, ia5, ia6
     &             , nave, isp1, ion, blank, iorig, irc, lb, im1, jm1, ip1, jp1, im, ip, jm, jp, m, ncx, ncy, nav, ii, jj !, iter

       parameter   ( iotty = 6 )

       real          fitsvers
       real*8        dx, dy, dx1, dy1, dx2, dy2, wave, xr, yr
     &             , crpix1, crpix2, crval1, crval2, rx, ry, ra, dec, beam, beam1, beam2, almostzero, funmin, funmax, crota1, crota2
     &             , equinox, bzero, bscale, datamin, datamax, cd11, cd12, cd21, cd22, expandvalue, wave1, wave2

       real*8 , allocatable :: original(:,:), maskimage(:,:), numbers(:,:), expanded(:,:)

       parameter   ( almostzero = 1.0d-20, dot = '•' )

       external      firstb, lastc, lasts, rfits, wfits, when, getfitshead, osystem, ftvers
!__________________________________________________________________________________________________________________________________
!
! Determine operating system

       call osystem ( lunix )
       call ftvers ( fitsvers )
       lb = 1
       write (cfitsversion,'(f6.3)') fitsvers
       if (cfitsversion .eq. ' 5.030') then
         lb = 2
       endif
       clibname = 'CFITSIO'
       cfpptime = __TIME__
       cfpptime = cfpptime(1:5)
       compda = __DATE__//' '//cfpptime

! Get command line parameters (file names).

       call getarg ( 1, arg1 )
       call getarg ( 2, arg2 )
       call getarg ( 3, arg3 )
       call getarg ( 4, arg4 )
       call getarg ( 5, arg5 )
       call getarg ( 6, arg6 )

       ia1 = lastc ( arg1 )
       ia2 = lastc ( arg2 )
       ia3 = lastc ( arg3 )
       ia4 = lastc ( arg4 )
       ia5 = lastc ( arg5 )
       ia6 = lastc ( arg6 )

       cverbose = '-verb2'
       if (ia6 .gt. 0 .and. arg6(1:ia6-1) .eq. '-verb') cverbose = arg6(1:ia6)
       if (ia5 .gt. 0 .and. arg5(1:ia5-1) .eq. '-verb') cverbose = arg5(1:ia5)
       if (ia4 .gt. 0 .and. arg4(1:ia4-1) .eq. '-verb') cverbose = arg4(1:ia4)
       if (ia3 .gt. 0 .and. arg3(1:ia3-1) .eq. '-verb') cverbose = arg3(1:ia3)

       if (cverbose .eq. '-verb2') then
         write (iotty,'( )')
         if (arg1(1:ia1) .eq. ':') write (iotty,'(a$)') '  '
         write (iotty,'(a$)') ' EXPANDA '//dot//' Expand Masked Image Area '//dot//' '//compda
       endif
       if (arg1(1:ia1) .eq. ':') stop
       if (cverbose .eq. '-verb2') write (iotty,'()')
       if (cverbose .eq. '-verb2') write (iotty,'(a)') ' Alexander Men’shchikov, DAp IRFU CEA Saclay, France.'
       if (cverbose .eq. '-verb2') write (iotty,'(a)') ' Using '//clibname(lb:7)//' library version'//cfitsversion
     &                                               //' by William D Pence.'

       if (ia3 .eq. 0) then
         write (iotty,'(/a)') ' USAGE: expanda <nave> <origname> <sourcemask> [-o <outname>] [-verb{0|1|2}]'
         stop 99
       endif
       read (arg1(1:ia1),*,err=9) nave
!!       if (nave .eq. 0) then
!!         write (iotty,'(/a)') '   EXPANDA: ERROR: Parameter NAVE must be greater than 0.'
!!         stop 99
!!       endif
       goto 10
  9    continue       
         write (iotty,'(/a)') '   EXPANDA: ERROR: Trouble reading command-line parameter NAVE.'
         stop 99
 10    continue
       origname = arg2(1:ia2)
       sourcemask = arg3(1:ia3)
       iorig = lastc ( origname )
       iobjs = lastc ( sourcemask )

! Check if the input file exists and open it.

       inquire ( file=origname(1:iorig), exist=lfnam1 )
       if (.not.lfnam1 .and. origname(iorig-4:iorig) .ne. '.fits') then
         origname = origname(1:iorig)//'.fits'
         iorig = lastc ( origname )
         inquire ( file=origname(1:iorig), exist=lfnam1 )
       endif

! Check if the input file exists and open it.

       inquire ( file=sourcemask(1:iobjs), exist=lfnam2 )
       if (.not.lfnam2 .and. sourcemask(iobjs-4:iobjs) .ne. '.fits') then
         sourcemask = sourcemask(1:iobjs)//'.fits'
         iobjs = lastc ( sourcemask )
         inquire ( file=sourcemask(1:iobjs), exist=lfnam2 )
       endif

       isp1 = lasts ( origname ) + 1

       outname = origname(1:max(iorig-5,1))//'.ea.fits'
       if (arg4(1:2) .eq. '-o') outname = arg5
       ion = lastc ( outname )
       if (outname(max(ion-4,1):ion) .ne. '.fits') then
         outname = outname(1:ion)//'.fits'
         ion = lastc ( outname )
       endif

       if (cverbose .eq. '-verb1') write (iotty,'(/a)') ' EXPANDA: '//arg1(1:ia1)//' '//origname(1:iorig-5)
     &                                                //' '//sourcemask(1:iobjs-5)
       if (.not.lfnam1) then
         write (iotty,'(/a)') '   EXPANDA: ERROR: File '''//origname(1:iorig-5)//''' not found.'
         stop 1
       endif
       if (.not.lfnam2) then
         write (iotty,'(/a)') '   EXPANDA: ERROR: File '''//sourcemask(1:iobjs-5)//''' not found.'
         stop 2
       endif

! Determine numbers of pixels in the FITS image.

       call getfitshead ( origname(1:iorig), nx1, ny1, dx1, dy1, bunit )

! Allocate the image arrays.
       
       allocate ( original(nx1,ny1), stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a)') '   EXPANDA: ERROR: Trouble allocating memory (1).'
         stop 12
       endif

       write (cx,'(i7)') nx1
       write (cy,'(i7)') ny1
       ncx = int ( log10 ( dble ( nx1 ) ) ) + 1
       ncy = int ( log10 ( dble ( ny1 ) ) ) + 1

       if (cverbose .eq. '-verb2') write (iotty,'(/a)') '   Reading ('//cx(7-ncx+1:7)//' x '//cy(7-ncy+1:7)//') '''
     &                                                //origname(1:iorig)//''''
       
       call rfits ( nx1, ny1, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, original, dx1, dy1, object, ra, dec
     &            , origname(1:iorig), iotty, 0, creator, beam1, funmin, funmax, blank, crota1, crota2, cd11, cd12
     &            , cd21, cd22, equinox, bzero, bscale, wave1, datamin, datamax, history, cverbose )

       if (cverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7))') '   Minmax values in the image:', funmin, funmax

! Determine numbers of pixels in the FITS image.

       call getfitshead ( sourcemask(1:iobjs), nx2, ny2, dx2, dy2, bunit )

! Allocate the image arrays.
       
       allocate ( maskimage(nx2,ny2), stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a)') '   EXPANDA: ERROR: Trouble allocating memory (2).'
         stop 13
       endif
       
       write (cx,'(i7)') nx2
       write (cy,'(i7)') ny2
       ncx = int ( log10 ( dble ( nx2 ) ) ) + 1
       ncy = int ( log10 ( dble ( ny2 ) ) ) + 1

       if (cverbose .eq. '-verb2') write (iotty,'(a)') '   Reading ('//cx(7-ncx+1:7)//' x '//cy(7-ncy+1:7)//') '''
     &                                               //sourcemask(1:iobjs)//''''

       call rfits ( nx2, ny2, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, maskimage, dx2, dy2, object, ra, dec
     &            , sourcemask(1:iobjs), iotty, 0, creator, beam2, funmin, funmax, blank, crota1, crota2, cd11
     &            , cd12, cd21, cd22, equinox, bzero, bscale, wave2, datamin, datamax, history, cverbose )

       if (cverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7))') '   Minmax values in the image:', funmin, funmax

       if (beam1 .eq. beam2) then
         beam = beam1
       else
         beam = 0.0
       endif

       if (nx1 .ne. nx2 .or. ny1 .ne. ny2 .or. abs ( dx1 - dx2 ) .gt. 0.001d0 * dx2 .or. abs ( dy1 - dy2 ) .gt. 0.001d0 * dy2) then
         write (iotty, '(/a)') '   EXPANDA: ERROR: The images have different parameters.'
         stop 99
       endif
       nx = nx1
       ny = ny1
       dx = dx1
       dy = dy1
       rx = crpix1
       ry = crpix2
       xr = crval1
       yr = crval2
       wave = wave1

       if (cverbose .eq. '-verb2') write (iotty,'(a)') '   Expanding masked area of '''//origname(1:iorig)//''''

! Allocate the image arrays.
       
       allocate ( numbers(nx,ny), expanded(nx,ny), stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a)') '   EXPANDA: ERROR: Trouble allocating memory (3).'
         stop 15
       endif
       
!!       do iter=1,1
       
         do j=1,ny
           do i=1,nx
             numbers(i,j) = 0.0d0
             expanded(i,j) = original(i,j)
             if (maskimage(i,j) .lt. almostzero) expanded(i,j) = 0.0d0
           enddo
         enddo             
         
         do j=1,ny
           do i=1,nx
             if (maskimage(i,j) .gt. almostzero) then
               im1 = max ( i - 1, 1 )
               jm1 = max ( j - 1, 1 )
               ip1 = min ( i + 1, nx )
               jp1 = min ( j + 1, ny )
               ldir1p  = .false.
               ldir1m  = .false.
               ldir2p  = .false.
               ldir2m  = .false.
               ldir3pp = .false.
               ldir3mm = .false.
               ldir4pm = .false.
               ldir4mp = .false.
               if (maskimage(ip1,  j) .lt. almostzero) ldir1p  = .true.
               if (maskimage(im1,  j) .lt. almostzero) ldir1m  = .true.
               if (maskimage(i  ,jp1) .lt. almostzero) ldir2p  = .true.
               if (maskimage(i  ,jm1) .lt. almostzero) ldir2m  = .true.
               if (maskimage(ip1,jp1) .lt. almostzero) ldir3pp = .true.
               if (maskimage(im1,jm1) .lt. almostzero) ldir3mm = .true.
               if (maskimage(ip1,jm1) .lt. almostzero) ldir4pm = .true.
               if (maskimage(im1,jp1) .lt. almostzero) ldir4mp = .true.
    
               nav = 0
               expandvalue = 0.0d0
  
               do jj=max(1,j-nave),min(ny,j+nave)
                 do ii=max(1,i-nave),min(nx,i+nave)
                   if (maskimage(ii,jj) .gt. almostzero) then 
                     expandvalue = expandvalue + original(ii,jj)
                     nav = nav + 1
                   endif
                 enddo
               enddo
               if (nav .gt. 0) then 
                 expandvalue = expandvalue / dble ( nav )
               endif
    
               if (ldir1p) then
                 do ip=ip1,nx
                   numbers(ip,j) = numbers(ip,j) + 1.0d0
                   expanded(ip,j) = expanded(ip,j) + expandvalue
                   if (maskimage(ip,j) .lt. almostzero .and. maskimage(min(ip+1,nx),j) .gt. almostzero) exit
                 enddo
               endif
               if (ldir1m) then
                 do im=im1,1,-1
                   numbers(im,j) = numbers(im,j) + 1.0d0
                   expanded(im,j) = expanded(im,j) + expandvalue
                   if (maskimage(im,j) .lt. almostzero .and. maskimage(max(im-1,1),j) .gt. almostzero) exit
                 enddo
               endif
  
               if (ldir2p) then
                 do jp=jp1,ny
                   numbers(i,jp) = numbers(i,jp) + 1.0d0
                   expanded(i,jp) = expanded(i,jp) + expandvalue
                   if (maskimage(i,jp) .lt. almostzero .and. maskimage(i,min(jp+1,ny)) .gt. almostzero) exit
                 enddo
               endif
               if (ldir2m) then
                 do jm=jm1,1,-1
                   numbers(i,jm) = numbers(i,jm) + 1.0d0
                   expanded(i,jm) = expanded(i,jm) + expandvalue
                   if (maskimage(i,jm) .lt. almostzero .and. maskimage(i,max(jm-1,1)) .gt. almostzero) exit
                 enddo
               endif
  
               if (ldir3pp) then
                 do m=1,min(nx-i,ny-j)
                   numbers(i+m,j+m) = numbers(i+m,j+m) + 1.0d0
                   expanded(i+m,j+m) = expanded(i+m,j+m) + expandvalue
                   if (maskimage(i+m,j+m) .lt. almostzero .and. maskimage(min(ip1+m,nx),min(jp1+m,ny)) .gt. almostzero) exit
                 enddo
               endif
               if (ldir3mm) then
                 do m=1,min(i-1,j-1)
                   numbers(i-m,j-m) = numbers(i-m,j-m) + 1.0d0
                   expanded(i-m,j-m) = expanded(i-m,j-m) + expandvalue
                   if (maskimage(i-m,j-m) .lt. almostzero .and. maskimage(max(im1-m,1),max(jm1-m,1)) .gt. almostzero) exit
                 enddo
               endif
  
               if (ldir4pm) then
                 do m=1,min(nx-i,j-1)
                   numbers(i+m,j-m) = numbers(i+m,j-m) + 1.0d0
                   expanded(i+m,j-m) = expanded(i+m,j-m) + expandvalue
                   if (maskimage(i+m,j-m) .lt. almostzero .and. maskimage(min(ip1+m,nx),max(jm1-m,1)) .gt. almostzero) exit
                 enddo
               endif
               if (ldir4mp) then
                 do m=1,min(i-1,ny-j)
                   numbers(i-m,j+m) = numbers(i-m,j+m) + 1.0d0
                   expanded(i-m,j+m) = expanded(i-m,j+m) + expandvalue
                   if (maskimage(i-m,j+m) .lt. almostzero .and. maskimage(max(im1-m,1),min(jp1+m,ny)) .gt. almostzero) exit
                 enddo
               endif
             endif
           enddo
         enddo
  
         do j=1,ny
           do i=1,nx
             if (numbers(i,j) .gt. almostzero) then
               expanded(i,j) = expanded(i,j) / numbers(i,j)
             endif
           enddo
         enddo
         
!!         if (iter .eq. 1) then
!!           do j=1,ny
!!             do i=1,nx
!!               original(i,j) = expanded(i,j)
!!               maskimage(i,j) = max ( maskimage(i,j), numbers(i,j) )
!!             enddo
!!           enddo
!!         endif
!!       enddo
       
! Get the current date and time.

       call when ( lunix, ctime, cdate, ndate, 4 )

       if (cverbose .eq. '-verb2') write (iotty,'(a)') '   Writing expanded image '''//outname(1:ion)//''''

       call wfits ( cfitsversion, nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, expanded, dx, dy, object, crval1
     &            , crval2, outname(1:ion), cdate, ctime, creator, beam, blank, crota1, crota2, cd11, cd12, cd21
     &            , cd22, equinox, bzero, bscale, wave, datamin, datamax, history )

       deallocate ( original, maskimage, numbers, expanded )

       if (cverbose .eq. '-verb2') write (iotty,'(/a)') ' Done.'

!!       stop !<-- commented out because it would lead to run-time messages about denormalized values, when using gfortran.
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine rfits 
       
     &            ( nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname, iotty, iolog, creator
     &            , beam, funmin, funmax, blank, rot1, rot2, cd11, cd12, cd21, cd22, equinox, bzero, bscale, wave, datamn, datamx
     &            , history, cverbose )
!__________________________________________________________________________________________________________________________________
!
! GETSF • Multi-Scale Multi-Wavelength Source & Filament Extraction • Alexander Men'shchikov, DAp IRFU CEA Saclay      
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend, anynull, myfile
       character*(*) ctype1, ctype2, creator, fname, bunit, object, history, cverbose
       character*3   cnan
       character*25  cdatetime
       character*80  comment
       integer       i, j, nx, ny, stat, blksize, group, naxis, pcount, gcount, bitpix, unit, iotty, iolog, maxdim
     &             , rw, nfound, blank, stat05, stat06, stat07, stat08, stat09, stat10, stat11, stat12, stat13, stat14, stat15
     &             , stat16, stat17, stat18, stat19, stat20, stat21, stat22, stat23, stat24, stat25, stat26, stat27, stat28, stat29
     &             , stat30, stat31, stat32, stat0
       parameter   ( maxdim = 2 )
       integer       naxes(maxdim)
       real*8        dx, dy, ra, dec, beam, funmin, funmax, xr, yr, rp1, rp2, rot1, rot2, epoch, equinox, bzero, bscale, wave
     &             , dxdeg, dydeg, cd11, cd12, cd21, cd22, almostzero, as2deg, datamx, datamn, sign1, arg1, arg2, argm
     &             , sinrota, cosrota, pi, angle
       real*8        fun(nx,ny)
       parameter   ( pi = 3.14159265358979d0, as2deg = 1.0d0 / 3600.0d0, almostzero = 1.0d-30 )
       external      ftopen, ftgkne, ftghpr, ftgkys, ftgkyd, ftgkye, ftg2de, ftclos, printerr, ftgkyj, ftg2dd
!__________________________________________________________________________________________________________________________________
!
! Initialize variables.

       stat = 0
       unit   = 20
       group  = 1
       rw     = 0
       stat05 = 0
       stat06 = 0
       stat07 = 0
       stat08 = 0
       stat09 = 0
       stat10 = 0
       stat11 = 0
       stat12 = 0
       stat13 = 0
       stat14 = 0
       stat15 = 0
       stat16 = 0
       stat17 = 0
       stat18 = 0
       stat19 = 0
       stat20 = 0
       stat21 = 0
       stat22 = 0
       stat23 = 0
       stat24 = 0
       stat25 = 0
       stat26 = 0
       stat27 = 0
       stat28 = 0
       stat29 = 0
       stat30 = 0
       stat31 = 0
       stat32 = 0

! Open the FITS file.

       call ftopen ( unit, fname, rw, blksize, stat )

       if (stat .ne. 0) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         stop 99
       endif

! Determine the size of the image.

       call ftgkne ( unit, 'NAXIS', 1, 2, naxes, nfound, stat )

! Check that it found both NAXIS1 and NAXIS2 keywords.

       if (nfound .ne. 2) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         stop 99
       endif

! Get the main primary header keywords which define the array structure.

       call ftghpr ( unit, maxdim, simple, bitpix, naxis, naxes, pcount, gcount, extend, stat )

! Read all parameters.

       call ftgkys ( unit, 'CREATOR', creator, comment, stat )
       
       if (stat .eq. 0 .and. (comment(1:21) .eq. 'Alexander Menshchikov' .or. comment(1:22) .eq. 'Alexander Men''shchikov')) then
         myfile = .true.
       else
         myfile = .false.
         stat = 0
         creator = 'EXPANDA'
       endif

! I don't understant why blank is assigned to my array values (if blank is not zero) - I have no undefined values in arrays. Maybe
! there is a bug in the version of FITSIO I use. For now I will ignore the blank value when reading FITS files.

       blank = 0

       datamx = 0.0d0
       datamn = 0.0d0
       bunit = ' '
       ctype1 = ' '
       ctype2 = ' '
       rp1 = 0.0d0
       rp2 = 0.0d0
       rot1 = 0.0d0
       rot2 = 0.0d0
       cd11 = 0.0d0
       cd12 = 0.0d0
       cd21 = 0.0d0
       cd22 = 0.0d0
       xr = 0.0d0
       yr = 0.0d0
       dx = 0.0d0
       dy = 0.0d0
       object = ' '
       ra = 0.0d0
       dec = 0.0d0
       epoch = 0.0d0
       equinox = 0.0d0
       beam = 0.0d0
       bzero = 0.0d0
       bscale = 1.0d0
       wave = 1.0d2
       history = ' '

       call ftgkys ( unit, 'DATE', cdatetime , comment, stat05 )
       call ftgkyd ( unit, 'DATAMAX', datamx , comment, stat06 )
       call ftgkyd ( unit, 'DATAMIN', datamn , comment, stat07 )
       call ftgkyj ( unit, 'BLANK'  , blank  , comment, stat08 )
       call ftgkys ( unit, 'BUNIT'  , bunit  , comment, stat09 )
       call ftgkyd ( unit, 'BZERO'  , bzero  , comment, stat10 )
       call ftgkyd ( unit, 'BSCALE' , bscale , comment, stat11 )
       call ftgkys ( unit, 'CTYPE1' , ctype1 , comment, stat12 )
       call ftgkys ( unit, 'CTYPE2' , ctype2 , comment, stat13 )
       call ftgkyd ( unit, 'CRPIX1' , rp1    , comment, stat14 )
       call ftgkyd ( unit, 'CRPIX2' , rp2    , comment, stat15 )
       call ftgkyd ( unit, 'CROTA1' , rot1   , comment, stat16 )
       call ftgkyd ( unit, 'CROTA2' , rot2   , comment, stat17 )
       call ftgkyd ( unit, 'CD1_1'  , cd11   , comment, stat18 )
       call ftgkyd ( unit, 'CD1_2'  , cd12   , comment, stat19 )
       call ftgkyd ( unit, 'CD2_1'  , cd21   , comment, stat20 )
       call ftgkyd ( unit, 'CD2_2'  , cd22   , comment, stat21 )
       call ftgkyd ( unit, 'CRVAL1' , xr     , comment, stat22 )
       call ftgkyd ( unit, 'CRVAL2' , yr     , comment, stat23 )
       call ftgkyd ( unit, 'CDELT1' , dxdeg  , comment, stat24 )
       call ftgkyd ( unit, 'CDELT2' , dydeg  , comment, stat25 )
       call ftgkys ( unit, 'OBJECT' , object , comment, stat26 )
       call ftgkyd ( unit, 'RA'     , ra     , comment, stat27 )
       call ftgkyd ( unit, 'DEC'    , dec    , comment, stat28 )
       call ftgkyd ( unit, 'EPOCH'  , epoch  , comment, stat29 )
       call ftgkyd ( unit, 'EQUINOX', equinox, comment, stat30 )
       call ftgkyd ( unit, 'WAVE'   , wave   , comment, stat31 )
       call ftgkyd ( unit, 'BEAM'   , beam   , comment, stat )
       call ftgkys ( unit, 'HISTORY', history, history, stat32 )

       if (stat .ne. 0) then
         beam = 0.0d0
         stat = 0
       endif

       stat0 = stat05 + stat06 + stat07 + stat09 + stat10 + stat11 + stat12 + stat13 + stat14 + stat15 + stat16 + stat17 + stat18
     &       + stat19 + stat20 + stat21 + stat22 + stat23 + stat24 + stat25 + stat26 + stat27 + stat28 + stat30 + stat31

       if (cverbose .ne. '-verb0') then
         if (stat0  .gt. 0) write (iotty,'()')
         if (stat05 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATE absent or error'
         if (stat06 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMAX absent or error'
         if (stat07 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMIN absent or error'
         if (stat09 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BUNIT absent or error'
         if (stat10 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BZERO absent or error'
         if (stat11 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BSCALE absent or error'
         if (stat12 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE1 absent or error'
         if (stat13 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE2 absent or error'
         if (stat14 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX1 absent or error'
         if (stat15 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX2 absent or error'
         if (stat16 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA1 absent or error'
         if (stat17 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA2 absent or error'
         if (stat18 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_1 absent or error'
         if (stat19 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_2 absent or error'
         if (stat20 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_1 absent or error'
         if (stat21 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_2 absent or error'
         if (stat22 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL1 absent or error'
         if (stat23 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL2 absent or error'
         if (stat24 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT1 absent or error'
         if (stat25 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT2 absent or error'
         if (stat26 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword OBJECT absent or error'
         if (stat27 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword RA absent or error'
         if (stat28 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DEC absent or error'
         if (stat30 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword EQUINOX absent or error'
         if (stat31 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword WAVE absent or error'
       endif
       if (stat26 .ne. 0 .or. object(1:10) .eq. '          ') object = '<Name>'

! Read the FITS file into the 2D array.

       call ftg2dd ( unit, group, dble ( blank ), naxes(1), naxes(1), naxes(2), fun, anynull, stat )

       if (stat .ne. 0) write (iotty,'(a)') ' Error reading FITS file'

! Close the file and free the unit number.

       call ftclos ( unit, stat )

! Check for any error, and if so print out error messages.

       if (stat .gt. 0) call printerr ( stat )

       if (stat17 .ne. 0) then
         rot2 = 0.0d0
       endif
       rot1 = rot2

       if (stat29 .eq. 0 .and. stat30 .ne. 0) then
         equinox = epoch
       endif

! CDELT* is absent or error.

       if (stat24 .ne. 0 .or. stat25 .ne. 0) then

! CD1* and CD2* are defined correctly: determine CDELT* and ROTA*.

         if (stat18 .eq. 0 .and. stat19 .eq. 0 .and. stat20 .eq. 0 .and. stat21 .eq. 0) then
           sign1 = sign ( 1.0d0, cd11 * cd22 - cd12 * cd21 )
           dxdeg = sign1 * sqrt ( cd11**2 + cd21**2 )
           dydeg = sqrt ( cd12**2 + cd22**2 )
           arg1 = sign1 * cd12 / cd22
           arg2 = -sign1 * cd21 / cd11
           argm = (arg1 + arg2) / 2.0d0
           if (abs ( arg1 - arg2 ) / argm .gt. 1.0d-10) then
             write (iotty,'(/a/)') '   RFITS: WARNING: Coordinate axes are not orthogonal'
           endif
           rot2 = atan ( argm )
           sinrota = sin ( rot2 )
           cosrota = cos ( rot2 )
           rot2 = rot2 * 180.0d0 / pi
           if (cosrota .ge. 0.0d0 .and. sinrota .ge. 0.0d0) then
             rot1 = rot2
           elseif (cosrota .lt. 0.0d0 .and. sinrota .ge. 0.0d0) then
             rot1 = 180.0d0 - rot2
           elseif (cosrota .lt. 0.0d0 .and. sinrota .lt. 0.0d0) then
             rot1 = 180.0d0 + rot2
           elseif (cosrota .ge. 0.0d0 .and. sinrota .lt. 0.0d0) then
             rot1 = - rot2
           endif
           rot2 = rot1
         endif
       endif

! CD1* and CD2* are absent or error.

       if (stat18 .ne. 0 .or. stat19 .ne. 0 .or. stat20 .ne. 0 .or. stat21 .ne. 0) then

! CDELT* are defined correctly: determine CD1* and CD2*.

         if (stat24 .eq. 0 .and. stat25 .eq. 0) then
           angle = rot2 * pi / 180.0d0
           cd11 = dxdeg * cos ( angle )
           cd12 = abs ( dydeg ) * sign ( 1.0d0, dxdeg ) * sin ( angle )
           cd21 = - abs ( dxdeg ) * sign ( 1.0d0, dydeg ) * sin ( angle )
           cd22 = dydeg * cos ( angle )
           if (abs ( cd12 ) .lt. almostzero) cd12 = 0.0d0
           if (abs ( cd21 ) .lt. almostzero) cd21 = 0.0d0
         endif
       endif

       dx = abs ( dxdeg ) / as2deg
       dy = abs ( dydeg ) / as2deg

! Print out the min and max values.

       nx = naxes(1)
       ny = naxes(2)
       funmin = 1.0d30
       funmax =-1.0d30
       cnan='NAN'
       do j=1,ny
         do i=1,nx
           if (.not.isnan ( fun(i,j) )) then  
             funmin = min ( funmin, fun(i,j) )
             funmax = max ( funmax, fun(i,j) )
           endif  
           if (int ( fun(i,j) ) .eq. blank .and. blank .ne. 0) then  
             read (cnan,*) fun(i,j)
           endif
         enddo
       enddo

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine wfits 
       
     &            ( cfitsversion, nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname, cdate
     &            , ctime, creator, beam, blank, rot1, rot2, cd11, cd12, cd21, cd22, equinox, bzero, bscale, wave, datamn, datamx
     &            , history )
!__________________________________________________________________________________________________________________________________
!
! GETSF • Multi-Scale Multi-Wavelength Source & Filament Extraction • Alexander Men'shchikov, DAp IRFU CEA Saclay      
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend
       character*(*) fname, ctype1, ctype2, object, creator, ctime, cdate, bunit, cfitsversion, history
       character*19  cdatetime
       integer       i, j, nx, ny, status, blocksze, group, naxis, pcount, gcount, bitpix, unit, naxes(2), blank
       real*8        xr, yr, dx, dy, rp1, rp2, ra, dec, beam, bzero, bscale, as2deg, rot1, rot2, equinox, datamn
     &             , datamx, dxdeg, dydeg, wave, cd11, cd12, cd21, cd22, almostzero, pi
       real*8        fun(nx,ny)
       parameter   ( pi = 3.14159265358979d0, as2deg = 1.0d0 / 3600.0d0, almostzero = 1.0d-30 )
       external      fdelete, ftgiou, ftinit, ftphpr, ftpkys, ftpcom, ftpkyd, ftpkyj, ftphis, ftp2dd, ftclos, ftfiou, printerr
!__________________________________________________________________________________________________________________________________
!
! Create a FITS primary array containing a 2-D image. Initialize parameters about the FITS image.

       status   = 0
       blocksze = 1
       simple   = .true.
       extend   = .false.
       group    = 1
       naxis    = 2
       pcount   = 0
       gcount   = 1
       naxes(1) = nx
       naxes(2) = ny

! Compute offset and scaling factor so as to represent real numbers most efficiently and accurately with given 'bitpix'.

       datamn = 1.0d30
       datamx =-1.0d30
       do j=1,ny
         do i=1,nx
           if (.not.isnan ( fun(i,j) )) then
             if (fun(i,j) .lt. datamn ) datamn = fun(i,j)
             if (fun(i,j) .gt. datamx ) datamx = fun(i,j)
           endif
         enddo
       enddo

! When packing numbers using the formulas below, I got values of the function different in the 5-th digit from DATAMIN and
!  DATAMAX placed in the header when writing the data. From now on I will use bzero = 0 bscale = 1 and bitpix = -32, as these
! seem to cure the problem.

cc!      bitpix = 16
cc!      bzero  = 0.5d0 * (funmin + funmax)
cc!      bscale = dabs ( funmax - bzero ) / 32767.0d0

       bitpix = -32     !<-- single precision floating-point values
       bzero  = 0.0d0
       bscale = 1.0d0
       blank = blank

       cdatetime = cdate(7:10)//'-'//cdate(4:5)//'-'//cdate(1:2)//'T'//ctime

! Delete the file if it already exists.

       if (cfitsversion .eq. '5.030') then
         call fdelete ( fname, status )
       endif

! Get an unused Logical Unit Number to use to open the FITS file.

       call ftgiou ( unit, status )

! Create the new empty FITS file.

       if (cfitsversion .eq. '5.030') then
         call ftinit ( unit, fname, blocksze, status )
       else
         call ftinit ( unit, '!'//fname, blocksze, status )
       endif

! Write the required header keywords.

       call ftphpr ( unit, simple, bitpix, naxis, naxes, pcount, gcount, extend, status )

! Write an optional comment.

       call ftpkys ( unit, 'CREATOR', creator,'Alexander Menshchikov, DAp IRFU CEA Saclay', status )
       call ftpcom ( unit,' ',status)
       call ftpkys ( unit, 'DATE'  , cdatetime, 'creation date and time', status )

! Write all parameters.

       dxdeg = - dx * as2deg
       dydeg = dy * as2deg

       call ftpkyd ( unit, 'BZERO'  , bzero , 13 , 'zero point in scaling equation', status )
       call ftpkyd ( unit, 'BSCALE' , bscale, 13 , 'linear factor in scaling equation' , status )
       call ftpkyd ( unit, 'DATAMAX', datamx, 13 , 'maximum data value', status )
       call ftpkyd ( unit, 'DATAMIN', datamn, 13 , 'minimum data value', status )
!!       if  (blank .gt. 0)
!!     & call ftpkyj ( unit, 'BLANK'  , blank      , 'value used for undefined array elements', status )
       call ftpkys ( unit, 'BUNIT'  , bunit      , 'physical units of the array values', status )
       call ftpkys ( unit, 'CTYPE1' , ctype1     , 'name of the coordinate axis', status )
       call ftpkys ( unit, 'CTYPE2' , ctype2     , 'name of the coordinate axis', status )
       call ftpkyd ( unit, 'CRPIX1' , rp1   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CRPIX2' , rp2   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CROTA1' , rot1  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CROTA2' , rot2  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CD1_1'  , cd11  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD1_2'  , cd12  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_1'  , cd21  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_2'  , cd22  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CRVAL1' , xr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CRVAL2' , yr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CDELT1' , dxdeg , 13 , 'coordinate increment along axis', status )
       call ftpkyd ( unit, 'CDELT2' , dydeg , 13 , 'coordinate increment along axis', status )
       call ftpkys ( unit, 'OBJECT' , object     , 'object identifier', status )
       call ftpkyd ( unit, 'RA'     , ra     , 13, 'right ascension', status )
       call ftpkyd ( unit, 'DEC'    , dec    , 13, 'declination', status )
       call ftpkyd ( unit, 'EQUINOX', equinox, 13, 'equinox', status )
       call ftpkyd ( unit, 'WAVE'   , wave   , 13, 'wavelength (microns)', status )
       call ftphis ( unit, history, status )

       if (beam .ne. 0.0d0) call ftpkyd ( unit, 'BEAM', beam, 13, 'convolved with the beam (arcsec)', status )

! Write the array to the FITS file.

       call ftp2dd ( unit, group, nx, nx, ny, fun, status )

! Close the file.

       call ftclos ( unit, status )

! Free the unit number.

       call ftfiou ( unit, status )

! Check for any error, and if so print out error messages

       if (status .gt. 0) call printerr ( status )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||