
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       program makecube
!__________________________________________________________________________________________________________________________________
!
! Split three-dimensional data cube along one of its dimensions, producing a set of numbered two-dimensional FITS images.
!
! GETSF • Multi-Scale Multi-Wavelength Source & Filament Extraction • Alexander Men'shchikov, DAp IRFU CEA Saclay
!__________________________________________________________________________________________________________________________________
!                 
       implicit      none

       logical       lunix, lfname

       character*3   dot
       character*4   ck
       character*6   cfitsversion, cverbose, cverb
       character*7   clibname, cx, cy, cz
       character*8   ctime, cfpptime
       character*10  cdate
       character*21  compda
       character*80  object, creator, bunit, ctype1, ctype2, history, record
       character*500 filecube, filename(1000)

       integer       firstb, lastc, lastbs, ifn(1000), i, j, k, nxcube, nycube, nzcube, ndate, iotty, ifc, blank, lb, status
     &             , readwrite, nkeys, blocksize, nspace, irc, naxis, ncx, ncy, ncz, nfiles, icv, unit
       parameter   ( iotty = 6, dot = '•' )

       real          fitsvers 
       real*8        dx, dy, wave, xr, yr, rp1, rp2, ra, dec, beam, rot1, rot2, equinox, funmin, funmax, bzero, bscale, datamin
     &             , datamax, cd11, cd12, cd21, cd22
       real*8, allocatable ::  fun2(:,:), fun3(:,:,:)

       external      firstb, lastc, rfits, wfits, when, osystem, ftvers, lastbs, ftgiou, ftopen, ftghsp, printerr, ftclos, ftgrec
!__________________________________________________________________________________________________________________________________
!
! Determine operating system

       call osystem ( lunix )
       call ftvers ( fitsvers )
       lb = 1
       write (cfitsversion,'(f6.3)') fitsvers
       if (cfitsversion .eq. ' 5.030') then
         lb = 2
       endif

       clibname = 'CFITSIO'
       cfpptime = __TIME__
       cfpptime = cfpptime(1:5)
       compda = __DATE__//' '//cfpptime

! Get command line parameters (file names).

       nfiles = 0
       do i=1,500
         call getarg ( i, filename(i) )
         ifn(i) = lastc ( filename(i) )
         if (ifn(i) .gt. 0 .and. filename(i)(1:ifn(i)) .eq. '-o') exit
         nfiles = i
       enddo

       cverbose = '-verb2'
       call getarg ( nfiles+2, filecube )
       call getarg ( nfiles+3, cverb )
       ifc = lastc ( filecube )
       icv = lastc ( cverb )
       if (icv .gt. 0) cverbose = cverb

       if (cverbose .eq. '-verb2') then
         write (iotty,'( )')
         if (filename(1)(1:ifn(1)) .eq. ':') write (iotty,'(a$)') ' '
         write (iotty,'(a$)') ' MAKECUBE '//dot//' Create FITS Data Cube '//dot//' '//compda
       endif
       if (filename(1)(1:ifn(1)) .eq. ':') stop
       if (cverbose .eq. '-verb2') write (iotty,'()')
       if (cverbose .eq. '-verb2') write (iotty,'( a)') ' Alexander Men’shchikov, DAp IRFU CEA Saclay, France.'
       if (cverbose .eq. '-verb2') write (iotty,'( a)') ' Using '//clibname(lb:7)//' library version'//cfitsversion
     &                                                //' by William D Pence.'

       if (ifn(1) .eq. 0) then
         write (iotty,'(/a)') ' USAGE: makecube <image1> <image2> [...<imageN>] -o <datacube> [-verb{0|1|2}]'
         stop 99
       endif

       if (cverbose .eq. '-verb1') then
         write (iotty,'(/a)') ' MAKECUBE: '//filename(1)(1:ifn(1))//'...'//filename(nfiles)(1:ifn(nfiles))//' -o '//filecube(1:ifc)
       endif

! Check if the input file exists and open it.

       do i=1,nfiles
         inquire ( file=filename(i)(1:ifn(i)), exist=lfname )
         if (.not.lfname .and. filename(i)(ifn(i)-4:ifn(i)) .ne. '.fits') then
           inquire ( file=filename(i)(1:ifn(i))//'.fits', exist=lfname )
           if (.not.lfname) then
             write (iotty,'(/a)') '   MAKECUBE: ERROR: File '''//filename(i)(1:ifn(i))//''' not found.'
             stop 1
           endif
           filename(i) = filename(i)(1:ifn(i))//'.fits'
           ifn(i) = lastc ( filename(i) )
         endif
       enddo
       if (filecube(ifc-4:ifc) .ne. '.fits') then
         filecube = filecube(1:ifc)//'.fits'
         ifc = lastc ( filecube )
       endif

       status = 0
       readwrite = 0
       call ftgiou ( unit, status )
       call ftopen ( unit, filename(1)(1:ifn(1)), readwrite, blocksize, status )
       call ftghsp ( unit, nkeys, nspace, status )
       if (status .gt. 0) then
         call printerr ( status )
         stop 99
       endif
       naxis = 0
       nxcube = 0
       nycube = 0
       nzcube = nfiles
       do i=1,10
         call ftgrec ( unit, i, record, status )
         if (index ( record, 'NAXIS ' ) .eq. 1) read (record(10:31),*) naxis
         if (index ( record, 'NAXIS1' ) .eq. 1) read (record(10:31),*) nxcube
         if (index ( record, 'NAXIS2' ) .eq. 1) read (record(10:31),*) nycube
       enddo
       call ftclos ( unit, status )

       write (cx,'(i7)') nxcube
       write (cy,'(i7)') nycube
       write (cz,'(i7)') nzcube
       ncx = int ( log10 ( dble ( nxcube ) ) ) + 1
       ncy = int ( log10 ( dble ( nycube ) ) ) + 1
       ncz = int ( log10 ( dble ( nzcube ) ) ) + 1
       
       if (cverbose .eq. '-verb2') then
         write (iotty,'(/a)') '   Creating cube from separate images along axis #3'
         write (iotty,'( a)') '   Output cube dimensions: '//cx(7-ncx+1:7)//' x '//cy(7-ncy+1:7)//' x '//cz(7-ncz+1:7)//' pixels'
       endif
       if (nxcube .eq. 0 .or. nycube .eq. 0 .or. nzcube .eq. 0) then
         write (iotty,'(/a)') '   MAKECUBE: ERROR: Trouble getting numbers of pixels in '''//filename(1)(1:ifn(1))//'''.'
         stop 99
       endif

       allocate ( fun2(nxcube,nycube), fun3(nxcube,nycube,1), stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a)') '   MAKECUBE: ERROR: Trouble allocating memory (5).'
         stop 5
       endif

! Get current date and time.

       call when ( lunix, ctime, cdate, ndate, 4 )
       
       datamin = 1.0d30
       datamax =-1.0d30
       
       do k=1,nzcube
         write (ck,'(i4)') k
         if (ck(1:1) .eq. ' ') ck(1:1) = '0'
         if (ck(2:2) .eq. ' ') ck(2:2) = '0'
         if (ck(3:3) .eq. ' ') ck(3:3) = '0'

         if (cverbose .eq. '-verb2') write (iotty,'(/a)') '   Reading image '''//filename(k)(1:ifn(k))//''''

         call rfits ( nxcube, nycube, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun2, dx, dy, object, ra, dec, filename(k)(1:ifn(k))
     &              , iotty, 0, creator, beam, funmin, funmax, blank, rot1, rot2, cd11, cd12, cd21, cd22, equinox, bzero, bscale
     &              , wave, history, cverbose )
     
         do j=1,nycube
           do i=1,nxcube
             fun3(i,j,1) = fun2(i,j)             
           enddo
         enddo
         
         if (cverbose .eq. '-verb2') write (iotty,'(a)') '   Writing image '//ck//' to cube '''//filecube(1:ifc)//''''
         
         call wfits ( cfitsversion, nxcube, nycube, nzcube, k, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun3, dx, dy, object, xr
     &              , yr, filecube(1:ifc), cdate, ctime, creator, beam, blank, rot1, rot2, cd11, cd12, cd21, cd22, equinox, bzero
     &              , bscale, wave, datamin, datamax, history, unit )
       enddo

       deallocate ( fun2, fun3 )
       
       if (cverbose .eq. '-verb2') write (iotty,'(/a)') ' Done.'

!!       stop !<-- commented out because it would lead to run-time messages about denormalized values, when using gfortran.
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine rfits 
       
     &            ( nxcube, nycube, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname, iotty, iolog
     &            , creator, beam, funmin, funmax, blank, rot1, rot2, cd11, cd12, cd21, cd22, equinox, bzero, bscale, wave
     &            , history, cverbose )
!__________________________________________________________________________________________________________________________________
!
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend, myfile, anynull
       character*(*) ctype1, ctype2, creator, fname, bunit, object, history, cverbose
       character*3   cnan
       character*25  cdatetime
       character*80  comment
       integer       i, j, stat, blksize, group, naxis, pcount, gcount, bitpix, unit, iotty, iolog, maxdim, rw, nfound
     &             , blank, stat05, stat06, stat07, stat08, stat09, stat10, stat11, stat12, stat13, stat14, stat15, stat16, stat17
     &             , stat18, stat19, stat20, stat21, stat22, stat23, stat24, stat25, stat26, stat27, stat28, stat29, stat30, stat31
     &             , stat32, stat0, nxcube, nycube
       parameter   ( maxdim = 2 )
       integer       naxes(maxdim)
       real*8        dx, dy, ra, dec, beam, funmin, funmax, xr, yr, rp1, rp2, rot1, rot2, epoch, equinox, bzero, bscale, wave
     &             , dxdeg, dydeg, cd11, cd12, cd21, cd22, almostzero, as2deg, datamx, datamn, sign1, arg1, arg2, argm, sinrota
     &             , cosrota, pi, angle
       real*8        fun(nxcube,nycube)
       parameter   ( pi = 3.14159265358979d0, as2deg = 1.0d0 / 3600.0d0, almostzero = 1.0d-30 )
       external      ftopen, ftgkne, ftghpr, ftgkys, ftgkyd, ftgkye, ftg2de, ftclos, printerr, ftgkyj, ftg2dd, ftgsvd
!__________________________________________________________________________________________________________________________________
!
! Initialize variables.

       stat = 0
       unit   = 20
       group  = 1
       rw     = 0
       stat05 = 0
       stat06 = 0
       stat07 = 0
       stat08 = 0
       stat09 = 0
       stat10 = 0
       stat11 = 0
       stat12 = 0
       stat13 = 0
       stat14 = 0
       stat15 = 0
       stat16 = 0
       stat17 = 0
       stat18 = 0
       stat19 = 0
       stat20 = 0
       stat21 = 0
       stat22 = 0
       stat23 = 0
       stat24 = 0
       stat25 = 0
       stat26 = 0
       stat27 = 0
       stat28 = 0
       stat29 = 0
       stat30 = 0
       stat31 = 0
       stat32 = 0

! Open the FITS file.

       call ftopen ( unit, fname, rw, blksize, stat )

       if (stat .ne. 0) then
         if (iotty .gt. 0) write (iotty,'(/a)') '   RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         if (iolog .gt. 0) write (iolog,'(/a)') '   RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         stop 99
       endif

! Determine the size of the image.

       call ftgkne ( unit, 'NAXIS', 1, 2, naxes, nfound, stat )

! Check that it found both NAXIS1 and NAXIS2 keywords.

       if (nfound .ne. 2) then
         if (iotty .gt. 0) write (iotty,'(/a)') '   RFITS: ERROR: Failed to read the NAXISn keywords.'
         if (iolog .gt. 0) write (iolog,'(/a)') '   RFITS: ERROR: Failed to read the NAXISn keywords.'
         stop 99
       endif

! Get the main primary header keywords which define the array structure.

!!       call ftgphx ( unit, maxdim, simple, bitpix, naxis, naxes, pcount, gcount, extend, bscale, bzero, blank, nblank, stat )

       call ftghpr ( unit, maxdim, simple, bitpix, naxis, naxes, pcount, gcount, extend, stat )

! Read all parameters.

       call ftgkys ( unit, 'CREATOR', creator, comment, stat )
       
       if (stat .eq. 0 .and. (comment(1:21) .eq. 'Alexander Menshchikov' .or. comment(1:22) .eq. 'Alexander Men''shchikov')) then
         myfile = .true.
       else
         myfile = .false.
         stat = 0
         creator = 'MAKECUBE'
       endif

! I don't understant why blank is assigned to my array values (if blank is not zero) - I have no undefined values in arrays. Maybe
! there is a bug in the version of FITSIO I use. For now I will ignore the blank value when reading FITS files.

       blank = 0

       datamx = 0.0d0
       datamn = 0.0d0
       bunit = ' '
       ctype1 = ' '
       ctype2 = ' '
       rp1 = 0.0d0
       rp2 = 0.0d0
       rot1 = 0.0d0
       rot2 = 0.0d0
       cd11 = 0.0d0
       cd12 = 0.0d0
       cd21 = 0.0d0
       cd22 = 0.0d0
       xr = 0.0d0
       yr = 0.0d0
       dx = 0.0d0
       dy = 0.0d0
       object = ' '
       ra = 0.0d0
       dec = 0.0d0
       epoch = 0.0d0
       equinox = 0.0d0
       beam = 0.0d0
       bzero = 0.0d0
       bscale = 1.0d0
       wave = 1.0d2
       history = ' '

       call ftgkys ( unit, 'DATE', cdatetime , comment, stat05 )
       call ftgkyd ( unit, 'DATAMAX', datamx , comment, stat06 )
       call ftgkyd ( unit, 'DATAMIN', datamn , comment, stat07 )
       call ftgkyj ( unit, 'BLANK'  , blank  , comment, stat08 )
       call ftgkys ( unit, 'BUNIT'  , bunit  , comment, stat09 )
       call ftgkyd ( unit, 'BZERO'  , bzero  , comment, stat10 )
       call ftgkyd ( unit, 'BSCALE' , bscale , comment, stat11 )
       call ftgkys ( unit, 'CTYPE1' , ctype1 , comment, stat12 )
       call ftgkys ( unit, 'CTYPE2' , ctype2 , comment, stat13 )
       call ftgkyd ( unit, 'CRPIX1' , rp1    , comment, stat14 )
       call ftgkyd ( unit, 'CRPIX2' , rp2    , comment, stat15 )
       call ftgkyd ( unit, 'CROTA1' , rot1   , comment, stat16 )
       call ftgkyd ( unit, 'CROTA2' , rot2   , comment, stat17 )
       call ftgkyd ( unit, 'CD1_1'  , cd11   , comment, stat18 )
       call ftgkyd ( unit, 'CD1_2'  , cd12   , comment, stat19 )
       call ftgkyd ( unit, 'CD2_1'  , cd21   , comment, stat20 )
       call ftgkyd ( unit, 'CD2_2'  , cd22   , comment, stat21 )
       call ftgkyd ( unit, 'CRVAL1' , xr     , comment, stat22 )
       call ftgkyd ( unit, 'CRVAL2' , yr     , comment, stat23 )
       call ftgkyd ( unit, 'CDELT1' , dxdeg  , comment, stat24 )
       call ftgkyd ( unit, 'CDELT2' , dydeg  , comment, stat25 )
       call ftgkys ( unit, 'OBJECT' , object , comment, stat26 )
       call ftgkyd ( unit, 'RA'     , ra     , comment, stat27 )
       call ftgkyd ( unit, 'DEC'    , dec    , comment, stat28 )
       call ftgkyd ( unit, 'EPOCH'  , epoch  , comment, stat29 )
       call ftgkyd ( unit, 'EQUINOX', equinox, comment, stat30 )
       call ftgkyd ( unit, 'WAVE'   , wave   , comment, stat31 )
       call ftgkyd ( unit, 'BEAM'   , beam   , comment, stat )
       call ftgkys ( unit, 'HISTORY', history, history, stat32 )

       if (stat .ne. 0) then
         beam = 0.0d0
         stat = 0
       endif

       stat0 = stat05 + stat06 + stat07 + stat09 + stat10 + stat11 + stat12 + stat13 + stat14 + stat15 + stat16 + stat17 + stat18
     &       + stat19 + stat20 + stat21 + stat22 + stat23 + stat24 + stat25 + stat26 + stat27 + stat28 + stat30 + stat31

       if (cverbose .ne. '-verb0') then
         if (stat0  .gt. 0) write (iotty,'()')
         if (stat05 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATE absent or error'
         if (stat06 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMAX absent or error'
         if (stat07 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMIN absent or error'
         if (stat09 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BUNIT absent or error'
         if (stat10 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BZERO absent or error'
         if (stat11 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BSCALE absent or error'
         if (stat12 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE1 absent or error'
         if (stat13 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE2 absent or error'
         if (stat14 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX1 absent or error'
         if (stat15 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX2 absent or error'
         if (stat16 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA1 absent or error'
         if (stat17 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA2 absent or error'
         if (stat18 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_1 absent or error'
         if (stat19 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_2 absent or error'
         if (stat20 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_1 absent or error'
         if (stat21 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_2 absent or error'
         if (stat22 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL1 absent or error'
         if (stat23 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL2 absent or error'
         if (stat24 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT1 absent or error'
         if (stat25 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT2 absent or error'
         if (stat26 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword OBJECT absent or error'
         if (stat27 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword RA absent or error'
         if (stat28 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DEC absent or error'
         if (stat30 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword EQUINOX absent or error'
         if (stat31 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword WAVE absent or error'
       endif
       if (stat26 .ne. 0 .or. object(1:10) .eq. '          ') object = '<Name>'

! Read a FITS file into a 2D array.

!!       call ftgsvd ( unit, group, naxis, naxes, fpixels, lpixels, incs, dble ( blank ), fun, anynull, stat )
!!       call ftg3dd ( unit, group, dble ( blank ), nxmx, nymx, naxes(1), naxes(2), naxes(3), fun, anynull, stat )

       call ftg2dd ( unit, group, dble ( blank ), naxes(1), naxes(1), naxes(2), fun, anynull, stat )

       if (stat .ne. 0) write (iotty,'(a)') ' Error reading FITS file'

! Close the file and free the unit number.

       call ftclos ( unit, stat )

! Check for any error, and if so print out error messages.

       if (stat .gt. 0) call printerr ( stat )

       if (stat17 .ne. 0) then
         rot2 = 0.0d0
       endif
       rot1 = rot2

       if (stat29 .eq. 0 .and. stat30 .ne. 0) then
         equinox = epoch
       endif

! CDELT* is absent or error.

       if (stat24 .ne. 0 .or. stat25 .ne. 0) then

! CD1* and CD2* are defined correctly: determine CDELT* and ROTA*.

         if (stat18 .eq. 0 .and. stat19 .eq. 0 .and. stat20 .eq. 0 .and. stat21 .eq. 0) then
           sign1 = sign ( 1.0d0, cd11 * cd22 - cd12 * cd21 )
           dxdeg = sign1 * sqrt ( cd11**2 + cd21**2 )
           dydeg = sqrt ( cd12**2 + cd22**2 )
           arg1 = sign1 * cd12 / cd22
           arg2 = -sign1 * cd21 / cd11
           argm = (arg1 + arg2) / 2.0d0
           if (abs ( arg1 - arg2 ) / argm .gt. 1.0d-10) then
             write (iotty,'(/a/)') '   RFITS: WARNING: Coordinate axes are not orthogonal'
           endif
           rot2 = atan ( argm )
           sinrota = sin ( rot2 )
           cosrota = cos ( rot2 )
           rot2 = rot2 * 180.0d0 / pi
           if (cosrota .ge. 0.0d0 .and. sinrota .ge. 0.0d0) then
             rot1 = rot2
           elseif (cosrota .lt. 0.0d0 .and. sinrota .ge. 0.0d0) then
             rot1 = 180.0d0 - rot2
           elseif (cosrota .lt. 0.0d0 .and. sinrota .lt. 0.0d0) then
             rot1 = 180.0d0 + rot2
           elseif (cosrota .ge. 0.0d0 .and. sinrota .lt. 0.0d0) then
             rot1 = - rot2
           endif
           rot2 = rot1
         endif
       endif

! CD1* and CD2* are absent or error.

       if (stat18 .ne. 0 .or. stat19 .ne. 0 .or. stat20 .ne. 0 .or. stat21 .ne. 0) then

! CDELT* are defined correctly: determine CD1* and CD2*.

         if (stat24 .eq. 0 .and. stat25 .eq. 0) then
           angle = rot2 * pi / 180.0d0
           cd11 = dxdeg * cos ( angle )
           cd12 = abs ( dydeg ) * sign ( 1.0d0, dxdeg ) * sin ( angle )
           cd21 = - abs ( dxdeg ) * sign ( 1.0d0, dydeg ) * sin ( angle )
           cd22 = dydeg * cos ( angle )
           if (abs ( cd12 ) .lt. almostzero) cd12 = 0.0d0
           if (abs ( cd21 ) .lt. almostzero) cd21 = 0.0d0
         endif
       endif

       dx = abs ( dxdeg ) / as2deg
       dy = abs ( dydeg ) / as2deg

! Print out the min and max values.

       funmin = 1.0d30
       funmax =-1.0d30
       cnan = 'NAN'
       do j=1,naxes(2)
         do i=1,naxes(1)
           if (.not.isnan ( fun(i,j) )) then  
             funmin = min ( funmin, fun(i,j) )
             funmax = max ( funmax, fun(i,j) )
           endif  
           if (int ( fun(i,j) ) .eq. blank .and. blank .ne. 0) then  
             read (cnan,*) fun(i,j)
           endif
         enddo
       enddo
       
       if (iotty .gt. 0) write (iotty,'(a,1x,2(1pe14.7,1x))') '   RFITS: Minmax values:', funmin, funmax
       if (iolog .gt. 0) write (iolog,'(a,1x,2(1pe14.7,1x))') '   RFITS: Minmax values:', funmin, funmax

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine wfits 
       
     &            ( cfitsversion, nx, ny, nz, k, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname, cdate
     &            , ctime, creator, beam, blank, rot1, rot2, cd11, cd12, cd21, cd22, equinox, bzero, bscale, wave, datamn, datamx
     &            , history, unit )
!__________________________________________________________________________________________________________________________________
!
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       extend, simple
       character*(*) fname, ctype1, ctype2, object, creator, ctime, cdate, bunit, cfitsversion, history
       character*4   ctype3
       character*19  cdatetime
       integer       i, j, k, nx, ny, nz, status, blocksze, group, naxis, pcount, gcount, bitpix, unit, blank, maxdim
       parameter   ( maxdim = 3 )
       integer       naxes(maxdim), fpixels(maxdim), lpixels(maxdim)
       real*8        xr, yr, dx, dy, rp1, rp2, ra, dec, beam, bzero, bscale, as2deg, rot1, rot2, equinox, datamn !!, rot3, zr
     &             , datamx, dxdeg, dydeg, wave, cd11, cd12, cd21, cd22, almostzero, pi  !!, dzdeg, rp3
       real*8        fun(nx,ny,1)
       parameter   ( pi = 3.14159265358979d0, as2deg = 1.0d0 / 3600.0d0, almostzero = 1.0d-30 )
       external      fdelete, ftgiou, ftinit, ftphpr, ftpkys, ftpcom, ftpkyd, ftpkyj, ftphis, ftpssd, ftclos, ftfiou, printerr
!__________________________________________________________________________________________________________________________________
!
! Create a FITS primary array containing a 3D image. Initialize parameters about the FITS image.

       status   = 0
       blocksze = 1
       simple   = .true.
       extend   = .false.
       group    = 1
       naxis    = 3
       pcount   = 0
       gcount   = 1
       naxes(1) = nx
       naxes(2) = ny
       naxes(3) = nz
       ctype3 = 'VRAD'
!!       rot3 = rot2
!!       rp3 = 1.0d0
!!       zr = 1.0d0

! Compute offset and scaling factor so as to represent real numbers most efficiently and accurately with given 'bitpix'.

       do j=1,ny
         do i=1,nx
           if (.not.isnan ( fun(i,j,1) )) then
             if (fun(i,j,1) .lt. datamn) datamn = fun(i,j,1)
             if (fun(i,j,1) .gt. datamx) datamx = fun(i,j,1)
           endif
         enddo
       enddo
       
       if (k .eq. 1) then
         bitpix = -32     !<-- single precision floating-point values
         bzero  = 0.0d0
         bscale = 1.0d0
         blank = blank
  
         cdatetime = cdate(7:10)//'-'//cdate(4:5)//'-'//cdate(1:2)//'T'//ctime

! Delete the file if it already exists.

         if (cfitsversion .eq. '5.030') then
           call fdelete ( fname, status )
         endif

! Get an unused Logical Unit Number to use to open the FITS file.

         call ftgiou ( unit, status )

! Create the new empty FITS file.

         if (cfitsversion .eq. '5.030') then
           call ftinit ( unit, fname, blocksze, status )
         else
           call ftinit ( unit, '!'//fname, blocksze, status )
         endif

! Write the required header keywords.

         call ftphpr ( unit, simple, bitpix, naxis, naxes, pcount, gcount, extend, status )

! Write an optional comment.

         call ftpkys ( unit, 'CREATOR', creator,'Alexander Menshchikov, DAp IRFU CEA Saclay', status )
         call ftpcom ( unit,' ',status)
         call ftpkys ( unit, 'DATE'  , cdatetime, 'creation date and time', status )
         call ftpkyd ( unit, 'BZERO'  , bzero , 13 , 'zero point in scaling equation', status )
         call ftpkyd ( unit, 'BSCALE' , bscale, 13 , 'linear factor in scaling equation' , status )
       endif

! Write the array to the FITS file.

       fpixels(1) = 1
       lpixels(1) = naxes(1)
       fpixels(2) = 1
       lpixels(2) = naxes(2)
       fpixels(3) = k
       lpixels(3) = fpixels(3)
       
! Write a slice from the FITS file into a 3D array.

       call ftpssd ( unit, group, naxis, naxes, fpixels, lpixels, fun, status )
       
! Write all parameters.

       if (k .eq. nz) then
         dxdeg = - dx * as2deg
         dydeg = dy * as2deg
!!         dzdeg = 1.0d0 * as2deg
         
         call ftpkyd ( unit, 'DATAMAX', datamx, 13 , 'maximum data value', status )
         call ftpkyd ( unit, 'DATAMIN', datamn, 13 , 'minimum data value', status )
!!         if  (blank .gt. 0)
!!       & call ftpkyj ( unit, 'BLANK'  , blank      , 'value used for undefined array elements', status )
         call ftpkys ( unit, 'BUNIT'  , bunit      , 'physical units of the array values', status )
         call ftpkys ( unit, 'CTYPE1' , ctype1     , 'name of the coordinate axis', status )
         call ftpkys ( unit, 'CTYPE2' , ctype2     , 'name of the coordinate axis', status )
         call ftpkys ( unit, 'CTYPE3' , ctype3     , 'name of the coordinate axis', status )
         call ftpkyd ( unit, 'CRPIX1' , rp1   , 13 , 'coordinate system reference pixel', status )
         call ftpkyd ( unit, 'CRPIX2' , rp2   , 13 , 'coordinate system reference pixel', status )
!!         call ftpkyd ( unit, 'CRPIX3' , rp3   , 13 , 'coordinate system reference pixel', status )
         call ftpkyd ( unit, 'CROTA1' , rot1  , 13 , 'coordinate system rotation angle', status )
         call ftpkyd ( unit, 'CROTA2' , rot2  , 13 , 'coordinate system rotation angle', status )
!!         call ftpkyd ( unit, 'CROTA3' , rot3  , 13 , 'coordinate system rotation angle', status )
         call ftpkyd ( unit, 'CD1_1'  , cd11  , 13 , 'linear projection matrix', status )
         call ftpkyd ( unit, 'CD1_2'  , cd12  , 13 , 'linear projection matrix', status )
         call ftpkyd ( unit, 'CD2_1'  , cd21  , 13 , 'linear projection matrix', status )
         call ftpkyd ( unit, 'CD2_2'  , cd22  , 13 , 'linear projection matrix', status )
         call ftpkyd ( unit, 'CRVAL1' , xr    , 13 , 'coordinate value at reference pixel', status )
         call ftpkyd ( unit, 'CRVAL2' , yr    , 13 , 'coordinate value at reference pixel', status )
!!         call ftpkyd ( unit, 'CRVAL3' , zr    , 13 , 'coordinate value at reference pixel', status )
         call ftpkyd ( unit, 'CDELT1' , dxdeg , 13 , 'coordinate increment along axis', status )
         call ftpkyd ( unit, 'CDELT2' , dydeg , 13 , 'coordinate increment along axis', status )
!!         call ftpkyd ( unit, 'CDELT3' , dzdeg , 13 , 'coordinate increment along axis', status )
         call ftpkys ( unit, 'OBJECT' , object     , 'object identifier', status )
         call ftpkyd ( unit, 'RA'     , ra     , 13, 'right ascension', status )
         call ftpkyd ( unit, 'DEC'    , dec    , 13, 'declination', status )
         call ftpkyd ( unit, 'EQUINOX', equinox, 13, 'equinox', status )
         call ftpkyd ( unit, 'WAVE'   , wave   , 13, 'wavelength (microns)', status )
         call ftphis ( unit, history, status )
        
         if (beam .ne. 0.0d0) call ftpkyd ( unit, 'BEAM', beam, 13, 'convolved with the beam (arcsec)', status )

! Close the file.

         call ftclos ( unit, status )

! Free the unit number.

         call ftfiou ( unit, status )
       endif

! Check for any error, and if so print out error messages

       if (status .gt. 0) call printerr ( status )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||