!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       program smeasure
!__________________________________________________________________________________________________________________________________
!
! Measure fluxes and shapes for each identified source.
!
! GETSF • Multi-Scale Multi-Wavelength Source & Filament Extraction • Alexander Men'shchikov, DAp IRFU CEA Saclay
!
! Return codes:
!
!   10-29 ... problems with the input parameters
!   30-49 ... non-existing files or directories
!   50-69 ... problems with reading/writing files or values
!   70-89 ... problems with memory allocation
!   90-99 ... other code-specific problems
!__________________________________________________________________________________________________________________________________
!
       implicit      none

       integer       iotty, ncolmax

       parameter   ( ncolmax = 100, iotty = 6 )

       logical       inellipse, lfnam1, lfnam2, lfnam3, lfnam4, lfnam5, lfnam7, lfnam9, lunix, lfirstcall, ldevelopment, lfnwcs
     &             , liter_phase1, liter_phase2, liter_phase3, lfinishing_mode, lann1usable, lann2usable, lann3usable, ltiming
     &             , ltime, lonlypositives, lnoexpansion, lnoshrinkage, laccelerate, lexist
       
       logical     , allocatable :: lfootprint(:), lsourcetodeblend(:), lnotbad(:)

       character*1   capcorr
       character*3   cskipflg, cskipfluxes, cskipcoords, cskipthetam, cflg, dot
       character*5   cn5, cnextrbad, cnextrgood, coosystem
       character*6   cfitsversion, cverbose, czno
       character*7   clibname, cx, cy
       character*8   ctime, cfpptime
       character*10  cdate, fmtwcs
       character*21  compda, intensunits
       character*40  cdistpc, cbeam, cowave
       character*80  object, creator, ctype1, ctype2, bunit, history
       character*128 csnt, cnwd, csnr, cfto
       character*500 dirmono, icatname, ocatname, obsname, obsfsname, obsbsname, obscbname, arg1, arg2, arg3, arg4, arg5, arg6
     &             , arg7, arg8, arg9, argA, argB, argC, argD, argE, argF, maskname, originpath, wcsfile, filewithnames
     &             , annulusname
       
       character*1 , allocatable :: flag(:), badsource(:)
       character*5 , allocatable :: cflm(:)
       character*12000, allocatable :: catline(:), headline(:)

       integer       firstb, lastc, lasts, i, ic, nx, ny, ia1, ia2, ia3, ia4, ia5, ia6, ia7, ia8, ia9, iaA, iaB, iaC, iaD, iaE
     &             , l, m, n, nhead, nextr, iicat, iocat, blank, icp, icb, iu, id, ncolumn, nobs, iaF, irc, nfg
     &             , nfluxes, nxcopix, npeakfx, ndate, idot2, idot3, idot4, ibad, igood, nbad, isw, iobs, iobsfs, ncolumn1, k
     &             , isnt, isnr, ian, nsx, inwd, l1, l2, m1, m2, nr, imsk, ifdn, iopa, icw, ica, idmon, idot, lb, iann
     &             , ir, ntry, n1ymin, n2ymax, iwcs, l1x, ifto, l2x, m1x, m2x, nwavesdetect, wchours, wcmins, wcsecs, cpuhours
     &             , cpumins, cpusecs, ibav1, ibav2, ibrms1, ibrms2, nrz, l1z, l2z, m1z, m2z, ibav3, ibrms3, ngood, iobsfsbs
     &             , nthetam, ipcat, iobsfscb, ncx, ncy
       
       integer     , allocatable :: flg(:), flm(:), n1x(:), n2x(:), nxco(:), nyco(:), nkover(:), nkoverlap(:,:)

       real          fitsvers

       real*8        timer, dx, dy, wave, crpix1, crpix2, crval1, crval2, ra, dec, beamd, funmin, funmax, crota1, crota2
     &             , bzero, bscale, datamin, datamax, pixel, beam, sq_arcsecs_per_sterad, pi, sq_arcsecs_per_beam_area, convftot
     &             , sqpixel, sqreightlogtwo, convtodegrees, dbll, dblm, sigmarms, factor, factoffset, stentative_nouse, owave, xmu
     &             , sreliable_nouse, convtoradians, major1, minor1, major2, minor2, areapix1, areapix2, thickpix, rad2annit, dlog4h
     &             , rad2min, radmax, semimajor1, semiminor1, semimajor2, semiminor2, factortotal, almostzero, newpixval, au2cm
     &             , aaa, bbb, ccc, sterad_per_beam_area, cd11, cd12, cd21, cd22, annulusareait, deficit, major1a, equinox, minor1a
     &             , semimajor1a, semiminor1a, semimajor2a, semiminor2a, major2a, minor2a, absobsfsatpos, rad2, xMsun, distpc, xmH
     &             , convftotm

       real*8        cpu_initialize, wal_initialize, cpu_iteratepeaks, wal_iteratepeaks, cpu_secondmoments, wal_secondmoments
     &             , cpu_totalfluxes, wal_totalfluxes, cpu_fittedfluxes, wal_fittedfluxes, cpu_createannuli, wal_createannuli
     &             , cpu_significances, wal_significances, cpu_fluctuations1, cpu_fluctuations2, cpu_fluctuations3, cpu_fwhmsizes
     &             , cpu_nearestannulus, cpu_secondannulus, cpu_extendannulus, wal_fluctuations1, cput_fwhmsizes
     &             , wal_fluctuations2, wal_fwhmsizes, wal_fluctuations3, wal_nearestannulus, wal_secondannulus
     &             , wal_extendannulus, walt_fwhmsizes, cpu_footprints, wal_footprints, cpu_testaccel, wal_testaccel
     
       real*8        cput_initialize, walt_initialize, cput_iteratepeaks, walt_iteratepeaks, cput_secondmoments, walt_secondmoments
     &             , cput_totalfluxes, walt_totalfluxes, cput_fittedfluxes, walt_fittedfluxes, cput_createannuli, walt_createannuli
     &             , cput_significances, walt_significances, cpu_sum, wal_sum, cput_fluctuations1, cput_fluctuations2, cputot
     &             , cput_fluctuations3, walt_fluctuations1, walt_fluctuations2, walt_fluctuations3, wctot, cput_nearestannulus
     &             , cput_secondannulus, cput_extendannulus, walt_nearestannulus, walt_secondannulus, walt_extendannulus
     &             , cput_footprints, walt_footprints, cput_testaccel, walt_testaccel

       real*8      , allocatable :: obsfsimage(:,:), obsfsbsimage(:,:), obsfscbimage(:,:), momxco(:), momyco(:), afwhm(:), bfwhm(:)
     &             , theta(:), fluxpeakbest(:), fluxtotfit(:), ccd(:), fluxtotbest(:), equivrad(:), sigmono(:), scalensx(:)
     &             , sigglob(:), apcorrect(:), backrms(:), fluxperror(:), fluxterror(:), dxcon2(:,:), backrms3(:), dycon2(:,:)
     &             , footfact(:), foomajor(:), momxx(:), momyy(:), momxy(:), foominor(:), maskimage(:,:), bigfoot(:,:)
     &             , smallfoot(:,:), good(:), annulus(:,:), wcsxco(:), wcsyco(:), bigfootclone(:,:), goodmono(:)
     &             , aax(:), bbx(:), ccx(:), aad(:), bbd(:), backav1(:), backav2(:), backav3(:), slope(:), backrms1(:), backrms2(:)
     &             , annulusarea(:), dist2pks(:,:), dist2foo(:,:), asize(:), bsize(:), momxxh(:), momyyh(:), momxyh(:)
     &             , footfacto(:), foomajoro(:), foominoro(:), thetao(:), convdelta(:,:)

       parameter   ( sq_arcsecs_per_sterad = 3282.80635d0 * 3600.0d0**2, pi = 3.14159265358979d0, convtodegrees = 180.0d0 / pi 
     &             , convtoradians = pi / 180.0d0, dlog4h = log ( 4.0d0 ) / 2.0d0, factor = 1.0d0 / 2.0d0**2 / dlog4h
     &             , almostzero = 1.0d-30, factoffset = 1.05d0, ltiming = .false., dot = '•', newpixval = 1.0d0, ntry = 20 )

       parameter   ( xmu = 2.8, xmH = 1.6726E-24, au2cm = 1.495978707e13, xMsun = 1.98892e+33 )   !!, pc2cm = 3.085678e+18
       
       external      firstb, lastc, lasts, timer, inellipse, rfits, wfits, osystem, when, getfitshead, ftvers, skipcomm
     &             , linescolumns, auxiliaryarrays, testacceleration, monosignificances, deblendpeaks, intensmoments, momentsizes
     &             , fitshapesfluxes, integratedfluxes, createannuli, expandfootprints, interpolation
!__________________________________________________________________________________________________________________________________
!                                     
       lnoexpansion = .false.
       sqreightlogtwo = sqrt ( 8.0d0 * log ( 2.0d0 ) )
       
! Determine operating system

       call osystem ( lunix )
       call when ( lunix, ctime, cdate, ndate, 4 )
       call ftvers ( fitsvers )
       lb = 1
       write (cfitsversion,'(f6.3)') fitsvers
       if (cfitsversion .eq. ' 5.030') then
         lb = 2
       endif
       clibname = 'CFITSIO'
       cfpptime = __TIME__
       cfpptime = cfpptime(1:5)
       compda = __DATE__//' '//cfpptime

! Get command line parameters (file names).

       call getarg (  1, arg1 )
       call getarg (  2, arg2 )
       call getarg (  3, arg3 )
       call getarg (  4, arg4 )
       call getarg (  5, arg5 )
       call getarg (  6, arg6 )
       call getarg (  7, arg7 )
       call getarg (  8, arg8 )
       call getarg (  9, arg9 )
       call getarg ( 10, argA )
       call getarg ( 11, argB )
       call getarg ( 12, argC )
       call getarg ( 13, argD )
       call getarg ( 14, argE )
       call getarg ( 15, argF )

       ia1 = lastc ( arg1 )
       ia2 = lastc ( arg2 )
       ia3 = lastc ( arg3 )
       ia4 = lastc ( arg4 )
       ia5 = lastc ( arg5 )
       ia6 = lastc ( arg6 )
       ia7 = lastc ( arg7 )
       ia8 = lastc ( arg8 )
       ia9 = lastc ( arg9 )
       iaA = lastc ( argA )
       iaB = lastc ( argB )
       iaC = lastc ( argC )
       iaD = lastc ( argD )
       iaE = lastc ( argE )
       iaF = lastc ( argF )

       cverbose = '-verb2'
       if (iaF .gt. 0 .and. argF(1:iaF-1) .eq. '-verb') cverbose = argF(1:iaF)
       if (iaE .gt. 0 .and. argE(1:iaE-1) .eq. '-verb') cverbose = argE(1:iaE)
       if (iaD .gt. 0 .and. argD(1:iaD-1) .eq. '-verb') cverbose = argD(1:iaD)

       if (cverbose .eq. '-verb2') then
         write (iotty,'( )')
         if (arg1(1:ia1) .eq. ':') write (iotty,'(a$)') ' '
         write (iotty,'(a$)') ' SMEASURE '//dot//' Measure & Catalog Sources '//dot//' '//compda
       endif
       if (arg1(1:ia1) .eq. ':') stop
       if (cverbose .eq. '-verb2') write (iotty,'()')
       if (cverbose .eq. '-verb2') write (iotty,'( a)') ' Alexander Men’shchikov, DAp IRFU CEA Saclay, France.'
       if (cverbose .eq. '-verb2') write (iotty,'( a)') ' Using '//clibname(lb:7)//' library version'//cfitsversion
     &                                                //' by William D Pence.'
       if (ia9 .eq. 0) then
         write (iotty,'(/a)') ' USAGE: smeasure <distpc> <owave> <beam> <apcorr> <snt> <snr> <nwdetect> <factortotal>'
     &                      //' <originpath> <dirmono> <obsname> <icatname> [-o <ocatname>] [-verb{0|1|2}]'
         stop 11
       endif

       icp  = ia1
       icw  = ia2
       icb  = ia3
       ica  = ia4
       isnt = ia5
       isnr = ia6
       inwd = ia7
       ifto = ia8
       iopa = ia9
       cdistpc = arg1(1:icp)
       cowave  = arg2(1:icw)
       cbeam   = arg3(1:icb)
       capcorr = arg4(1:ica)
       csnt    = arg5(1:isnt)
       csnr    = arg6(1:isnr)
       cnwd    = arg7(1:inwd)
       cfto    = arg8(1:ifto)
       originpath = arg9(1:iopa)
       areapix1 = 0.0d0
       areapix2 = 0.0d0
       
       read (cdistpc(1:icp),*,end=11,err=11) distpc    !<-- distance (pc)
       read (cowave(1:icw),*,end=12,err=12) owave      !<-- wavelength (microns)
       read (cbeam(1:icb), *,end=13,err=13) beam       !<-- beam size (arcsec)
       read (csnt(1:isnt),*,end=4 ,err=4 ) stentative_nouse
       read (csnr(1:isnr),*,end=5 ,err=5 ) sreliable_nouse
       read (cnwd(1:inwd),*,end=6 ,err=6 ) nwavesdetect
       read (cfto(1:ifto),*,end=63,err=63) factortotal

       lonlypositives = .false.
       if (cowave(1:1) .eq. '+') then
         lonlypositives = .true.
         cowave = cowave(2:icw)//' '
         icw = lastc ( cowave )
       endif

       liter_phase1 = .false.
       if (csnt(1:1) .eq. '+') then
         liter_phase1 = .true.
       endif
       liter_phase2 = .false.
       if (csnr(1:1) .eq. '+') then
         liter_phase2 = .true.
       endif
       liter_phase3 = .false.
       if (cnwd(1:1) .eq. '+' .or. cnwd(1:1) .eq. '-') then
         liter_phase3 = .true.
       endif
       lfinishing_mode = .false.
       if (cnwd(1:1) .eq. '-') then
         lfinishing_mode = .true.
       endif
       ldevelopment = .false.
       if (lfinishing_mode .and. csnr(1:1) .eq. '+') then
         ldevelopment = .true.
       endif

       if (csnt(1:1) .eq. '+') then
         csnt = csnt(2:isnt)//' '
       endif
       if (csnt(1:1) .eq. '-') then
         stentative_nouse = abs ( stentative_nouse )
         csnt = csnt(2:isnt)//' '
       endif
       if (csnr(1:1) .eq. '+') then
         csnr = csnr(2:isnr)//' '
       endif
       if (csnr(1:1) .eq. '-') then
         sreliable_nouse = abs ( sreliable_nouse )
         csnr = csnr(2:isnr)//' '
       endif
       if (cnwd(1:1) .eq. '+') then
         cnwd = cnwd(2:inwd)//' '
       endif
       if (cnwd(1:1) .eq. '-') then
         nwavesdetect = abs ( nwavesdetect )
         cnwd = cnwd(2:inwd)//' '
       endif

       isnt = lastc ( csnt )
       isnr = lastc ( csnr )
       inwd = lastc ( cnwd )

       dirmono = argA(1:iaA)
       obsname = argB(1:iaB)
       icatname = argC(1:iaC)
       idmon = lastc ( dirmono )
       iobs = lastc ( obsname )
       iicat = lastc ( icatname  )

       isw = index ( icatname(1:iicat), '.det.cat' )
       lfirstcall = .false.
       if (isw .gt. 0) lfirstcall = .true.

       if (ltiming) then
         inquire ( file='+timing.smeasure', exist=ltime )
         open ( 60, file='+timing.smeasure', status='unknown' )
         if (ltime) then
           do 
             read (60,'()',end=61)
           enddo
  61       backspace ( 60 )
         endif
         if (lfirstcall) then
           write (60,'(/a)') ' SMEASURE started: '//cdate//' '//ctime//' Image: '//obsname(1:iobs)//' <~ 1ST ITERATION'
         else
           write (60,'(/a)') ' SMEASURE started: '//cdate//' '//ctime//' Image: '//obsname(1:iobs)
         endif
         write (60,'(a)') ' ------------------------------------------------------'
         
         cput_initialize = 0.0d0
         walt_initialize = 0.0d0
         cput_iteratepeaks = 0.0d0
         walt_iteratepeaks = 0.0d0
         cput_secondmoments = 0.0d0
         walt_secondmoments = 0.0d0
         cput_fwhmsizes = 0.0d0
         walt_fwhmsizes = 0.0d0
         cput_totalfluxes = 0.0d0
         walt_totalfluxes = 0.0d0
         cput_fittedfluxes = 0.0d0
         walt_fittedfluxes = 0.0d0
         cput_createannuli = 0.0d0
         walt_createannuli = 0.0d0
         cput_fluctuations1 = 0.0d0
         cput_fluctuations2 = 0.0d0
         cput_fluctuations3 = 0.0d0
         cput_nearestannulus = 0.0d0
         cput_secondannulus = 0.0d0
         cput_extendannulus = 0.0d0
         walt_fluctuations1 = 0.0d0
         walt_fluctuations2 = 0.0d0
         walt_fluctuations3 = 0.0d0
         walt_nearestannulus = 0.0d0
         walt_secondannulus = 0.0d0
         walt_extendannulus = 0.0d0
         cput_significances = 0.0d0
         walt_significances = 0.0d0
         cput_footprints = 0.0d0
         walt_footprints = 0.0d0
         cput_testaccel = 0.0d0
         walt_testaccel = 0.0d0
         cpu_sum = 0.0d0
         wal_sum = 0.0d0
       endif

       goto 15
  4    write (iotty,'(/a)') '   SMEASURE: ERROR: Trouble reading STENTATIVE: '//csnt(1:isnt)
       stop 51
  5    write (iotty,'(/a)') '   SMEASURE: ERROR: Trouble reading SRELIABLE: '//csnr(1:isnr)
       stop 52 
  6    write (iotty,'(/a)') '   SMEASURE: ERROR: Trouble reading NWAVESDETECT: '//cnwd(1:inwd)
       stop 53
 63    write (iotty,'(/a)') '   SMEASURE: ERROR: Trouble reading FACTORTOTAL: '//cfto(1:ifto)
       stop 54
 11    write (iotty,'(/a)') '   SMEASURE: ERROR: Trouble reading DISTANCE: '//cdistpc(1:icp)
       stop 55
 12    write (iotty,'(/a)') '   SMEASURE: ERROR: Trouble reading OBSWAVE: '//cowave(1:icw)
       stop 56
 13    write (iotty,'(/a)') '   SMEASURE: ERROR: Trouble reading BEAMSIZE: '//cbeam(1:icb)
       stop 57
 15    continue
       if (capcorr .ne. 'y' .and. capcorr .ne. 'n') then
         write (iotty,'(/a)') '   SMEASURE: ERROR: Invalid aperture correction parameter: '//capcorr
         stop 59
       endif
 
       sq_arcsecs_per_beam_area = pi * (beam / 2.0)**2 / log ( 2.0d0 )
       sterad_per_beam_area = sq_arcsecs_per_beam_area / sq_arcsecs_per_sterad

       iwcs = 0
       if (lfirstcall) then
         idot = index ( icatname(1:iicat), '.' )
         idot2 = index ( icatname(idot+1:iicat), '.' ) + idot
         idot3 = index ( icatname(idot2+1:iicat), '.' ) + idot2
         wcsfile = icatname(1:idot3-1)//'.det.wcs'
         iwcs = lastc ( wcsfile )
       endif

       ipcat = index ( icatname(1:iicat), '.cat' )
       inquire ( file=icatname(1:ipcat-1)//'.previous.cat', exist=laccelerate )

! Check if the observed filament-subtracted image exists.

       inquire ( file=obsname(1:iobs), exist=lfnam1 )
       if (.not.lfnam1 .and. obsname(iobs-4:iobs) .ne. '.fits') then
         obsname = obsname(1:iobs)//'.fits'
         iobs = lastc ( obsname )
         inquire ( file=obsname(1:iobs), exist=lfnam1 )
       endif
       
! The image OBSFSNAME is used only in calculation of standard deviations in annuli around sources.
       
       obsfsname = obsname
!!       obsfsname = obsname(1:iobs-5)//'.sources.n.fits'
       iobsfs = lastc ( obsfsname )
       inquire ( file=obsfsname(1:iobsfs), exist=lfnam2 )
       
! The image OBSCBNAME is used only in expanding and shrinking source footprints.
  
!!       obscbname = obsname(1:iobs-5)//'.cb.fits'
       obscbname = obsname(1:iobs-5)//'.sources.n.cb.fits'
       iobsfscb = lastc ( obscbname )
       inquire ( file=obscbname(1:iobsfscb), exist=lfnam3 )

! The image OBSBSNAME is used only in measurements of sizes and peak and total fluxes.

       obsbsname = obsname(1:iobs-5)//'.bs.fits'
       iobsfsbs = lastc ( obsbsname )
       inquire ( file=obsbsname(1:iobsfsbs), exist=lfnam4 )

! Determine name for single wave-combined detection image.

       idot = index ( obsname, '.' )
       nobs = index ( obsname, '.obs' )

! Determine name for the observational mask image.

       maskname = obsname(1:nobs-1)//'.bs.omask.fits'
       imsk = lastc ( maskname )

       inquire ( file=maskname(1:imsk), exist=lfnam9 )

! Define name for the observed background image.

       idot2 = index ( obsname(idot +1:nobs), '.' ) + idot
       idot3 = index ( obsname(idot2+1:nobs), '.' ) + idot2
       idot4 = index ( obsname(idot3+1:nobs), '.' ) + idot3

! Define name of a file listing all file names.

       filewithnames = dirmono(1:idmon)//'/'//obsname(1:nobs-1)//'.det.separateso'
       ifdn = lastc ( filewithnames )

       inquire ( file=filewithnames(1:ifdn), exist=lfnam7 )
       inquire ( file=icatname(1:iicat), exist=lfnam5 )

       ocatname = icatname(1:iicat)
       if (argD(1:2) .eq. '-o') ocatname = argE
       iocat = lastc ( ocatname )

       if (cverbose .eq. '-verb1') write (iotty,'(/a)') ' SMEASURE: '//cdistpc(1:icp)//' '//cbeam(1:icb)//' '//icatname(1:iicat)
     &                                                               //' '//ocatname(1:iocat)

       if (.not.lfnam1) then
         write (iotty,'(/a)') '   SMEASURE: ERROR: File '''//obsname(1:iobs)//''' not found.'
         stop 31
       endif
       if (.not.lfnam2) then
         write (iotty,'(/a)') '   SMEASURE: ERROR: File '''//obsfsname(1:iobsfs)//''' not found.'
         stop 32
       endif
       if (.not.lfnam3) then
         write (iotty,'(/a)') '   SMEASURE: ERROR: File '''//obscbname(1:iobsfscb)//''' not found.'
         stop 33
       endif
       if (.not.lfnam4) then
         write (iotty,'(/a)') '   SMEASURE: ERROR: File '''//obsbsname(1:iobsfsbs)//''' not found.'
         stop 34
       endif
       if (.not.lfnam5) then
         write (iotty,'(/a)') '   SMEASURE: ERROR: File '''//icatname(1:iicat)//''' not found.'
         stop 35
       endif
       if (.not.lfnam7) then
         write (iotty,'(/a)') '   SMEASURE: ERROR: File '''//filewithnames(1:ifdn)//''' not found.'
         stop 36
       endif
       if (.not.lfnam9) then
         write (iotty,'(/a)') '   SMEASURE: ERROR: File '''//maskname(1:imsk)//''' not found.'
         stop 37
       endif

       if (laccelerate) then
         
         if (cverbose .eq. '-verb2')
     &       write (iotty,'(/a)') '   Reading previous catalog '''//icatname(1:ipcat-1)//'.previous.cat'''
       
         open ( 14, file=icatname(1:ipcat-1)//'.previous.cat', status='old' )

         call skipcomm     ( 14, icatname(1:ipcat-1)//'.previous.cat', nhead )
         call linescolumns ( 14, icatname(1:ipcat-1)//'.previous.cat', nhead, nextr )
       endif

       if (cverbose .eq. '-verb2')
     &     write (iotty,'(/a)') '   Reading input catalog '''//icatname(1:iicat)//''''
       
       open ( 15, file=icatname(1:iicat), status='old' )

       call skipcomm     ( 15, icatname(1:iicat), nhead )
       call linescolumns ( 15, icatname(1:iicat), nhead, nextr )
       
       rewind ( 15 )

! Determine numbers of pixels in the FITS image.

       call getfitshead ( obsname(1:iobs), nx, ny, dx, dy, bunit )
       
       sqpixel = dx * dy
       pixel = sqrt ( sqpixel )

       allocate ( momxco(nextr), momyco(nextr), momxx(nextr), momyy(nextr), momxy(nextr), afwhm(nextr), bfwhm(nextr), nxco(nextr)
     &          , theta(nextr), fluxtotfit(nextr), fluxpeakbest(nextr), fluxtotbest(nextr), equivrad(nextr), sigmono(nextr)
     &          , sigglob(nextr), scalensx(nextr), catline(nextr), flag(nextr), footfacto(nextr), nyco(nextr), badsource(nextr)
     &          , backrms(nextr), fluxperror(nextr), apcorrect(nextr), fluxterror(nextr), dxcon2(nextr,nx), dycon2(nextr,ny)
     &          , foomajor(nextr), foominor(nextr), flg(nextr), lsourcetodeblend(nextr), footfact(nextr), lfootprint(nextr)
     &          , wcsxco(nextr), wcsyco(nextr), slope(nextr), good(nextr), backav1(nextr), cflm(nextr), backav2(nextr)
     &          , backav3(nextr), backrms1(nextr), backrms2(nextr), backrms3(nextr), aax(nextr), bbx(nextr), ccx(nextr)
     &          , aad(nextr), bbd(nextr), ccd(nextr), annulusarea(nextr), dist2pks(nextr,nextr), dist2foo(nextr,nextr)
     &          , asize(nextr), bsize(nextr), momxxh(nextr), foomajoro(nextr), foominoro(nextr), thetao(nextr), nkover(nextr)
     &          , flm(nextr), momyyh(nextr), momxyh(nextr), nkoverlap(nextr,nextr), lnotbad(nextr), goodmono(nextr), headline(1000)
     &          , stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a)') '   SMEASURE: ERROR: Trouble allocating memory (71).'
         stop 71
       endif

       nfg = 0
       nfluxes = 0
       nxcopix = 0
       npeakfx = 0
       nthetam = 0
       
       do i=1,nhead
         read (15,'(a)') headline(i)
         ncolumn = index ( headline(i), '  XCO_PIXEL  ' )
         if (ncolumn .gt. 0) nxcopix = ncolumn
         ncolumn = index ( headline(i), '  FG  ' )
         if (ncolumn .gt. 0) nfg = ncolumn
         ncolumn = index ( headline(i), '  FXP_BST    ' )
         if (ncolumn .gt. 0) nfluxes = ncolumn
         ncolumn = index ( headline(i), '  THETA_MOM  ' )
         if (ncolumn .gt. 0) nthetam = ncolumn
         ncolumn = index ( headline(i), '  PEAKFLUX   ' )
         if (ncolumn .gt. 0) npeakfx = ncolumn
       enddo
       if (nfg .eq. 0) then
         write (iotty,'(/a)') '   SMEASURE: ERROR: Column FG not found.'
         stop 12
       endif
       if (nfluxes .eq. 0) then
         write (iotty,'(/a)') '   SMEASURE: ERROR: Column FXP_BST not found.'
         stop 13
       endif
       if (nxcopix .eq. 0) then
         write (iotty,'(/a)') '   SMEASURE: ERROR: Column XCO_PIXEL not found.'
         stop 14
       endif
       if (npeakfx .eq. 0) then
         write (iotty,'(/a)') '   SMEASURE: ERROR: Column PEAKFLUX not found.'
         stop 15
       endif
       if (nthetam .eq. 0) then
         write (iotty,'(/a)') '   SMEASURE: ERROR: Column THETA_MOM not found.'
         stop 16
       endif
       write (cskipcoords,'(i3)') nxcopix - 1
       write (cskipflg,'(i3)') nfg
       write (cskipfluxes,'(i3)') nfluxes - 2
       write (cskipthetam,'(i3)') nthetam - 1
       nbad = 0
       
! Loop over all extracted sources.

       do n=1,nextr

         if (laccelerate) then
           read (14,'(a)') catline(n)
           read (catline(n),'('//cskipthetam//'x,4(1pe11.3))') thetao(n), footfacto(n), foomajoro(n), foominoro(n)
         endif

         read (15,'(a)') catline(n)
         read (catline(n),'(12x,i3,2x,a1)') nsx, flag(n)
         if (nsx .eq. 0 .and. flag(n) .eq. ' ') then
           write (iotty,'(/a)') '   SMEASURE: ERROR: Some sources have not been footprinted.'
           stop 91
         endif

         read (catline(n),'(a1)') badsource(n)

         read (catline(n),'('//cskipcoords//'x,2(f11.4),2(f13.7))') momxco(n), momyco(n), wcsxco(n), wcsyco(n)

         read (catline(n),'('//cskipflg//'x,i3,2(1pe11.3),i5,2(1pe11.3))') flg(n), sigglob(n), good(n), flm(n), sigmono(n)
     &                   , goodmono(n)

         read (catline(n),'('//cskipfluxes//'x,14(1pe11.3))') fluxpeakbest(n), fluxperror(n), fluxtotbest(n), fluxterror(n)
     &                   , fluxtotfit(n), scalensx(n), afwhm(n), bfwhm(n), asize(n), bsize(n), theta(n), footfact(n)
     &                   , foomajor(n), foominor(n)
     
         if (flag(n) .ne. 'n' .and. badsource(n) .ne. '!') then
           lnotbad(n) = .true.
         else
           lnotbad(n) = .false.
         endif
         if (badsource(n) .eq. '!') nbad = nbad + 1

         if (lfirstcall) then
           fluxpeakbest(n) = 0.0d0
           fluxtotbest(n) = 0.0d0
           fluxperror(n) = 0.0d0
           fluxterror(n) = 0.0d0
           fluxtotfit(n) = 0.0d0
           sigmono(n) = 0.0d0
           goodmono(n) = 0.0d0
           footfact(n) = factortotal
           afwhm(n) = max ( scalensx(n), beam )
           bfwhm(n) = max ( scalensx(n), beam )
           asize(n) = afwhm(n)
           bsize(n) = bfwhm(n)
           theta(n) = 0.0d0
           foomajor(n) = factortotal * afwhm(n)
           foominor(n) = factortotal * bfwhm(n)
           lfootprint(n) = .true.
         else

! Check if there is a signal (NEGATIVE FOOTPRINTS!) from FINALCAT that a source is not good in some way
! and therefore it should not be considered. NOTE: ABSOLUTE VALUES of the footprints must be used.

           if (foomajor(n) .gt. almostzero .and. foominor(n) .gt. almostzero) then
             lfootprint(n) = .true.
           else
             lfootprint(n) = .false.
             foomajor(n) = abs ( foomajor(n) )
             foominor(n) = abs ( foominor(n) )
           endif
         endif
         
! Measurements of sizes and fluxes will be done within fixed circular apertures, if requested.

         if (capcorr .eq. 'y') then
           apcorrect(n) = 1.0d0
         else
           apcorrect(n) = 0.0d0
         endif

         nxco(n) = nint ( momxco(n))
         nyco(n) = nint ( momyco(n))
         lsourcetodeblend(n) = .true.
         do k=1,nextr
           nkover(n) = k
           nkoverlap(n,nkover(n)) = k
         enddo
       enddo
       
       if (laccelerate) close ( 14 )
       close ( 15 )

! Allocate the image arrays.
       
       allocate ( obsfsimage(nx,ny), obsfsbsimage(nx,ny), obsfscbimage(nx,ny), maskimage(nx,ny), convdelta(nextr,99), stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a)') '   SMEASURE: ERROR: Trouble allocating memory (72).'
         stop 72
       endif

!-----------------------------------------------------------------------------------------------------------------------------------

       if (ltiming) then
         cpu_initialize = timer ( 'cpu', 0.0d0 )
         wal_initialize = timer ( 'wal', 0.0d0 )
       endif

       write (cx,'(i7)') nx
       write (cy,'(i7)') ny
       ncx = int ( log10 ( dble ( nx ) ) ) + 1
       ncy = int ( log10 ( dble ( ny ) ) ) + 1

       if (cverbose .eq. '-verb2')
     &     write (iotty,'(/a)') '   Reading filament-subtracted ('//cx(7-ncx+1:7)//' x '//cy(7-ncy+1:7)//') '''
     &                        //obsname(1:iobs)//''''
      
       call rfits ( nx, ny, intensunits, ctype1, ctype2, crpix1, crpix2, crval1, crval2, obsfsimage, dx, dy, object, ra, dec
     &            , obsfsname(1:iobsfs), iotty, 0, creator, beamd, funmin, funmax, blank, crota1, crota2, cd11, cd12, cd21, cd22
     &            , equinox, bzero, bscale, wave, datamin, datamax, history, cverbose )

       if (cverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7))') '   Minmax values in the image:', funmin, funmax
       
       iu = lastc ( intensunits )

       if (intensunits(1:iu) .ne. 'MJy/sr' .and. intensunits(1:iu) .ne. 'H2/cm^2') then
         iu = lastc ( intensunits )
         write (iotty,'(/a)') '   SMEASURE: ERROR: Unknown intensity units in the image.'
         write (iotty,'(a )') '   BUNIT must be ''MJy/sr'' or ''H2/cm^2''.'
         stop 92
       endif
       
       if (intensunits(1:iu) .eq. 'MJy/sr') then
         convftot = sqpixel * 1.0e6 / sq_arcsecs_per_sterad            !<~ intensity in MJy/sr to flux in Jy
         convftotm = convftot
       endif       
       if (intensunits(1:iu) .eq. 'H2/cm^2') then
         convftot = sqpixel                                            !<~ surface density in H2/cm^2 to flux in H2/cm^2*arcsec^2
         convftotm = sqpixel * xmu *xmH * (au2cm * distpc)**2 / xMsun  !<~ surface density in H2/cm^2 to flux in Msun
       endif       
       
       if (cverbose .eq. '-verb2')
     &     write (iotty,'(/a)') '   Reading clean background ('//cx(7-ncx+1:7)//' x '//cy(7-ncy+1:7)//') '''
     &                        //obscbname(1:iobsfscb)//''''
      
       call rfits ( nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, obsfscbimage, dx, dy, object, ra, dec
     &            , obscbname(1:iobsfscb), iotty, 0, creator, beamd, funmin, funmax, blank, crota1, crota2, cd11, cd12, cd21
     &            , cd22, equinox, bzero, bscale, wave, datamin, datamax, history, cverbose )

       if (cverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7))') '   Minmax values in the image:', funmin, funmax
       
       if (cverbose .eq. '-verb2')
     &     write (iotty,'(/a)') '   Reading backg-subtracted ('//cx(7-ncx+1:7)//' x '//cy(7-ncy+1:7)//') '''
     &                        //obsbsname(1:iobsfsbs)//''''
      
       call rfits ( nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, obsfsbsimage, dx, dy, object, ra, dec
     &            , obsbsname(1:iobsfsbs), iotty, 0, creator, beamd, funmin, funmax, blank, crota1, crota2, cd11, cd12, cd21
     &            , cd22, equinox, bzero, bscale, wave, datamin, datamax, history, cverbose )

       if (cverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7))') '   Minmax values in the image:', funmin, funmax
       
       if (cverbose .eq. '-verb2')
     &     write (iotty,'(/a)') '   Reading observational mask ('//cx(7-ncx+1:7)//' x '//cy(7-ncy+1:7)//') '''
     &                        //maskname(1:imsk)//''''
       
       call rfits ( nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, maskimage, dx, dy, object, ra, dec
     &            , maskname(1:imsk), iotty, 0, creator, beamd, funmin, funmax, blank, crota1, crota2, cd11, cd12, cd21, cd22
     &            , equinox, bzero, bscale, wave, datamin, datamax, history, cverbose )

       if (cverbose .eq. '-verb2') write (iotty,'(a,1x,2(1pe14.7))') '   Minmax values in the image:', funmin, funmax

!-----------------------------------------------------------------------------------------------------------------------------------

       if (cverbose .eq. '-verb2')
     &   write (iotty,'(a)') '   Preparing auxiliary arrays to accelerate'

       call auxiliaryarrays ( nx, ny, nextr, flag, momxco, momyco, afwhm, bfwhm, theta, sqpixel, almostzero
     &                      , convtoradians, aax, bbx, ccx, aad, bbd, ccd, dxcon2, dycon2, foomajor, dist2pks, dist2foo )

!-----------------------------------------------------------------------------------------------------------------------------------

       if (cverbose .eq. '-verb2')
     &    write (iotty,'(a)') '   Determining which sources allow acceleration'

       if (ltiming) then
         cpu_testaccel = timer ( 'cpu', 0.0d0 )
         wal_testaccel = timer ( 'wal', 0.0d0 )
       endif

       call testacceleration ( nx, ny, dx, dy, nextr, lnotbad, momxco, momyco, nxco, nyco, foomajor, foominor, theta, footfacto
     &                       , foomajoro, foominoro, thetao, lfootprint, maskimage, dist2pks, dist2foo, laccelerate
     &                       , lsourcetodeblend, nkover, nkoverlap )

       if (ltiming) then
         cput_testaccel = cput_testaccel + timer ( 'cpu', cpu_testaccel )
         walt_testaccel = walt_testaccel + timer ( 'wal', wal_testaccel )
         write (60,'(a,2(1pe9.2,a))') ' Total ''ACCELERATION''    CPU',cput_testaccel,' WALL',walt_testaccel,' sec'
       endif

!-----------------------------------------------------------------------------------------------------------------------------------

       if (cverbose .eq. '-verb2')
     &   write (iotty,'(a)') '   Computing monochromatic characteristic sizes'

       call monosignificances ( iotty, nx, ny, nextr, filewithnames(1:ifdn), beam, catline, lnotbad, cverbose, lfirstcall
     &                        , lfinishing_mode, liter_phase2, liter_phase3, lonlypositives, almostzero, scalensx, sigmono
     &                        , fluxpeakbest, fluxperror, maskimage, nxco, nyco, foomajor, foominor, theta, factortotal, flm
     &                        , momxco, momyco, aad, bbd, ccd, lfootprint, cdate, ctime, compda, clibname, cfitsversion, dxcon2
     &                        , dycon2, cflm, lsourcetodeblend, nkover, nkoverlap, flag, convdelta )

       if (ltiming) then
         cput_initialize = cput_initialize + timer ( 'cpu', cpu_initialize )
         walt_initialize = walt_initialize + timer ( 'wal', wal_initialize )
         write (60,'(a,2(1pe9.2,a))') ' Total ''READING IMAGES''  CPU',cput_initialize,' WALL',walt_initialize,' sec'
       endif

!-----------------------------------------------------------------------------------------------------------------------------------

       if (cverbose .eq. '-verb2')
     &    write (iotty,'(a)') '   Deblending peak intensities of overlapping sources'

       if (ltiming) then
         cpu_iteratepeaks = timer ( 'cpu', 0.0d0 )
         wal_iteratepeaks = timer ( 'wal', 0.0d0 )
       endif

       call deblendpeaks ( nx, ny, dx, dy, nextr, lnotbad, fluxpeakbest, maskimage, flm, momxco, momyco, nxco, nyco, foominor
     &                   , foomajor, theta, aad, bbd, ccd, obsname(1:iobs), obsfsbsimage, lfootprint, cdate, ctime, compda
     &                   , clibname(lb:7), cfitsversion, dxcon2, dycon2, cflm, almostzero, lsourcetodeblend, nkover, nkoverlap
     &                   , flag, .false., convdelta )

       if (ltiming) then
         cput_iteratepeaks = cput_iteratepeaks + timer ( 'cpu', cpu_iteratepeaks )
         walt_iteratepeaks = walt_iteratepeaks + timer ( 'wal', wal_iteratepeaks )
         write (60,'(a,2(1pe9.2,a))') ' Total ''ITERATE PEAKS''   CPU',cput_iteratepeaks,' WALL',walt_iteratepeaks,' sec'
       endif

!-----------------------------------------------------------------------------------------------------------------------------------

       if (cverbose .eq. '-verb2')
     &    write (iotty,'(a)') '   Computing deblended intensity moments for each source'

       if (ltiming) then
         cpu_secondmoments = timer ( 'cpu', 0.0d0 )
         wal_secondmoments = timer ( 'wal', 0.0d0 )
       endif

       call intensmoments ( nx, ny, dx, dy, nextr, lnotbad, sigmono, fluxpeakbest, momxco, momyco, nxco, nyco, foominor !,fluxperror
     &                    , foomajor, aad, bbd, ccd, dxcon2, dycon2, lfootprint, sqpixel, almostzero, momxx, momxy, momyy
     &                    , maskimage, obsfsbsimage, equivrad, pi, lsourcetodeblend, nkover, nkoverlap, momxxh, momxyh, momyyh
     &                    , afwhm, bfwhm, theta )

       if (ltiming) then
         cput_secondmoments = cput_secondmoments + timer ( 'cpu', cpu_secondmoments )
         walt_secondmoments = walt_secondmoments + timer ( 'wal', wal_secondmoments )
         write (60,'(a,2(1pe9.2,a))') ' Total ''SECOND MOMENTS''  CPU',cput_secondmoments,' WALL',walt_secondmoments,' sec'
       endif

!-----------------------------------------------------------------------------------------------------------------------------------

       if (cverbose .eq. '-verb2')
     &    write (iotty,'(a)') '   Computing sizes of sources from intensity moments'
       
       if (ltiming) then
         cpu_fwhmsizes = timer ( 'cpu', 0.0d0 )
         wal_fwhmsizes = timer ( 'wal', 0.0d0 )
       endif

       call momentsizes ( nextr, pixel, lnotbad, lfootprint, momxx, momyy, momxy, afwhm, bfwhm, momxxh, momyyh, momxyh !!, beam
     &                  , asize, bsize, theta, sqreightlogtwo, convtodegrees, lsourcetodeblend, almostzero ) !!, lfinishing_mode 

       if (ltiming) then
         cput_fwhmsizes = cput_fwhmsizes + timer ( 'cpu', cpu_fwhmsizes )
         walt_fwhmsizes = walt_fwhmsizes + timer ( 'wal', wal_fwhmsizes )
         write (60,'(a,2(1pe9.2,a))') ' Total ''SOURCE SIZES''    CPU',cput_fwhmsizes,' WALL',walt_fwhmsizes,' sec'
       endif

!-----------------------------------------------------------------------------------------------------------------------------------

       if (cverbose .eq. '-verb2')
     &    write (iotty,'(a)') '   Integrating total fluxes of fitted elliptical shapes'

       if (ltiming) then
         cpu_fittedfluxes = timer ( 'cpu', 0.0d0 )
         wal_fittedfluxes = timer ( 'wal', 0.0d0 )
       endif

       call fitshapesfluxes ( nextr, lnotbad, lfootprint, nx, ny, dx, dy, momxco, momyco, nxco, nyco, afwhm, bfwhm, theta, foomajor
     &                      , foominor, fluxpeakbest, fluxtotfit, convtoradians, lsourcetodeblend )

       if (ltiming) then
         cput_fittedfluxes = cput_fittedfluxes + timer ( 'cpu', cpu_fittedfluxes )
         walt_fittedfluxes = walt_fittedfluxes + timer ( 'wal', wal_fittedfluxes )
         write (60,'(a,2(1pe9.2,a))') ' Total ''FITTED FLUXES''   CPU',cput_fittedfluxes,' WALL',walt_fittedfluxes,' sec'
       endif

!-----------------------------------------------------------------------------------------------------------------------------------

       if (cverbose .eq. '-verb2')
     &    write (iotty,'(a)') '   Integrating total fluxes of each deblended source'
       
       if (ltiming) then
         cpu_totalfluxes = timer ( 'cpu', 0.0d0 )
         wal_totalfluxes = timer ( 'wal', 0.0d0 )
       endif

       call integratedfluxes ( iotty, nx, ny, dx, dy, nextr, lnotbad, fluxpeakbest, momxco, momyco, nxco, nyco, foominor, foomajor
     &                       , theta, aax, bbx, ccx, maskimage, lfootprint, dxcon2, dycon2, convftot, convftotm, intensunits(1:iu)
     &                       , capcorr, cowave, equivrad, apcorrect, almostzero, fluxtotbest, fluxtotfit  !!, beam, afwhm, bfwhm
     &                       , obsfsbsimage, lsourcetodeblend, nkover, nkoverlap )   !!, lfinishing_mode

       if (ltiming) then
         cput_totalfluxes = cput_totalfluxes + timer ( 'cpu', cpu_totalfluxes )
         walt_totalfluxes = walt_totalfluxes + timer ( 'wal', wal_totalfluxes )
         write (60,'(a,2(1pe9.2,a))') ' Total ''TOTAL FLUXES''    CPU',cput_totalfluxes,' WALL',walt_totalfluxes,' sec'
       endif

!-----------------------------------------------------------------------------------------------------------------------------------

       if (cverbose .eq. '-verb2')
     &    write (iotty,'(a)') '   Computing local standard deviations for each source'

       if (ltiming) then
         cpu_createannuli = timer ( 'cpu', 0.0d0 )
         wal_createannuli = timer ( 'wal', 0.0d0 )
       endif

! Allocate the image arrays.

       allocate ( bigfoot(nx,ny), smallfoot(nx,ny), bigfootclone(nx,ny), annulus(nx,ny), n1x(ny), n2x(ny), stat=irc )
       
       if (irc .ne. 0) then
         write (iotty,'(/a)') '   SMEASURE: ERROR: Trouble allocating memory (77).'
         stop 77
       endif

       call createannuli ( nx, ny, dx, dy, nextr, lnotbad, fluxperror, momxco, momyco, nxco, nyco, foominor, foomajor, theta
     &                   , maskimage, lfootprint, n1x, n2x, sqpixel, factoffset, sq_arcsecs_per_beam_area, pi, n1ymin, n2ymax
     &                   , factortotal, footfact, annulusarea, smallfoot, bigfoot, annulus )

       if (ltiming) then
         cput_createannuli = cput_createannuli + timer ( 'cpu', cpu_createannuli )
         walt_createannuli = walt_createannuli + timer ( 'wal', wal_createannuli )
         write (60,'(a,2(1pe9.2,a))') ' Total ''CREATE ANNULI''   CPU', cput_createannuli, ' WALL', walt_createannuli, ' sec'
       endif

       call wfits ( cfitsversion, nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, bigfoot, dx, dy, object, crval1
     &            , crval2, 'bigfoot.'//cowave(1:icw)//'.fits', cdate, ctime, creator, beam, blank, crota1, crota2, cd11, cd12
     &            , cd21, cd22, equinox, bzero, bscale, wave, datamin, datamax, history )
                    
!==================================================================================================================================

!!       if (liter_phase2 .or. liter_phase3) then

         if (ldevelopment) then
           open ( 17, file='sm'//obsname(idot:idot2)//'snratios.annuli.apcorrs.dat', status='unknown' )
        
           write (17,'(a)') '# Date: '//cdate//' Time: '//ctime//'  Created by SMEASURE compiled on '//compda
           write (17,'(a)') '# Alexander Men''shchikov, DAp IRFU CEA Saclay, France (alexander.menshchikov@cea.fr)'
           write (17,'(a)') '# Used '//clibname(lb:7)//' library version '//cfitsversion//' by William D Pence '
     &                    //'(william.d.pence@nasa.gov)'
           write (17,'(a)') '#'
           write (17,'(a)') '#  N  XCO_PIXEL  YCO_PIXEL  FP/SIGMA  ANNULPIX ANNULPIX1 ANNULPIX2  MAJORPIX  MINORPIX  THICKPIX'
     &                    //'  APCORRECT'
           write (17,'(a)') '#'
         endif

         do n=1,nextr
           
           if (lnotbad(n) .and. ((lsourcetodeblend(n) .and. lfootprint(n)) .or.
     &         (maskimage(nxco(n),nyco(n)) .gt. 0.9d0 .and. fluxperror(n) .lt. 1.0d-50))) then
             
             czno = catline(n)(1:6)
             if (czno(2:2) .eq. ' ') czno(2:2) = '0'
             if (czno(3:3) .eq. ' ') czno(3:3) = '0'
             if (czno(4:4) .eq. ' ') czno(4:4) = '0'
             if (czno(5:5) .eq. ' ') czno(5:5) = '0'

!----------------------------------------------------------------------------------------------------------------------------------

             if (ltiming) then
               cpu_fluctuations1 = timer ( 'cpu', 0.0d0 )
               wal_fluctuations1 = timer ( 'wal', 0.0d0 )
             endif

! Shortest distance (squared) from source position to bigfoot pixels.

             radmax = foominor(n)
             
             do ir=1,ntry

               nr = nint ( radmax / min ( dx, dy ) ) + 1
               l1x = max ( nxco(n) - nr, 1 )
               l2x = min ( nxco(n) + nr, nx )
               m1x = max ( nyco(n) - nr, 1 )
               m2x = min ( nyco(n) + nr, ny )
               rad2min = 1.0d30
               
               do m=m1x,m2x
                 do l=l1x,l2x
                   if (annulus(l,m) .gt. almostzero) then
                     rad2min = min ( rad2min, dxcon2(n,l) + dycon2(n,m) )
                   endif
                 enddo
               enddo
               
               if (rad2min .gt. 9.9d29) then
                 radmax = 2.0d0 * radmax
               else
                 exit
               endif
             enddo
             
             if (ltiming) then
               cput_fluctuations1 = cput_fluctuations1 + timer ( 'cpu', cpu_fluctuations1 )
               walt_fluctuations1 = walt_fluctuations1 + timer ( 'wal', wal_fluctuations1 )
             endif

!----------------------------------------------------------------------------------------------------------------------------------

             if (ltiming) then
               cpu_fluctuations2 = timer ( 'cpu', 0.0d0 )
               wal_fluctuations2 = timer ( 'wal', 0.0d0 )
             endif

! This absolute peak value is used below to ensure that all chosen annulus pixels are fainter than source peak.
             
             absobsfsatpos = abs ( obsfsimage(nxco(n),nyco(n)) )

             deficit = 0.0d0
             annulusareait = annulusarea(n)
             
             do ir=1,ntry
               
               rad2annit = rad2min + annulusareait / pi
               nr = nint ( sqrt ( rad2annit ) / min ( dx, dy ) ) + 1
               l1 = max ( nxco(n) - nr, 1 )
               l2 = min ( nxco(n) + nr, nx )
               m1 = max ( nyco(n) - nr, 1 )
               m2 = min ( nyco(n) + nr, ny )
               iann = 0
               
               do m=m1,m2
                 do l=l1,l2
                   if (smallfoot(l,m) .lt. almostzero) then
                     rad2 = dxcon2(n,l) + dycon2(n,m)
                     if (rad2 .le. rad2annit .and. abs ( obsfsimage(l,m) ) .lt. absobsfsatpos) then
                       iann = iann + 1
                     endif
                   endif
                 enddo
               enddo
               
               deficit = annulusarea(n) - iann * sqpixel
               
               if (deficit .ge. almostzero) then
                 annulusareait = 1.2d0 * annulusareait
               else
                 exit
               endif
             enddo
             
             if (deficit / annulusarea(n) .lt. 0.5d0) then
               lann3usable = .true.
             else
               lann3usable = .false.
             endif
             
             if (ltiming) then
               cput_fluctuations2 = cput_fluctuations2 + timer ( 'cpu', cpu_fluctuations2 )
               walt_fluctuations2 = walt_fluctuations2 + timer ( 'wal', wal_fluctuations2 )
             endif

!----------------------------------------------------------------------------------------------------------------------------------

             if (ltiming) then
               cpu_fluctuations3 = timer ( 'cpu', 0.0d0 )
               wal_fluctuations3 = timer ( 'wal', 0.0d0 )
             endif
             
             do m=m1,m2
               do l=l1,l2
                 if (smallfoot(l,m) .lt. almostzero) then
                   rad2 = dxcon2(n,l) + dycon2(n,m)
                   if (rad2 .le. rad2annit) then
                     annulus(l,m) = newpixval
                   endif
                 endif
               enddo
             enddo
             
             if (ltiming) then
               cput_fluctuations3 = cput_fluctuations3 + timer ( 'cpu', cpu_fluctuations3 )
               walt_fluctuations3 = walt_fluctuations3 + timer ( 'wal', wal_fluctuations3 )
             endif
             
!----------------------------------------------------------------------------------------------------------------------------------

             if (ltiming) then
               cpu_nearestannulus = timer ( 'cpu', 0.0d0 )
               wal_nearestannulus = timer ( 'wal', 0.0d0 )
             endif

             do m=n1ymin,n2ymax
               do l=n1x(m),n2x(m)
                 bigfootclone(l,m) = bigfoot(l,m)
               enddo
             enddo

! The nearest elliptical annulus around the source.

             major1 = factoffset * foomajor(n)
             minor1 = factoffset * foominor(n)
             semimajor1 = major1 / 2.0d0
             semiminor1 = minor1 / 2.0d0
             aaa = 1.0d0
             bbb = semimajor1 - semiminor1
             annulusareait = annulusarea(n)

             do ir=1,ntry
               
               ccc = -annulusareait / pi - semiminor1 * semimajor1
               semiminor2 = (-bbb + sqrt ( bbb**2 - 4.0d0 * aaa * ccc )) / (2.0d0 * aaa)
               semimajor2 = semimajor1 + semiminor2 - semiminor1
               major2 = semimajor2 * 2.0d0
               minor2 = semiminor2 * 2.0d0

! Compute standard deviation over an annulus around the source; other nearby sources may contaminate the annulus pixels.
          
               nrz = nint ( semimajor2 / min ( dx, dy ) ) + 1
               l1z = max ( nxco(n) - nrz, 1 )
               l2z = min ( nxco(n) + nrz, nx )
               m1z = max ( nyco(n) - nrz, 1 )
               m2z = min ( nyco(n) + nrz, ny )

               ibav1 = 0
               backav1(n) = 0.0d0
              
               do m=m1z,m2z
                 dblm = dble ( m )
                 do l=l1z,l2z
                   if (maskimage(l,m) .gt. 0.9d0) then
                     dbll = dble ( l )
                     
                     if (inellipse ( dbll, dblm, dx, dy, momxco(n), momyco(n), major2, minor2, theta(n) ) .and.
     &                   .not.inellipse ( dbll, dblm, dx, dy, momxco(n), momyco(n), major1, minor1, theta(n) )) then

                       if (abs ( obsfsimage(l,m) ) .lt. absobsfsatpos) then
                         ibav1 = ibav1 + 1
                         backav1(n) = backav1(n) + obsfsimage(l,m)
                         bigfootclone(l,m) = newpixval
                       else
                         bigfootclone(l,m) = 0.7d0
                       endif
                     endif
                   endif
                 enddo
               enddo
               
               deficit = annulusarea(n) - ibav1 * sqpixel
               
               if (deficit .ge. almostzero) then
                 annulusareait = 1.2d0 * annulusareait
               else
                 exit
               endif
             enddo
             
             if (ibav1 .gt. 0) backav1(n) = backav1(n) / dble ( ibav1 )

             if (deficit / annulusarea(n) .lt. 0.5d0) then
               lann1usable = .true.
             else
               lann1usable = .false.
             endif

             ibrms1 = 0
             backrms1(n) = 0.0d0
             
             do m=m1z,m2z
               do l=l1z,l2z
                 if (bigfootclone(l,m) .gt. 0.9d0) then
                   ibrms1 = ibrms1 + 1
                   backrms1(n) = backrms1(n) + (obsfsimage(l,m) - backav1(n))**2
                 endif
               enddo
             enddo
             
             if (ibrms1 .gt. 1) backrms1(n) = sqrt ( backrms1(n) / dble ( ibrms1 - 1 ) )
             if (.not.lann1usable .or. backrms1(n) .lt. almostzero) then
               backrms1(n) = 9.999e+29
             endif
             
             if (ltiming) then
               cput_nearestannulus = cput_nearestannulus + timer ( 'cpu', cpu_nearestannulus )
               walt_nearestannulus = walt_nearestannulus + timer ( 'wal', wal_nearestannulus )
             endif

!----------------------------------------------------------------------------------------------------------------------------------

             if (ltiming) then
               cpu_secondannulus = timer ( 'cpu', 0.0d0 )
               wal_secondannulus = timer ( 'wal', 0.0d0 )
             endif

! Additional (twice more distant) elliptical annulus around the source.

             do m=n1ymin,n2ymax
               do l=n1x(m),n2x(m)
                 bigfootclone(l,m) = bigfoot(l,m)
               enddo
             enddo

             major1a = (2.0d0 * factoffset) * foomajor(n)
             minor1a = (2.0d0 * factoffset) * foominor(n)
             semimajor1a = major1a / 2.0d0
             semiminor1a = minor1a / 2.0d0
             annulusareait = annulusarea(n)
             aaa = 1.0d0
             bbb = semimajor1a - semiminor1a

             do ir=1,ntry

               ccc = -annulusareait / pi - semiminor1a * semimajor1a
               semiminor2a = (-bbb + sqrt ( bbb**2 - 4.0d0 * aaa * ccc )) / (2.0d0 * aaa)
               semimajor2a = semimajor1a + semiminor2a - semiminor1a
               major2a = semimajor2a * 2.0d0
               minor2a = semiminor2a * 2.0d0

! Compute standard deviation over an annulus around the source; other nearby sources may contaminate the annulus pixels.
          
               nrz = nint ( semimajor2a / min ( dx, dy ) ) + 1
               l1z = max ( nxco(n) - nrz, 1 )
               l2z = min ( nxco(n) + nrz, nx )
               m1z = max ( nyco(n) - nrz, 1 )
               m2z = min ( nyco(n) + nrz, ny )
               
               ibav2 = 0
               backav2(n) = 0.0d0
            
               do m=m1z,m2z
                 dblm = dble ( m )
                 do l=l1z,l2z
                   if (maskimage(l,m) .gt. 0.9d0) then
                     dbll = dble ( l )
              
                     if (inellipse ( dbll, dblm, dx, dy, momxco(n), momyco(n), major2a, minor2a, theta(n) ) .and.
     &                   .not.inellipse ( dbll, dblm, dx, dy, momxco(n), momyco(n), major1a, minor1a, theta(n) )) then
              
                       if (abs ( obsfsimage(l,m) ) .lt. absobsfsatpos) then
                         ibav2 = ibav2 + 1
                         backav2(n) = backav2(n) + obsfsimage(l,m)
                         bigfootclone(l,m) = newpixval
                       else
                         bigfootclone(l,m) = 0.7d0
                       endif
                     endif
                   endif
                 enddo
               enddo

               deficit = annulusarea(n) - ibav2 * sqpixel
               
               if (deficit .ge. almostzero) then
                 annulusareait = 1.2d0 * annulusareait
               else
                 exit
               endif
             enddo

             if (ibav2 .gt. 0) backav2(n) = backav2(n) / dble ( ibav2 )

             if (deficit / annulusarea(n) .lt. 0.5d0) then
               lann2usable = .true.
             else
               lann2usable = .false.
             endif

             ibrms2 = 0
             backrms2(n) = 0.0d0
             
             do m=m1z,m2z
               do l=l1z,l2z
                 if (bigfootclone(l,m) .gt. 0.9d0) then
                   ibrms2 = ibrms2 + 1
                   backrms2(n) = backrms2(n) + (obsfsimage(l,m) - backav2(n))**2
                 endif
               enddo
             enddo
             
             if (ibrms2 .gt. 1) backrms2(n) = sqrt ( backrms2(n) / dble ( ibrms2 - 1 ) )
             if (.not.lann2usable .or. backrms2(n) .lt. almostzero) then
               backrms2(n) = 9.999e+29
             endif

             if (ltiming) then
               cput_secondannulus = cput_secondannulus + timer ( 'cpu', cpu_secondannulus )
               walt_secondannulus = walt_secondannulus + timer ( 'wal', wal_secondannulus )
             endif

!----------------------------------------------------------------------------------------------------------------------------------

             if (ltiming) then
               cpu_extendannulus = timer ( 'cpu', 0.0d0 )
               wal_extendannulus = timer ( 'wal', 0.0d0 )
             endif

! Compute standard deviation over a (possibly larger) environment of the source, extending the annulus to avoid other sources.
          
             ibav3 = 0
             ibrms3 = 0
             backav3(n) = 0.0d0
             backrms3(n) = 0.0d0
             
             do m=m1,m2
               do l=l1,l2
                 if (maskimage(l,m) .gt. 0.9d0 .and. annulus(l,m) .gt. 0.9d0) then
                   if (abs ( obsfsimage(l,m) ) .lt. absobsfsatpos) then
                     ibav3 = ibav3 + 1
                     backav3(n) = backav3(n) + obsfsimage(l,m)
                   else
                     annulus(l,m) = 0.3d0
                   endif
                 endif
               enddo
             enddo

             if (ibav3 .gt. 0) backav3(n) = backav3(n) / dble ( ibav3 )
             
             do m=m1,m2
               do l=l1,l2
                 if (maskimage(l,m) .gt. 0.9d0 .and. annulus(l,m) .gt. 0.9d0) then
                   ibrms3 = ibrms3 + 1
                   backrms3(n) = backrms3(n) + (obsfsimage(l,m) - backav3(n))**2
                 endif
               enddo
             enddo
             
             if (ibrms3 .gt. 1) backrms3(n) = sqrt ( backrms3(n) / dble ( ibrms3 - 1 ) )
             if (.not.lann3usable .or. backrms3(n) .lt. almostzero) then
               backrms3(n) = 9.999e+29
             endif

! The final standard deviation is a minimum of the three estimates; this must be more accurate than either one.

             backrms(n) = min ( backrms1(n), backrms2(n), backrms3(n) )
          
             if (backrms(n) .gt. almostzero .and. backrms(n) .lt. 9.0e+29) then
               sigmarms = backrms(n)
             else
               sigmarms = 1.111d-31
             endif

             if (ltiming) then
               cput_extendannulus = cput_extendannulus + timer ( 'cpu', cpu_extendannulus )
               walt_extendannulus = walt_extendannulus + timer ( 'wal', wal_extendannulus )
             endif

!----------------------------------------------------------------------------------------------------------------------------------

             if (ldevelopment) then
               
               if (ibrms2 .lt. 80 .and. cverbose .eq. '-verb2') then
                 id = int ( log10 ( dble ( n ) ) ) + 1
                 write (cn5,'(i5)') n
                 areapix1 = pi * semimajor1 * semiminor1 / sqpixel
                 areapix2 = pi * semimajor2 * semiminor2 / sqpixel
                 thickpix = (semiminor2 - semiminor1) / sqrt ( dx * dy )
                 write (iotty,'(a)') '   SMEASURE: WARNING: Computing rms with too few pixels for source # '//cn5(5-id+1:5)
                 write (iotty,*) nxco(n), nyco(n), areapix2 - areapix1, ibrms2 
               endif
          
               write (17,'(i4,10(1pe10.2))') n, momxco(n), momyco(n), fluxpeakbest(n) / max ( backrms(n), almostzero)
     &                                     , areapix2 - areapix1, dble ( ibrms1 ), dble ( ibrms2 ), major2 / sqrt ( dx * dy )
     &                                     , minor2 / sqrt ( dx * dy ), thickpix, apcorrect(n)

               annulusname = originpath(1:iopa)//'/+iterations/+annuli/'//'sm.annulus.'//cowave(1:icw)//'.no'//czno(2:6)//'.fits'
               ian = lastc ( annulusname )
            
               if (backrms1(n) .lt. backrms3(n) .or. backrms2(n) .lt. backrms3(n)) then
                 
                 call wfits ( cfitsversion, nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, bigfootclone, dx, dy
     &                      , object, crval1, crval2, annulusname(1:ian), cdate, ctime, creator, beam, blank, crota1, crota2
     &                      , cd11, cd12, cd21, cd22, equinox, bzero, bscale, wave, datamin, datamax, history )
               else
                 call wfits ( cfitsversion, nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, annulus, dx, dy, object
     &                      , crval1, crval2, annulusname(1:ian), cdate, ctime, creator, beam, blank, crota1, crota2, cd11, cd12
     &                      , cd21, cd22, equinox, bzero, bscale, wave, datamin, datamax, history )
               endif
             endif

! If the footprint area contains N times the beam area, then the rms error of the sum of intensities over the footprint area 
! will be the square root of the quadratic sum of the individual rms errors, since the N beam measurements are statistically 
! independent. If the individual rms errors are assumed identical and equal to errFpeak, then errFtot = errFpeak * sqrt ( N ). 
! Since N = footprint area / beam area, sqrt ( N ) = sqrt ( footprint area / beam area ).
! Convert flux errors to proper units before cataloging them.
       
             fluxperror(n) = sigmarms
             fluxterror(n) = fluxperror(n) * sqrt ( foomajor(n) * foominor(n) ) / (factortotal * beam)
             fluxterror(n) = fluxterror(n) * convftot / sqpixel * sq_arcsecs_per_beam_area
           endif
         enddo

         if (ltiming) then
           write (60,'(a,2(1pe9.2,a))') ' Total ''FLUCTUATIONS1''   CPU',cput_fluctuations1, ' WALL', walt_fluctuations1, ' sec'
           write (60,'(a,2(1pe9.2,a))') ' Total ''FLUCTUATIONS2''   CPU',cput_fluctuations2, ' WALL', walt_fluctuations2, ' sec'
           write (60,'(a,2(1pe9.2,a))') ' Total ''FLUCTUATIONS3''   CPU',cput_fluctuations3, ' WALL', walt_fluctuations3, ' sec'
           write (60,'(a,2(1pe9.2,a))') ' Total ''NEARESTANNULUS''  CPU',cput_nearestannulus,' WALL', walt_nearestannulus,' sec'
           write (60,'(a,2(1pe9.2,a))') ' Total ''SECONDANNULUS''   CPU',cput_secondannulus, ' WALL', walt_secondannulus, ' sec'
           write (60,'(a,2(1pe9.2,a))') ' Total ''EXTENDEDANNULUS'' CPU',cput_extendannulus, ' WALL', walt_extendannulus, ' sec'
         endif
       
         if (ldevelopment) then
           close ( 17 )
         endif
!!       endif

!-----------------------------------------------------------------------------------------------------------------------------------

       if (cverbose .eq. '-verb2')
     &    write (iotty,'(a)') '   Computing monochromatic significances'

       if (ltiming) then
         cpu_significances = timer ( 'cpu', 0.0d0 )
         wal_significances = timer ( 'wal', 0.0d0 )
       endif

       call monosignificances ( iotty, nx, ny, nextr, filewithnames(1:ifdn), beam, catline, lnotbad, cverbose, .false.
     &                        , lfinishing_mode, liter_phase2, liter_phase3, lonlypositives, almostzero, scalensx, sigmono
     &                        , fluxpeakbest, fluxperror, maskimage, nxco, nyco, foomajor, foominor, theta, factortotal, flm
     &                        , momxco, momyco, aad, bbd, ccd, lfootprint, cdate, ctime, compda, clibname, cfitsversion, dxcon2
     &                        , dycon2, cflm, lsourcetodeblend, nkover, nkoverlap, flag, convdelta )
       
       if (ltiming) then
         cput_significances = cput_significances + timer ( 'cpu', cpu_significances )
         walt_significances = walt_significances + timer ( 'wal', wal_significances )
         write (60,'(a,2(1pe9.2,a))') ' Total ''SIGNIFICANCES''   CPU',cput_significances,' WALL',walt_significances,' sec'
       endif
       
!----------------------------------------------------------------------------------------------------------------------------------
        
       if (cverbose .eq. '-verb2')
     &    write (iotty,'(a)') '   Improving footprints of non-Gaussian sources'

       if (ltiming) then
         cpu_footprints = timer ( 'cpu', 0.0d0 )
         wal_footprints = timer ( 'wal', 0.0d0 )
       endif

       call expandfootprints ( iotty, nx, ny, dx, dy, nextr, lnotbad, momxco, momyco, nxco, nyco, fluxpeakbest, fluxperror
     &                       , fluxtotbest, fluxterror, scalensx, afwhm, bfwhm, asize, bsize, beam, sigglob, sigmono, footfact
     &                       , obsname(idot:idot2-1), cdate, ctime, compda, clibname(lb:7), cfitsversion, lfirstcall, lfootprint
     &                       , foomajor, foominor, theta, maskimage, almostzero, obsfscbimage, factortotal, lnoexpansion
     &                       , lnoshrinkage )

!!       if (.not.lfirstcall) then
!!
!!         call wfits ( cfitsversion, nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, obsfscbimage, dx, dy, object
!!     &              , crval1, crval2, 'ann.footx.back.'//obscbname(1:iobsfscb), cdate, ctime, creator, beam
!!     &              , blank, crota1, crota2, cd11, cd12, cd21, cd22, equinox, bzero, bscale, wave, datamin, datamax, history )
!!     
!!         call wfits ( cfitsversion, nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, obsfscbimage, dx, dy, object
!!     &              , crval1, crval2, 'ann.footx.surr.'//obsname(1:iobs), cdate, ctime, creator, beam
!!     &              , blank, crota1, crota2, cd11, cd12, cd21, cd22, equinox, bzero, bscale, wave, datamin, datamax, history )
!!
!!         stop 99
!!       endif

       if (ltiming) then
         cput_footprints = cput_footprints + timer ( 'cpu', cpu_footprints )
         walt_footprints = walt_footprints + timer ( 'wal', wal_footprints )
         write (60,'(a,2(1pe9.2,a))') ' Total ''FOOTPRINTS''      CPU',cput_footprints,' WALL',walt_footprints,' sec'
       endif

       deallocate ( obsfsimage, obsfsbsimage, obsfscbimage, maskimage, bigfoot, smallfoot, bigfootclone, n1x, n2x, annulus
     &            , annulusarea, convdelta )

!-----------------------------------------------------------------------------------------------------------------------------------

       if (ltiming) then
         cpu_sum = cpu_sum + cput_initialize + cput_iteratepeaks + cput_testaccel + cput_secondmoments + cput_fwhmsizes 
     &                     + cput_totalfluxes + cput_fittedfluxes + cput_createannuli + cput_fluctuations1 + cput_fluctuations2
     &                     + cput_fluctuations3 + cput_nearestannulus + cput_secondannulus + cput_extendannulus + cput_significances
     &                     + cput_footprints
         wal_sum = wal_sum + walt_initialize + walt_iteratepeaks + walt_testaccel + walt_secondmoments + walt_fwhmsizes 
     &                     + walt_totalfluxes + walt_fittedfluxes + walt_createannuli + walt_fluctuations1 + walt_fluctuations2
     &                     + walt_fluctuations3 + walt_nearestannulus + walt_secondannulus + walt_extendannulus + walt_significances
     &                     + walt_footprints

! Total cpu usage, expressed in hours, minutes, and seconds.

         cputot = cpu_sum
         cpuhours = int ( cputot / 3600.0d0 )
         cpumins  = int ( cputot / 60.0d0 ) -  60 * cpuhours
         cpusecs  = int ( cputot + 0.5d0 ) - 3600 * cpuhours - 60 * cpumins

! Final wall clock time, expressed in hours, minutes, and seconds.

         wctot  = wal_sum
         wchours  = int ( wctot / 3600.0d0 )
         wcmins   = int ( wctot / 60.0d0 ) -  60 * wchours
         wcsecs   = int ( wctot + 0.5d0 ) - 3600 * wchours - 60 * wcmins

         write (60,'(a)') ' ------------------------------------------------------'
         write (60,'(a,2(1pe9.2,a))') ' ACCUMULATED CPU & WALL TIME', cpu_sum, ' sec ', wal_sum, ' sec'
         write (60,'(a,i3.2,a,i2.2,a,i2.2,a,i3.2,a,i2.2,a,i2.2,a,f7.2,a)') ' ---------------------------'
     &            , cpuhours,':',cpumins,':',cpusecs, ' ----', wchours,':',wcmins,':',wcsecs, ' ---'
  
         call when ( lunix, ctime, cdate, ndate, 4 )
  
         write (60,'(a)') ' SMEASURE finished: '//cdate//' '//ctime
         close ( 60 )
       endif

!-----------------------------------------------------------------------------------------------------------------------------------

       if (cverbose .eq. '-verb2')
     &     write (iotty,'(a)') '   Writing output catalog '''//ocatname(1:iocat)//''''
                                     
       fmtwcs='2(0pf13.7)'

       if (lfirstcall) then
         inquire ( file=wcsfile(1:iwcs), exist=lfnwcs )
         if (lfnwcs) then
           open ( 88, file=wcsfile(1:iwcs), status='old' )
           do n=1,nextr
             read (88,*,end=198,err=199) wcsxco(n), wcsyco(n), coosystem
           enddo
           close ( 88 )
           goto 200
  198      continue
             lfnwcs=.false.
             goto 200
  199      continue
             write (iotty,'(/a)') '   SMEASURE: ERROR: Trouble reading WCS coordinates from '''//wcsfile(1:iwcs)//''''
             stop 60
  200      continue
         endif
       endif

       open ( 15, file=ocatname(1:iocat), status='unknown' )

! Copy the original catalog header lines.

       ibad = 1
       igood = 1
       ibad = int ( log10 ( dble ( max ( nbad, 1 ) ) ) ) + 1
       write (cnextrbad,'(i5)') nbad
       ngood = nextr - nbad
       igood = int ( log10 ( dble ( max ( ngood, 1 ) ) ) ) + 1
       write (cnextrgood,'(i5)') ngood

       ncolumn1 = 0
       
       do i=1,nhead
         ic = lastc ( headline(i) )
         ncolumn = index ( headline(i)(1:ic), '# Tabulated quantities' )
         if (ncolumn .eq. 0) then
           ncolumn = index ( headline(i)(1:ic), '# TABULATED QUANTITIES' )
         endif
         if (ncolumn .gt. 0) then
           headline(i) = headline(i)(1:ic-1)//' ('//cnextrgood(5-igood+1:5)//' good + '//cnextrbad(5-ibad+1:5)//' bad):'
           ic = lastc ( headline(i) )
         endif
         ncolumn  = index ( headline(i)(1:ic), '#         measurement image ' )
         if (ncolumn1 .gt. 0 .and. ncolumn .eq. 0) then
           write (15,'(a)') '#         measurement image '''//obsname(1:iobs)//''''
         endif
         ncolumn1 = index ( headline(i)(1:ic), '# DETECTION TABLE for image ' )
         if (lfnwcs) then
           ncolumn = index ( headline(i)(1:ic), '#   XCO_PIXRO .....' )
           if (ncolumn .gt. 0) then
             headline(i) = '#   WCS_ACOOR ..... WCS ('//coosystem//') alpha-coordinate of current source ........... (degrees)'
           endif
           ncolumn = index ( headline(i)(1:ic), '#   YCO_PIXRO .....' )
           if (ncolumn .gt. 0) then
             headline(i) = '#   WCS_DCOOR ..... WCS ('//coosystem//') delta-coordinate of current source ........... (degrees)'
           endif
           ncolumn = index ( headline(i)(1:ic), 'XCO_PIXRO    YCO_PIXRO' )
           if (ncolumn .gt. 0) then
             headline(i) = headline(i)(1:ncolumn-1)//'WCS_ACOOR    WCS_DCOOR'//headline(i)(ncolumn+22:ic+9)
           endif
         endif
         write (15,'(a)') headline(i)(1:ic)
       enddo                                     

       do n=1,nextr
         ic = lastc ( catline(n) )
         write (cflg,'(i3)') flg(n)
!!         if (.not.lsourcetodeblend(n)) cflg(2:2) = '-'
         
         write (15,'(a,a,2(f11.4),'//fmtwcs//',a,2(1pe11.3),a,16(1pe11.3),a)')
     &         badsource(n), catline(n)(2:18), momxco(n), momyco(n), wcsxco(n), wcsyco(n), cflg, sigglob(n), good(n), cflm(n)
     &       , sigmono(n), goodmono(n), fluxpeakbest(n), fluxperror(n), fluxtotbest(n), fluxterror(n), fluxtotfit(n), scalensx(n)
     &       , afwhm(n), bfwhm(n), asize(n), bsize(n), theta(n), footfact(n), foomajor(n), foominor(n), catline(n)(npeakfx:ic)
       enddo

       close ( 15 )

       deallocate ( momxco, momyco, momxx, momyy, momxy, afwhm, bfwhm, theta, fluxtotfit, fluxpeakbest, fluxtotbest, equivrad
     &            , sigmono, sigglob, scalensx, catline, flag, badsource, backrms, fluxperror, cflm, apcorrect, fluxterror, dxcon2
     &            , dycon2, foomajor, foominor, flg, flm, lfootprint, wcsxco, wcsyco, slope, good, footfact, backav1, backav2
     &            , backav3, backrms1, backrms2, backrms3, aax, bbx, ccx, aad, bbd, ccd, dist2pks, dist2foo, asize, bsize
     &            , lsourcetodeblend, footfacto, foomajoro, momxxh, momyyh, momxyh, foominoro, thetao, nxco, nyco, nkover
     &            , nkoverlap, lnotbad, goodmono, headline )
       
       if (cverbose .eq. '-verb2') then
         write (iotty,'(/a)') ' Done.'
       endif

       if (lnoexpansion .and. lnoshrinkage) then
         inquire ( file='+wave.'//cowave(1:icw)//'.foot', exist=lexist )
         if (.not.lexist) then
           open ( 9, file='+wave.'//cowave(1:icw)//'.foot', status='unknown' )
           write (9,'(a)') 'Footprint expansion and shrinkage for the wave '//cowave(1:icw)//' has completed'
           close ( 9 )
         endif
       endif

!!       stop !<-- commented out because it would lead to run-time messages about denormalized values, when using gfortran.
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine auxiliaryarrays

     &            ( nx, ny, nextr, flag, momxco, momyco, afwhm, bfwhm, theta, sqpixel, almostzero, convtoradians, aax, bbx, ccx
     &            , aad, bbd, ccd, dxcon2, dycon2, foomajor, dist2pks, dist2foo)
!__________________________________________________________________________________________________________________________________
!
! Define auxiliary arrays.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       
       character*(*) flag(*)
       
       integer       n, nx, ny, nextr, l, m, k
       
       real*8        dbll, dblm, sigmaa2, sigmab2, sigmda2, sigmdb2, angle, cosa2, sina2, sin2a, sqpixel, almostzero, convtoradians
       
       real*8        momxco(*), momyco(*), afwhm(*), bfwhm(*), theta(*), aax(*), bbx(*), ccx(*), aad(*), bbd(*), ccd(*)
     &             , dxcon2(nextr,*), dycon2(nextr,*), dist2pks(nextr,*), dist2foo(nextr,*), foomajor(*)
!__________________________________________________________________________________________________________________________________
!
       do n=1,nextr
         
! Arrays for deblending.

         aad(n) = 0.0d0
         bbd(n) = 0.0d0
         ccd(n) = 0.0d0

! Standard arrays.

         aax(n) = 0.0d0
         bbx(n) = 0.0d0
         ccx(n) = 0.0d0

! Standard definition for an elliptical shape.

         sigmda2 = (afwhm(n) / 2.0d0)**2
         sigmdb2 = (bfwhm(n) / 2.0d0)**2
         sigmaa2 = (afwhm(n) / 2.0d0)**2
         sigmab2 = (bfwhm(n) / 2.0d0)**2

         if (flag(n) .ne. 'n') then
           angle = (90.0d0 - theta(n)) * convtoradians
           cosa2  = cos ( angle )**2
           sina2  = sin ( angle )**2
           sin2a  = sin ( 2.0d0 * angle )

! Arrays for deblending.

           aad(n) =  cosa2 / (sigmda2 + almostzero) + sina2 / (sigmdb2 + almostzero)
           bbd(n) = -sin2a / (sigmda2 + almostzero) + sin2a / (sigmdb2 + almostzero)
           ccd(n) =  sina2 / (sigmda2 + almostzero) + cosa2 / (sigmdb2 + almostzero)

! Standard definition.
           
           aax(n) =  cosa2 / (sigmaa2 + almostzero) + sina2 / (sigmab2 + almostzero)
           bbx(n) = -sin2a / (sigmaa2 + almostzero) + sin2a / (sigmab2 + almostzero)
           ccx(n) =  sina2 / (sigmaa2 + almostzero) + cosa2 / (sigmab2 + almostzero)
         endif

! Squares of distances of all pixels (in X and Y) from source position.

         do l=1,nx
           dbll = dble ( l )
           dxcon2(n,l) = (dbll - momxco(n))**2 * sqpixel
         enddo
         do m=1,ny
           dblm = dble ( m )
           dycon2(n,m) = (dblm - momyco(n))**2 * sqpixel
         enddo

! Squares of actual distances between source positions.
         
         do k=1,nextr
           dist2pks(n,k) = (momxco(n) - momxco(k))**2 + (momyco(n) - momyco(k))**2
           dist2foo(n,k) = (foomajor(n) + foomajor(k))**2 / (4.0d0 * sqpixel)
         enddo
       enddo
       
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine testacceleration

     &            ( nx, ny, dx, dy, nextr, lnotbad, momxco, momyco, nxco, nyco, foomajor, foominor, theta, footfacto, foomajoro
     &            , foominoro, thetao, lfootprint, maskimage, dist2pks, dist2foo, laccelerate, lsourcetodeblend, nkover, nkoverlap )
!__________________________________________________________________________________________________________________________________
!
! Decide if sources need to be remeasured or not. If they are isolated and footprint is the same, no need to remeasure.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       
       logical       inellipse, lfootprint(*), lsourcetodeblend(*), lnotbad(*), laccelerate
       
       integer       nx, ny, k, n, nr, l1, l2, m1, m2, m, l, kk, j, nextr, nkacc, nxco(*), nyco(*), nkover(*), nkoverlap(nextr,*)
     &             , nkaccel(nextr)
       
       real*8        dx, dy, dblm, dbll, semimajor, foomajorx, foominorx, dxdy2
       
       real*8        footfacto(*), foominor(*), foomajor(*), theta(*), foominoro(*), foomajoro(*), thetao(*), momxco(*)
     &             , dist2pks(nextr,*), dist2foo(nextr,*), momyco(*), maskimage(nx,ny)
       
       external      inellipse
!__________________________________________________________________________________________________________________________________
!
       dxdy2 = 2.0d0 * (dx + dy)
       
       do n=1,nextr
         
         lsourcetodeblend(n) = .true.
         
         if (laccelerate) then
           
           if (lnotbad(n) .and. lfootprint(n) .and. footfacto(n) .lt. 0.0d0 .and.
     &         2.0d0 * abs ( foomajor(n) - foomajoro(n) ) / (foomajor(n) + foomajoro(n)) .lt. 0.01d0 .and. 
     &         2.0d0 * abs ( foominor(n) - foominoro(n) ) / (foominor(n) + foominoro(n)) .lt. 0.01d0 .and. 
     &         2.0d0 * abs (    theta(n) - thetao(n) ) / (   theta(n) + thetao(n)) .lt. 0.01d0) then
             
             foomajorx = foomajor(n) + dxdy2
             foominorx = foominor(n) + dxdy2
             semimajor = foomajorx / 2.0d0
          
             nr = nint ( semimajor / min ( dx, dy ) )
             l1 = max ( nxco(n) - nr, 1 )
             l2 = min ( nxco(n) + nr, nx )
             m1 = max ( nyco(n) - nr, 1 )
             m2 = min ( nyco(n) + nr, ny )
          
             lsourcetodeblend(n) = .false.
             
             do m=m1,m2
               dblm = dble ( m )
               do l=l1,l2
                 dbll = dble ( l )
                   
                 if (maskimage(l,m) .gt. 0.9d0) then
                   if (inellipse ( dbll, dblm, dx, dy, momxco(n), momyco(n), foomajorx, foominorx, theta(n) )) then
                     do k=1,nextr
                       if (k .ne. n .and. lnotbad(k) .and. lfootprint(k)) then
                         if (dist2pks(n,k) .lt. dist2foo(n,k)) then
                           if (inellipse ( dbll, dblm, dx, dy, momxco(k), momyco(k), foomajor(k), foominor(k), theta(k) )) then
                             lsourcetodeblend(n) = .true.
                             goto 10
                           endif
                         endif
                       endif
                     enddo
                   endif
                 endif
               enddo
             enddo
           endif
         endif
  10     continue         
       enddo
       
! Identify and store overlapping source numbers to accelerate measurements.

       do n=1,nextr
         
         nkover(n) = 0
       
         if (lnotbad(n) .and. lfootprint(n)) then

           nkacc = 0
           do k=1,nextr
             if (k .ne. n .and. lnotbad(k) .and. lfootprint(k)) then
               if (maskimage(nxco(k),nyco(k)) .gt. 0.9d0 .and. dist2pks(n,k) .lt. dist2foo(n,k)) then
                 nkacc = nkacc + 1
                 nkaccel(nkacc) = k
               endif
             endif
           enddo
           
           semimajor = foomajor(n) / 2.0d0
           nr = nint ( semimajor / min ( dx, dy ) )
           l1 = max ( nxco(n) - nr, 1 )
           l2 = min ( nxco(n) + nr, nx )
           m1 = max ( nyco(n) - nr, 1 )
           m2 = min ( nyco(n) + nr, ny )
        
           do m=m1,m2
             dblm = dble ( m )
             do l=l1,l2
               dbll = dble ( l )
                 
               if (maskimage(l,m) .gt. 0.9d0) then
                 if (inellipse ( dbll, dblm, dx, dy, momxco(n), momyco(n), foomajor(n), foominor(n), theta(n) )) then
                   do j=1,nkacc
                     k = nkaccel(j)
                     do kk=1,nkover(n)
                       if (k .eq. nkoverlap(n,kk)) goto 20
                     enddo
                     if (inellipse ( dbll, dblm, dx, dy, momxco(k), momyco(k), foomajor(k), foominor(k), theta(k) )) then
                       nkover(n) = nkover(n) + 1
                       nkoverlap(n,nkover(n)) = k
                     endif
  20                 continue
                   enddo
                 endif
               endif
             enddo
           enddo
!!           write (*,'(100i5)') n, nkacc, (nkaccel(k), k=1,nkacc)
!!           write (*,'(100i5)') n, nkover(n), (nkoverlap(n,k), k=1,nkover(n))
         endif
       enddo

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine deblendpeaks

     &            ( nx, ny, dx, dy, nextr, lnotbad, fluxpeakbest, maskimage, flm, momxco, momyco, nxco, nyco, foominor, foomajor
     &            , theta, aad, bbd, ccd, obsname, obsfsbsimage, lfootprint, cdate, ctime, compda, clibname, cfitsversion
     &            , dxcon2, dycon2, cflm, almostzero, lsourcetodeblend, nkover, nkoverlap, flag, lmonosignificances, convdelta )
!__________________________________________________________________________________________________________________________________
!
! Deblend peak intensities of overlapping sources (in iterations).
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       
       logical       inellipse, lallsmall, lhavenonzeros, lmonosignificances, lfootprint(*), lsourcetodeblend(*), lnotbad(*)
     &             , lconverged(nextr)  !! , lexist
       
       character*2   cniter
       character*(*) flag(*), cflm(*), cfitsversion, ctime, cdate, compda, obsname, clibname
       
       integer       lastc, nx, ny, k, n, j, nextr, nxco(*), nyco(*), iter, idot, idot2, iobs, i, nit, nkacc, flm(*), nkover(*)
     &             , nkoverlap(nextr,*), nkaccel(nextr), ian  !!, nn, l1, l2, m1, m2, nr, l, m  !!, np
       
       real*8        dx, dy, almostzero, changemax, qsign, funn, funksum, argexp, argexpm, gaussfun, relerror, deblendfactor, dlog4h
!!     &             , semimajor, dbll, dblm, selectk
     
       parameter   ( relerror = 0.001d0, changemax = 0.2d0 )
       
       real*8        fluxpeakbest(*), momxco(*), momyco(*), fluxpiter(nextr), foominor(*), foomajor(*), theta(*), aad(*), bbd(*)
     &             , ccd(*), dxcon2(nextr,*), dycon2(nextr,*), fluxpkini(nextr), obsfsbsimage(nx,ny), convdelta(nextr,*)
     &             , maskimage(nx,ny) !!, corrneg(nextr) !!, funx(nx*ny)

       parameter   ( dlog4h = log ( 4.0d0 ) / 2.0d0 )
       
       external      lastc, inellipse  !!, selectk
!__________________________________________________________________________________________________________________________________
!
       do n=1,nextr
         
! Attempt to correct for negative bg-subtracted peak values for very faint noisy sources.

!!         semimajor = foomajor(n) / 2.0d0
!!         nr = nint ( semimajor / min ( dx, dy ) )
!!         l1 = max ( nxco(n) - nr, 1 )
!!         l2 = min ( nxco(n) + nr, nx )
!!         m1 = max ( nyco(n) - nr, 1 )
!!         m2 = min ( nyco(n) + nr, ny )
!!         corrneg(n) = 0.0d0
!!         nn = 0
!!
!!         do m=m1,m2
!!           dblm = dble ( m )
!!           do l=l1,l2
!!             dbll = dble ( l )
!!  
!!             if (maskimage(l,m) .gt. 0.9d0) then
!!               if (inellipse ( dbll, dblm, dx, dy, momxco(n), momyco(n), foomajor(n), foominor(n), theta(n) )) then
!!                 if (obsfsbsimage(l,m) .lt. 0.0d0) then
!!                   nn = nn + 1
!!                   corrneg(n) = corrneg(n) + abs ( obsfsbsimage(l,m) )
!!!!                   funx(nn) = abs ( obsfsbsimage(l,m) )
!!                 endif
!!               endif
!!             endif
!!           enddo
!!         enddo
!!         if (nn .gt. 1) corrneg(n) = corrneg(n) / dble ( nn )
!!!!         corrneg(n) = selectk ( (nn + 1) / 2, nn, funx )
!!
!!! Somewhat smaller corrections seem to gove better results.
!!
!!         corrneg(n) = 0.5d0 * corrneg(n)

         if (flag(n) .ne. 'n') then
           fluxpiter(n) = obsfsbsimage(nxco(n),nyco(n))
         else
           fluxpiter(n) = 0.0d0
         endif
!!         fluxpiter(n) = fluxpiter(n) + corrneg(n)
         fluxpkini(n) = fluxpiter(n)
         lconverged(n) = .false.
       enddo
       
       iter = 0
  111  continue
       iter = iter + 1

       do n=1,nextr
         
         if (lnotbad(n) .and. lfootprint(n) .and. .not.lconverged(n)) then
           if (lsourcetodeblend(n) .or. lmonosignificances) then

             nkacc = 0
             do j=1,nkover(n)
               k = nkoverlap(n,j)
               if (fluxpiter(k) .gt. almostzero) then
                 nkacc = nkacc + 1
                 nkaccel(nkacc) = k
               endif
             enddo
             fluxpeakbest(n) = 0.0d0
             
             if (maskimage(nxco(n),nyco(n)) .gt. 0.9d0) then
               funn = fluxpiter(n)
               funksum = 0.0d0
               do j=1,nkacc
                 k = nkaccel(j)
                 if (inellipse ( momxco(n), momyco(n), dx, dy, momxco(k), momyco(k), foomajor(k), foominor(k), theta(k) )) then
                   argexp = aad(k) * dxcon2(k,nxco(n)) + bbd(k) * sqrt ( dxcon2(k,nxco(n)) * dycon2(k,nyco(n)) )
     &                    + ccd(k) * dycon2(k,nyco(n))
                   argexpm = min ( argexp * dlog4h, 100.0d0 )
                   gaussfun = fluxpiter(k) * exp ( - argexpm )
                   funksum = funksum + gaussfun
                 endif
               enddo
               deblendfactor = 1.0d0
               if (funn + funksum .ne. 0.0d0) then
                 deblendfactor = funn / (funn + funksum)
               endif
               fluxpeakbest(n) = fluxpkini(n) * deblendfactor
             endif
           endif
         endif
       enddo

! Check convergence, limiting possible drastic changes in the quantities to reasonable values.

       lallsmall = .true.
       
       do n=1,nextr
         convdelta(n,iter) = 0.0d0
         if (.not.lconverged(n) .and. lsourcetodeblend(n)) then
           lconverged(n) = .true.
           if (lnotbad(n) .and. lfootprint(n)) then
             if (maskimage(nxco(n),nyco(n)) .gt. 0.9d0) then
               if (abs ( fluxpeakbest(n) - fluxpiter(n) ) .gt. changemax * abs ( fluxpiter(n) )) then
                 qsign = sign ( 1.0d0, fluxpeakbest(n) - fluxpiter(n) )
                 convdelta(n,iter) = qsign * changemax
                 fluxpiter(n) = (1.0d0 + qsign * changemax) * fluxpiter(n)
               else
                 convdelta(n,iter) = (fluxpeakbest(n) - fluxpiter(n)) / (fluxpiter(n) + almostzero)
                 fluxpiter(n) = fluxpeakbest(n)
               endif
               if (abs ( convdelta(n,iter) ) .gt. relerror) then
                 lconverged(n) = .false.
                 lallsmall = .false.
               endif
             endif
           endif
         endif
       enddo

       if (.not.lallsmall .and. iter .lt. 99) goto 111

! Monochromatic flag of a source will become negative, if the above deblending iterations have not converged in 99 iterations.

       if (.not.lmonosignificances) then

         do n=1,nextr
           if (.not.lconverged(n) .and. lsourcetodeblend(n)) then
             if (abs ( flm(n) ) .ge. 1) then
               write (cflm(n),'(i5)') -abs ( flm(n) )
             else
               cflm(n) = '   -0'
             endif
           else
             write (cflm(n),'(i5)') abs ( flm(n) )
           endif
         enddo
       endif
       
       lhavenonzeros = .false.
       
       do n=1,nextr
         if (lnotbad(n)) then
           if (maskimage(nxco(n),nyco(n)) .lt. 0.9d0) then
             fluxpeakbest(n) = 0.0d0
           endif
         endif
         if (lnotbad(n) .and. lfootprint(n)) then
           do i=1,iter
             if (abs ( convdelta(n,i) ) .gt. almostzero) lhavenonzeros = .true.
           enddo
         endif
       enddo
       
! Save deblending information (only when greater than zero).

       if (lhavenonzeros) then
         iobs = lastc ( obsname )
         idot = index ( obsname, '.' )
         idot2 = index ( obsname(idot+1:iobs), '.' ) + idot
         ian = index ( obsname, '.an.' )
         if (lmonosignificances) then
!!           inquire ( file=obsname(1:iobs)//'.fpeak.deblending.dat', exist=lexist )
!!           if (lexist) then
!!             open ( 17, file=obsname(1:iobs)//'.fpeak.deblending.dat', status='old' )
!!             close (17, status='delete' )
!!           endif
           open ( 17, file=obsname(1:iobs-5)//'.fpeak.deblending.dat', status='unknown' )
         else
           open ( 17, file='sm'//obsname(idot:idot2-1)//'.fpeak.deblending.dat', status='unknown' )
         endif
         nit = int ( log10 ( dble ( iter ) ) ) + 1
         write (cniter,'(i2)') iter
         write (17,'(a)') '# Date: '//cdate//' Time: '//ctime//'  Created by SMEASURE compiled on '//compda
         write (17,'(a)') '# Alexander Men''shchikov, DAp IRFU CEA Saclay, France (alexander.menshchikov@cea.fr)'
         write (17,'(a)') '# Used '//clibname//' library version '//cfitsversion//' by William D Pence (william.d.pence@nasa.gov)'
         write (17,'(a)') '#'
         write (17,'(a)') '# CONVERGENCE OF PEAK INTENSITY ITERATIONS (DEBLENDING OVERLAPPING SOURCES).'
         write (17,'(a)') '#'
         write (17,'(a)') '#   N .............. Source number'
         write (17,'(a)') '#   PEAKINI ........ Initial peak intensity'
         write (17,'(a)') '#   PEAKDEB ........ Deblended peak intensity'
         write (17,'(a)') '#   {1 2 3 etc.} ... Iteration number for source N'
         write (17,'(a)') '#   REL_DELTA_i .... Relative difference of peak intensity between two iterations'
         write (17,'(a)') '#'
         write (17,'(a,1pe10.3)') '# NOTE: Maximum of 99 iterations down to a relative difference of', relerror
         write (17,'(a)') '#'
         write (17,'(a)') '#                      •~~~~~~>  RELATIVE_DELTA_i (i=1,MAXITER), MAXITER = '//cniter(2-nit+1:2)
         write (17,'(a)') '#'
         write (17,'(a,'//cniter//'i10)') '#    N   PEAKINI    PEAKDEB         1         2', (i, i=3,iter)
         write (17,'(a)') '#'
        
         do n=1,nextr
           
           if (lnotbad(n) .and. lfootprint(n)) then
             lhavenonzeros = .false.
             do i=1,iter
               if (abs ( convdelta(n,i) ) .gt. almostzero) lhavenonzeros = .true.
             enddo
             if (maskimage(nxco(n),nyco(n)) .gt. 0.9d0 .and. lhavenonzeros) then
               write (17,'(i6,2(1pe11.3),4x,'//cniter//'(1pg10.2))') n, fluxpkini(n), fluxpeakbest(n), (convdelta(n,i),i=1,iter)
             endif
           endif
         enddo
        
         close ( 17 )
       endif

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine intensmoments

     &            ( nx, ny, dx, dy, nextr, lnotbad, sigmono, fluxpeakbest, momxco, momyco, nxco, nyco, foominor  !!, fluxperror
     &            , foomajor, aad, bbd, ccd, dxcon2, dycon2, lfootprint, sqpixel, almostzero, momxx, momxy, momyy, maskimage
     &            , obsfsbsimage, equivrad, pi, lsourcetodeblend, nkover, nkoverlap, momxxh, momxyh, momyyh, afwhm, bfwhm, theta )
!__________________________________________________________________________________________________________________________________
!
! Compute intensity moments, deblending overlapping sources.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       
       logical       inellipse, lcond1, lcond2, lcond3, lcond4, lcond5, lcond6, lcond7, lcond8, lfootprint(*), lsourcetodeblend(*)
     &             , lnotbad(*), lpositive, lnegative
       
       integer       nx, ny, k, n, nr, nextr, l1, l2, m1, m2, m, l, j, numpix, nkacc, nhalfmax, lm1, lp1, mm1, mp1, nn, nkover(*)
     &             , nkoverlap(nextr,*), nxco(*), nyco(*), nkaccel(nextr)
       
       real*8        dx, dy, almostzero, funksum, argexp, semimajor, dblm, dbll, intd, xcorc, ycorc, intxx, intxxh, funn
     &             , intyy, intyyh, intxy, intxyh, int0m, deblended, sqpixel, pi, xxcorc, yycorc, xycorc, deblendfactor, int0hm
     &             , int0, int0h, halfpeak, deblended1, deblended2, deblended3, deblended4, deblended5, deblended6, deblended7
     &             , deblended8, dbll2, dbmm2, radius2, drad2, fwhm, hwhm, radhalf, dlogim, sigm2, funksum1, funksum2, funksum3
     &             , funksum4, funksum5, funksum6, funksum7, funksum8, dbllm1, dbllp1, dblmm1, dblmp1, argexpm, deblendfactor1
     &             , deblendfactor2, deblendfactor3, deblendfactor4, deblendfactor5, deblendfactor6, deblendfactor7, deblendfactor8
     &             , funn1, funn2, funn3, funn4, funn5, funn6, funn7, funn8, dlog4h, deblendedabs  !!, xhalf, yhalf, xcent, ycent
       
       real*8        momxx(*), momxy(*), momyy(*), momxco(*), momyco(*), foominor(*), foomajor(*), aad(*), bbd(*)
     &             , ccd(*), dxcon2(nextr,*), dycon2(nextr,*), fluxpeakbest(*), equivrad(*), momxxh(*), momxyh(*), momyyh(*)
     &             , obsfsbsimage(nx,ny), maskimage(nx,ny), afwhm(*), bfwhm(*), theta(*), sigmono(*) !!, fluxperror(*)

       parameter   ( dlog4h = log ( 4.0d0 ) / 2.0d0 )
       
       external      inellipse
!__________________________________________________________________________________________________________________________________
!
       do n=1,nextr
         
         intxx = 0.0d0
         intyy = 0.0d0
         intxy = 0.0d0           
         int0 = 0.0d0
         intxxh = 0.0d0
         intyyh = 0.0d0
         intxyh = 0.0d0           
         int0h = 0.0d0
         
         if (lnotbad(n) .and. lfootprint(n) .and. lsourcetodeblend(n)) then
           
           numpix = 0
           semimajor = foomajor(n) / 2.0d0
           nr = nint ( semimajor / min ( dx, dy ) )
           
           if (maskimage(nxco(n),nyco(n)) .gt. 0.9d0) then

             nkacc = 0
             do j=1,nkover(n)
               k = nkoverlap(n,j)
               if (fluxpeakbest(k) .gt. almostzero) then
                 nkacc = nkacc + 1
                 nkaccel(nkacc) = k
               endif
             enddo

             hwhm = 0.0d0
             fwhm = 0.0d0
!!             xcent = 0.0d0
!!             ycent = 0.0d0
             halfpeak = fluxpeakbest(n) / 2.0d0
             nhalfmax = 0
             
             l1 = max ( nxco(n) - nr, 1 )
             l2 = min ( nxco(n) + nr, nx )
             m1 = max ( nyco(n) - nr, 1 )
             m2 = min ( nyco(n) + nr, ny )

             do m=m1,m2
               dblm = dble ( m )
               ycorc = dblm - momyco(n)
               dbmm2 = ycorc**2
               do l=l1,l2
                 dbll = dble ( l )
                 xcorc = dbll - momxco(n)
                 dbll2 = xcorc**2

                 if (maskimage(l,m) .gt. 0.9d0) then    !! .and. obsfsbsimage(l,m) .gt. 0.0d0
                   
                   if (inellipse ( dbll, dblm, dx, dy, momxco(n), momyco(n), foomajor(n), foominor(n), theta(n) )) then
                     
                     lm1 = max ( l - 1, 1 )
                     mm1 = max ( m - 1, 1 )
                     lp1 = min ( l + 1, nx )
                     mp1 = min ( m + 1, ny )
                     dbllm1 = dble ( lm1 )
                     dblmm1 = dble ( mm1 )
                     dbllp1 = dble ( lp1 )
                     dblmp1 = dble ( mp1 )
                   
                     argexp = aad(n) * dxcon2(n,l) + bbd(n) * sqrt ( dxcon2(n,l) * dycon2(n,m) ) + ccd(n) * dycon2(n,m)
                     argexpm = min ( argexp * dlog4h, 100.0d0 )
                     funn = fluxpeakbest(n) * exp ( - argexpm )
                     argexp = aad(n) * dxcon2(n,lm1) + bbd(n) * sqrt ( dxcon2(n,lm1) * dycon2(n,m) ) + ccd(n) * dycon2(n,m)
                     argexpm = min ( argexp * dlog4h, 100.0d0 )
                     funn1 = fluxpeakbest(n) * exp ( - argexpm )
                     argexp = aad(n) * dxcon2(n,lp1) + bbd(n) * sqrt ( dxcon2(n,lp1) * dycon2(n,m) ) + ccd(n) * dycon2(n,m)
                     argexpm = min ( argexp * dlog4h, 100.0d0 )
                     funn2 = fluxpeakbest(n) * exp ( - argexpm )
                     argexp = aad(n) * dxcon2(n,l) + bbd(n) * sqrt ( dxcon2(n,l) * dycon2(n,mm1) ) + ccd(n) * dycon2(n,mm1)
                     argexpm = min ( argexp * dlog4h, 100.0d0 )
                     funn3 = fluxpeakbest(n) * exp ( - argexpm )
                     argexp = aad(n) * dxcon2(n,l) + bbd(n) * sqrt ( dxcon2(n,l) * dycon2(n,mp1) ) + ccd(n) * dycon2(n,mp1)
                     argexpm = min ( argexp * dlog4h, 100.0d0 )
                     funn4 = fluxpeakbest(n) * exp ( - argexpm )
                     argexp = aad(n) * dxcon2(n,lm1) + bbd(n) * sqrt ( dxcon2(n,lm1) * dycon2(n,mm1) ) + ccd(n) * dycon2(n,mm1)
                     argexpm = min ( argexp * dlog4h, 100.0d0 )
                     funn5 = fluxpeakbest(n) * exp ( - argexpm )
                     argexp = aad(n) * dxcon2(n,lm1) + bbd(n) * sqrt ( dxcon2(n,lm1) * dycon2(n,mp1) ) + ccd(n) * dycon2(n,mp1)
                     argexpm = min ( argexp * dlog4h, 100.0d0 )
                     funn6 = fluxpeakbest(n) * exp ( - argexpm )
                     argexp = aad(n) * dxcon2(n,lp1) + bbd(n) * sqrt ( dxcon2(n,lp1) * dycon2(n,mm1) ) + ccd(n) * dycon2(n,mm1)
                     argexpm = min ( argexp * dlog4h, 100.0d0 )
                     funn7 = fluxpeakbest(n) * exp ( - argexpm )
                     argexp = aad(n) * dxcon2(n,lp1) + bbd(n) * sqrt ( dxcon2(n,lp1) * dycon2(n,mp1) ) + ccd(n) * dycon2(n,mp1)
                     argexpm = min ( argexp * dlog4h, 100.0d0 )
                     funn8 = fluxpeakbest(n) * exp ( - argexpm )

                     funksum  = 0.0d0
                     funksum1 = 0.0d0
                     funksum2 = 0.0d0
                     funksum3 = 0.0d0
                     funksum4 = 0.0d0
                     funksum5 = 0.0d0
                     funksum6 = 0.0d0
                     funksum7 = 0.0d0
                     funksum8 = 0.0d0
                     
                     do j=1,nkacc
                       k = nkaccel(j)
                       if (inellipse ( dbll, dblm, dx, dy, momxco(k), momyco(k), foomajor(k), foominor(k), theta(k) )) then
                         argexp = aad(k) * dxcon2(k,l) + bbd(k) * sqrt ( dxcon2(k,l) * dycon2(k,m) ) + ccd(k) * dycon2(k,m)
                         argexpm = min ( argexp * dlog4h, 100.0d0 )
                         funksum = funksum + fluxpeakbest(k) * exp ( - argexpm )
                       endif
                       if (inellipse ( dbllm1, dblm, dx, dy, momxco(k), momyco(k), foomajor(k), foominor(k), theta(k) )) then
                         argexp = aad(k) * dxcon2(k,lm1) + bbd(k) * sqrt ( dxcon2(k,lm1) * dycon2(k,m) ) + ccd(k) * dycon2(k,m)
                         argexpm = min ( argexp * dlog4h, 100.0d0 )
                         funksum1 = funksum1 + fluxpeakbest(k) * exp ( - argexpm )
                       endif
                       if (inellipse ( dbllp1, dblm, dx, dy, momxco(k), momyco(k), foomajor(k), foominor(k), theta(k) )) then
                         argexp = aad(k) * dxcon2(k,lp1) + bbd(k) * sqrt ( dxcon2(k,lp1) * dycon2(k,m) ) + ccd(k) * dycon2(k,m)
                         argexpm = min ( argexp * dlog4h, 100.0d0 )
                         funksum2 = funksum2 + fluxpeakbest(k) * exp ( - argexpm )
                       endif
                       if (inellipse ( dbll, dblmm1, dx, dy, momxco(k), momyco(k), foomajor(k), foominor(k), theta(k) )) then
                         argexp = aad(k) * dxcon2(k,l) + bbd(k) * sqrt ( dxcon2(k,l) * dycon2(k,mm1) ) + ccd(k) * dycon2(k,mm1)
                         argexpm = min ( argexp * dlog4h, 100.0d0 )
                         funksum3 = funksum3 + fluxpeakbest(k) * exp ( - argexpm )
                       endif
                       if (inellipse ( dbll, dblmp1, dx, dy, momxco(k), momyco(k), foomajor(k), foominor(k), theta(k) )) then
                         argexp = aad(k) * dxcon2(k,l) + bbd(k) * sqrt ( dxcon2(k,l) * dycon2(k,mp1) ) + ccd(k) * dycon2(k,mp1)
                         argexpm = min ( argexp * dlog4h, 100.0d0 )
                         funksum4 = funksum4 + fluxpeakbest(k) * exp ( - argexpm )
                       endif
                       if (inellipse ( dbllm1, dblmm1, dx, dy, momxco(k), momyco(k), foomajor(k), foominor(k), theta(k) )) then
                         argexp = aad(k) * dxcon2(k,lm1) + bbd(k) * sqrt ( dxcon2(k,lm1) * dycon2(k,mm1) ) + ccd(k) * dycon2(k,mm1)
                         argexpm = min ( argexp * dlog4h, 100.0d0 )
                         funksum5 = funksum5 + fluxpeakbest(k) * exp ( - argexpm )
                       endif
                       if (inellipse ( dbllm1, dblmp1, dx, dy, momxco(k), momyco(k), foomajor(k), foominor(k), theta(k) )) then
                         argexp = aad(k) * dxcon2(k,lm1) + bbd(k) * sqrt ( dxcon2(k,lm1) * dycon2(k,mp1) ) + ccd(k) * dycon2(k,mp1)
                         argexpm = min ( argexp * dlog4h, 100.0d0 )
                         funksum6 = funksum6 + fluxpeakbest(k) * exp ( - argexpm )
                       endif
                       if (inellipse ( dbllp1, dblmm1, dx, dy, momxco(k), momyco(k), foomajor(k), foominor(k), theta(k) )) then
                         argexp = aad(k) * dxcon2(k,lp1) + bbd(k) * sqrt ( dxcon2(k,lp1) * dycon2(k,mm1) ) + ccd(k) * dycon2(k,mm1)
                         argexpm = min ( argexp * dlog4h, 100.0d0 )
                         funksum7 = funksum7 + fluxpeakbest(k) * exp ( - argexpm )
                       endif
                       if (inellipse ( dbllp1, dblmp1, dx, dy, momxco(k), momyco(k), foomajor(k), foominor(k), theta(k) )) then
                         argexp = aad(k) * dxcon2(k,lp1) + bbd(k) * sqrt ( dxcon2(k,lp1) * dycon2(k,mp1) ) + ccd(k) * dycon2(k,mp1)
                         argexpm = min ( argexp * dlog4h, 100.0d0 )
                         funksum8 = funksum8 + fluxpeakbest(k) * exp ( - argexpm )
                       endif
                     enddo
                     
                     deblendfactor = 1.0d0
                     if (funn + funksum .ne. 0.0d0) then
                       deblendfactor  = funn / (funn + funksum )
                     endif
                     deblendfactor1 = 1.0d0
                     if (funn1 + funksum1 .ne. 0.0d0) then
                       deblendfactor1  = funn1 / (funn1 + funksum1 )
                     endif
                     deblendfactor2 = 1.0d0
                     if (funn2 + funksum2 .ne. 0.0d0) then
                       deblendfactor2  = funn2 / (funn2 + funksum2 )
                     endif
                     deblendfactor3 = 1.0d0
                     if (funn3 + funksum3 .ne. 0.0d0) then
                       deblendfactor3  = funn3 / (funn3 + funksum3 )
                     endif
                     deblendfactor4 = 1.0d0
                     if (funn4 + funksum4 .ne. 0.0d0) then
                       deblendfactor4  = funn4 / (funn4 + funksum4 )
                     endif
                     deblendfactor5 = 1.0d0
                     if (funn5 + funksum5 .ne. 0.0d0) then
                       deblendfactor5  = funn5 / (funn5 + funksum5 )
                     endif
                     deblendfactor6 = 1.0d0
                     if (funn6 + funksum6 .ne. 0.0d0) then
                       deblendfactor6  = funn6 / (funn6 + funksum6 )
                     endif
                     deblendfactor7 = 1.0d0
                     if (funn7 + funksum7 .ne. 0.0d0) then
                       deblendfactor7  = funn7 / (funn7 + funksum7 )
                     endif
                     deblendfactor8 = 1.0d0
                     if (funn8 + funksum8 .ne. 0.0d0) then
                       deblendfactor8  = funn8 / (funn8 + funksum8 )
                     endif
                     deblended  = obsfsbsimage(l,m    ) * deblendfactor
                     deblended1 = obsfsbsimage(lm1,m  ) * deblendfactor1
                     deblended2 = obsfsbsimage(lp1,m  ) * deblendfactor2
                     deblended3 = obsfsbsimage(l  ,mm1) * deblendfactor3
                     deblended4 = obsfsbsimage(l  ,mp1) * deblendfactor4
                     deblended5 = obsfsbsimage(lm1,mm1) * deblendfactor5
                     deblended6 = obsfsbsimage(lm1,mp1) * deblendfactor6
                     deblended7 = obsfsbsimage(lp1,mm1) * deblendfactor7
                     deblended8 = obsfsbsimage(lp1,mp1) * deblendfactor8
                     
                     deblendedabs = abs ( deblended )

! Signal that source is not measurable (just noise) and its footprint must be removed at this wavelength,
! because some pixels within central part of the footprint have an opposite sign with respect to the peak.
                     
                     radius2 = (dbll2 + dbmm2) * sqpixel
                     if (radius2 .le. (semimajor / 3.0d0)**2) then
                       if ((fluxpeakbest(n) .gt. 0.0d0 .and. deblended .lt. 0.0d0) .or.
     &                     (fluxpeakbest(n) .lt. 0.0d0 .and. deblended .gt. 0.0d0)) then
                         sigmono(n) = 5.555d-31
                         nhalfmax = 0
                         goto 333
                       endif
                     endif
                     
                     lpositive = fluxpeakbest(n) .gt. 0.0d0 .and. deblended .ge. halfpeak
                     lnegative = fluxpeakbest(n) .lt. 0.0d0 .and. deblended .le. halfpeak

! Direct measurements of the minimum and maximum sizes (at peak half-maximum) for an arbitrary intensity distribution.

                     if (lpositive .or. lnegative) then

                       if (lpositive) then
                         lcond1 = deblended1 .lt. halfpeak .and. deblended1 .gt. almostzero
                         lcond2 = deblended2 .lt. halfpeak .and. deblended2 .gt. almostzero
                         lcond3 = deblended3 .lt. halfpeak .and. deblended3 .gt. almostzero
                         lcond4 = deblended4 .lt. halfpeak .and. deblended4 .gt. almostzero
                         lcond5 = deblended5 .lt. halfpeak .and. deblended5 .gt. almostzero
                         lcond6 = deblended6 .lt. halfpeak .and. deblended6 .gt. almostzero
                         lcond7 = deblended7 .lt. halfpeak .and. deblended7 .gt. almostzero
                         lcond8 = deblended8 .lt. halfpeak .and. deblended8 .gt. almostzero
                       endif
                       if (lnegative) then
                         lcond1 = deblended1 .gt. halfpeak .and. deblended1 .lt. -almostzero
                         lcond2 = deblended2 .gt. halfpeak .and. deblended2 .lt. -almostzero
                         lcond3 = deblended3 .gt. halfpeak .and. deblended3 .lt. -almostzero
                         lcond4 = deblended4 .gt. halfpeak .and. deblended4 .lt. -almostzero
                         lcond5 = deblended5 .gt. halfpeak .and. deblended5 .lt. -almostzero
                         lcond6 = deblended6 .gt. halfpeak .and. deblended6 .lt. -almostzero
                         lcond7 = deblended7 .gt. halfpeak .and. deblended7 .lt. -almostzero
                         lcond8 = deblended8 .gt. halfpeak .and. deblended8 .lt. -almostzero
                       endif

                       if (lcond1 .or. lcond2 .or. lcond3 .or. lcond4 .or. lcond5 .or. lcond6 .or. lcond7 .or. lcond8) then
                         nhalfmax = nhalfmax + 1
                         dlogim = log ( deblendedabs ) - log ( abs ( halfpeak ) )
                         radhalf = 0.0d0
!!                         xhalf = 0.0d0
!!                         yhalf = 0.0d0
                         nn = 0

! Gaussian interpolation to obtain accurate values of full width at half-maximum for Gaussian-like shapes.
! From two equations: ln(G1) - ln(G2) = -x1^2 + x2^2 / (2 sigma^2) and ln(G1) - ln(Gh) = -x1^2 + xh^2 / (2 sigma^2):
! (2 sigma^2) = (x2^2 - x1^2) / (ln(G1) - ln(G2)) and xh = sqrt ( x1^2 + (2 sigma^2) * (ln(G1) - ln(Gh)) )
                   
                         if (lcond1) then
                           nn = nn + 1
                           drad2 = (dble ( lm1 - nxco(n) )**2 + dbmm2) * sqpixel - radius2
                           sigm2 = drad2 / (log ( deblendedabs ) - log ( abs ( deblended1 ) ))
                           radhalf = radhalf + sqrt ( radius2 + sigm2 * dlogim )
!!                           xhalf = xhalf + dble ( lm1 )
!!                           yhalf = yhalf + dble ( m )
                         endif
                         if (lcond2) then
                           nn = nn + 1
                           drad2 = (dble ( lp1 - nxco(n) )**2 + dbmm2) * sqpixel - radius2
                           sigm2 = drad2 / (log ( deblendedabs ) - log ( abs ( deblended2 ) ))
                           radhalf = radhalf + sqrt ( radius2 + sigm2 * dlogim )
!!                           xhalf = xhalf + dble ( lp1 )
!!                           yhalf = yhalf + dble ( m )
                         endif
                         if (lcond3) then
                           nn = nn + 1
                           drad2 = (dbll2 + dble ( mm1 - nyco(n) )**2) * sqpixel - radius2
                           sigm2 = drad2 / (log ( deblendedabs ) - log ( abs ( deblended3 ) ))
                           radhalf = radhalf + sqrt ( radius2 + sigm2 * dlogim )
!!                           xhalf = xhalf + dble ( l )
!!                           yhalf = yhalf + dble ( mm1 )
                         endif
                         if (lcond4) then
                           nn = nn + 1
                           drad2 = (dbll2 + dble ( mp1 - nyco(n) )**2) * sqpixel - radius2
                           sigm2 = drad2 / (log ( deblendedabs ) - log ( abs ( deblended4 ) ))
                           radhalf = radhalf + sqrt ( radius2 + sigm2 * dlogim )
!!                           xhalf = xhalf + dble ( l )
!!                           yhalf = yhalf + dble ( mp1 )
                         endif
                         if (lcond5) then
                           nn = nn + 1
                           drad2 = (dble ( lm1 - nxco(n) )**2 + dble ( mm1 - nyco(n) )**2) * sqpixel - radius2
                           sigm2 = drad2 / (log ( deblendedabs ) - log ( abs ( deblended5 ) ))
                           radhalf = radhalf + sqrt ( radius2 + sigm2 * dlogim )
!!                           xhalf = xhalf + dble ( lm1 )
!!                           yhalf = yhalf + dble ( mm1 )
                         endif
                         if (lcond6) then
                           nn = nn + 1
                           drad2 = (dble ( lm1 - nxco(n) )**2 + dble ( mp1 - nyco(n) )**2) * sqpixel - radius2
                           sigm2 = drad2 / (log ( deblendedabs ) - log ( abs ( deblended6 ) ))
                           radhalf = radhalf + sqrt ( radius2 + sigm2 * dlogim )
!!                           xhalf = xhalf + dble ( lm1 )
!!                           yhalf = yhalf + dble ( mp1 )
                         endif
                         if (lcond7) then
                           nn = nn + 1
                           drad2 = (dble ( lp1 - nxco(n) )**2 + dble ( mm1 - nyco(n) )**2) * sqpixel - radius2
                           sigm2 = drad2 / (log ( deblendedabs ) - log ( abs ( deblended7 ) ))
                           radhalf = radhalf + sqrt ( radius2 + sigm2 * dlogim )
!!                           xhalf = xhalf + dble ( lp1 )
!!                           yhalf = yhalf + dble ( mm1 )
                         endif
                         if (lcond8) then
                           nn = nn + 1
                           drad2 = (dble ( lp1 - nxco(n) )**2 + dble ( mp1 - nyco(n) )**2) * sqpixel - radius2
                           sigm2 = drad2 / (log ( deblendedabs ) - log ( abs ( deblended8 ) ))
                           radhalf = radhalf + sqrt ( radius2 + sigm2 * dlogim )
!!                           xhalf = xhalf + dble ( lp1 )
!!                           yhalf = yhalf + dble ( mp1 )
                         endif
                         if (nn .gt. 0) then
                           radhalf = radhalf / dble ( nn )
!!                           xhalf = xhalf / dble ( nn )
!!                           yhalf = yhalf / dble ( nn )
                         endif
                         hwhm = hwhm + radhalf
!!                         xcent = xcent + xhalf
!!                         ycent = ycent + yhalf
                       endif
                     endif

! Intensity moments work fine for negative peaks (in contrast to the above direct sizes), no need in taking absolute values.
                     
                     numpix = numpix + 1
                     intd = deblended * sqpixel
                     xxcorc = xcorc * xcorc * intd
                     yycorc = ycorc * ycorc * intd
                     xycorc = xcorc * ycorc * intd
                     intxx = intxx + xxcorc
                     intyy = intyy + yycorc
                     intxy = intxy + xycorc
                     int0 = int0 + intd

! Next special quantities will be used in MOMENTSIZES for elongation calculation only.
                     
                     lpositive = fluxpeakbest(n) .gt. 0.0d0 .and. deblended .ge. 0.1d0 * fluxpeakbest(n)
                     lnegative = fluxpeakbest(n) .lt. 0.0d0 .and. deblended .le. 0.1d0 * fluxpeakbest(n)
                     
                     if (lpositive .or. lnegative) then
                       intd = deblended * sqpixel
                       xxcorc = xcorc * xcorc * intd
                       yycorc = ycorc * ycorc * intd
                       xycorc = xcorc * ycorc * intd
                       intxxh = intxxh + xxcorc
                       intyyh = intyyh + yycorc
                       intxyh = intxyh + xycorc
                       int0h = int0h + intd
                     endif
                   endif
                 endif
               enddo
             enddo

 333         continue             

             if (nhalfmax .ge. 4) then
               hwhm = hwhm / dble ( nhalfmax )
!!               xcent = xcent / dble ( nhalfmax )
!!               ycent = ycent / dble ( nhalfmax )
               fwhm = 2.0d0 * hwhm
               afwhm(n) = fwhm
               bfwhm(n) = fwhm
               theta(n) = 0.0d0
             else
               theta(n) = 999.0d0
             endif
           endif
           
           int0m = int0 + almostzero
           momxx(n) = intxx / int0m
           momyy(n) = intyy / int0m
           momxy(n) = intxy / int0m
           int0hm = int0h + almostzero
           momxxh(n) = intxxh / int0hm
           momyyh(n) = intyyh / int0hm
           momxyh(n) = intxyh / int0hm
           equivrad(n) = sqrt ( dble ( numpix ) * sqpixel / pi )
         else
           momxx(n) = 0.0d0
           momyy(n) = 0.0d0
           momxy(n) = 0.0d0
           momxxh(n) = 0.0d0
           momyyh(n) = 0.0d0
           momxyh(n) = 0.0d0
         endif
       enddo

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine momentsizes
       
     &            ( nextr, pixel, lnotbad, lfootprint, momxx, momyy, momxy, afwhm, bfwhm, momxxh, momyyh, momxyh, asize  !!, beam
     &            , bsize, theta, sqreightlogtwo, convtodegrees, lsourcetodeblend, almostzero )  !!, lfinishing_mode
!__________________________________________________________________________________________________________________________________
!
! Compute FWHM sizes from intensity moments.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       
       logical       lfootprint(*), lsourcetodeblend(*), lnotbad(*)      !! lfinishing_mode, 
       
       integer       n, nextr
       
       real*8        pixel, sqreightlogtwo, momp2, momm2, sqrmo, asig, bsig, convtodegrees, elongation, almostzero, correction
     &             , elongationh, asigh, bsigh !!, beam, corrfactsize
       
       real*8        afwhm(*), bfwhm(*), momxx(*), momyy(*), momxy(*), momxxh(*), momyyh(*), momxyh(*), asize(*), bsize(*), theta(*)
!__________________________________________________________________________________________________________________________________
!
! Loop over all extracted sources.

       asig = 0.0d0
       bsig = 0.0d0
       asigh = 0.0d0
       bsigh = 0.0d0
       
       do n=1,nextr

         if (lnotbad(n) .and. lfootprint(n) .and. lsourcetodeblend(n)) then
           
! These sizes from moments over the entire footprint are valid for Gaussian sources only. 
! They may be heavily overestimated for the power-law sources.

           momp2 = (momxx(n) + momyy(n)) / 2.0d0
           momm2 = (momxx(n) - momyy(n)) / 2.0d0

           if (abs ( momxy(n) ) .lt. almostzero) then
             if (momxx(n) .gt. momyy(n)) then
               momp2 = momxx(n)
               momm2 = momyy(n)
               theta(n) = 90.0d0
             else
               momp2 = momyy(n)
               momm2 = momxx(n)
               theta(n) = 0.0d0
             endif
             sqrmo = sqrt ( momm2**2 + momxy(n)**2 )
             asig  = sqrt ( momp2 + sqrmo ) * pixel
             if (momp2 .gt. sqrmo) then 
               bsig = sqrt ( momp2 - sqrmo ) * pixel
             else
               asig = 0.0d0
               bsig = 0.0d0
             endif
           else       
             sqrmo = sqrt ( momm2**2 + momxy(n)**2 )
             if (momp2 + sqrmo .ge. 0.0d0) then 
               asig = sqrt ( momp2 + sqrmo ) * pixel
             endif
             if (momp2 - sqrmo .ge. 0.0d0) then 
               bsig = sqrt ( momp2 - sqrmo ) * pixel
             endif
             if (momp2 + sqrmo .lt. 0.0d0 .or. momp2 - sqrmo .lt. 0.0d0) then 
               asig = 0.0d0
               bsig = 0.0d0
             endif
             theta(n) = 180.0d0 - atan2 ( sqrmo + momm2, momxy(n) ) * convtodegrees
           endif                                        

           asize(n) = max ( sqreightlogtwo * asig, pixel )
           bsize(n) = max ( sqreightlogtwo * bsig, pixel )
           elongation = asize(n) / bsize(n)

! Next computes elongation from moments from a restricted range of intensities (above FPEAK/10).

           momp2 = (momxxh(n) + momyyh(n)) / 2.0d0
           momm2 = (momxxh(n) - momyyh(n)) / 2.0d0

           if (abs ( momxyh(n) ) .lt. almostzero) then
             if (momxxh(n) .gt. momyyh(n)) then
               momp2 = momxxh(n)
               momm2 = momyyh(n)
               theta(n) = 90.0d0
             else
               momp2 = momyyh(n)
               momm2 = momxxh(n)
             endif
             sqrmo = sqrt ( momm2**2 + momxyh(n)**2 )
             asigh = sqrt ( momp2 + sqrmo ) * pixel
             if (momp2 .gt. sqrmo) then 
               bsigh = sqrt ( momp2 - sqrmo ) * pixel
             else
               asigh = 0.0d0
               bsigh = 0.0d0
             endif
           else       
             sqrmo = sqrt ( momm2**2 + momxyh(n)**2 )
             if (momp2 + sqrmo .ge. 0.0d0) then 
               asigh = sqrt ( momp2 + sqrmo ) * pixel
             endif
             if (momp2 - sqrmo .ge. 0.0d0) then 
               bsigh = sqrt ( momp2 - sqrmo ) * pixel
             endif
             if (momp2 + sqrmo .lt. 0.0d0 .or. momp2 - sqrmo .lt. 0.0d0) then 
               asigh = 0.0d0
               bsigh = 0.0d0
             endif
           endif                                        

! A mean size (AFWHM=BFWHM) comes from the direct FWHM measurement performed in a call to INTENSMOMENTS.
! The FWHM sizes of an ellipse are reconstructed here using the elongation computed from the above intensity moments
! assuming that the mean HWHM equals the radius of a circle with the same area as the ellipse.

! The mean half-maximum radius is slightly larger than the equivalent-area radius for small ellipse elongations.
! For increasing elongations, the ratio MEANRADIUS / EQUIVAREARADIUS increases as EXP (( ELONGATION - 1 ) / 20), as I find
! empirically. Hence, I introduce the correction factor to make sure that the interpolated radius is equal to EQUIVAREARADIUS
! for all elongations.

           elongationh = max ( sqreightlogtwo * asigh, pixel ) / max ( sqreightlogtwo * bsigh, pixel )
           elongationh = min ( max ( elongationh, 1.0d0 ), 3.0d0 )
           correction = exp ( (elongationh - 1.0d0) / 20.0d0 )

           bfwhm(n) = (afwhm(n) / correction) / sqrt ( elongationh )
           afwhm(n) = bfwhm(n) * elongationh

!!! The following correction factor fixes the problem of overestimated source background that leaves a pedestal and leads to
!!! underestimated FWHM size, peak flux, and total flux. For unresolved sources, it makes FWHM smaller than the beam size, which
!!! clearly indicates the problem of oversubtracted background. The correction factors are found empirically by considering a
!!! Gaussian source and subtracting its faintest parts of varying thickness.
!!
!!           if (lfinishing_mode) then
!!             corrfactsize = beam / min ( sqrt ( afwhm(n) * bfwhm(n) ), beam )
!!             afwhm(n) = afwhm(n) * corrfactsize
!!             bfwhm(n) = bfwhm(n) * corrfactsize
!!           endif
         endif
       enddo
       
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine fitshapesfluxes
       
     &            ( nextr, lnotbad, lfootprint, nx, ny, dx, dy, momxco, momyco, nxco, nyco, afwhm, bfwhm, theta, foomajor, foominor
     &            , fluxpeakbest, fluxtotfit, convtoradians, lsourcetodeblend )
!__________________________________________________________________________________________________________________________________
!
! Compute total fluxes assuming an intensity distribution of an elliptical sources with known peak flux and size.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       
       logical       inellipse, lfootprint(*), lsourcetodeblend(*), lnotbad(*)
       
       integer       n, nextr, nx, ny, l, m, nr, l1, l2, m1, m2, nxco(*), nyco(*)
       
       real*8        dbll, dblm, angle, dxcon, dycon, aa, bb, cc, dx, dy, argexp, argexpm, gaussfun, convtoradians, dlog4h
     &             , sigmaa2, sigmab2, cosa2, sina2, sin2a, dxcon2, dycon2, semimajor
       
       real*8        fluxpeakbest(*), fluxtotfit(*), afwhm(*), bfwhm(*), theta(*), momxco(*), momyco(*), foomajor(*), foominor(*)

       parameter   ( dlog4h = log ( 4.0d0 ) / 2.0d0 )
       
       external      inellipse
!__________________________________________________________________________________________________________________________________
!                  
       do n=1,nextr
         
         if (lnotbad(n) .and. lfootprint(n) .and. lsourcetodeblend(n)) then
           
           if (afwhm(n) .gt. 0.0d0 .and. bfwhm(n) .gt. 0.0d0) then
             
             fluxtotfit(n) = 0.0d0
             sigmaa2 = (afwhm(n) / 2.0d0)**2
             sigmab2 = (bfwhm(n) / 2.0d0)**2
             angle = (90.0d0 - theta(n)) * convtoradians
             cosa2 = cos ( angle )**2       
             sina2 = sin ( angle )**2       
             sin2a = sin ( 2.0d0 * angle )  
             aa =  cosa2 / sigmaa2 + sina2 / sigmab2
             bb = -sin2a / sigmaa2 + sin2a / sigmab2
             cc =  sina2 / sigmaa2 + cosa2 / sigmab2
             semimajor = foomajor(n) / 2.0d0
             nr = nint ( semimajor / min ( dx, dy ) )   
             l1 = max ( nxco(n) - nr, 1 )     
             l2 = min ( nxco(n) + nr, nx )    
             m1 = max ( nyco(n) - nr, 1 )     
             m2 = min ( nyco(n) + nr, ny )    

             do m=m1,m2                     
               dblm = dble ( m )            
               dycon = (dblm - momyco(n)) * dy
               dycon2 = dycon**2            
               
               do l=l1,l2                   
                 dbll = dble ( l )          
                 dxcon = (dbll - momxco(n)) * dx
                 dxcon2 = dxcon**2          
                 
                 if (inellipse ( dbll, dblm, dx, dy, momxco(n), momyco(n), foomajor(n), foominor(n), theta(n) )) then
                   argexp = aa * dxcon2 + bb * dxcon * dycon + cc * dycon2
                   argexpm = min ( argexp * dlog4h, 100.0d0 )
                   gaussfun = fluxpeakbest(n) * exp ( - argexpm )
                   fluxtotfit(n) = fluxtotfit(n) + gaussfun
                 endif
               enddo
             enddo
           endif
         endif
       enddo

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine integratedfluxes

     &            ( iotty, nx, ny, dx, dy, nextr, lnotbad, fluxpeakbest, momxco, momyco, nxco, nyco, foominor, foomajor, theta
     &            , aax, bbx, ccx, maskimage, lfootprint, dxcon2, dycon2, convftot, convftotm, intensunits !!, afwhm, bfwhm, beam
     &            , capcorr, cowave, equivrad, apcorrect, almostzero, fluxtotbest, fluxtotfit, obsfsbsimage !!, lfinishing_mode
     &            , lsourcetodeblend, nkover, nkoverlap )
!__________________________________________________________________________________________________________________________________
!
! Compute total fluxes of sources.
!__________________________________________________________________________________________________________________________________
!
       implicit      none

       logical       inellipse, lfootprint(*), lsourcetodeblend(*), lnotbad(*)  !!, lfinishing_mode

       character*(*) capcorr, intensunits, cowave

       integer       lastc, iotty, nx, ny, k, n, nextr, i, iu, na070, na100, na160, na250, na350, na500, nr, l1, l2, m1, m2
     &             , m, l, j, icw, nxco(*), nyco(*), nkover(*), nkoverlap(nextr,*), nkacc, nkaccel(nextr)  !!, nn

       real*8        dx, dy, almostzero, funn, funksum, argexp, argexpm, semimajor, deblended, dblm, dbll, convftot, frac, dlog4h
     &             , deblendfactor, convftotm !!, sizeratio, corrfactpeak, corrfactflux  !!, beam !!, selectk

       real*8        fluxpeakbest(*), fluxtotbest(*), maskimage(nx,ny), momxco(*), momyco(*), foominor(*), foomajor(*), theta(*)
     &             , aax(*), bbx(*), ccx(*), equivrad(*), apcorrect(*), fluxtotfit(*), dxcon2(nextr,*), dycon2(nextr,*)
     &             , obsfsbsimage(nx,ny)  !!, afwhm(*), bfwhm(*) !!, funx(nx*ny)

! Tables for the aperture corrections at all Herschel wavelengths. The PACS tables were taken in the early days of March 2011 from 
! the Herschel Science Centre's web site. The SPIRE tables were computed in early March 2011 by Philippe Andre using the images of 
! measured PSFs for SPIRE. 

       parameter   ( na070 = 141, na100 = 141, na160 = 141, na250 = 700, na350 = 700, na500 = 700, dlog4h = log ( 4.0d0 ) / 2.0d0 )

       real*8        ar070(na070), af070(na070), ar100(na100), af100(na100), ar160(na160), af160(na160)
     &             , ar250(na250), af250(na250), ar350(na350), af350(na350), ar500(na500), af500(na500)

       data (ar070(i),i=1,na070)
     &    / 1.0417, 1.0853, 1.1306, 1.1779, 1.2271, 1.2784, 1.3318, 1.3875, 1.4455, 1.5059, 1.5689, 1.6344, 1.7027, 1.7739, 1.8480,
     &      1.9253, 2.0057, 2.0896, 2.1769, 2.2679, 2.3627, 2.4614, 2.5643, 2.6715, 2.7831, 2.8994, 3.0206, 3.1468, 3.2784, 3.4154,
     &      3.5581, 3.7068, 3.8618, 4.0231, 4.1913, 4.3665, 4.5490, 4.7391, 4.9371, 5.1435, 5.3584, 5.5824, 5.8157, 6.0587, 6.3120,
     &      6.5758, 6.8506, 7.1369, 7.4352, 7.7459, 8.0696, 8.4069, 8.7583, 9.1243, 9.5056, 9.9029, 10.316, 10.748, 11.197, 11.665,
     &      12.152, 12.660, 13.189, 13.740, 14.315, 14.913, 15.536, 16.186, 16.862, 17.567, 18.301, 19.066, 19.863, 20.693, 21.558,
     &      22.459, 23.397, 24.375, 25.394, 26.455, 27.561, 28.713, 29.913, 31.163, 32.465, 33.822, 35.236, 36.709, 38.243, 39.841,
     &      41.506, 43.241, 45.048, 46.931, 48.892, 50.936, 53.064, 55.282, 57.593, 60.000, 63.095, 67.608, 72.443, 77.624, 83.176,
     &      89.125, 95.499, 102.32, 109.64, 117.48, 125.89, 134.89, 144.54, 154.88, 165.95, 177.82, 190.54, 204.17, 218.77, 234.42,
     &      251.18, 269.15, 288.40, 309.02, 331.13, 354.81, 380.18, 407.38, 436.51, 467.73, 501.18, 537.03, 575.43, 616.59, 660.69,
     &      707.94, 758.57, 812.83, 870.96, 933.25, 1000.0 /
       data (af070(i),i=1,na070)
     &    / 0.0568, 0.0614, 0.0664, 0.0718, 0.0776, 0.0840, 0.0908, 0.0983, 0.1064, 0.1151, 0.1239, 0.1330, 0.1426, 0.1527, 0.1634,
     &      0.1748, 0.1869, 0.1998, 0.2136, 0.2283, 0.2441, 0.2611, 0.2785, 0.2951, 0.3119, 0.3290, 0.3466, 0.3650, 0.3842, 0.4044,
     &      0.4248, 0.4433, 0.4609, 0.4785, 0.4961, 0.5139, 0.5319, 0.5476, 0.5622, 0.5764, 0.5905, 0.6046, 0.6170, 0.6292, 0.6413,
     &      0.6536, 0.6653, 0.6771, 0.6891, 0.7009, 0.7124, 0.7239, 0.7348, 0.7449, 0.7547, 0.7633, 0.7711, 0.7783, 0.7845, 0.7903,
     &      0.7955, 0.8004, 0.8052, 0.8099, 0.8148, 0.8198, 0.8251, 0.8304, 0.8358, 0.8411, 0.8460, 0.8504, 0.8544, 0.8580, 0.8613,
     &      0.8644, 0.8674, 0.8704, 0.8734, 0.8764, 0.8793, 0.8821, 0.8848, 0.8875, 0.8901, 0.8927, 0.8953, 0.8978, 0.9004, 0.9030,
     &      0.9056, 0.9081, 0.9107, 0.9132, 0.9158, 0.9184, 0.9211, 0.9239, 0.9268, 0.9297, 0.9319, 0.9362, 0.9404, 0.9446, 0.9486,
     &      0.9522, 0.9556, 0.9589, 0.9624, 0.9660, 0.9694, 0.9726, 0.9754, 0.9779, 0.9802, 0.9823, 0.9842, 0.9860, 0.9875, 0.9889,
     &      0.9901, 0.9912, 0.9921, 0.9930, 0.9937, 0.9944, 0.9950, 0.9956, 0.9960, 0.9965, 0.9970, 0.9974, 0.9978, 0.9983, 0.9987,
     &      0.9990, 0.9993, 0.9995, 0.9997, 0.9998, 1.0000 /   
       data (ar100(i),i=1,na100)
     &    / 1.0417, 1.0853, 1.1306, 1.1779, 1.2271, 1.2784, 1.3318, 1.3875, 1.4455, 1.5059, 1.5689, 1.6344, 1.7027, 1.7739, 1.8480,
     &      1.9253, 2.0057, 2.0896, 2.1769, 2.2679, 2.3627, 2.4614, 2.5643, 2.6715, 2.7831, 2.8994, 3.0206, 3.1468, 3.2784, 3.4154,
     &      3.5581, 3.7068, 3.8618, 4.0231, 4.1913, 4.3665, 4.5490, 4.7391, 4.9371, 5.1435, 5.3584, 5.5824, 5.8157, 6.0587, 6.3120,
     &      6.5758, 6.8506, 7.1369, 7.4352, 7.7459, 8.0696, 8.4069, 8.7583, 9.1243, 9.5056, 9.9029, 10.316, 10.748, 11.197, 11.665,
     &      12.152, 12.660, 13.189, 13.740, 14.315, 14.913, 15.536, 16.186, 16.862, 17.567, 18.301, 19.066, 19.863, 20.693, 21.558,
     &      22.459, 23.397, 24.375, 25.394, 26.455, 27.561, 28.713, 29.913, 31.163, 32.465, 33.822, 35.236, 36.709, 38.243, 39.841,
     &      41.506, 43.241, 45.048, 46.931, 48.892, 50.936, 53.064, 55.282, 57.593, 60.000, 63.095, 67.608, 72.443, 77.624, 83.176,
     &      89.125, 95.499, 102.32, 109.64, 117.48, 125.89, 134.89, 144.54, 154.88, 165.95, 177.82, 190.54, 204.17, 218.77, 234.42,
     &      251.18, 269.15, 288.40, 309.02, 331.13, 354.81, 380.18, 407.38, 436.51, 467.73, 501.18, 537.03, 575.43, 616.59, 660.69,
     &      707.94, 758.57, 812.83, 870.96, 933.25, 1000.0 /
       data (af100(i),i=1,na100)
     &    / 0.0405, 0.0439, 0.0475, 0.0515, 0.0558, 0.0604, 0.0654, 0.0708, 0.0767, 0.0831, 0.0897, 0.0967, 0.1041, 0.1120, 0.1204,
     &      0.1295, 0.1391, 0.1495, 0.1607, 0.1727, 0.1856, 0.1995, 0.2140, 0.2285, 0.2435, 0.2592, 0.2756, 0.2929, 0.3113, 0.3308,
     &      0.3509, 0.3701, 0.3892, 0.4087, 0.4286, 0.4492, 0.4703, 0.4894, 0.5076, 0.5254, 0.5432, 0.5607, 0.5757, 0.5896, 0.6029,
     &      0.6158, 0.6268, 0.6370, 0.6469, 0.6563, 0.6653, 0.6744, 0.6838, 0.6935, 0.7037, 0.7144, 0.7254, 0.7368, 0.7479, 0.7589,
     &      0.7691, 0.7787, 0.7872, 0.7948, 0.8013, 0.8070, 0.8120, 0.8163, 0.8202, 0.8239, 0.8275, 0.8312, 0.8352, 0.8396, 0.8443,
     &      0.8492, 0.8539, 0.8584, 0.8627, 0.8665, 0.8700, 0.8732, 0.8761, 0.8788, 0.8814, 0.8840, 0.8865, 0.8891, 0.8917, 0.8943,
     &      0.8969, 0.8994, 0.9018, 0.9042, 0.9066, 0.9090, 0.9114, 0.9140, 0.9167, 0.9194, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000 /
       data (ar160(i),i=1,na160)
     &    / 1.0417, 1.0853, 1.1306, 1.1779, 1.2271, 1.2784, 1.3318, 1.3875, 1.4455, 1.5059, 1.5689, 1.6344, 1.7027, 1.7739, 1.8480,
     &      1.9253, 2.0057, 2.0896, 2.1769, 2.2679, 2.3627, 2.4614, 2.5643, 2.6715, 2.7831, 2.8994, 3.0206, 3.1468, 3.2784, 3.4154,
     &      3.5581, 3.7068, 3.8618, 4.0231, 4.1913, 4.3665, 4.5490, 4.7391, 4.9371, 5.1435, 5.3584, 5.5824, 5.8157, 6.0587, 6.3120,
     &      6.5758, 6.8506, 7.1369, 7.4352, 7.7459, 8.0696, 8.4069, 8.7583, 9.1243, 9.5056, 9.9029, 10.316, 10.748, 11.197, 11.665,
     &      12.152, 12.660, 13.189, 13.740, 14.315, 14.913, 15.536, 16.186, 16.862, 17.567, 18.301, 19.066, 19.863, 20.693, 21.558,
     &      22.459, 23.397, 24.375, 25.394, 26.455, 27.561, 28.713, 29.913, 31.163, 32.465, 33.822, 35.236, 36.709, 38.243, 39.841,
     &      41.506, 43.241, 45.048, 46.931, 48.892, 50.936, 53.064, 55.282, 57.593, 60.000, 63.095, 67.608, 72.443, 77.624, 83.176,
     &      89.125, 95.499, 102.32, 109.64, 117.48, 125.89, 134.89, 144.54, 154.88, 165.95, 177.82, 190.54, 204.17, 218.77, 234.42,
     &      251.18, 269.15, 288.40, 309.02, 331.13, 354.81, 380.18, 407.38, 436.51, 467.73, 501.18, 537.03, 575.43, 616.59, 660.69,
     &      707.94, 758.57, 812.83, 870.96, 933.25, 1000.0 /
       data (af160(i),i=1,na160)
     &    / 0.0163, 0.0177, 0.0192, 0.0208, 0.0226, 0.0245, 0.0266, 0.0288, 0.0313, 0.0339, 0.0367, 0.0398, 0.0430, 0.0465, 0.0503,
     &      0.0544, 0.0588, 0.0636, 0.0688, 0.0743, 0.0804, 0.0869, 0.0939, 0.1012, 0.1090, 0.1174, 0.1264, 0.1360, 0.1463, 0.1575,
     &      0.1693, 0.1816, 0.1945, 0.2081, 0.2226, 0.2381, 0.2544, 0.2710, 0.2882, 0.3061, 0.3249, 0.3446, 0.3640, 0.3837, 0.4041,
     &      0.4251, 0.4453, 0.4655, 0.4860, 0.5059, 0.5248, 0.5436, 0.5615, 0.5782, 0.5944, 0.6092, 0.6231, 0.6363, 0.6482, 0.6597,
     &      0.6703, 0.6807, 0.6907, 0.7008, 0.7108, 0.7210, 0.7314, 0.7419, 0.7524, 0.7628, 0.7728, 0.7824, 0.7913, 0.7995, 0.8069,
     &      0.8134, 0.8193, 0.8246, 0.8293, 0.8337, 0.8378, 0.8419, 0.8460, 0.8503, 0.8547, 0.8593, 0.8638, 0.8682, 0.8724, 0.8763,
     &      0.8801, 0.8836, 0.8870, 0.8903, 0.8936, 0.8968, 0.8998, 0.9027, 0.9054, 0.9081, 0.9109, 0.9151, 0.9191, 0.9231, 0.9270,
     &      0.9308, 0.9345, 0.9381, 0.9414, 0.9446, 0.9477, 0.9507, 0.9537, 0.9567, 0.9595, 0.9623, 0.9650, 0.9677, 0.9704, 0.9730,
     &      0.9754, 0.9778, 0.9801, 0.9823, 0.9844, 0.9864, 0.9883, 0.9900, 0.9916, 0.9930, 0.9945, 0.9958, 0.9970, 0.9980, 0.9989,
     &      0.9995, 0.9998, 1.0000, 0.9999, 0.9997, 0.9996 /
       data (ar250(i),i=1,na250)
     &    / 1.0000, 2.0000, 3.0000, 4.0000, 5.0000, 6.0000, 7.0000, 8.0000, 9.0000, 10.000, 11.000, 12.000, 13.000, 14.000, 15.000,
     &      16.000, 17.000, 18.000, 19.000, 20.000, 21.000, 22.000, 23.000, 24.000, 25.000, 26.000, 27.000, 28.000, 29.000, 30.000,
     &      31.000, 32.000, 33.000, 34.000, 35.000, 36.000, 37.000, 38.000, 39.000, 40.000, 41.000, 42.000, 43.000, 44.000, 45.000,
     &      46.000, 47.000, 48.000, 49.000, 50.000, 51.000, 52.000, 53.000, 54.000, 55.000, 56.000, 57.000, 58.000, 59.000, 60.000,
     &      61.000, 62.000, 63.000, 64.000, 65.000, 66.000, 67.000, 68.000, 69.000, 70.000, 71.000, 72.000, 73.000, 74.000, 75.000,
     &      76.000, 77.000, 78.000, 79.000, 80.000, 81.000, 82.000, 83.000, 84.000, 85.000, 86.000, 87.000, 88.000, 89.000, 90.000,
     &      91.000, 92.000, 93.000, 94.000, 95.000, 96.000, 97.000, 98.000, 99.000, 100.00, 101.00, 102.00, 103.00, 104.00, 105.00,
     &      106.00, 107.00, 108.00, 109.00, 110.00, 111.00, 112.00, 113.00, 114.00, 115.00, 116.00, 117.00, 118.00, 119.00, 120.00,
     &      121.00, 122.00, 123.00, 124.00, 125.00, 126.00, 127.00, 128.00, 129.00, 130.00, 131.00, 132.00, 133.00, 134.00, 135.00,
     &      136.00, 137.00, 138.00, 139.00, 140.00, 141.00, 142.00, 143.00, 144.00, 145.00, 146.00, 147.00, 148.00, 149.00, 150.00,
     &      151.00, 152.00, 153.00, 154.00, 155.00, 156.00, 157.00, 158.00, 159.00, 160.00, 161.00, 162.00, 163.00, 164.00, 165.00,
     &      166.00, 167.00, 168.00, 169.00, 170.00, 171.00, 172.00, 173.00, 174.00, 175.00, 176.00, 177.00, 178.00, 179.00, 180.00,
     &      181.00, 182.00, 183.00, 184.00, 185.00, 186.00, 187.00, 188.00, 189.00, 190.00, 191.00, 192.00, 193.00, 194.00, 195.00,
     &      196.00, 197.00, 198.00, 199.00, 200.00, 201.00, 202.00, 203.00, 204.00, 205.00, 206.00, 207.00, 208.00, 209.00, 210.00,
     &      211.00, 212.00, 213.00, 214.00, 215.00, 216.00, 217.00, 218.00, 219.00, 220.00, 221.00, 222.00, 223.00, 224.00, 225.00,
     &      226.00, 227.00, 228.00, 229.00, 230.00, 231.00, 232.00, 233.00, 234.00, 235.00, 236.00, 237.00, 238.00, 239.00, 240.00,
     &      241.00, 242.00, 243.00, 244.00, 245.00, 246.00, 247.00, 248.00, 249.00, 250.00, 251.00, 252.00, 253.00, 254.00, 255.00,
     &      256.00, 257.00, 258.00, 259.00, 260.00, 261.00, 262.00, 263.00, 264.00, 265.00, 266.00, 267.00, 268.00, 269.00, 270.00,
     &      271.00, 272.00, 273.00, 274.00, 275.00, 276.00, 277.00, 278.00, 279.00, 280.00, 281.00, 282.00, 283.00, 284.00, 285.00,
     &      286.00, 287.00, 288.00, 289.00, 290.00, 291.00, 292.00, 293.00, 294.00, 295.00, 296.00, 297.00, 298.00, 299.00, 300.00,
     &      301.00, 302.00, 303.00, 304.00, 305.00, 306.00, 307.00, 308.00, 309.00, 310.00, 311.00, 312.00, 313.00, 314.00, 315.00,
     &      316.00, 317.00, 318.00, 319.00, 320.00, 321.00, 322.00, 323.00, 324.00, 325.00, 326.00, 327.00, 328.00, 329.00, 330.00,
     &      331.00, 332.00, 333.00, 334.00, 335.00, 336.00, 337.00, 338.00, 339.00, 340.00, 341.00, 342.00, 343.00, 344.00, 345.00,
     &      346.00, 347.00, 348.00, 349.00, 350.00, 351.00, 352.00, 353.00, 354.00, 355.00, 356.00, 357.00, 358.00, 359.00, 360.00,
     &      361.00, 362.00, 363.00, 364.00, 365.00, 366.00, 367.00, 368.00, 369.00, 370.00, 371.00, 372.00, 373.00, 374.00, 375.00,
     &      376.00, 377.00, 378.00, 379.00, 380.00, 381.00, 382.00, 383.00, 384.00, 385.00, 386.00, 387.00, 388.00, 389.00, 390.00,
     &      391.00, 392.00, 393.00, 394.00, 395.00, 396.00, 397.00, 398.00, 399.00, 400.00, 401.00, 402.00, 403.00, 404.00, 405.00,
     &      406.00, 407.00, 408.00, 409.00, 410.00, 411.00, 412.00, 413.00, 414.00, 415.00, 416.00, 417.00, 418.00, 419.00, 420.00,
     &      421.00, 422.00, 423.00, 424.00, 425.00, 426.00, 427.00, 428.00, 429.00, 430.00, 431.00, 432.00, 433.00, 434.00, 435.00,
     &      436.00, 437.00, 438.00, 439.00, 440.00, 441.00, 442.00, 443.00, 444.00, 445.00, 446.00, 447.00, 448.00, 449.00, 450.00,
     &      451.00, 452.00, 453.00, 454.00, 455.00, 456.00, 457.00, 458.00, 459.00, 460.00, 461.00, 462.00, 463.00, 464.00, 465.00,
     &      466.00, 467.00, 468.00, 469.00, 470.00, 471.00, 472.00, 473.00, 474.00, 475.00, 476.00, 477.00, 478.00, 479.00, 480.00,
     &      481.00, 482.00, 483.00, 484.00, 485.00, 486.00, 487.00, 488.00, 489.00, 490.00, 491.00, 492.00, 493.00, 494.00, 495.00,
     &      496.00, 497.00, 498.00, 499.00, 500.00, 501.00, 502.00, 503.00, 504.00, 505.00, 506.00, 507.00, 508.00, 509.00, 510.00,
     &      511.00, 512.00, 513.00, 514.00, 515.00, 516.00, 517.00, 518.00, 519.00, 520.00, 521.00, 522.00, 523.00, 524.00, 525.00,
     &      526.00, 527.00, 528.00, 529.00, 530.00, 531.00, 532.00, 533.00, 534.00, 535.00, 536.00, 537.00, 538.00, 539.00, 540.00,
     &      541.00, 542.00, 543.00, 544.00, 545.00, 546.00, 547.00, 548.00, 549.00, 550.00, 551.00, 552.00, 553.00, 554.00, 555.00,
     &      556.00, 557.00, 558.00, 559.00, 560.00, 561.00, 562.00, 563.00, 564.00, 565.00, 566.00, 567.00, 568.00, 569.00, 570.00,
     &      571.00, 572.00, 573.00, 574.00, 575.00, 576.00, 577.00, 578.00, 579.00, 580.00, 581.00, 582.00, 583.00, 584.00, 585.00,
     &      586.00, 587.00, 588.00, 589.00, 590.00, 591.00, 592.00, 593.00, 594.00, 595.00, 596.00, 597.00, 598.00, 599.00, 600.00,
     &      601.00, 602.00, 603.00, 604.00, 605.00, 606.00, 607.00, 608.00, 609.00, 610.00, 611.00, 612.00, 613.00, 614.00, 615.00,
     &      616.00, 617.00, 618.00, 619.00, 620.00, 621.00, 622.00, 623.00, 624.00, 625.00, 626.00, 627.00, 628.00, 629.00, 630.00,
     &      631.00, 632.00, 633.00, 634.00, 635.00, 636.00, 637.00, 638.00, 639.00, 640.00, 641.00, 642.00, 643.00, 644.00, 645.00,
     &      646.00, 647.00, 648.00, 649.00, 650.00, 651.00, 652.00, 653.00, 654.00, 655.00, 656.00, 657.00, 658.00, 659.00, 660.00,
     &      661.00, 662.00, 663.00, 664.00, 665.00, 666.00, 667.00, 668.00, 669.00, 670.00, 671.00, 672.00, 673.00, 674.00, 675.00,
     &      676.00, 677.00, 678.00, 679.00, 680.00, 681.00, 682.00, 683.00, 684.00, 685.00, 686.00, 687.00, 688.00, 689.00, 690.00,
     &      691.00, 692.00, 693.00, 694.00, 695.00, 696.00, 697.00, 698.00, 699.00, 700.00 /
       data (af250(i),i=1,na250)
     &    / 0.0205, 0.0493, 0.0959, 0.1476, 0.2065, 0.2717, 0.3397, 0.4065, 0.4731, 0.5331, 0.5894, 0.6371, 0.6780, 0.7115, 0.7386,
     &      0.7598, 0.7757, 0.7869, 0.7954, 0.8022, 0.8076, 0.8126, 0.8176, 0.8232, 0.8295, 0.8362, 0.8437, 0.8517, 0.8595, 0.8674,
     &      0.8745, 0.8818, 0.8881, 0.8938, 0.8989, 0.9032, 0.9070, 0.9100, 0.9127, 0.9149, 0.9168, 0.9185, 0.9200, 0.9214, 0.9226,
     &      0.9238, 0.9250, 0.9262, 0.9275, 0.9287, 0.9301, 0.9315, 0.9330, 0.9346, 0.9362, 0.9378, 0.9396, 0.9413, 0.9430, 0.9448,
     &      0.9465, 0.9481, 0.9497, 0.9512, 0.9527, 0.9540, 0.9552, 0.9564, 0.9574, 0.9583, 0.9592, 0.9600, 0.9607, 0.9613, 0.9619,
     &      0.9625, 0.9630, 0.9635, 0.9640, 0.9645, 0.9649, 0.9654, 0.9658, 0.9663, 0.9667, 0.9671, 0.9676, 0.9680, 0.9684, 0.9688,
     &      0.9693, 0.9697, 0.9700, 0.9704, 0.9708, 0.9712, 0.9715, 0.9719, 0.9722, 0.9726, 0.9729, 0.9733, 0.9736, 0.9739, 0.9743,
     &      0.9746, 0.9749, 0.9752, 0.9755, 0.9759, 0.9762, 0.9765, 0.9768, 0.9771, 0.9774, 0.9777, 0.9780, 0.9783, 0.9786, 0.9789,
     &      0.9792, 0.9794, 0.9797, 0.9800, 0.9802, 0.9805, 0.9807, 0.9810, 0.9812, 0.9815, 0.9817, 0.9819, 0.9822, 0.9824, 0.9827,
     &      0.9829, 0.9831, 0.9834, 0.9836, 0.9838, 0.9841, 0.9843, 0.9845, 0.9848, 0.9850, 0.9852, 0.9855, 0.9857, 0.9859, 0.9861,
     &      0.9863, 0.9865, 0.9867, 0.9869, 0.9871, 0.9873, 0.9875, 0.9877, 0.9879, 0.9881, 0.9882, 0.9884, 0.9886, 0.9887, 0.9889,
     &      0.9891, 0.9892, 0.9894, 0.9895, 0.9897, 0.9899, 0.9900, 0.9901, 0.9903, 0.9904, 0.9906, 0.9907, 0.9908, 0.9909, 0.9911,
     &      0.9912, 0.9913, 0.9914, 0.9915, 0.9916, 0.9917, 0.9918, 0.9919, 0.9920, 0.9921, 0.9922, 0.9923, 0.9924, 0.9925, 0.9925,
     &      0.9926, 0.9927, 0.9928, 0.9929, 0.9930, 0.9931, 0.9931, 0.9932, 0.9933, 0.9934, 0.9935, 0.9936, 0.9937, 0.9938, 0.9939,
     &      0.9940, 0.9940, 0.9941, 0.9942, 0.9943, 0.9944, 0.9945, 0.9946, 0.9946, 0.9947, 0.9948, 0.9948, 0.9949, 0.9950, 0.9950,
     &      0.9951, 0.9951, 0.9952, 0.9952, 0.9953, 0.9953, 0.9954, 0.9954, 0.9955, 0.9955, 0.9955, 0.9956, 0.9956, 0.9957, 0.9957,
     &      0.9958, 0.9958, 0.9959, 0.9959, 0.9960, 0.9961, 0.9961, 0.9962, 0.9962, 0.9963, 0.9963, 0.9964, 0.9965, 0.9965, 0.9966,
     &      0.9966, 0.9966, 0.9967, 0.9967, 0.9968, 0.9968, 0.9968, 0.9969, 0.9969, 0.9969, 0.9969, 0.9969, 0.9970, 0.9970, 0.9970,
     &      0.9970, 0.9970, 0.9970, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9972,
     &      0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9973, 0.9973,
     &      0.9973, 0.9973, 0.9973, 0.9973, 0.9974, 0.9974, 0.9974, 0.9974, 0.9974, 0.9975, 0.9975, 0.9975, 0.9975, 0.9976, 0.9976,
     &      0.9976, 0.9976, 0.9976, 0.9977, 0.9977, 0.9977, 0.9977, 0.9978, 0.9978, 0.9978, 0.9978, 0.9979, 0.9979, 0.9979, 0.9980,
     &      0.9980, 0.9980, 0.9981, 0.9981, 0.9981, 0.9981, 0.9981, 0.9982, 0.9982, 0.9982, 0.9982, 0.9982, 0.9982, 0.9982, 0.9982,
     &      0.9982, 0.9982, 0.9982, 0.9982, 0.9982, 0.9982, 0.9982, 0.9982, 0.9982, 0.9982, 0.9982, 0.9982, 0.9981, 0.9981, 0.9981,
     &      0.9981, 0.9981, 0.9981, 0.9981, 0.9981, 0.9981, 0.9981, 0.9981, 0.9981, 0.9981, 0.9981, 0.9981, 0.9981, 0.9981, 0.9981,
     &      0.9982, 0.9982, 0.9982, 0.9982, 0.9982, 0.9983, 0.9983, 0.9983, 0.9984, 0.9984, 0.9984, 0.9985, 0.9985, 0.9986, 0.9986,
     &      0.9987, 0.9987, 0.9988, 0.9989, 0.9989, 0.9990, 0.9991, 0.9992, 0.9992, 0.9993, 0.9994, 0.9995, 0.9996, 0.9997, 0.9997,
     &      0.9998, 0.9999, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000 /
       data (ar350(i),i=1,na350)
     &    / 1.0000, 2.0000, 3.0000, 4.0000, 5.0000, 6.0000, 7.0000, 8.0000, 9.0000, 10.000, 11.000, 12.000, 13.000, 14.000, 15.000,
     &      16.000, 17.000, 18.000, 19.000, 20.000, 21.000, 22.000, 23.000, 24.000, 25.000, 26.000, 27.000, 28.000, 29.000, 30.000,
     &      31.000, 32.000, 33.000, 34.000, 35.000, 36.000, 37.000, 38.000, 39.000, 40.000, 41.000, 42.000, 43.000, 44.000, 45.000,
     &      46.000, 47.000, 48.000, 49.000, 50.000, 51.000, 52.000, 53.000, 54.000, 55.000, 56.000, 57.000, 58.000, 59.000, 60.000,
     &      61.000, 62.000, 63.000, 64.000, 65.000, 66.000, 67.000, 68.000, 69.000, 70.000, 71.000, 72.000, 73.000, 74.000, 75.000,
     &      76.000, 77.000, 78.000, 79.000, 80.000, 81.000, 82.000, 83.000, 84.000, 85.000, 86.000, 87.000, 88.000, 89.000, 90.000,
     &      91.000, 92.000, 93.000, 94.000, 95.000, 96.000, 97.000, 98.000, 99.000, 100.00, 101.00, 102.00, 103.00, 104.00, 105.00,
     &      106.00, 107.00, 108.00, 109.00, 110.00, 111.00, 112.00, 113.00, 114.00, 115.00, 116.00, 117.00, 118.00, 119.00, 120.00,
     &      121.00, 122.00, 123.00, 124.00, 125.00, 126.00, 127.00, 128.00, 129.00, 130.00, 131.00, 132.00, 133.00, 134.00, 135.00,
     &      136.00, 137.00, 138.00, 139.00, 140.00, 141.00, 142.00, 143.00, 144.00, 145.00, 146.00, 147.00, 148.00, 149.00, 150.00,
     &      151.00, 152.00, 153.00, 154.00, 155.00, 156.00, 157.00, 158.00, 159.00, 160.00, 161.00, 162.00, 163.00, 164.00, 165.00,
     &      166.00, 167.00, 168.00, 169.00, 170.00, 171.00, 172.00, 173.00, 174.00, 175.00, 176.00, 177.00, 178.00, 179.00, 180.00,
     &      181.00, 182.00, 183.00, 184.00, 185.00, 186.00, 187.00, 188.00, 189.00, 190.00, 191.00, 192.00, 193.00, 194.00, 195.00,
     &      196.00, 197.00, 198.00, 199.00, 200.00, 201.00, 202.00, 203.00, 204.00, 205.00, 206.00, 207.00, 208.00, 209.00, 210.00,
     &      211.00, 212.00, 213.00, 214.00, 215.00, 216.00, 217.00, 218.00, 219.00, 220.00, 221.00, 222.00, 223.00, 224.00, 225.00,
     &      226.00, 227.00, 228.00, 229.00, 230.00, 231.00, 232.00, 233.00, 234.00, 235.00, 236.00, 237.00, 238.00, 239.00, 240.00,
     &      241.00, 242.00, 243.00, 244.00, 245.00, 246.00, 247.00, 248.00, 249.00, 250.00, 251.00, 252.00, 253.00, 254.00, 255.00,
     &      256.00, 257.00, 258.00, 259.00, 260.00, 261.00, 262.00, 263.00, 264.00, 265.00, 266.00, 267.00, 268.00, 269.00, 270.00,
     &      271.00, 272.00, 273.00, 274.00, 275.00, 276.00, 277.00, 278.00, 279.00, 280.00, 281.00, 282.00, 283.00, 284.00, 285.00,
     &      286.00, 287.00, 288.00, 289.00, 290.00, 291.00, 292.00, 293.00, 294.00, 295.00, 296.00, 297.00, 298.00, 299.00, 300.00,
     &      301.00, 302.00, 303.00, 304.00, 305.00, 306.00, 307.00, 308.00, 309.00, 310.00, 311.00, 312.00, 313.00, 314.00, 315.00,
     &      316.00, 317.00, 318.00, 319.00, 320.00, 321.00, 322.00, 323.00, 324.00, 325.00, 326.00, 327.00, 328.00, 329.00, 330.00,
     &      331.00, 332.00, 333.00, 334.00, 335.00, 336.00, 337.00, 338.00, 339.00, 340.00, 341.00, 342.00, 343.00, 344.00, 345.00,
     &      346.00, 347.00, 348.00, 349.00, 350.00, 351.00, 352.00, 353.00, 354.00, 355.00, 356.00, 357.00, 358.00, 359.00, 360.00,
     &      361.00, 362.00, 363.00, 364.00, 365.00, 366.00, 367.00, 368.00, 369.00, 370.00, 371.00, 372.00, 373.00, 374.00, 375.00,
     &      376.00, 377.00, 378.00, 379.00, 380.00, 381.00, 382.00, 383.00, 384.00, 385.00, 386.00, 387.00, 388.00, 389.00, 390.00,
     &      391.00, 392.00, 393.00, 394.00, 395.00, 396.00, 397.00, 398.00, 399.00, 400.00, 401.00, 402.00, 403.00, 404.00, 405.00,
     &      406.00, 407.00, 408.00, 409.00, 410.00, 411.00, 412.00, 413.00, 414.00, 415.00, 416.00, 417.00, 418.00, 419.00, 420.00,
     &      421.00, 422.00, 423.00, 424.00, 425.00, 426.00, 427.00, 428.00, 429.00, 430.00, 431.00, 432.00, 433.00, 434.00, 435.00,
     &      436.00, 437.00, 438.00, 439.00, 440.00, 441.00, 442.00, 443.00, 444.00, 445.00, 446.00, 447.00, 448.00, 449.00, 450.00,
     &      451.00, 452.00, 453.00, 454.00, 455.00, 456.00, 457.00, 458.00, 459.00, 460.00, 461.00, 462.00, 463.00, 464.00, 465.00,
     &      466.00, 467.00, 468.00, 469.00, 470.00, 471.00, 472.00, 473.00, 474.00, 475.00, 476.00, 477.00, 478.00, 479.00, 480.00,
     &      481.00, 482.00, 483.00, 484.00, 485.00, 486.00, 487.00, 488.00, 489.00, 490.00, 491.00, 492.00, 493.00, 494.00, 495.00,
     &      496.00, 497.00, 498.00, 499.00, 500.00, 501.00, 502.00, 503.00, 504.00, 505.00, 506.00, 507.00, 508.00, 509.00, 510.00,
     &      511.00, 512.00, 513.00, 514.00, 515.00, 516.00, 517.00, 518.00, 519.00, 520.00, 521.00, 522.00, 523.00, 524.00, 525.00,
     &      526.00, 527.00, 528.00, 529.00, 530.00, 531.00, 532.00, 533.00, 534.00, 535.00, 536.00, 537.00, 538.00, 539.00, 540.00,
     &      541.00, 542.00, 543.00, 544.00, 545.00, 546.00, 547.00, 548.00, 549.00, 550.00, 551.00, 552.00, 553.00, 554.00, 555.00,
     &      556.00, 557.00, 558.00, 559.00, 560.00, 561.00, 562.00, 563.00, 564.00, 565.00, 566.00, 567.00, 568.00, 569.00, 570.00,
     &      571.00, 572.00, 573.00, 574.00, 575.00, 576.00, 577.00, 578.00, 579.00, 580.00, 581.00, 582.00, 583.00, 584.00, 585.00,
     &      586.00, 587.00, 588.00, 589.00, 590.00, 591.00, 592.00, 593.00, 594.00, 595.00, 596.00, 597.00, 598.00, 599.00, 600.00,
     &      601.00, 602.00, 603.00, 604.00, 605.00, 606.00, 607.00, 608.00, 609.00, 610.00, 611.00, 612.00, 613.00, 614.00, 615.00,
     &      616.00, 617.00, 618.00, 619.00, 620.00, 621.00, 622.00, 623.00, 624.00, 625.00, 626.00, 627.00, 628.00, 629.00, 630.00,
     &      631.00, 632.00, 633.00, 634.00, 635.00, 636.00, 637.00, 638.00, 639.00, 640.00, 641.00, 642.00, 643.00, 644.00, 645.00,
     &      646.00, 647.00, 648.00, 649.00, 650.00, 651.00, 652.00, 653.00, 654.00, 655.00, 656.00, 657.00, 658.00, 659.00, 660.00,
     &      661.00, 662.00, 663.00, 664.00, 665.00, 666.00, 667.00, 668.00, 669.00, 670.00, 671.00, 672.00, 673.00, 674.00, 675.00,
     &      676.00, 677.00, 678.00, 679.00, 680.00, 681.00, 682.00, 683.00, 684.00, 685.00, 686.00, 687.00, 688.00, 689.00, 690.00,
     &      691.00, 692.00, 693.00, 694.00, 695.00, 696.00, 697.00, 698.00, 699.00, 700.00 /
       data (af350(i),i=1,na350)
     &    / 0.0115, 0.0292, 0.0538, 0.0862, 0.1238, 0.1641, 0.2145, 0.2634, 0.3139, 0.3636, 0.4168, 0.4697, 0.5168, 0.5621, 0.6050,
     &      0.6443, 0.6784, 0.7096, 0.7364, 0.7586, 0.7771, 0.7922, 0.8043, 0.8143, 0.8216, 0.8274, 0.8318, 0.8355, 0.8387, 0.8416,
     &      0.8447, 0.8479, 0.8516, 0.8554, 0.8598, 0.8649, 0.8700, 0.8756, 0.8810, 0.8868, 0.8926, 0.8982, 0.9036, 0.9086, 0.9134,
     &      0.9177, 0.9216, 0.9253, 0.9284, 0.9312, 0.9336, 0.9357, 0.9375, 0.9390, 0.9404, 0.9414, 0.9424, 0.9433, 0.9440, 0.9447,
     &      0.9453, 0.9459, 0.9464, 0.9470, 0.9476, 0.9481, 0.9488, 0.9494, 0.9501, 0.9509, 0.9516, 0.9525, 0.9534, 0.9543, 0.9553,
     &      0.9563, 0.9574, 0.9585, 0.9596, 0.9608, 0.9620, 0.9631, 0.9643, 0.9654, 0.9665, 0.9677, 0.9687, 0.9698, 0.9707, 0.9717,
     &      0.9726, 0.9734, 0.9742, 0.9750, 0.9756, 0.9763, 0.9769, 0.9774, 0.9780, 0.9784, 0.9789, 0.9793, 0.9797, 0.9801, 0.9804,
     &      0.9807, 0.9810, 0.9813, 0.9816, 0.9819, 0.9822, 0.9824, 0.9827, 0.9830, 0.9832, 0.9835, 0.9837, 0.9840, 0.9842, 0.9844,
     &      0.9847, 0.9849, 0.9851, 0.9854, 0.9856, 0.9858, 0.9860, 0.9862, 0.9864, 0.9866, 0.9868, 0.9870, 0.9872, 0.9874, 0.9876,
     &      0.9878, 0.9879, 0.9881, 0.9883, 0.9885, 0.9886, 0.9888, 0.9890, 0.9891, 0.9893, 0.9894, 0.9896, 0.9897, 0.9899, 0.9900,
     &      0.9902, 0.9903, 0.9905, 0.9906, 0.9907, 0.9908, 0.9910, 0.9911, 0.9912, 0.9913, 0.9915, 0.9916, 0.9917, 0.9918, 0.9919,
     &      0.9921, 0.9922, 0.9923, 0.9924, 0.9925, 0.9926, 0.9927, 0.9929, 0.9930, 0.9931, 0.9932, 0.9933, 0.9934, 0.9935, 0.9936,
     &      0.9937, 0.9938, 0.9939, 0.9940, 0.9941, 0.9942, 0.9944, 0.9945, 0.9946, 0.9947, 0.9948, 0.9949, 0.9950, 0.9951, 0.9952,
     &      0.9953, 0.9954, 0.9955, 0.9956, 0.9957, 0.9958, 0.9959, 0.9960, 0.9961, 0.9963, 0.9964, 0.9965, 0.9966, 0.9967, 0.9968,
     &      0.9969, 0.9970, 0.9971, 0.9972, 0.9973, 0.9974, 0.9975, 0.9976, 0.9977, 0.9977, 0.9978, 0.9979, 0.9979, 0.9980, 0.9980,
     &      0.9981, 0.9981, 0.9981, 0.9982, 0.9982, 0.9982, 0.9983, 0.9983, 0.9983, 0.9983, 0.9983, 0.9983, 0.9983, 0.9984, 0.9984,
     &      0.9984, 0.9984, 0.9984, 0.9984, 0.9984, 0.9985, 0.9985, 0.9985, 0.9985, 0.9985, 0.9985, 0.9985, 0.9985, 0.9985, 0.9986,
     &      0.9986, 0.9986, 0.9986, 0.9986, 0.9986, 0.9985, 0.9985, 0.9985, 0.9985, 0.9985, 0.9984, 0.9984, 0.9984, 0.9983, 0.9983,
     &      0.9983, 0.9982, 0.9982, 0.9981, 0.9981, 0.9980, 0.9980, 0.9980, 0.9979, 0.9979, 0.9978, 0.9978, 0.9977, 0.9977, 0.9976,
     &      0.9976, 0.9976, 0.9975, 0.9975, 0.9974, 0.9974, 0.9974, 0.9973, 0.9973, 0.9973, 0.9972, 0.9972, 0.9972, 0.9972, 0.9971,
     &      0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9970, 0.9970, 0.9970, 0.9970, 0.9970, 0.9970, 0.9970,
     &      0.9970, 0.9970, 0.9970, 0.9970, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971,
     &      0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9972,
     &      0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9972, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971, 0.9971,
     &      0.9971, 0.9970, 0.9970, 0.9970, 0.9970, 0.9970, 0.9970, 0.9970, 0.9970, 0.9970, 0.9970, 0.9970, 0.9970, 0.9970, 0.9970,
     &      0.9970, 0.9971, 0.9971, 0.9971, 0.9971, 0.9972, 0.9972, 0.9972, 0.9973, 0.9973, 0.9974, 0.9974, 0.9975, 0.9975, 0.9976,
     &      0.9977, 0.9977, 0.9978, 0.9979, 0.9980, 0.9981, 0.9981, 0.9982, 0.9983, 0.9984, 0.9985, 0.9986, 0.9987, 0.9988, 0.9989,
     &      0.9990, 0.9990, 0.9991, 0.9992, 0.9993, 0.9994, 0.9994, 0.9995, 0.9995, 0.9996, 0.9997, 0.9997, 0.9998, 0.9998, 0.9999,
     &      0.9999, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000 /
       data (ar500(i),i=1,na500)
     &    / 1.0000, 2.0000, 3.0000, 4.0000, 5.0000, 6.0000, 7.0000, 8.0000, 9.0000, 10.000, 11.000, 12.000, 13.000, 14.000, 15.000,
     &      16.000, 17.000, 18.000, 19.000, 20.000, 21.000, 22.000, 23.000, 24.000, 25.000, 26.000, 27.000, 28.000, 29.000, 30.000,
     &      31.000, 32.000, 33.000, 34.000, 35.000, 36.000, 37.000, 38.000, 39.000, 40.000, 41.000, 42.000, 43.000, 44.000, 45.000,
     &      46.000, 47.000, 48.000, 49.000, 50.000, 51.000, 52.000, 53.000, 54.000, 55.000, 56.000, 57.000, 58.000, 59.000, 60.000,
     &      61.000, 62.000, 63.000, 64.000, 65.000, 66.000, 67.000, 68.000, 69.000, 70.000, 71.000, 72.000, 73.000, 74.000, 75.000,
     &      76.000, 77.000, 78.000, 79.000, 80.000, 81.000, 82.000, 83.000, 84.000, 85.000, 86.000, 87.000, 88.000, 89.000, 90.000,
     &      91.000, 92.000, 93.000, 94.000, 95.000, 96.000, 97.000, 98.000, 99.000, 100.00, 101.00, 102.00, 103.00, 104.00, 105.00,
     &      106.00, 107.00, 108.00, 109.00, 110.00, 111.00, 112.00, 113.00, 114.00, 115.00, 116.00, 117.00, 118.00, 119.00, 120.00,
     &      121.00, 122.00, 123.00, 124.00, 125.00, 126.00, 127.00, 128.00, 129.00, 130.00, 131.00, 132.00, 133.00, 134.00, 135.00,
     &      136.00, 137.00, 138.00, 139.00, 140.00, 141.00, 142.00, 143.00, 144.00, 145.00, 146.00, 147.00, 148.00, 149.00, 150.00,
     &      151.00, 152.00, 153.00, 154.00, 155.00, 156.00, 157.00, 158.00, 159.00, 160.00, 161.00, 162.00, 163.00, 164.00, 165.00,
     &      166.00, 167.00, 168.00, 169.00, 170.00, 171.00, 172.00, 173.00, 174.00, 175.00, 176.00, 177.00, 178.00, 179.00, 180.00,
     &      181.00, 182.00, 183.00, 184.00, 185.00, 186.00, 187.00, 188.00, 189.00, 190.00, 191.00, 192.00, 193.00, 194.00, 195.00,
     &      196.00, 197.00, 198.00, 199.00, 200.00, 201.00, 202.00, 203.00, 204.00, 205.00, 206.00, 207.00, 208.00, 209.00, 210.00,
     &      211.00, 212.00, 213.00, 214.00, 215.00, 216.00, 217.00, 218.00, 219.00, 220.00, 221.00, 222.00, 223.00, 224.00, 225.00,
     &      226.00, 227.00, 228.00, 229.00, 230.00, 231.00, 232.00, 233.00, 234.00, 235.00, 236.00, 237.00, 238.00, 239.00, 240.00,
     &      241.00, 242.00, 243.00, 244.00, 245.00, 246.00, 247.00, 248.00, 249.00, 250.00, 251.00, 252.00, 253.00, 254.00, 255.00,
     &      256.00, 257.00, 258.00, 259.00, 260.00, 261.00, 262.00, 263.00, 264.00, 265.00, 266.00, 267.00, 268.00, 269.00, 270.00,
     &      271.00, 272.00, 273.00, 274.00, 275.00, 276.00, 277.00, 278.00, 279.00, 280.00, 281.00, 282.00, 283.00, 284.00, 285.00,
     &      286.00, 287.00, 288.00, 289.00, 290.00, 291.00, 292.00, 293.00, 294.00, 295.00, 296.00, 297.00, 298.00, 299.00, 300.00,
     &      301.00, 302.00, 303.00, 304.00, 305.00, 306.00, 307.00, 308.00, 309.00, 310.00, 311.00, 312.00, 313.00, 314.00, 315.00,
     &      316.00, 317.00, 318.00, 319.00, 320.00, 321.00, 322.00, 323.00, 324.00, 325.00, 326.00, 327.00, 328.00, 329.00, 330.00,
     &      331.00, 332.00, 333.00, 334.00, 335.00, 336.00, 337.00, 338.00, 339.00, 340.00, 341.00, 342.00, 343.00, 344.00, 345.00,
     &      346.00, 347.00, 348.00, 349.00, 350.00, 351.00, 352.00, 353.00, 354.00, 355.00, 356.00, 357.00, 358.00, 359.00, 360.00,
     &      361.00, 362.00, 363.00, 364.00, 365.00, 366.00, 367.00, 368.00, 369.00, 370.00, 371.00, 372.00, 373.00, 374.00, 375.00,
     &      376.00, 377.00, 378.00, 379.00, 380.00, 381.00, 382.00, 383.00, 384.00, 385.00, 386.00, 387.00, 388.00, 389.00, 390.00,
     &      391.00, 392.00, 393.00, 394.00, 395.00, 396.00, 397.00, 398.00, 399.00, 400.00, 401.00, 402.00, 403.00, 404.00, 405.00,
     &      406.00, 407.00, 408.00, 409.00, 410.00, 411.00, 412.00, 413.00, 414.00, 415.00, 416.00, 417.00, 418.00, 419.00, 420.00,
     &      421.00, 422.00, 423.00, 424.00, 425.00, 426.00, 427.00, 428.00, 429.00, 430.00, 431.00, 432.00, 433.00, 434.00, 435.00,
     &      436.00, 437.00, 438.00, 439.00, 440.00, 441.00, 442.00, 443.00, 444.00, 445.00, 446.00, 447.00, 448.00, 449.00, 450.00,
     &      451.00, 452.00, 453.00, 454.00, 455.00, 456.00, 457.00, 458.00, 459.00, 460.00, 461.00, 462.00, 463.00, 464.00, 465.00,
     &      466.00, 467.00, 468.00, 469.00, 470.00, 471.00, 472.00, 473.00, 474.00, 475.00, 476.00, 477.00, 478.00, 479.00, 480.00,
     &      481.00, 482.00, 483.00, 484.00, 485.00, 486.00, 487.00, 488.00, 489.00, 490.00, 491.00, 492.00, 493.00, 494.00, 495.00,
     &      496.00, 497.00, 498.00, 499.00, 500.00, 501.00, 502.00, 503.00, 504.00, 505.00, 506.00, 507.00, 508.00, 509.00, 510.00,
     &      511.00, 512.00, 513.00, 514.00, 515.00, 516.00, 517.00, 518.00, 519.00, 520.00, 521.00, 522.00, 523.00, 524.00, 525.00,
     &      526.00, 527.00, 528.00, 529.00, 530.00, 531.00, 532.00, 533.00, 534.00, 535.00, 536.00, 537.00, 538.00, 539.00, 540.00,
     &      541.00, 542.00, 543.00, 544.00, 545.00, 546.00, 547.00, 548.00, 549.00, 550.00, 551.00, 552.00, 553.00, 554.00, 555.00,
     &      556.00, 557.00, 558.00, 559.00, 560.00, 561.00, 562.00, 563.00, 564.00, 565.00, 566.00, 567.00, 568.00, 569.00, 570.00,
     &      571.00, 572.00, 573.00, 574.00, 575.00, 576.00, 577.00, 578.00, 579.00, 580.00, 581.00, 582.00, 583.00, 584.00, 585.00,
     &      586.00, 587.00, 588.00, 589.00, 590.00, 591.00, 592.00, 593.00, 594.00, 595.00, 596.00, 597.00, 598.00, 599.00, 600.00,
     &      601.00, 602.00, 603.00, 604.00, 605.00, 606.00, 607.00, 608.00, 609.00, 610.00, 611.00, 612.00, 613.00, 614.00, 615.00,
     &      616.00, 617.00, 618.00, 619.00, 620.00, 621.00, 622.00, 623.00, 624.00, 625.00, 626.00, 627.00, 628.00, 629.00, 630.00,
     &      631.00, 632.00, 633.00, 634.00, 635.00, 636.00, 637.00, 638.00, 639.00, 640.00, 641.00, 642.00, 643.00, 644.00, 645.00,
     &      646.00, 647.00, 648.00, 649.00, 650.00, 651.00, 652.00, 653.00, 654.00, 655.00, 656.00, 657.00, 658.00, 659.00, 660.00,
     &      661.00, 662.00, 663.00, 664.00, 665.00, 666.00, 667.00, 668.00, 669.00, 670.00, 671.00, 672.00, 673.00, 674.00, 675.00,
     &      676.00, 677.00, 678.00, 679.00, 680.00, 681.00, 682.00, 683.00, 684.00, 685.00, 686.00, 687.00, 688.00, 689.00, 690.00,
     &      691.00, 692.00, 693.00, 694.00, 695.00, 696.00, 697.00, 698.00, 699.00, 700.00 /
       data (af500(i),i=1,na500)
     &    / 0.0054, 0.0151, 0.0258, 0.0417, 0.0603, 0.0822, 0.1085, 0.1347, 0.1645, 0.1947, 0.2282, 0.2643, 0.2980, 0.3345, 0.3697,
     &      0.4050, 0.4412, 0.4761, 0.5109, 0.5426, 0.5732, 0.6013, 0.6287, 0.6555, 0.6791, 0.7007, 0.7207, 0.7395, 0.7559, 0.7703,
     &      0.7832, 0.7944, 0.8042, 0.8125, 0.8197, 0.8260, 0.8312, 0.8357, 0.8395, 0.8429, 0.8460, 0.8486, 0.8512, 0.8535, 0.8559,
     &      0.8583, 0.8607, 0.8634, 0.8661, 0.8690, 0.8719, 0.8751, 0.8784, 0.8818, 0.8853, 0.8888, 0.8925, 0.8962, 0.8998, 0.9034,
     &      0.9069, 0.9104, 0.9138, 0.9170, 0.9201, 0.9231, 0.9259, 0.9286, 0.9310, 0.9334, 0.9355, 0.9375, 0.9393, 0.9410, 0.9425,
     &      0.9439, 0.9452, 0.9463, 0.9474, 0.9483, 0.9492, 0.9500, 0.9507, 0.9514, 0.9521, 0.9527, 0.9532, 0.9538, 0.9543, 0.9548,
     &      0.9553, 0.9558, 0.9563, 0.9568, 0.9573, 0.9578, 0.9583, 0.9589, 0.9594, 0.9600, 0.9606, 0.9612, 0.9618, 0.9624, 0.9630,
     &      0.9637, 0.9643, 0.9650, 0.9657, 0.9664, 0.9671, 0.9678, 0.9685, 0.9692, 0.9699, 0.9706, 0.9713, 0.9720, 0.9727, 0.9734,
     &      0.9741, 0.9747, 0.9754, 0.9760, 0.9766, 0.9772, 0.9778, 0.9784, 0.9789, 0.9794, 0.9800, 0.9805, 0.9809, 0.9814, 0.9819,
     &      0.9823, 0.9827, 0.9831, 0.9835, 0.9838, 0.9842, 0.9845, 0.9849, 0.9852, 0.9855, 0.9857, 0.9860, 0.9862, 0.9865, 0.9867, 
     &      0.9869, 0.9871, 0.9873, 0.9875, 0.9877, 0.9879, 0.9880, 0.9882, 0.9884, 0.9885, 0.9886, 0.9888, 0.9889, 0.9890, 0.9892,
     &      0.9893, 0.9894, 0.9895, 0.9897, 0.9898, 0.9899, 0.9900, 0.9901, 0.9902, 0.9903, 0.9904, 0.9905, 0.9906, 0.9907, 0.9908,
     &      0.9909, 0.9910, 0.9911, 0.9912, 0.9913, 0.9914, 0.9915, 0.9916, 0.9917, 0.9917, 0.9918, 0.9919, 0.9920, 0.9921, 0.9922,
     &      0.9923, 0.9924, 0.9924, 0.9925, 0.9926, 0.9927, 0.9928, 0.9929, 0.9930, 0.9930, 0.9931, 0.9932, 0.9933, 0.9934, 0.9934,
     &      0.9935, 0.9936, 0.9937, 0.9937, 0.9938, 0.9939, 0.9939, 0.9940, 0.9940, 0.9941, 0.9941, 0.9942, 0.9942, 0.9943, 0.9943,
     &      0.9943, 0.9944, 0.9944, 0.9945, 0.9945, 0.9945, 0.9945, 0.9946, 0.9946, 0.9946, 0.9946, 0.9947, 0.9947, 0.9947, 0.9947,
     &      0.9947, 0.9948, 0.9948, 0.9948, 0.9948, 0.9948, 0.9949, 0.9949, 0.9949, 0.9949, 0.9949, 0.9949, 0.9949, 0.9950, 0.9950,
     &      0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950,
     &      0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9949, 0.9949, 0.9949, 0.9949, 0.9949, 0.9949, 0.9949, 0.9949, 0.9948, 0.9948,
     &      0.9948, 0.9948, 0.9948, 0.9948, 0.9948, 0.9948, 0.9948, 0.9948, 0.9948, 0.9948, 0.9947, 0.9948, 0.9948, 0.9947, 0.9947,
     &      0.9948, 0.9948, 0.9947, 0.9947, 0.9947, 0.9948, 0.9948, 0.9948, 0.9948, 0.9948, 0.9948, 0.9948, 0.9948, 0.9948, 0.9948,
     &      0.9948, 0.9948, 0.9949, 0.9949, 0.9949, 0.9949, 0.9949, 0.9949, 0.9949, 0.9949, 0.9949, 0.9950, 0.9950, 0.9950, 0.9950,
     &      0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950, 0.9950,
     &      0.9949, 0.9949, 0.9949, 0.9949, 0.9949, 0.9949, 0.9949, 0.9949, 0.9949, 0.9948, 0.9948, 0.9948, 0.9948, 0.9948, 0.9948,
     &      0.9948, 0.9948, 0.9948, 0.9948, 0.9947, 0.9947, 0.9947, 0.9947, 0.9947, 0.9947, 0.9947, 0.9947, 0.9948, 0.9948, 0.9948,
     &      0.9948, 0.9948, 0.9948, 0.9949, 0.9949, 0.9949, 0.9949, 0.9950, 0.9950, 0.9950, 0.9951, 0.9951, 0.9952, 0.9952, 0.9953,
     &      0.9953, 0.9954, 0.9955, 0.9955, 0.9956, 0.9957, 0.9957, 0.9958, 0.9959, 0.9959, 0.9960, 0.9961, 0.9962, 0.9962, 0.9963,
     &      0.9964, 0.9965, 0.9965, 0.9966, 0.9967, 0.9967, 0.9968, 0.9969, 0.9970, 0.9970, 0.9971, 0.9971, 0.9972, 0.9973, 0.9973,
     &      0.9974, 0.9975, 0.9975, 0.9976, 0.9976, 0.9977, 0.9978, 0.9978, 0.9979, 0.9979, 0.9980, 0.9980, 0.9981, 0.9982, 0.9982,
     &      0.9983, 0.9983, 0.9984, 0.9984, 0.9985, 0.9986, 0.9986, 0.9987, 0.9987, 0.9988, 0.9988, 0.9989, 0.9989, 0.9990, 0.9990,
     &      0.9990, 0.9991, 0.9991, 0.9992, 0.9992, 0.9992, 0.9993, 0.9993, 0.9993, 0.9994, 0.9994, 0.9994, 0.9995, 0.9995, 0.9995,
     &      0.9995, 0.9996, 0.9996, 0.9996, 0.9996, 0.9997, 0.9997, 0.9997, 0.9997, 0.9997, 0.9997, 0.9997, 0.9997, 0.9998, 0.9998,
     &      0.9998, 0.9998, 0.9998, 0.9998, 0.9998, 0.9998, 0.9999, 0.9999, 0.9999, 0.9999, 0.9999, 0.9999, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000,
     &      1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000 /

       external      inellipse, lastc, interpolation  !!, selectk
!__________________________________________________________________________________________________________________________________
!
       do n=1,nextr
         
         if (lnotbad(n) .and. lfootprint(n) .and. lsourcetodeblend(n)) then
           
! Attempt to correct for negative bg-subtracted values for very faint noisy sources.

           semimajor = foomajor(n) / 2.0d0
           nr = nint ( semimajor / min ( dx, dy ) )
           l1 = max ( nxco(n) - nr, 1 )
           l2 = min ( nxco(n) + nr, nx )
           m1 = max ( nyco(n) - nr, 1 )
           m2 = min ( nyco(n) + nr, ny )
           fluxtotbest(n) = 0.0d0

!!           nn = 0
!!           
!!           do m=m1,m2
!!             dblm = dble ( m )
!!             do l=l1,l2
!!               dbll = dble ( l )
!!           
!!               if (maskimage(l,m) .gt. 0.9d0 .and. obsfsbsimage(l,m) .lt. 0.0d0) then
!!                 if (inellipse ( dbll, dblm, dx, dy, momxco(n), momyco(n), foomajor(n), foominor(n), theta(n) )) then
!!                   nn = nn + 1
!!!!                   funx(nn) = abs ( obsfsbsimage(l,m) )
!!                   fluxtotbest(n) = fluxtotbest(n) + abs ( obsfsbsimage(l,m) )
!!                 endif
!!               endif
!!             enddo
!!           enddo
!!           
!!!!           if (nn .gt. 0) fluxtotbest(n) = selectk ( (nn + 1) / 2, nn, funx )
!!           if (nn .gt. 1) fluxtotbest(n) = fluxtotbest(n) / dble ( nn )
!!           fluxtotbest(n) = 0.5d0 * fluxtotbest(n)

           nkacc = 0
           do j=1,nkover(n)
             k = nkoverlap(n,j)
             if (fluxpeakbest(k) .gt. almostzero) then
               nkacc = nkacc + 1
               nkaccel(nkacc) = k
             endif
           enddo
           
           semimajor = foomajor(n) / 2.0d0
           nr = nint ( semimajor / min ( dx, dy ) )
           l1 = max ( nxco(n) - nr, 1 )
           l2 = min ( nxco(n) + nr, nx )
           m1 = max ( nyco(n) - nr, 1 )
           m2 = min ( nyco(n) + nr, ny )

           do m=m1,m2
             dblm = dble ( m )
             do l=l1,l2
               dbll = dble ( l )
    
               if (maskimage(l,m) .gt. 0.9d0) then
                 
                 if (inellipse ( dbll, dblm, dx, dy, momxco(n), momyco(n), foomajor(n), foominor(n), theta(n) )) then
                   argexp = aax(n) * dxcon2(n,l) + bbx(n) * sqrt ( dxcon2(n,l) * dycon2(n,m) ) + ccx(n) * dycon2(n,m)
                   argexpm = min ( argexp * dlog4h, 100.0d0 )
                   funn = fluxpeakbest(n) * exp ( - argexpm )
                   funksum = 0.0d0
                   do j=1,nkacc
                     k = nkaccel(j)
                     if (inellipse ( dbll, dblm, dx, dy, momxco(k), momyco(k), foomajor(k), foominor(k), theta(k) )) then
                       argexp = aax(k) * dxcon2(k,l) + bbx(k) * sqrt ( dxcon2(k,l) * dycon2(k,m) ) + ccx(k) * dycon2(k,m)
                       argexpm = min ( argexp * dlog4h, 100.0d0 )
                       funksum = funksum + fluxpeakbest(k) * exp ( - argexpm )
                     endif
                   enddo
                   
                   deblendfactor = 1.0d0
                   if (funn + funksum .ne. 0.0d0) then
                     deblendfactor = funn / (funn + funksum)
                   endif
                   
                   if (fluxpeakbest(n) .ge. 0.0d0) then
!!                     deblended = max ( obsfsbsimage(l,m), 0.0d0 ) * deblendfactor
                     deblended = obsfsbsimage(l,m) * deblendfactor
                   else
!!                     deblended = min ( obsfsbsimage(l,m), 0.0d0 ) * deblendfactor
                     deblended = obsfsbsimage(l,m) * deblendfactor
                   endif
                   fluxtotbest(n) = fluxtotbest(n) + deblended
                 endif
               endif
             enddo
           enddo
         endif
       enddo
           
! Aperture corrections for Herschel wavelengths.
           
       do n=1,nextr
         
         if (lnotbad(n) .and. lfootprint(n) .and. lsourcetodeblend(n)) then
           
           icw = lastc ( cowave )
           
           if (capcorr .eq. 'y' .and. intensunits .eq. 'MJy/sr' .and. 
     &         (cowave(1:icw) .eq. '70' .or.
     &          cowave(1:icw) .eq. '070' .or. cowave(1:icw) .eq. '0070' .or. 
     &          cowave(1:icw) .eq. '100' .or. cowave(1:icw) .eq. '0100' .or. 
     &          cowave(1:icw) .eq. '160' .or. cowave(1:icw) .eq. '0160' .or. 
     &          cowave(1:icw) .eq. '250' .or. cowave(1:icw) .eq. '0250' .or. 
     &          cowave(1:icw) .eq. '350' .or. cowave(1:icw) .eq. '0350' .or. 
     &          cowave(1:icw) .eq. '500' .or. cowave(1:icw) .eq. '0500')) then
             frac = 1.0d0
             if (cowave(1:icw) .eq. '070' .or. cowave(1:icw) .eq. '0070' .or. cowave(1:icw) .eq. '70') then
               if (equivrad(n) .ge. ar070(1) .and. equivrad(n) .le. ar070(na070)) then
                 call interpolation ( iotty, equivrad(n), na070, ar070, af070, frac )
               endif
             endif
             if (cowave(1:icw) .eq. '100' .or. cowave(1:icw) .eq. '0100') then
               if (equivrad(n) .ge. ar100(1) .and. equivrad(n) .le. ar100(na100)) then
                 call interpolation ( iotty, equivrad(n), na100, ar100, af100, frac )
               endif
             endif
             if (cowave(1:icw) .eq. '160' .or. cowave(1:icw) .eq. '0160') then
               if (equivrad(n) .ge. ar160(1) .and. equivrad(n) .le. ar160(na160)) then
                 call interpolation ( iotty, equivrad(n), na160, ar160, af160, frac )
               endif
             endif
             if (cowave(1:icw) .eq. '250' .or. cowave(1:icw) .eq. '0250') then
               if (equivrad(n) .ge. ar250(1) .and. equivrad(n) .le. ar250(na250)) then
                 call interpolation ( iotty, equivrad(n), na250, ar250, af250, frac )
               endif
             endif
             if (cowave(1:icw) .eq. '350' .or. cowave(1:icw) .eq. '0350') then
               if (equivrad(n) .ge. ar350(1) .and. equivrad(n) .le. ar350(na350)) then
                 call interpolation ( iotty, equivrad(n), na350, ar350, af350, frac )
               endif
             endif
             if (cowave(1:icw) .eq. '500' .or. cowave(1:icw) .eq. '0500') then
               if (equivrad(n) .ge. ar500(1) .and. equivrad(n) .le. ar500(na500)) then
                 call interpolation ( iotty, equivrad(n), na500, ar500, af500, frac )
               endif
             endif
             apcorrect(n) = 1.0d0 / frac

             fluxpeakbest(n) = fluxpeakbest(n) * apcorrect(n)
             fluxtotbest(n) = fluxtotbest(n) * apcorrect(n)
             fluxtotfit(n) = fluxtotfit(n) * apcorrect(n)
           endif
                 
           if (maskimage(nxco(n),nyco(n)) .lt. 0.9d0) then
             fluxtotbest(n) = 0.0d0
             fluxtotfit(n) = 0.0d0
           endif

!!! The following correction factor fixes the problem of overestimated source background that leaves a pedestal and leads to
!!! underestimated FWHM size, peak flux, and total flux. For unresolved sources, it makes FWHM smaller than the beam size, which
!!! clearly indicates the problem of oversubtracted background. The correction factors are found empirically by considering a
!!! Gaussian source and subtracting its faintest parts of varying thickness.
!!
!!           if (lfinishing_mode) then
!!             sizeratio = beam / min ( sqrt ( afwhm(n) * bfwhm(n) ), beam )
!!             corrfactpeak = sizeratio**(1.75d0) + 0.35d0 * (1.0d0 - sizeratio)
!!             corrfactflux = 1.15d0*sizeratio**(3.87d0) - 0.082d0 * (sizeratio - 0.04d0)**(-15)
!!             fluxpeakbest(n) = corrfactpeak * fluxpeakbest(n)
!!             fluxtotbest(n) = corrfactflux * fluxtotbest(n)
!!           endif
           
! Convert flux units for extracted sources.

           iu = lastc ( intensunits )
           if (intensunits(1:iu) .eq. 'H2/cm^2') then
             fluxtotfit(n) = fluxtotbest(n) * convftotm    !<-- convert to {Jy|Msun}
           else
             fluxtotfit(n) = fluxtotfit(n) * convftotm     !<-- convert to {Jy|Msun}
           endif
           fluxtotbest(n) = fluxtotbest(n) * convftot   !<-- convert to {Jy|H2/cm^2*arcsec^2}
         endif
       enddo

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine createannuli

     &            ( nx, ny, dx, dy, nextr, lnotbad, fluxperror, momxco, momyco, nxco, nyco, foominor, foomajor, theta, maskimage
     &            , lfootprint, n1x, n2x, sqpixel, factoffset, sq_arcsecs_per_beam_area, pi, n1ymin, n2ymax, factortotal, footfact
     &            , annulusarea, smallfoot, bigfoot, annulus )
!__________________________________________________________________________________________________________________________________
!
! Create annuli around sources.
!__________________________________________________________________________________________________________________________________
!
       implicit      none

       logical       inellipse, lfootprint(*), lnotbad(*)

       integer       nx, ny, n, nr, nextr, l1, l2, m1, m2, m, l, n1xmin, n2xmax, n1ymin, n2ymax, n1y, n2y, n1x(ny), n2x(ny)
     &             , nxco(*), nyco(*)

       real*8        dx, dy, dblm, dbll, sqpixel, aaa, bbb, ccc, pi, minor1, major1, minor2, major2, factoffset, semiminor1
     &             , semimajor1, semiminor2, semimajor2, sq_arcsecs_per_beam_area, factortotal    

       real*8        maskimage(nx,ny), smallfoot(nx,ny), bigfoot(nx,ny), annulus(nx,ny), momxco(*), momyco(*), foominor(*)
     &             , foomajor(*), theta(*), fluxperror(*), footfact(*), annulusarea(*)
      
       external      inellipse
!__________________________________________________________________________________________________________________________________
!
! Prepare the limiting boxes to accelerate.

       n1xmin = nx
       n2xmax = 1
       do m=1,ny
         n1x(m) = 0
         do l=1,nx
           if (maskimage(l,m) .gt. 0.9d0) then
             n1x(m) = max ( l - 2, 1 )
             if (n1x(m) .lt. n1xmin) n1xmin = n1x(m)
             exit
           endif
         enddo
         n2x(m) = 0
         do l=nx,1,-1
           if (maskimage(l,m) .gt. 0.9d0) then
             n2x(m) = min ( l + 2, nx )
             if (n2x(m) .gt. n2xmax) n2xmax = n2x(m)
             exit
           endif
         enddo
       enddo
       
       n1ymin = ny
       n2ymax = 1
       do l=1,nx
         do m=1,ny
           if (maskimage(l,m) .gt. 0.9d0) then
             n1y = max ( m - 2, 1 )
             if (n1y .lt. n1ymin) n1ymin = n1y
             exit
           endif
         enddo
         do m=ny,1,-1
           if (maskimage(l,m) .gt. 0.9d0) then
             n2y = min ( m + 2, ny )
             if (n2y .gt. n2ymax) n2ymax = n2y
             exit
           endif
         enddo
       enddo
       
       if (n1ymin .eq. ny) n1ymin = 1
       if (n2ymax .eq. 1 ) n2ymax = 1
       do m=1,ny
         if (n1x(m) .eq. 0) n1x(m) = 1
         if (n2x(m) .eq. 0) n2x(m) = 1
       enddo
       
       do m=1,ny
         do l=1,nx
           smallfoot(l,m) = 0.0d0
           bigfoot(l,m) = 0.0d0
           annulus(l,m) = 0.0d0
         enddo
       enddo

       aaa = 1.0d0

! Prepare the accumulated footprints image for the inner sides of the annuli.
! The factor (footfact(n) / factortotal) is here to expand the annulus area in proportion to the expanding footprint area.

       do n=1,nextr
         
         annulusarea(n) = 20.0d0 * max ( sq_arcsecs_per_beam_area * (abs ( footfact(n) ) / factortotal)  !!!**2
     &                                 , pi * sqpixel / log ( 2.0d0 ) )

! Do not use lsourcetodeblend in the next condition, because otherwise the annuli around the 
! accelerated sources would not exist in the BIGFOOT image, making the error estimates wrong.
         
         if (lnotbad(n) .and. (lfootprint(n) .or. (maskimage(nxco(n),nyco(n)) .gt. 0.9d0 .and. fluxperror(n) .lt. 1.0d-50))) then
           
           major1 = factoffset * foomajor(n)
           minor1 = factoffset * foominor(n)
           semimajor1 = major1 / 2.0d0

! Limit the area to accelerate.
           
           nr = nint ( semimajor1 / min ( dx, dy ) ) + 1
           l1 = max ( nxco(n) - nr, 1 )
           l2 = min ( nxco(n) + nr, nx )
           m1 = max ( nyco(n) - nr, 1 )
           m2 = min ( nyco(n) + nr, ny )
           
           do m=m1,m2
             dblm = dble ( m )
             do l=l1,l2
               dbll = dble ( l )
                 
               if (maskimage(l,m) .gt. 0.9d0) then
                 if (inellipse ( dbll, dblm, dx, dy, momxco(n), momyco(n), major1, minor1, theta(n) )) then
                   smallfoot(l,m) = 0.1d0
                 endif
               endif
             enddo
           enddo
         endif
       enddo
       
! Prepare the accumulated footprints image for the annuli.

       do n=1,nextr
         
! Do not use lsourcetodeblend in the next condition, because otherwise the annuli around the 
! accelerated sources would not exist in the BIGFOOT image, making the error estimates wrong.
         
         if (lnotbad(n) .and. (lfootprint(n) .or. (maskimage(nxco(n),nyco(n)) .gt. 0.9d0 .and. fluxperror(n) .lt. 1.0d-50))) then
           
           major1 = factoffset * foomajor(n)
           minor1 = factoffset * foominor(n)
           semimajor1 = major1 / 2.0d0
           semiminor1 = minor1 / 2.0d0
           
           bbb = semimajor1 - semiminor1
           ccc = -annulusarea(n) / pi - semiminor1 * semimajor1
           semiminor2 = (-bbb + sqrt ( bbb**2 - 4.0d0 * aaa * ccc )) / (2.0d0 * aaa)
           semimajor2 = semimajor1 + semiminor2 - semiminor1
           major2 = semimajor2 * 2.0d0
           minor2 = semiminor2 * 2.0d0

! Limit the area to accelerate.
           
           nr = nint ( semimajor2 / min ( dx, dy ) ) + 1
           l1 = max ( nxco(n) - nr, 1 )
           l2 = min ( nxco(n) + nr, nx )
           m1 = max ( nyco(n) - nr, 1 )
           m2 = min ( nyco(n) + nr, ny )
           
           do m=m1,m2
             dblm = dble ( m )
             do l=l1,l2
               dbll = dble ( l )
                 
               if (maskimage(l,m) .gt. 0.9d0) then
                 if (inellipse ( dbll, dblm, dx, dy, momxco(n), momyco(n), major2, minor2, theta(n) )) then
                   bigfoot(l,m) = max ( 0.1d0 - smallfoot(l,m), 0.0d0 )
                   annulus(l,m) = bigfoot(l,m)
                 endif
               endif
             enddo
           enddo
         endif
       enddo
       
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine expandfootprints

     &            ( iotty, nx, ny, dx, dy, nextr, lnotbad, momxco, momyco, nxco, nyco, fluxpeakbest, fluxperror, fluxtotbest
     &            , fluxterror, scalensx, afwhm, bfwhm, asize, bsize, beam, sigglob, sigmono, footfact, obsname, cdate, ctime
     &            , compda, clibname, cfitsversion, lfirstcall, lfootprint, foomajor, foominor, theta, maskimage, almostzero
     &            , obsfscbimage, factortotal, lnoexpansion, lnoshrinkage )
!__________________________________________________________________________________________________________________________________
!
! Expand footprints of power-law sources footprints.
!__________________________________________________________________________________________________________________________________
!
       implicit      none

       logical       inellipse, lfirstcall, lexpand, lshrink, lnoexpansion, lnoshrinkage, lallowexpansion, lallowshrinkage
     &             , lgoodtoexpand, lfootprint(*), lnotbad(*)

       character*(*) cfitsversion, ctime, cdate, compda, obsname, clibname

       integer       nx, ny, n, nextr, nr, nxco(*), nyco(*), l, m, l1, l2, m1, m2, npixels, npixelsx, iotty, irc, numpix
     
       real*8        selectk, dx, dy, beam, almostzero, snpeak, snflux, snpsigg, snfsigg, snpsigm, snfsigm, dbll, dblm, fexpansion
     &             , semimajor2, foomajor2, foominor2, factout1, foomajor1, foominor1, factortotal, foomajor0, foominor0
     &             , upperlimit, ffestimated, pixsize, sobackground, surroundings, reldifftol, factout2, semimajor, elongation
     &             , sizeratio, fshrinkage, lowerlimit, meabground, measurround, stdsurround, medsurroundx, medsurrounds
     &             , medbgroundx, medbgrounds
     
       parameter   ( lallowexpansion = .true., lallowshrinkage = .true., fexpansion = 1.05d0, fshrinkage = 1.02d0
     &             , reldifftol = 0.0d0, lowerlimit = 2.1d0 )

       real*8        momxco(*), momyco(*), fluxpeakbest(*), fluxperror(*), fluxtotbest(*), fluxterror(*), footfact(*), scalensx(*)
     &             , sigglob(*), sigmono(*), afwhm(*), bfwhm(*), asize(*), bsize(*), foominor(*), theta(*), maskimage(nx,ny)
     &             , foomajor(*), obsfscbimage(nx,ny)

       real*8, allocatable :: fun(:)
       
       external      inellipse, selectk
!__________________________________________________________________________________________________________________________________
!
       if (.not.lfirstcall) then
       
         open ( 17, file='sm'//obsname//'.foot.expansion.dat', status='unknown' )
         write (17,'(a)') '# Date: '//cdate//' Time: '//ctime//'  Created by SMEASURE compiled on '//compda
         write (17,'(a)') '# Alexander Men''shchikov, DAp IRFU CEA Saclay, France (alexander.menshchikov@cea.fr)'
         write (17,'(a)') '# Used '//clibname//' library version '//cfitsversion//' by William D Pence '
     &                   //'(william.d.pence@nasa.gov)'
         write (17,'(a)') '#'
         write (17,'(a)') '# FOOTPRINT EXPANSION OR SHRINKAGE'
         write (17,'(a)') '#'
         write (17,'(3(a,f6.3))') '# RELDIFFTOL =', reldifftol, ', FEXPANSION =', fexpansion, ', FSHRINKAGE =', fshrinkage
         write (17,'(a)') '#'
         write (17,'(a)') '#  N XCO_PX YCO_PX   MEDBGRD MEDSURRD STDSURRD FBGROUND    NP X S   FOOFA SCALENSX   AFWHM'
     &                   //'    BFWHM    ASIZE    BSIZE    SIGN   SIGNMON  S/N_PEAK S/N_FLUX SNP/SIGG SNF_SIGG SNP/SIGM SNF/SIGM'
         write (17,'(a)') '#'
         
         lnoexpansion = .true.
         lnoshrinkage = .true.
         pixsize = sqrt ( dx * dy )

         allocate ( fun(nx*ny), stat=irc )
         if (irc .ne. 0) then
           write (iotty,'(/a)') '   SMEASURE: ERROR: Trouble allocating memory (150).'
           stop 150
         endif

! Define background values for footprint expansion over the pedestal area (within the footprint).
       
         do n=1,nextr
    
           meabground = 0.0d0
           medbgroundx = 0.0d0
           medbgrounds = 0.0d0
           measurround = 0.0d0
           medsurroundx = 0.0d0
           medsurrounds = 0.0d0
           stdsurround = 0.0d0
           sobackground = 0.0d0
           surroundings = 0.0d0
           lexpand = .false.
           lshrink = .false.
           lgoodtoexpand = .true.
           snpeak = 0.0d0
           snflux = 0.0d0
           snpsigg = 0.0d0
           snfsigg = 0.0d0
           snpsigm = 0.0d0
           snfsigm = 0.0d0
           numpix = -1
           upperlimit = 0.9d0 * min ( nx * dx, ny * dy ) / max ( scalensx(n), beam )

           if (lnotbad(n) .and. lfootprint(n) .and. maskimage(nxco(n),nyco(n)) .gt. 0.9d0 .and. footfact(n) .gt. 0.0d0 .and.
     &         sigmono(n) .gt. 1.0d-30 .and. fluxpeakbest(n) .gt. almostzero) then 

             semimajor = foomajor(n) / 2.0d0
             nr = nint ( semimajor / min ( dx, dy ) )
             l1 = max ( nxco(n) - nr, 1 )
             l2 = min ( nxco(n) + nr, nx )
             m1 = max ( nyco(n) - nr, 1 )
             m2 = min ( nyco(n) + nr, ny )
             npixels = 0
          
! Compute the mean value of all pixels in the entire footprint area.

             do m=m1,m2
               dblm = dble ( m )
               do l=l1,l2
                 dbll = dble ( l )
                 if (maskimage(l,m) .gt. 0.9d0) then
                   if (abs ( obsfscbimage(l,m) ) .gt. almostzero) then
                     if (inellipse ( dbll, dblm, dx, dy, momxco(n), momyco(n), foomajor(n), foominor(n), theta(n) )) then
                       npixels = npixels + 1
                       meabground = meabground + obsfscbimage(l,m)
                       fun(npixels) = obsfscbimage(l,m)
                     endif
                   endif
                 else
                   lgoodtoexpand = .false.
                 endif
               enddo
             enddo
             
             if (npixels .gt. 0) then
               meabground = meabground / dble ( npixels )
               medbgroundx = selectk ( (npixels + 1) / 3, npixels, fun )  !<~ for expansion
               medbgrounds = selectk ( (npixels + 1) / 4, npixels, fun )  !<~ for shrinkage
             else
               write (iotty,'(/a)') '   SMEASURE: ERROR: (1) Trouble with the annulus: no pixels?'
               write (iotty,'(3x,2i5,4(1pe11.3))') n, npixels, foomajor0, foominor0, foomajor(n), foominor(n)
               stop 99
             endif
             
! Compute the mean value of all pixels in the "outer" annulus adjacent to the edge of the footprint area.

!!             factout1 = (foominor(n) + 2.0d0 * beam) / foominor(n)
             factout1 = fexpansion
             foomajor1 = foomajor(n) * factout1
             foominor1 = foominor(n) * factout1
             factout2 = (foominor1 + 4.0d0 * beam) / foominor1
             foomajor2 = foomajor1 * factout2
             foominor2 = foominor1 * factout2

             semimajor2 = foomajor2 / 2.0d0
             nr = nint ( semimajor2 / min ( dx, dy ) )
             l1 = max ( nxco(n) - nr, 1 )
             l2 = min ( nxco(n) + nr, nx )
             m1 = max ( nyco(n) - nr, 1 )
             m2 = min ( nyco(n) + nr, ny )
             numpix = 0
             npixelsx = 0

             do m=m1,m2
               dblm = dble ( m )
               do l=l1,l2
                 dbll = dble ( l )
                 if (maskimage(l,m) .gt. 0.9d0) then
                   if (abs ( obsfscbimage(l,m) ) .gt. almostzero) then
                     if (inellipse ( dbll, dblm, dx, dy, momxco(n), momyco(n), foomajor2, foominor2, theta(n) )) then
                       if (.not.inellipse ( dbll, dblm, dx, dy, momxco(n), momyco(n), foomajor1, foominor1, theta(n) )) then
                         numpix = numpix + 1
                         measurround = measurround + obsfscbimage(l,m)
                         fun(numpix) = obsfscbimage(l,m)
                       endif
                     endif
                   endif
                 endif
               enddo
             enddo
             
             if (numpix .gt. 0) then
               measurround = measurround / dble ( numpix )
               medsurroundx = selectk ( (numpix + 1) / 4, numpix, fun )  !<~ Lower quartile value, not median!
               medsurrounds = selectk ( (numpix + 1) / 4, numpix, fun )  !<~ Lower quartile value, not median!
             else
               write (iotty,'(/a)') '   SMEASURE: ERROR: (3) Trouble with the annulus: no pixels?'
               write (iotty,'(3x,2i5,4(1pe11.3))') n, numpix, foomajor1, foominor1, foomajor2, foominor2
               stop 99
             endif

! Compute the standard deviations of all pixels in the "outer" annulus adjacent to the edge of the footprint area.
    
             do m=m1,m2
               dblm = dble ( m )
               do l=l1,l2
                 dbll = dble ( l )
                 if (maskimage(l,m) .gt. 0.9d0) then
                   if (inellipse ( dbll, dblm, dx, dy, momxco(n), momyco(n), foomajor2, foominor2, theta(n) )) then 
                     if (.not.inellipse ( dbll, dblm, dx, dy, momxco(n), momyco(n), foomajor1, foominor1, theta(n) )) then 
                       npixelsx = npixelsx + 1
                       stdsurround = stdsurround + ( obsfscbimage(l,m) - measurround)**2
                     endif
                   endif
                 endif
               enddo
             enddo
    
             if (npixelsx .gt. 1) then
               stdsurround = sqrt ( stdsurround / dble ( npixelsx - 1 ) ) 
             else
               write (iotty,'(/a)') '   SMEASURE: ERROR: (4) Trouble with annulus: no pixels?'
               write (iotty,'(3x,2i5,4(1pe11.3))') n, npixelsx, foomajor1, foominor1, foomajor2, foominor2
               stop 99
             endif

! Condition for expanding footprints: difference between the source background and surroundings > 0.

!!             sobackground = medbgroundx
!!             surroundings = medsurroundx + stdsurround
!!             sobackground = 1.3d0 * meabground     
             sobackground = 1.1d0 * medbgroundx
             surroundings = measurround + stdsurround  

             elongation = afwhm(n) / bfwhm(n)
             sizeratio = foomajor(n) / afwhm(n)
             ffestimated = footfact(n) * fexpansion
             snpeak = fluxpeakbest(n) / fluxperror(n)

             lexpand = lallowexpansion .and. lgoodtoexpand .and. sobackground .gt. surroundings .and.
     &                 ffestimated .le. upperlimit .and. elongation .le. 1.3d0 .and. sizeratio .ge. 1.5d0 .and.
     &                 (sigmono(n) .gt. 50.0d0 .or. snpeak .gt. 15.0d0) .and. footfact(n) .ge. factortotal

             if (lexpand) then
               footfact(n) = min ( ffestimated, upperlimit )
               lnoexpansion = .false.
             else

! When expansion has finished, some small pedestals still remain; somewhat increasing footprints will eliminate the pedestals.
! If the footprint expansion was not possible, try shrinking the footprints. 

               if (footfact(n) .gt. factortotal * 1.000001d0) then
                 footfact(n) = -min ( footfact(n) * (0.9d0 + 0.1d0 * footfact(n) / factortotal), upperlimit )
               else
!!                 sobackground = medbgrounds
!!                 surroundings = medsurrounds + stdsurround
!!                 sobackground = 1.3d0 * meabground 
                 sobackground = 1.1d0 * medbgrounds
                 surroundings = measurround + stdsurround  

                 ffestimated = footfact(n) / fshrinkage

                 lshrink = lallowshrinkage .and. sobackground .lt. surroundings .and. ffestimated .ge. lowerlimit

                 if (lshrink) then
                   footfact(n) = max ( ffestimated, lowerlimit )
                   lnoshrinkage = .false.
                 else
                   if (footfact(n) .le. factortotal) then
                     footfact(n) = -min ( footfact(n) * 1.1d0, factortotal )
                   endif
                 endif
               endif
             endif

! Write some information to a file.

             if (fluxperror(n) .gt. almostzero) then
               snpeak = fluxpeakbest(n) / fluxperror(n)
               if (sigglob(n) .gt. almostzero) snpsigg = snpeak / sigglob(n)
               if (sigmono(n) .gt. almostzero) snpsigm = snpeak / sigmono(n)
             endif
             
             if (fluxterror(n) .gt. almostzero) then
               snflux = fluxtotbest(n) / fluxterror(n)
               if (sigglob(n) .gt. almostzero) snfsigg = snflux / sigglob(n)
               if (sigmono(n) .gt. almostzero) snfsigm = snflux / sigmono(n)
             endif
           else
             if (footfact(n) .gt. 0.0d0) then
               footfact(n) = - factortotal
             endif
           endif
           
           write (17,'(i4,2(f7.1),1x,4(1pe9.2),i6,2(l),0pf8.3,13(1pe9.2))') n, momxco(n), momyco(n), medbgroundx, medsurroundx
     &              , stdsurround, sobackground - surroundings, numpix, lexpand, lshrink, footfact(n), scalensx(n), afwhm(n)
     &              , bfwhm(n), asize(n), bsize(n), sigglob(n), sigmono(n), snpeak, snflux, snpsigg, snfsigg, snpsigm, snfsigm 
         enddo

         deallocate ( fun )

         close ( 17 )
       endif
       
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine monosignificances

     &            ( iotty, nx, ny, nextr, filewithnames, beam, catline, lnotbad, cverbose, lfirstcall, lfinishing_mode, liter_phase2
     &            , liter_phase3, lonlypositives, almostzero, scalensx, sigmono, fluxpeakbest, fluxperror, maskimage, nxco, nyco
     &            , foomajor, foominor, theta, factortotal, flm, momxco, momyco, aad, bbd, ccd, lfootprint, cdate, ctime, compda
     &            , clibname, cfitsversion, dxcon2, dycon2, cflm, lsourcetodeblend, nkover, nkoverlap, flag, convdelta )

!__________________________________________________________________________________________________________________________________
!
! Compute monochromatic characteristic sizes and significances.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       
       logical       lfnam1, lfirstcall, lfinishing_mode, liter_phase2, liter_phase3, lonlypositives, lfootprint(*), lnotbad(*)
     &             , lsourcetodeblend(*), lmonosignificances
       
       character*(*) catline(*), cflm(*), flag(*), cverbose, filewithnames, cdate, ctime, compda, clibname, cfitsversion
       character*1   cfl
       character*80  object, creator, ctype1, ctype2, bunit, history
       character*500 sswavename(100)
       
       integer       lastc, lasts, nx, ny, iotty, k, kk, n, no, blank, irc, iss, isca, nheadf, nextr, nfiles, nfiles1
     &             , nfiles2, ns1, ns2, nsx, nxco(*), nyco(*), flm(*), nkover(*), nkoverlap(nextr,*)
       
       logical       ldosigmono(nextr)

       real*8        dx, dy, crpix1, crpix2, crval1, crval2, ra, dec, beamd, funmin, funmax, crota1, crota2, equinox
     &             , bzero, bscale, datamin, datamax, cd11, cd12, cd21, cd22, wave, almostzero, beam, factortotal
       
       real*8        sigma(100), cutoff(100), sigmas(100), sscale(100), scalensx(*), fluxpeakbest(*), fluxperror(*), sigmono(*)
     &             , foomajor(*), foominor(*), theta(*), maskimage(nx,ny), ssimagemax(nextr), fluxpksscale(nextr)
     &             , momxco(*), momyco(*), aad(*), bbd(*), ccd(*), dxcon2(nextr,*), dycon2(nextr,*), convdelta(nextr,*)
       
       real*8, allocatable :: ssimage(:,:)
       
       external      lastc, lasts, rfits, skipcomm, linescolumns, deblendpeaks
!__________________________________________________________________________________________________________________________________
!
! For sources in absorption, there is no way of computing accurate single-scale monochromatic significances.
! One has to replace the significance with a signal-to-noise ratio.

       if (.not.lfirstcall) then
         do n=1,nextr
           if (lfootprint(n) .and. fluxpeakbest(n) .lt. -almostzero) then
             if (fluxperror(n) .gt. almostzero) then
               if (lonlypositives) then
                 sigmono(n) = fluxpeakbest(n) / fluxperror(n)
               else
                 sigmono(n) = abs ( fluxpeakbest(n) ) / fluxperror(n)
               endif
             else
               sigmono(n) = 6.666d-31  !<-- Signals negative peaks and zero errors
             endif
           endif
         enddo
       endif

       allocate ( ssimage(nx,ny), stat=irc )

       if (irc .ne. 0) then
         write (iotty,'(/a)') '   SMEASURE: ERROR: Trouble allocating memory (79).'
         stop 79
       endif

       if (cverbose .eq. '-verb2') then
         write (iotty,'(a)') '   Reading names from '''//filewithnames//''''
       endif
       
       open ( 25, file=filewithnames, status='old' )
       
       call skipcomm     ( 25, filewithnames, nheadf )
       call linescolumns ( 25, filewithnames, nheadf, nfiles )
       
       rewind ( 25 )
       do k=1,nheadf
         read (25,'(a)')
       enddo

       do kk=1,nfiles
         read (25,*) sswavename(kk), sigma(kk), sigmas(kk), cutoff(kk)
         iss = lastc ( sswavename(kk) )
         isca = index ( sswavename(kk)(1:iss), '.c' )
         read (sswavename(kk)(isca+2:isca+7),*) sscale(kk)
       enddo

       close ( 25 )

       nfiles1 = 0
       nfiles2 = nfiles
       
       do kk=1,nfiles
         iss = lastc ( sswavename(kk) )
         inquire ( file=sswavename(kk)(1:iss), exist=lfnam1 )
         if (nfiles1 .eq. 0 .and. lfnam1 .and. cutoff(kk) .lt. 1.0d26) then
           nfiles1 = kk + 1
         endif
         if (nfiles1 .gt. 0 .and. .not.lfnam1) then
           nfiles2 = kk - 1
           exit
         endif
       enddo 
       
       if (nfiles1 .eq. 0 .or. nfiles2 .eq. 0) then
         write (iotty,'(/a)') '   SMEASURE: ERROR: Trouble finding files from '//filewithnames
         stop 38
       endif
       
       do n=1,nextr
         ssimagemax(n) = 0.0d0
         ldosigmono(n) = .true.
       enddo

! Compute single-scale significances.
 
       do kk=nfiles1,nfiles2
         
         iss = lastc ( sswavename(kk) )
         
         call rfits ( nx, ny, bunit, ctype1, ctype2, crpix1, crpix2, crval1, crval2, ssimage, dx, dy, object, ra, dec
     &              , sswavename(kk)(1:iss), iotty, 0, creator, beamd, funmin, funmax, blank, crota1, crota2, cd11, cd12, cd21
     &              , cd22, equinox, bzero, bscale, wave, datamin, datamax, history, cverbose )
       
         if (lfirstcall) then
           if (cverbose .eq. '-verb2') write (iotty,'(a,i2)') '   Computing monochromatic characteristic sizes on scale ', kk
         else
           if (cverbose .eq. '-verb2') write (iotty,'(a,i2)') '   Computing monochromatic significances on scale ', kk
         endif
         
         lmonosignificances = .true.

         call deblendpeaks ( nx, ny, dx, dy, nextr, lnotbad, fluxpksscale, maskimage, flm, momxco, momyco, nxco, nyco, foominor
     &                     , foomajor, theta, aad, bbd, ccd, sswavename(kk)(1:iss), ssimage, lfootprint, cdate, ctime, compda
     &                     , clibname, cfitsversion, dxcon2, dycon2, cflm, almostzero, lsourcetodeblend, nkover, nkoverlap, flag
     &                     , lmonosignificances, convdelta )
         do n=1,nextr

           read (catline(n)(2:18),'(i5,3i3,2x,a1)') no, ns1, ns2, nsx, cfl
       
           if (lnotbad(n) .and. lfootprint(n) .and. fluxpksscale(n) .gt. almostzero) then
                 
             if (kk .ge. ns1 .and. kk .le. ns2 .and. sscale(kk) .le. 1.15d0 * max ( scalensx(n), beam )) then
               
               if (ldosigmono(n) .and. fluxpksscale(n) .gt. ssimagemax(n)) then
                 ssimagemax(n) = fluxpksscale(n)
                 if (maskimage(nxco(n),nyco(n)) .gt. 0.9d0) then
                   sigmono(n) = fluxpksscale(n) / sigma(kk)
                 else
                   sigmono(n) = 0.0d0
                 endif
!!                 if (lfirstcall) then
!!                   scalensx(n) = sscale(kk)    !! max ( , beam )
!!                 endif
               else
                 ldosigmono(n) = .false.
               endif
             endif
           endif
         enddo
       enddo

! For sources in absorption, assign a temporary value of 100 to allow normal further processing. 
! In the last measurement iteration, it will be replaced by the signal-to-noise value.

       do n=1,nextr
         
         if (lnotbad(n) .and. lfootprint(n)) then
         
           if (maskimage(nxco(n),nyco(n)) .gt. 0.9d0) then

             if (lfirstcall) then
               scalensx(n) = max ( scalensx(n), beam )
!!               scalensx(n) = scalensx(n) / 1.05d0
               foomajor(n) = factortotal * scalensx(n)
               foominor(n) = foomajor(n)
             else
               if (abs ( fluxpeakbest(n) ) .lt. almostzero) then
                 sigmono(n) = 7.777d-31  !<-- Signals zero peak intensity (deconvolved)
               else
                 if (ssimagemax(n) .lt. almostzero) then
                   sigmono(n) = 9.999d-31  !<-- Signals zero clean single scale intensity at source position
                 endif
                 if (fluxpeakbest(n) .lt. -almostzero .and. .not.lfinishing_mode) then
                   sigmono(n) = 100.0d0
                   if ((liter_phase2 .or. liter_phase3) .and. lonlypositives) then
                     sigmono(n) = -100.0d0
                   endif
                 endif
               endif
             endif
           else
             sigmono(n) = 0.0d0
           endif
         endif
       enddo
       
       deallocate ( ssimage )
      
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine interpolation

     &            ( iotty, x, n, arg, fun, y )
!__________________________________________________________________________________________________________________________________
!
! Interpolation in table fun(arg) for a given argument x. arg(i), i=1,2,...,n  should be monotonically increasing.
!
!   Input parameters:
!      x   - argument value for interpolation
!      n   - dimension of argument and function vectors
!      arg - vector of argument values of a table
!      fun - vector of function values of a table
!
!   Output parameter:
!      y   - value of function y(x)
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       integer       i, j, k, n, iotty
       real*8        x, y, dx, arg(n), fun(n)
!__________________________________________________________________________________________________________________________________
!
       if ((x .lt. arg(1) .and. abs ( (x - arg(1)) / (arg(1) + 1.0d-30) ) .gt. 1.0d-13) .or. 
     &     (x .gt. arg(n) .and. abs ( (x - arg(n)) / (arg(n) + 1.0d-30) ) .gt. 1.0d-13)) then
         
         write (iotty,'(/'' ERROR: Invalid argument x='',1pd10.3)') x
         write (iotty,'(/'' Here are the input data (x,n,j,argj,funj):''/)')
         write (iotty,*) x,n
         do j=1,n
            write (iotty,*) j, arg(j), fun(j)
         enddo
         stop 99
       endif
       i = 1
       j = n + 1
 10    continue
         k = (i + j) / 2
         if (x .le. arg(k)) then
           j = k
         else
           i = k
         endif
       if (j .gt. i + 1) then
         goto 10
       endif
       if (i .eq. n) i = i - 1
       dx = (x - arg(i)) / (arg(i+1) - arg(i) + 1.0d-30)
       y = fun(i) * (1.0d0 - dx) + fun(i+1) * dx
     
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine rfits 
       
     &            ( nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname, iotty, iolog, creator
     &            , beam, funmin, funmax, blank, rot1, rot2, cd11, cd12, cd21, cd22, equinox, bzero, bscale, wave, datamn, datamx
     &            , history, cverbose )
!__________________________________________________________________________________________________________________________________
!
! GETSF • Multi-Scale Multi-Wavelength Source & Filament Extraction • Alexander Men'shchikov, DAp IRFU CEA Saclay
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend, anynull, myfile
       character*(*) ctype1, ctype2, creator, fname, bunit, object, history, cverbose
       character*3   cnan
       character*25  cdatetime
       character*80  comment
       integer       i, j, nx, ny, stat, blksize, group, naxis, pcount, gcount, bitpix, unit, iotty, iolog, maxdim, rw
     &             , nfound, blank, stat05, stat06, stat07, stat08, stat09, stat10, stat11, stat12, stat13, stat14, stat15, stat16
     &             , stat17, stat18, stat19, stat20, stat21, stat22, stat23, stat24, stat25, stat26, stat27, stat28, stat29, stat30
     &             , stat31, stat32, stat0
       parameter   ( maxdim = 2 )
       integer       naxes(maxdim)
       real*8        dx, dy, ra, dec, beam, funmin, funmax, xr, yr, rp1, rp2, rot1, rot2, epoch, equinox, bzero, bscale, wave
     &             , dxdeg, dydeg, cd11, cd12, cd21, cd22, almostzero, as2deg, datamx, datamn, sign1, arg1, arg2, argm
     &             , sinrota, cosrota, pi, angle
       real*8        fun(nx,ny)
       parameter   ( pi = 3.14159265358979d0, as2deg = 1.0d0 / 3600.0d0, almostzero = 1.0d-30 )
       external      ftopen, ftgkne, ftghpr, ftgkys, ftgkyd, ftgkye, ftg2de, ftclos, printerr, ftgkyj, ftg2dd
!__________________________________________________________________________________________________________________________________
!
! Initialize variables.

       stat = 0
       unit   = 20
       group  = 1
       rw     = 0
       stat05 = 0
       stat06 = 0
       stat07 = 0
       stat08 = 0
       stat09 = 0
       stat10 = 0
       stat11 = 0
       stat12 = 0
       stat13 = 0
       stat14 = 0
       stat15 = 0
       stat16 = 0
       stat17 = 0
       stat18 = 0
       stat19 = 0
       stat20 = 0
       stat21 = 0
       stat22 = 0
       stat23 = 0
       stat24 = 0
       stat25 = 0
       stat26 = 0
       stat27 = 0
       stat28 = 0
       stat29 = 0
       stat30 = 0
       stat31 = 0
       stat32 = 0

! Open the FITS file.

       call ftopen ( unit, fname, rw, blksize, stat )

       if (stat .ne. 0) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Trouble opening FITS file '''//fname//'''.'
         stop 99
       endif

! Determine the size of the image.

       call ftgkne ( unit, 'NAXIS', 1, 2, naxes, nfound, stat )

! Check that it found both NAXIS1 and NAXIS2 keywords.

       if (nfound .ne. 2) then
         if (iotty .gt. 0) write (iotty,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         if (iolog .gt. 0) write (iolog,'(/a)') ' RFITS: ERROR: Failed to read the NAXISn keywords.'
         stop 99
       endif

! Get the main primary header keywords which define the array structure.

!!       call ftgphx ( unit, maxdim, simple, bitpix, naxis, naxes, pcount, gcount, extend, bscale, bzero, blank, nblank, stat )

       call ftghpr ( unit, maxdim, simple, bitpix, naxis, naxes, pcount, gcount, extend, stat )

! Read all parameters.

       call ftgkys ( unit, 'CREATOR', creator, comment, stat )
       
       if (stat .eq. 0 .and. (comment(1:21) .eq. 'Alexander Menshchikov' .or. comment(1:22) .eq. 'Alexander Men''shchikov')) then
         myfile = .true.
       else
         myfile = .false.
         stat = 0
         creator = 'SMEASURE'
       endif

! I don't understant why blank is assigned to my array values (if blank is not zero) - I have no undefined values in arrays. Maybe
! there is a bug in the version of FITSIO I use. For now I will ignore the blank value when reading FITS files.

       blank = 0

       datamx = 0.0d0
       datamn = 0.0d0
       bunit = ' '
       ctype1 = ' '
       ctype2 = ' '
       rp1 = 0.0d0
       rp2 = 0.0d0
       rot1 = 0.0d0
       rot2 = 0.0d0
       cd11 = 0.0d0
       cd12 = 0.0d0
       cd21 = 0.0d0
       cd22 = 0.0d0
       xr = 0.0d0
       yr = 0.0d0
       dx = 0.0d0
       dy = 0.0d0
       object = ' '
       ra = 0.0d0
       dec = 0.0d0
       epoch = 0.0d0
       equinox = 0.0d0
       beam = 0.0d0
       bzero = 0.0d0
       bscale = 1.0d0
       wave = 1.0d2
       history = ' '

       call ftgkys ( unit, 'DATE', cdatetime , comment, stat05 )
       call ftgkyd ( unit, 'DATAMAX', datamx , comment, stat06 )
       call ftgkyd ( unit, 'DATAMIN', datamn , comment, stat07 )
       call ftgkyj ( unit, 'BLANK'  , blank  , comment, stat08 )
       call ftgkys ( unit, 'BUNIT'  , bunit  , comment, stat09 )
       call ftgkyd ( unit, 'BZERO'  , bzero  , comment, stat10 )
       call ftgkyd ( unit, 'BSCALE' , bscale , comment, stat11 )
       call ftgkys ( unit, 'CTYPE1' , ctype1 , comment, stat12 )
       call ftgkys ( unit, 'CTYPE2' , ctype2 , comment, stat13 )
       call ftgkyd ( unit, 'CRPIX1' , rp1    , comment, stat14 )
       call ftgkyd ( unit, 'CRPIX2' , rp2    , comment, stat15 )
       call ftgkyd ( unit, 'CROTA1' , rot1   , comment, stat16 )
       call ftgkyd ( unit, 'CROTA2' , rot2   , comment, stat17 )
       call ftgkyd ( unit, 'CD1_1'  , cd11   , comment, stat18 )
       call ftgkyd ( unit, 'CD1_2'  , cd12   , comment, stat19 )
       call ftgkyd ( unit, 'CD2_1'  , cd21   , comment, stat20 )
       call ftgkyd ( unit, 'CD2_2'  , cd22   , comment, stat21 )
       call ftgkyd ( unit, 'CRVAL1' , xr     , comment, stat22 )
       call ftgkyd ( unit, 'CRVAL2' , yr     , comment, stat23 )
       call ftgkyd ( unit, 'CDELT1' , dxdeg  , comment, stat24 )
       call ftgkyd ( unit, 'CDELT2' , dydeg  , comment, stat25 )
       call ftgkys ( unit, 'OBJECT' , object , comment, stat26 )
       call ftgkyd ( unit, 'RA'     , ra     , comment, stat27 )
       call ftgkyd ( unit, 'DEC'    , dec    , comment, stat28 )
       call ftgkyd ( unit, 'EPOCH'  , epoch  , comment, stat29 )
       call ftgkyd ( unit, 'EQUINOX', equinox, comment, stat30 )
       call ftgkyd ( unit, 'WAVE'   , wave   , comment, stat31 )
       call ftgkyd ( unit, 'BEAM'   , beam   , comment, stat )
       call ftgkys ( unit, 'HISTORY', history, history, stat32 )

       if (stat .ne. 0) then
         beam = 0.0d0
         stat = 0
       endif

       stat0 = stat05 + stat06 + stat07 + stat09 + stat10 + stat11 + stat12 + stat13 + stat14 + stat15 + stat16 + stat17 + stat18
     &       + stat19 + stat20 + stat21 + stat22 + stat23 + stat24 + stat25 + stat26 + stat27 + stat28 + stat30 + stat31

       if (cverbose .ne. '-verb0') then
         if (stat0  .gt. 0) write (iotty,'()')
         if (stat05 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATE absent or error'
         if (stat06 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMAX absent or error'
         if (stat07 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DATAMIN absent or error'
         if (stat09 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BUNIT absent or error'
         if (stat10 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BZERO absent or error'
         if (stat11 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword BSCALE absent or error'
         if (stat12 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE1 absent or error'
         if (stat13 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CTYPE2 absent or error'
         if (stat14 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX1 absent or error'
         if (stat15 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRPIX2 absent or error'
         if (stat16 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA1 absent or error'
         if (stat17 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CROTA2 absent or error'
         if (stat18 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_1 absent or error'
         if (stat19 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD1_2 absent or error'
         if (stat20 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_1 absent or error'
         if (stat21 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CD2_2 absent or error'
         if (stat22 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL1 absent or error'
         if (stat23 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CRVAL2 absent or error'
         if (stat24 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT1 absent or error'
         if (stat25 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword CDELT2 absent or error'
         if (stat26 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword OBJECT absent or error'
         if (stat27 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword RA absent or error'
         if (stat28 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword DEC absent or error'
         if (stat30 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword EQUINOX absent or error'
         if (stat31 .ne. 0) write (iotty,'(a)') '   RFITS: WARNING: Keyword WAVE absent or error'
       endif
       if (stat26 .ne. 0 .or. object(1:10) .eq. '          ') object = '<Name>'

! Read the FITS file into the 2D array.

       call ftg2dd ( unit, group, dble ( blank ), naxes(1), naxes(1), naxes(2), fun, anynull, stat )

       if (stat .ne. 0) write (iotty,'(a)') ' Error reading FITS file'

! Close the file and free the unit number.

       call ftclos ( unit, stat )

! Check for any error, and if so print out error messages.

       if (stat .gt. 0) call printerr ( stat )

       if (stat17 .ne. 0) then
         rot2 = 0.0d0
       endif
       rot1 = rot2

       if (stat29 .eq. 0 .and. stat30 .ne. 0) then
         equinox = epoch
       endif

! CDELT* is absent or error.

       if (stat24 .ne. 0 .or. stat25 .ne. 0) then

! CD1* and CD2* are defined correctly: determine CDELT* and ROTA*.

         if (stat18 .eq. 0 .and. stat19 .eq. 0 .and. stat20 .eq. 0 .and. stat21 .eq. 0) then
           sign1 = sign ( 1.0d0, cd11 * cd22 - cd12 * cd21 )
           dxdeg = sign1 * sqrt ( cd11**2 + cd21**2 )
           dydeg = sqrt ( cd12**2 + cd22**2 )
           arg1 = sign1 * cd12 / cd22
           arg2 = -sign1 * cd21 / cd11
           argm = (arg1 + arg2) / 2.0d0
           if (abs ( arg1 - arg2 ) / argm .gt. 1.0d-10) then
             write (iotty,'(/a/)') '   RFITS: WARNING: Coordinate axes are not orthogonal'
           endif
           rot2 = atan ( argm )
           sinrota = sin ( rot2 )
           cosrota = cos ( rot2 )
           rot2 = rot2 * 180.0d0 / pi
           if (cosrota .ge. 0.0d0 .and. sinrota .ge. 0.0d0) then
             rot1 = rot2
           elseif (cosrota .lt. 0.0d0 .and. sinrota .ge. 0.0d0) then
             rot1 = 180.0d0 - rot2
           elseif (cosrota .lt. 0.0d0 .and. sinrota .lt. 0.0d0) then
             rot1 = 180.0d0 + rot2
           elseif (cosrota .ge. 0.0d0 .and. sinrota .lt. 0.0d0) then
             rot1 = - rot2
           endif
           rot2 = rot1
         endif
       endif

! CD1* and CD2* are absent or error.

       if (stat18 .ne. 0 .or. stat19 .ne. 0 .or. stat20 .ne. 0 .or. stat21 .ne. 0) then

! CDELT* are defined correctly: determine CD1* and CD2*.

         if (stat24 .eq. 0 .and. stat25 .eq. 0) then
           angle = rot2 * pi / 180.0d0
           cd11 = dxdeg * cos ( angle )
           cd12 = abs ( dydeg ) * sign ( 1.0d0, dxdeg ) * sin ( angle )
           cd21 = - abs ( dxdeg ) * sign ( 1.0d0, dydeg ) * sin ( angle )
           cd22 = dydeg * cos ( angle )
           if (abs ( cd12 ) .lt. almostzero) cd12 = 0.0d0
           if (abs ( cd21 ) .lt. almostzero) cd21 = 0.0d0
         endif
       endif

       dx = abs ( dxdeg ) / as2deg
       dy = abs ( dydeg ) / as2deg

! Print out the min and max values.

       nx = naxes(1)
       ny = naxes(2)
       funmin = 1.0d30
       funmax =-1.0d30
       cnan = 'NAN'
       do j=1,ny
         do i=1,nx
           if (.not.isnan ( fun(i,j) )) then  
             funmin = min ( funmin, fun(i,j) )
             funmax = max ( funmax, fun(i,j) )
           endif  
           if (int ( fun(i,j) ) .eq. blank .and. blank .ne. 0) then  
             read (cnan,*) fun(i,j)
           endif
         enddo
       enddo

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine wfits 
       
     &            ( cfitsversion, nx, ny, bunit, ctype1, ctype2, rp1, rp2, xr, yr, fun, dx, dy, object, ra, dec, fname, cdate
     &            , ctime, creator, beam, blank, rot1, rot2, cd11, cd12, cd21, cd22, equinox, bzero, bscale, wave, datamn, datamx
     &            , history )
!__________________________________________________________________________________________________________________________________
!
! GETSF • Multi-Scale Multi-Wavelength Source & Filament Extraction • Alexander Men'shchikov, DAp IRFU CEA Saclay
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       simple, extend
       character*(*) fname, ctype1, ctype2, object, creator, ctime, cdate, bunit, cfitsversion, history
       character*19  cdatetime
       integer       i, j, nx, ny, status, blocksze, group, naxis, pcount, gcount, bitpix, unit, naxes(2), blank
       real*8        xr, yr, dx, dy, rp1, rp2, ra, dec, beam, bzero, bscale, as2deg, rot1, rot2, equinox, datamn
     &             , datamx, dxdeg, dydeg, wave, cd11, cd12, cd21, cd22, almostzero, pi
       real*8        fun(nx,ny)
       parameter   ( pi = 3.14159265358979d0, as2deg = 1.0d0 / 3600.0d0, almostzero = 1.0d-30 )
       external      fdelete, ftgiou, ftinit, ftphpr, ftpkys, ftpcom, ftpkyd, ftpkyj, ftphis, ftp2dd, ftclos, ftfiou, printerr
!__________________________________________________________________________________________________________________________________
!
! Create a FITS primary array containing a 2-D image. Initialize parameters about the FITS image.

       status   = 0
       blocksze = 1
       simple   = .true.
       extend   = .false.
       group    = 1
       naxis    = 2
       pcount   = 0
       gcount   = 1
       naxes(1) = nx
       naxes(2) = ny

! Compute offset and scaling factor so as to represent real numbers most efficiently and accurately with given 'bitpix'.

       datamn = 1.0d30
       datamx =-1.0d30
       do j=1,ny
         do i=1,nx
           if (.not.isnan ( fun(i,j) )) then
             if (fun(i,j) .lt. datamn ) datamn = fun(i,j)
             if (fun(i,j) .gt. datamx ) datamx = fun(i,j)
           endif
         enddo
       enddo

! When packing numbers using the formulas below, I got values of the function different in the 5-th digit from DATAMIN and
!  DATAMAX placed in the header when writing the data. From now on I will use bzero = 0 bscale = 1 and bitpix = -32, as these
! seem to cure the problem.

cc!      bitpix = 16
cc!      bzero  = 0.5d0 * (funmin + funmax)
cc!      bscale = dabs ( funmax - bzero ) / 32767.0d0

       bitpix = -32     !<-- single precision floating-point values
       bzero  = 0.0d0
       bscale = 1.0d0
       blank = blank

       cdatetime = cdate(7:10)//'-'//cdate(4:5)//'-'//cdate(1:2)//'T'//ctime

! Delete the file if it already exists.

       if (cfitsversion .eq. '5.030') then
         call fdelete ( fname, status )
       endif

! Get an unused Logical Unit Number to use to open the FITS file.

       call ftgiou ( unit, status )

! Create the new empty FITS file.

       if (cfitsversion .eq. '5.030') then
         call ftinit ( unit, fname, blocksze, status )
       else
         call ftinit ( unit, '!'//fname, blocksze, status )
       endif

! Write the required header keywords.

       call ftphpr ( unit, simple, bitpix, naxis, naxes, pcount, gcount, extend, status )

! Write an optional comment.

       call ftpkys ( unit, 'CREATOR', creator,'Alexander Menshchikov, DAp IRFU CEA Saclay', status )
       call ftpcom ( unit,' ',status)
       call ftpkys ( unit, 'DATE'  , cdatetime, 'creation date and time', status )

! Write all parameters.

       dxdeg = - dx * as2deg
       dydeg = dy * as2deg

       call ftpkyd ( unit, 'BZERO'  , bzero , 13 , 'zero point in scaling equation', status )
       call ftpkyd ( unit, 'BSCALE' , bscale, 13 , 'linear factor in scaling equation' , status )
       call ftpkyd ( unit, 'DATAMAX', datamx, 13 , 'maximum data value', status )
       call ftpkyd ( unit, 'DATAMIN', datamn, 13 , 'minimum data value', status )
!!       if  (blank .gt. 0)
!!     & call ftpkyj ( unit, 'BLANK'  , blank      , 'value used for undefined array elements', status )
       call ftpkys ( unit, 'BUNIT'  , bunit      , 'physical units of the array values', status )
       call ftpkys ( unit, 'CTYPE1' , ctype1     , 'name of the coordinate axis', status )
       call ftpkys ( unit, 'CTYPE2' , ctype2     , 'name of the coordinate axis', status )
       call ftpkyd ( unit, 'CRPIX1' , rp1   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CRPIX2' , rp2   , 13 , 'coordinate system reference pixel', status )
       call ftpkyd ( unit, 'CROTA1' , rot1  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CROTA2' , rot2  , 13 , 'coordinate system rotation angle', status )
       call ftpkyd ( unit, 'CD1_1'  , cd11  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD1_2'  , cd12  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_1'  , cd21  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CD2_2'  , cd22  , 13 , 'linear projection matrix', status )
       call ftpkyd ( unit, 'CRVAL1' , xr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CRVAL2' , yr    , 13 , 'coordinate value at reference pixel', status )
       call ftpkyd ( unit, 'CDELT1' , dxdeg , 13 , 'coordinate increment along axis', status )
       call ftpkyd ( unit, 'CDELT2' , dydeg , 13 , 'coordinate increment along axis', status )
       call ftpkys ( unit, 'OBJECT' , object     , 'object identifier', status )
       call ftpkyd ( unit, 'RA'     , ra     , 13, 'right ascension', status )
       call ftpkyd ( unit, 'DEC'    , dec    , 13, 'declination', status )
       call ftpkyd ( unit, 'EQUINOX', equinox, 13, 'equinox', status )
       call ftpkyd ( unit, 'WAVE'   , wave   , 13, 'wavelength (microns)', status )
       call ftphis ( unit, history, status )

       if (beam .ne. 0.0d0) call ftpkyd ( unit, 'BEAM', beam, 13, 'convolved with the beam (arcsec)', status )

! Write the array to the FITS file.

       call ftp2dd ( unit, group, nx, nx, ny, fun, status )

! Close the file.

       call ftclos ( unit, status )

! Free the unit number.

       call ftfiou ( unit, status )

! Check for any error, and if so print out error messages

       if (status .gt. 0) call printerr ( status )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
